//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.agent;

//----------------------------------------------------------------------

import fipaos.ont.fipa.FIPACONSTANTS;
import fipaos.ont.BASETOKENS;
import fipaos.ont.fipa.fipaman.AgentID;

import java.util.TreeMap;
import java.util.Iterator;
import java.util.TreeSet;

//----------------------------------------------------------------------

/**
 * This class implements a specific list type data structure to store lists
 * of agents of specific types. The idea is that this list structure provides
 * a convenient way of storing lists of agents that we are registered with, or
 * a list of agents that we are waiting on etc. The data structure lets you store
 * an agent name or AgentGUID object with an associated type/id/string. You can then
 * search the structure for the name or identifier and receive specific objects 
 * or lists of objects back.
 * <br>
 * <br>
 */
public class AgentList
{
     
//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// STATIC VARIABLES
//----------------------------------------------------------------------
		
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

	// Internal store for the list stuff
	private TreeMap _store;
	
	
//----------------------------------------------------------------------
// CONSTRUCTOR
//----------------------------------------------------------------------

	/**
	 * Constructs an empty agent list. Use the add, get and remove methods to access, set and retrieve
	 * data in the list.
	 **/
	public AgentList()
	{
		
		_store = new TreeMap( fipaos.util.Comparators.STRING_COMPARE );	
		
	}
	
	
//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------
	
//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------	
	
	/**
	 * Adds a new agent to the list. The name parameter is the agent name. This can be of type
	 * String or AgentID. If an object of a different type is passed into this method then it will
	 * return false.
	 *
	 * @param name A String or AgentGUID representing the agent to be stored in the list
	 * @param type A String that can hold whatever data you like. For example, it could be the type
	 *             of an agent (df, ams, oa, arb etc) or it could be some other data.
	 * @return True if the name parameter is of the correct type, false otherwise
	 **/
	public boolean add( Object name, String type )
	{
		
		if ( name instanceof String )
		{
			
			_store.put( name, type );
			
		}
		else if ( name instanceof AgentID )
		{
			
			_store.put( ( (AgentID) name ).getName(), type );	
			
		}
		
		return false;
		
	}
	
// -------------------------------------------------------------------------	
	
	/**
	 * Removes an agent from the list. As with add the name parameter can be of type String or AgentGUID.
	 *
	 * @param name The String or AgentGUID object to remove.
	 **/
	public void remove( Object name )
	{
		
		if ( name instanceof String )
		{
			
			_store.remove( name );	
			
		}
		else if ( name instanceof AgentID )
		{
			
			_store.remove( ( (AgentID) name ).getName() );	
			
		}
		
	}
	
// -------------------------------------------------------------------------	
	
	/**
	 * Removes all agents with the given type/id parameter. Any agents that have been stored with their 'type'
	 * set to the given parameter will be removed from the list.
	 *
	 * @param type The type/id/string of the agent names to remove
	 **/
	public void removeAll( String type )
	{
		
		Iterator keys = _store.keySet().iterator();
		Object key = null;
		
		while ( keys.hasNext() )
		{
			
			key = keys.next();
			
			if ( ( (String)_store.get( key ) ).compareTo( type ) == 0 )
			{
				
				_store.remove( key );
				
			}
						
		}
		
	}
	
// -------------------------------------------------------------------------	
	
	/**
	 * Removes every element in the list.
	 **/
	public void removeAll()
	{
		
		clear();
		
	}
	
// -------------------------------------------------------------------------	
	
	/**
	 * Removes every element in the list.
	 **/
	public void clear()
	{
		
		_store.clear();
		
	}
	
// -------------------------------------------------------------------------
	
	/**
	 * Indicates whether this list contains the agent name specified.
	 *
	 * @param name The agent name to look for in the list (as a String or AgentID)
	 * @return True if the name is in the list, false otherwise
	 **/
	public boolean contains( Object name )
	{
		
		if ( name instanceof String )
		{
			
			return _store.containsKey( name.toString() );	
			
		}
		else if ( name instanceof AgentID )
		{
			
			return _store.containsKey( ( (AgentID) name ).getName() );	
			
		}
		
		return false;
		
	}
	
// -------------------------------------------------------------------------
	
	/**
	 * Indicates whether this list contains any elements added with an associated string that
	 * matches the type parameter.
	 *
	 * @param type The type/id/string to look for in the list
	 * @return True if the type/id/string is in the list, false otherwise
	 **/
	public boolean containsType( String type )
	{
		
		return _store.containsValue( type );
		
	}	
		
// -------------------------------------------------------------------------	
	
	/**
	 * Returns the type/id/string of the specified agent name. The name parameter can have type
	 * AgentGUID or String. If the name cannot be found in the list then null is returned.
	 *
	 * @param name The String or AgentGUID of the agent of whose type you wish to find
	 * @return The type/id/string associated with the specified agent name, or null if the name could not be found
	 **/
	public String getType( Object name )
	{
		
		if ( name instanceof String )
		{
			
			return (String)_store.get( name.toString() );	
			
		}
		else if ( name instanceof AgentID )
		{
			
			return (String) _store.get( ( (AgentID) name ).getName() );	
			
		}

		return null;
		
	}
	
// -------------------------------------------------------------------------	
	
	/**
	 * Returns every agent name with the associated type specified by the type parameter in an array of Strings.
	 *
	 * @param type The type/id/string associated with the names that you wish to extract
	 * @return An array of agent id's as String objects
	 **/
	public String[] getNames( String type )
	{
		
		TreeSet out = new TreeSet( fipaos.util.Comparators.STRING_COMPARE );
		Iterator keys = _store.keySet().iterator();
		Object key = null;
		
		while ( keys.hasNext() )
		{
			
			key = keys.next();
			
			if ( ( (String)_store.get( key ) ).compareTo( type ) == 0 )
			{
				
				out.add( key );
				
			}
						
		}	
		
		return (String[])out.toArray( new String[0] );
		
	}
	
// -------------------------------------------------------------------------	
	
	/**
	 * Returns every agent name with the associated type specified by the type parameter in an array of AgentGUIDs.
	 * The names are returned in the order in which they were added to the list.
	 *
	 * @param type The type/id/string associated with the names that you wish to extract
	 * @return An array of agent names as AgentGUID objects
	 **/
	public AgentID[] getAIDs( String type )
	{
		
		String names[] = getNames( type );
		AgentID ids[] = new AgentID[ names.length ];
		int max = names.length;
		
		for ( int i=0; i<max; i++ )
		{
			
			try
			{
				ids[ i ] = new AgentID( names[ i ] );
			}
			catch( fipaos.parser.ParserException e )
			{
			}
			
		}
		
		return ids;
		
	}
	
// -------------------------------------------------------------------------	
	
	/**
	 * Returns a string representation of the contents of the agent list.
	 * <br>The output is in the form:
	 * <p>
	 * <code>[ {agent-name, type/id/string}, ... ]</code>
	 * <p>e.g.<br>
	 * [ {agent@iiop://45.45.45.45:9000/acc, df}, {agent2@iiop://67.45.34.564:90/Agency/FIPA_Agent_97, ams} ]
	 * 
	 * @return The string representation of this list
	 **/
	public String toString()
	{
		
		Iterator keys = _store.keySet().iterator();
		Object key = null;
		boolean first = true;
		StringBuffer output = new StringBuffer( BASETOKENS.OPENSQUAREBRACKET );
		output.append( BASETOKENS.SPACE );
		
		while ( keys.hasNext() )
		{
			
			key = keys.next();
			
			if ( ! first )
			{
			
				output.append( BASETOKENS.COMMA );
				output.append( BASETOKENS.SPACE );
				
			}
			else
			{
				
				first = false;	
				
			}
			
			output.append( BASETOKENS.OPENBRACE );
			output.append( key.toString() );
			output.append( BASETOKENS.COMMA );
			output.append( BASETOKENS.SPACE );
			output.append( _store.get( key ) );
			output.append( BASETOKENS.CLOSEBRACE );
						
		}			
		
		output.append( BASETOKENS.SPACE );
		output.append( BASETOKENS.CLOSESQUAREBRACKET );
		
		return output.toString();
		
	}
	
	
//----------------------------------------------------------------------
// MAIN/TESTING METHOD
//----------------------------------------------------------------------
  
	/*public static void main( String args[] )
	{
		
		AgentList al = new AgentList();
		
		al.add( "agent1@iiop://45.45.45.45:9000/acc", "df" );
		al.add( "agent2@iiop://45.45.45.45:9000/acc", "df" );
		al.add( "agent3@iiop://45.45.45.45:9000/acc", "df" );
		al.add( "agent4@iiop://45.45.45.45:9000/acc", "ams" );
		al.add( "agent5@iiop://45.45.45.45:9000/acc", "ams" );
		al.add( "agent6@iiop://45.45.45.45:9000/acc", "ams" );
		al.add( "agent7@iiop://45.45.45.45:9000/acc", "oa" );
		al.add( "agent8@iiop://45.45.45.45:9000/acc", "oa" );
		al.add( "agent9@iiop://45.45.45.45:9000/acc", "oa" );
		al.add( "agent10@iiop://45.45.45.45:9000/acc", "arb" );
		
		String names[] = al.getNames( "df" );
		
		for ( int i=0; i<names.length; i++ )
		{
			
			System.out.println( names[ i ] );	
			
		}
		
		AgentGUID guids[] = al.getGUIDs( "ams" );
		
		for ( int j=0; j<guids.length; j++ )
		{
			
			System.out.println( guids[ j ].toString() );	
			
		}
			
		System.out.println( al.toString() );	
		
	}*/
	
	
} // End AgentList

	
