//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Emorphia are Copyright (C) 2001 Emorphia Limited
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.agent.conversation;

//----------------------------------------------------------------------

import java.util.List;
import java.util.Map;
import java.util.Iterator;

import fipaos.ont.fipa.ACL;
import fipaos.ont.fipa.fipaman.AgentID;
import fipaos.ont.fipa.fipaman.Envelope;
import fipaos.parser.acl.ACLMessage;

import fipaos.agent.conversation.UnknownProtocolException;

//----------------------------------------------------------------------

	/**
	 * <p>This interface represents the notion of a Conversation which is comprised of
	 * several Message's, the order of which follow a certain protocol
	 * (e.g. FIPA-Request).  The methods it provides allow retrieval of information
	 * about the conversation.
	 */
public interface Conversation extends Cloneable, java.io.Serializable, fipaos.skill.db.DatabaseObject
{
	//----------------------------------------------------------------------
	// CONSTANTS
	//----------------------------------------------------------------------
	
	/**
	 * Conversation has ended - This state is entered once the last message in
	 * the conversation has been received, or the last message recieved has timed-out.
	 * In the former case, the last message could either have 
	 * a performative that was defined in the protocol as an end-point, or
	 * a "not-understood" was received, indicating the implicit end of the conversation
	 */
	int CONVERSATION_END = -1;
	
	
	/**
	 * Conversation requires no Agent interaction - This is generally the state
	 * whenever the last message that was part of a conversation was sent by
	 * the Agent that owns this Conversation object, or a conversation update
	 * has already been forwarded to the ConversationListener for the latest
	 * update
	 */
	int NO_AGENT_ACTION_REQ = -2;
	
	
	/**
	 * Conversation requires Agent interaction - This is generally the state
	 * whenever the latest message in the conversation is from an Agent other
	 * than the owner of the Conversation object, and the conversation update
	 * hasn't been propogated to a ConversationListener
	 */
	int AGENT_ACTION_REQ = -3;
	
	
	/**
	 * Wait for response before ending conversation - Indicates that a reply-by
	 * time was specified on the last message in the conversation, so this
	 * is a special case of NO_AGENT_ACTION_REQ.  Generally if the timeout expires
	 * the Conversation will move into the CONVERSATION_END state
	 */
	int WAIT_BEFORE_END = -4;

	//----------------------------------------------------------------------
	// PUBLIC METHODS
	//----------------------------------------------------------------------
	
	
	/** Pass back the index of the latest message in the conversation
	 * @return Index of the latest message, or -1 if there are no messages
	 *        in this conversation
	 */
	int getLatestMessageIndex();
	
	
	/**
	 * Pass-back the latest message in this conversation
	 */
	ACL getACL( int index );
	
	
	/**
	 * Get the Envelope of the latest message received
	 */
	Envelope getEnvelope( int index );
	
	
	
	/**
	 * Gets the sender of the message at the specified index
	 *
	 * @param index The index of the message to retrieve
	 *
	 * @return The name of the sender of the message at the given index, or
	 *         null if no such message exists
	 */
	String getSender(int index);
	
	
	/**
	 * Gets the sender of the message at the specified index
	 *
	 * @param index The index of the message to retrieve
	 *
	 * @return The id of the sender of the message at the given index, or
	 *         null if no such message exists
	 */
	AgentID getSenderAID(int index);
	
	
	/**
	 * Method to retrieve the time stamp of a given message
	 *
	 * @param index The index of the message for which to retrieve the timestamp
	 *
	 * @return The time the message was added in milliseconds since 1/1/1970 0:00 GMT,
	 *         or -1 if no such message exists
	 */
	long getTimeStamp( int index );
	
	
	/**
	 * Method to get the time at which the last message will time out
	 *
	 * @return The time the message will time out in milliseconds since 1/1/1970 0:00 GMT,
	 *         or -1 if there is no timeout associated with the last message
	 */
	long getTimeOut();
	
	
	/**
	 * Method to discover what performatives are allowed next in the protocol.
	 *
	 * @return A list of performatives, or an empty list if no messages can be sent next
	 **/
	List getNextPerformatives();
	
	
	/**
	 * Method to discover what types of messages can be sent next
	 *
	 * @return A list of possible ACLMessage's which could be sent next, or
	 *         return an empty list if none can
	 */
	List getNextMessages();
	
	
	
	/** Method to produce a filled in ACL - e.g. sender, conversation ID and
	 * protocol fields will be set appropriately for the conversation
	 * @return The filled in ACL object
	 */
	ACL getFilledInACL();
	
	
	/**
	 * Method to get the protocol this Conversation is following
	 */
	String getProtocol();
	
	/**
	 * Gets the conversation ID of the conversation
	 *
	 * @return The String representation of the Conversation's
	 *         ID, or null if one has not been assigned
	 */
	String getConversationID();
	
	
	/**
	 * Method to retrieve the state of the conversation (e.g. does the conversation
	 * require agent input)
	 *
	 * @return One of NO_AGENT_ACTION_REQ, AGENT_ACTION_REQ or CONVERSATION_END
	 */
	int getState();
	
} // Conversation
