//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Emorphia are Copyright (C) 2001 Emorphia Limited
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.agent.conversation.lightweight;

import java.util.List;
import java.util.LinkedList;
import java.util.Collections;
//import java.util.Map;
//import java.util.HashMap;
import java.util.Iterator;

import fipaos.ont.fipa.ACL;
import fipaos.ont.fipa.fipaman.AgentID;
import fipaos.ont.fipa.fipaman.Envelope;
import fipaos.parser.acl.ACLMessage;
import fipaos.mts.Message;
import fipaos.util.*;

import fipaos.ont.fipa.FIPACONSTANTS;
import fipaos.ont.BASETOKENS;

import java.io.Serializable;
/*
 import fipaos.skill.db.DatabaseObject;
import fipaos.parser.ParserException;
 */
import fipaos.agent.conversation.Conversation;
import fipaos.agent.conversation.UnknownProtocolException;
import fipaos.agent.conversation.ConversationManager;
import fipaos.agent.conversation.ProtocolErrorException;
import fipaos.agent.conversation.InvalidConversationIDException;
import fipaos.agent.conversation.ConversationFinishedException;

/**
 * This class is an lighter alternative implementation to the heavy
 * ConversationImpl. Functionality is also more limited, protocols are not
 * enforced, although there is intention for doing this in the future. All use that
 * depends on the protocol enforcing missing will be working as a side effect only.
 *
 * @todo protocol enforcement
 */
public class LightweightConversationImpl implements Conversation
{
	
	/**
	 * Agent-id of the Agent to whom this belongs
	 */
	protected AgentID _agent_id;
	
	/**
	 * Agent-name of the Agent to whom this Conversation object belongs
	 */
	protected String _agent_name;
	
	/**
	 * Conversation ID of this Conversation
	 */
	private String _conv_id;
	
	/**
	 * Contains the list of MessageDetails
	 */
	private List _messages;
	
	/**
	 * Other Agent in the conversations'
	 */
	private AgentID[] _participants;
	
	/**
	 * Name of the protocol this conversation is following
	 */
	protected String _protocol;
	
	/**
	 * Helps to indicate if this is a deserialised object
	 */
	private transient Object _serialised = new Object();
	
	/**
	 * Indicates if this conversation has been updated since the last time it was
	 * examined by the planner scheduler
	 */
	private boolean _updated;
	
	/**
	 * Timeout for WAIT_BEFORE_END
	 */
	private long _wait_before_end_timeout;
	
	/**
	 * Conversation requires Agent interaction
	 */
	public static final int AGENT_ACTION_REQ = -3;
	
	/**
	 * Conversation has ended
	 */
	public static final int CONVERSATION_END = -1;
	
	/**
	 * Conversation requires no Agent interaction
	 */
	public static final int NO_AGENT_ACTION_REQ = -2;
	
	/**
	 * Wait for response before ending conversation
	 */
	public static final int WAIT_BEFORE_END = -4;
	
/**
 * Shows the current state of the object - either ACTIVE or NOT_UNDESTOOD.
 */
	private int _current_state =  ACTIVE;
	
/**
 * Possible state for _current_state. The agent is in this state when the
 * conversation hasn't reached a NOT_UNDESTOOD.
 */
	private static final int ACTIVE =  0;
	
/**
 * Possible state for _current_state. This state is reached when either agent sends
 * or receives a not-understood message.
 */
	private static final int NOT_UNDERSTOOD =  1;
	
/**
 * This consctructor creates a new LightweightConversationImpl object.
 */
	public LightweightConversationImpl()
	{
		_messages = Collections.synchronizedList(new LinkedList());
		_participants = new AgentID[10];
	}
	
	public LightweightConversationImpl(final String protocol,final AgentID agent_id)
	{
		_agent_id = agent_id;
		_agent_name = agent_id.getName();
		_protocol = protocol;
		_messages = Collections.synchronizedList(new LinkedList());
		_participants = new AgentID[10];
		
	}
	
	/**
	 * Pass back the index of the latest message in the conversation.
	 *
	 * @return Index of the latest message, or -1 if there are no messages in this conversation
	 */
	public int getLatestMessageIndex()
	{
		return ( _messages.size()-1 );
	}
	
	/**
	 * Pass-back the latest message in this conversation.
	 *
	 * @param index index of the message in the message list
	 * @return the ACL message of the given index
	 */
	public fipaos.ont.fipa.ACL getACL(int index)
	{
		ACL acl = null;
		
		// Attempt to get the message, but catch an exception if the index is out of range
		try
		{
			// Get the Message object that this is refering to
			Message m = ( (MessageDetails) _messages.get( index ) )._msg;
			
			// Get the ACL object contained within the Message
			acl = m.getACL();
		}
		catch ( Throwable t )
		{
		}
		
		return acl;
	}
	
	/**
	 * Get the Envelope of a message received.
	 *
	 * @param index index of the message the envelope belongs to
	 * @return the envelope of the message
	 */
	public fipaos.ont.fipa.fipaman.Envelope getEnvelope(int index)
	{
		Envelope env = null;
		
		// Attempt to get the message, but catch an exception if the index is out of range
		try
		{
			// Get the Message object that this is refering to
			Message m = ( (MessageDetails) _messages.get( index ) )._msg;
			
			// Get the ACL object contained within the Message
			env = m.getEnvelope();
		}
		catch ( Throwable t )
		{
		}
		
		return env;
	}
	
	/**
	 * Gets the sender of the message at the specified index.
	 *
	 * @param index The index of the message to retrieve
	 * @return The name of the sender of the message at the given index, or null if no such message exists
	 */
	public java.lang.String getSender(int index)
	{
		String sender = null;
		
		// Attempt to get the sender, but catch an exception if the index is out of range
		try
		{
			int index2 = ( (MessageDetails) _messages.get( index ) )._sender_number;
			sender = _participants[ index2 ].toString();
		}
		catch ( Throwable t )
		{
		}
		
		return sender;
	}
	
/**
 * Returns the conversation id for this conversation.
 *
 * @return the current conversation id
 */
	public java.lang.String getConversationID()
	{
		return _conv_id;
	}
	
	/**
	 * Method to produce a filled in ACL - e.g. sender, conversation ID and
	 * protocol fields will be set appropriately for the conversation
	 *
	 * @return The filled in ACL object
	 */
	public fipaos.ont.fipa.ACL getFilledInACL()
	{
		ACL msg = new ACL();
		
		try
		{
			msg.setConversationID( new String( _conv_id ) );
			msg.setSenderAID( _agent_id );
			msg.setProtocol( new String( _protocol ) );
		}
		catch( Throwable t )
		{
		}
		
		return msg;
	}
	
	/** Method to produce a filled in ACLMessage - e.g. sender, conversation ID and
	 * protocol fields will be set appropriately for the conversation
	 *
	 * @return The filled in ACLMessage object
	 * @deprecated The ACL object should be used in favour of the ACLMessage object, hence the getFilledInACL() should be used in preference to this
	 */
	public fipaos.parser.acl.ACLMessage getFilledInMessage()
	{
		return new fipaos.parser.acl.ACLMessage();
	}
	
	/**
	 * Method to discover what types of messages can be sent next. This returns
	 * null since protocols are not forced.
	 *
	 * @return A list of possible ACLMessage's which could be sent next, or return an empty list if none can
	 */
	public java.util.List getNextMessages()
	{
		return null;
	}
	
	/**
	 * Method to discover what performatives are allowed next in the protocol.
	 * This returns null since protocols are not forced.
	 *
	 * @return A list of performatives, or an empty list if no messages can be sent next
	 **/
	public java.util.List getNextPerformatives()
	{
		return null;
	}
	
	/**
	 * Gets the conversation ID of the conversation.
	 *
	 * @return The String representation of the Conversation's ID, or null if one has not been assigned
	 */
	public java.lang.String getObjectID()
	{
		return _conv_id;
	}
	
	/**
	 * Method to get the protocol this Conversation is following.
	 *
	 * @return the protocol
	 */
	public java.lang.String getProtocol()
	{
		return _protocol;
	}
	
	/**
	 * Gets the sender of the message at the specified index
	 *
	 * @param index The index of the message to retrieve
	 * @return The id of the sender of the message at the given index, or null if no such message exists
	 */
	public fipaos.ont.fipa.fipaman.AgentID getSenderAID(int index)
	{
		AgentID aid = null;
		
		try
		{
			aid = new AgentID( getSender( index ) );
		}
		catch( Throwable t )
		{
			System.out.println( "Error when getting AID of sender of message");
		}
		
		return aid;
	}
	
	/**
	 * Method to retrieve the state of the conversation (e.g. does the conversation
	 * require agent input)
	 *
	 * @return One of NO_AGENT_ACTION_REQ, AGENT_ACTION_REQ or CONVERSATION_END
	 */
	public int getState()
	{
		// if the last send message is a not-understood
		// then the conversation is over
		if (_current_state == NOT_UNDERSTOOD)
		{
			return CONVERSATION_END;
		}
		else
		{
			// Check if we are waiting for a response before ending the conversation
			// ie. last message was sent by us
			if( getSender( _messages.size()-1 ).equals(_agent_name) )
			{
				if( getTimeOut() > System.currentTimeMillis() )
				{
					return CONVERSATION_END;
				}
				else
				{
					return NO_AGENT_ACTION_REQ;
				}
			}
			else
			{
				return AGENT_ACTION_REQ;
			}
		}
	}
	
	/**
	 * Method to get the time at which the last message will time out
	 *
	 * @return The time the message will time out in milliseconds since 1/1/1970 0:00 GMT, or -1 if there is no timeout associated with the last message
	 */
	public long getTimeOut()
	{
		long time = -1;
		
		// Attempt to get the message, but catch an exception if the index is out of range
		try
		{
			time = ( (MessageDetails) _messages.get( _messages.size() - 1 ) )._timeout;
		}
		catch ( Throwable e )
		{
		}
		
		return time;
	}
	
	/**
	 * Method to retrieve the time stamp of a given message
	 *
	 * @param index The index of the message for which to retrieve the timestamp
	 * @return The time the message was added in milliseconds since 1/1/1970 0:00 GMT, or -1 if no such message exists
	 */
	public long getTimeStamp(int index)
	{
		long time = -1;
		
		// Attempt to get the message, but catch an exception if the index is out of range
		try
		{
			time = ( (MessageDetails) _messages.get( index ) )._timestamp;
		}
		catch ( ArrayIndexOutOfBoundsException e )
		{
		}
		
		return time;
	}
	
/**
 * This method is for the ConversationManager to add messages into the
 * Conversation.
 *
 * @param message message to be added
 * @throws InvalidConversationIDException if the message has a conversation id not belonging to this conversation
 */
	void addMessage(Message message) throws InvalidConversationIDException
	{
		// Check that the ACL Message has a conversation ID and that
		// it matches this conversation's ID
		String msg_conv_id = message.getACL().getConversationID();
		
		if( msg_conv_id == null || !msg_conv_id.equals( _conv_id ) )
		{
			throw new InvalidConversationIDException( "ConversationID is invalid or not specified" );
		}
		
		// Copy the Message in order to decouple versions that are flying about
		message = (Message) message.clone();
		ACL mess = message.getACL();
		
		// Get the performative
		String performative = mess.getPerformative();
		
		
		// Check if the message is a not-understood
		if( performative.equals( FIPACONSTANTS.NOT_UNDERSTOOD ))
		{
			// Check if we are already in the not-understood state
			if( _current_state == NOT_UNDERSTOOD )
			{
				return;
			}
			else
			{
				int index = findParticipant( mess.getSenderAID() );
				
				// If message is not from a known participant, just ignore it
				if( index == -1 && !_agent_name.equals( mess.getSenderAID().getName() ) )
				{
					return;
				}
				
				// Add the message to the message list
				addMessageDetails( message );
				
				_current_state = NOT_UNDERSTOOD;
				_updated = true;
			}
		}
		else
		{
			//since the message is not not-understood and we're not forcing protocols
			//just add the message
			
			// Add the sender to the list of participants if not
			if( findParticipant( mess.getSenderAID() ) == -1 )
			{
				// Add the Agent name to the list of participants
				for( int i=0; i<_participants.length; i++ )
				{
					if( _participants[ i ] == null )
					{
						_participants[ i ] = mess.getSenderAID();
						break;
					}
				}
			}
			
			// Add the receiver to the list of participants if not
			AgentID recv_name = (AgentID) mess.getReceiverAIDs().get(0);
			if( findParticipant( recv_name ) == -1 )
			{
				// Add the Agent name to the list of participants
				for( int i=0; i<_participants.length; i++ )
				{
					if( _participants[ i ] == null )
					{
						_participants[ i ] = recv_name;
						break;
					}
				}
			}
			
			// Add the message to the list of messages
			addMessageDetails( message );
			
			// Set the updated flag
			_updated = true;
		}
	}
	
	/**
	 * Finds the given particpants number
	 *
	 * @param name The AID of the participant to lookup
	 * @return The participant's number, or -1 if they were not found
	 */
	private int findParticipant(AgentID name)
	{
		// Find the participants index, if it exists
		int index=-1;
		for( int i=0; i<_participants.length; i++ )
		{
			// Check the current index
			if( _participants[ i ] != null && name.getName().equals( _participants[ i ].getName() ) )
			{
				// Store the index, and quit the for loop
				index = i;
				break;
			}
		}
		
		return index;
	}
	
/**
 * Private method for adding additional message details to a message. These details
 * are only visible to the conversation itself.
 *
 * @param message the message to be processed
 */
	private void addMessageDetails(Message message)
	{
		// Create a MessageDetails object to store the message in
		MessageDetails details = new MessageDetails();
		
		// Find the participant number
		int part_num = findParticipant( message.getACL().getSenderAID() );
		
		// Fill in the relevant fields
		details._msg = message;
		details._sender_number = part_num;
		details._timestamp = System.currentTimeMillis();
		details._timeout = -1;
		
		// Get the timeout
		if( message.getACL().getReplyByUTC() != null )
		{
			// Get the reply-by value
			UTCContainer reply_by = message.getACL().getReplyByUTC();
			
			// Determine if its a UTCTime or UTCDuration, and take appropriate measures
			if( reply_by instanceof UTCTime )
			{
				long current_time = (new UTCTime() ).getZTime().getTimeInMillis();
				long duration = ( (UTCTime) reply_by ).getZTime().getTimeInMillis() - current_time;
				details._timeout = details._timestamp + duration;
			}
			else if( reply_by instanceof UTCDuration )
			{
				details._timeout = details._timestamp +
				((UTCDuration) reply_by).getDurationInMillis();
			}
		}
		
		// Add the message to the message list
		_messages.add( details );
	}
	
	/**
	 * Sets the conversation ID of the Conversation - should only be
	 * called directly after creation of a Conversation object from
	 * the ConversationManagerImpl
	 *
	 * @param conv_id The Conversation ID that this Conversation object should use
	 */
	void setConversationID(String conv_id)
	{
		_conv_id = conv_id;
	}
	
	/**
	 * Checks if the conversation has been updated.
	 *
	 * @return True if the conversation has been updated
	 */
	synchronized boolean isUpdated()
	{
		return _updated;
	}
	
/**
 * Sets all references to null to aid garbage collection - ONLY USE WHEN
 * OBJECT HAS BEEN FINISHED WITH!
 */
	synchronized void removeReferences()
	{
		if( _messages != null )
		{
			_messages.clear();
			_messages = null;
		}
		
		if( _participants != null )
		{
			for( int i = 0; i < _participants.length; i++ )
			{
				_participants[ i ] = null;
			}
			_participants = null;
		}
		
		_agent_id = null;
		_agent_name = null;
		_conv_id = null;
		_protocol = null;
		
	}
	
/**
 * Resets the update flag to false.
 */
	synchronized void resetUpdate()
	{
		_updated = false;
	}
	
	/**
	 * Creates a copy of this Conversation object, and sets the update state
	 * to false
	 *
	 * @return A copy of the current Conversation, with update state set to false
	 */
	synchronized LightweightConversationImpl cloneAndResetUpdate()
	{
		LightweightConversationImpl conv;
		
		// Clone this
		conv = (LightweightConversationImpl) this.clone();
		
		// Change the state of the updated flag
		_updated = false;
		
		return conv;
	}
	
	/**
	 * Clone method to make duplicates of this Conversation object
	 *
	 * @return An independant copy of this Conversation object
	 *
	 * @exception CloneNotSupportedException Thrown if the object cannot be cloned
	 */
	public synchronized Object clone()
	{
		try
		{
			// Duplicate Conversation object
			LightweightConversationImpl conv = (LightweightConversationImpl) super.clone();
			
			// Clone parent AID
			conv._agent_id = (AgentID) _agent_id.clone();
			
			// Clone participants
			conv._participants = new AgentID[ _participants.length ];
			for( int i = 0; i < _participants.length; i++ )
			{
				if( _participants[ i ] != null )
				{
					conv._participants[ i ] = (AgentID) _participants[ i ].clone();
				}
			}
			
			// Clone messages
			if( _messages != null )
			{
				conv._messages = new LinkedList();
				Iterator i = _messages.iterator();
				while( i.hasNext() )
				{
					conv._messages.add( ( (MessageDetails) i.next() ).clone() );
				}
			}
			
			// Ensure serialisation state information is maintaied for copy
			conv._serialised = _serialised;
			
			return conv;
		}
		catch( CloneNotSupportedException e )
		{
			return null;
		}
	}
	
	/**
	 * Class to contain information about a message
	 */
	class MessageDetails implements Serializable, Cloneable
	{
/**
 * Message that this object is representing
 */
		public Message _msg;
		
/**
 * Participant number of sender
 */
		public int _sender_number;
		
/**
 * Local time of message receipt
 */
		public long _timestamp;
		
/**
 * Duration of the timeout for the last message added
 */
		public long _timeout = -1;
		
/**
 * Overrides default toString().
 *
 * @return stringified object
 */
		public String toString()
		{
			
			StringBuffer out = new StringBuffer();
			
			out.append( BASETOKENS.OPENBRACE );
			out.append( BASETOKENS.SPACE );
			out.append( _msg.toString() );
			out.append( BASETOKENS.COMMA );
			out.append( BASETOKENS.SPACE );
			out.append( _sender_number );
			out.append( BASETOKENS.COMMA );
			out.append( BASETOKENS.SPACE );
			out.append( _timestamp );
			out.append( BASETOKENS.COMMA );
			out.append( BASETOKENS.SPACE );
			out.append( _timeout );
			out.append( BASETOKENS.SPACE );
			out.append( BASETOKENS.CLOSEBRACE );
			
			return out.toString();
			
		}
		
/**
 * Overriding clone().
 *
 * @return cloned object
 */
		public Object clone()
		{
			try
			{
				MessageDetails md = (MessageDetails) super.clone();
				md._msg = (Message) _msg.clone();
				return md;
			}
			catch( CloneNotSupportedException e )
			{
				return null;
			}
		}
		
		
	}
	
}
