//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Emorphia are Copyright (C) 2001 Emorphia Limited
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.agent.conversation.lightweight;

import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.mts.*;
import fipaos.mts.service.*;
import fipaos.agent.*;
import fipaos.agent.conversation.ConversationListener;
import fipaos.agent.conversation.Conversation;
import fipaos.agent.conversation.ProtocolErrorException;
import fipaos.agent.conversation.ConversationFinishedException;
import fipaos.agent.conversation.ConversationManagerException;
import fipaos.agent.conversation.UnknownProtocolException;
import fipaos.agent.conversation.InvalidConversationIDException;
import fipaos.agent.profile.*;
import fipaos.skill.db.*;
import fipaos.util.*;
import fipaos.parser.acl.ACLMessage;
import fipaos.ont.BASETOKENS;

import java.util.*;
import java.awt.*;
import fipaos.agent.conversation.ConversationManager;


public class LightweightConversationManagerImpl implements ConversationManager, Runnable
{
	
	/**
	 * Map of active conversations
	 */
	private Map _active_list;
	
	/**
	 * Agent id
	 */
	private AgentID _agent_id;
	
	/**
	 * Agent name
	 */
	private String _agent_name;
	
	/**
	 * Reference to the object that Conversations should be handed onto.
	 * Typically this will be the TaskManager.
	 */
	private ConversationListener _conversation_listener;
	
	/**
	 * Counts the number of messages sent by this ConversationManagerImpl (ensures
	 * local uniqueness)
	 */
	private int _count = 0;
	
	/**
	 * Database in use to archive finished conversations
	 */
	private Database _database;
	
	/**
	 * Monitor running within this CM
	 */
	private Monitor _monitor;
	
	/**
	 * Reference to the owner of this
	 */
	private FIPAOSAgent _owner;
	
	/**
	 * Cach of protocols
	 */
	private Map _protocol_cache;
	
	/**
	 * Contains the mappings between protocol names and their implementing
	 * classes
	 */
	private static ProtocolProfile _protocol_profile;
	
	/**
	 * Reference to MessageSender used to send messages
	 */
	private MessageSender _sender;
	
	/**
	 * Indicates if the CM has terminated yet
	 */
	private boolean _terminate = false;
	
	/**
	 * Thread running within the CM
	 */
	private Thread _thread;
	
	/**
	 * Reference to the FIFOQueue containing updates to be dealt with
	 */
	private fipaos.util.Queue _updates;
	
	/**
	 * Create and initalise the CM.  This will invoked the setMessageSender()
	 * method on the ConversationListener passed as part of the initialisation
	 */
	public LightweightConversationManagerImpl(ConversationListener cl,MessageSender ms,FIPAOSAgent agent,DatabaseProfile db_profile,ProtocolProfile protocol_profile)
	{
		// Save references passed
		_owner = agent;
		_conversation_listener = cl;
		_agent_id = agent.getAID();
		_agent_name = agent.getAID().getName();
		_sender = ms;
		
		_protocol_profile = protocol_profile;
		
		// Create Map of active conversations.
		_active_list = new TreeMap( fipaos.util.Comparators.STRING_COMPARE );
		
		// Initialise the protocol cache
		_protocol_cache = new TreeMap( fipaos.util.Comparators.STRING_COMPARE );
		
		// Create a DB for the Agent's conversations
		String database_name = agent.getAID().getName() + "_convdb";
		_database = DatabaseFactory.getDatabase( database_name, db_profile );
		if( _database == null )
		{
			System.out.println("CM couldn't create a new database, " +	db_profile.getDatabaseType() + " couldn't be instantiated");
			return;
		}
		
		// Open the Database
		try
		{
			_database.openDatabase();
		}
		catch (DatabaseException e)
		{
			System.out.println( "CM error while creating opening the database");
		}
		
		// Start main handling thread
		_thread = new Thread(this, "CM-" + _agent_name );
		_thread.start();
		
		// Create monitor for updated queue
		_updates = new FIFOQueue();
		_monitor = new Monitor( _updates, new MonitorListener()
		{
			public void monitorUpdate( Object o )
			{
				Conversation conv = (Conversation) o;
				checkState( conv );
			}
		});
		
		// Indicate that the message sender should send messages via this
		_sender.setMessageReceiver( this );
		
		// Set this as the default MessageSender for the ConversationListener
		_conversation_listener.setMessageSender( this );
	}
	
	/**
	 * Method to create a new conversation given a protcol, and add it to the
	 * active list.
	 */
	public Conversation addConversation(String protocol) throws UnknownProtocolException
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return null;
		}
		
		synchronized( _active_list )
		{
			LightweightConversationImpl conv = createConversation( protocol, _agent_id );
			_active_list.put( conv.getConversationID(), conv );
			return (Conversation) conv.clone();
		}
	}
	
	/**
	 * Retrieves an active/previous Conversation that an Agent has had
	 */
	public Conversation getConversation(String conversation_id) throws InvalidConversationIDException
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return null;
		}
		
		Conversation conversation = null;
		
		synchronized( _active_list )
		{
			conversation = (Conversation) ((LightweightConversationImpl) _active_list.get(conversation_id) ).clone();
		}
		
		// Check if the conversation was found
		if( conversation == null )
		{
			try
			{
				// Open database for read-only transaction
				_database.startTransaction( true );
				
				conversation = (Conversation) _database.findObject( conversation_id );
			}
			catch( Throwable t )
			{
				System.out.println( "CM error when accessing Database");
			}
			finally
			{
				try
				{
					_database.endTransaction();
				}
				catch( Throwable t )
				{
				}
			}
		}
		
		// If conversation still hasn't been found, indicate that the conversation ID is unknown
		if( conversation == null )
		{
			throw new InvalidConversationIDException("Conversation " + conversation_id + " isn't active");
		}
		
		return conversation;
	}
	
	/**
	 * Removes the given conversation from the ConversationManager - this may be
	 * necessary in order to end conversations that have no defined end-point
	 *
	 * @param id The conversation-id of the conversation to remove
	 */
	public void removeConversation(String id) throws InvalidConversationIDException
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return;
		}
		
		LightweightConversationImpl conv = null;
		
		synchronized( _active_list )
		{
			// Remove the conversation
			conv = (LightweightConversationImpl) _active_list.remove(id);
		}
	}
	
	/**
	 * Adds a new protocol to the CM - this implementation will always return
	 * true since protocols are not forced.
	 *
	 * @param protocol_name The name of the protocol to store (e.g. fipa-new-protocol)
	 * @param protocol_defn The RDF definition of the protocol
	 * @return always returns true
	 */
	public boolean addProtocol(String protocol_name,String protocol_defn)
	{
		return true;
	}
	
	/**
	 * Indicates whether the specified protocol is already in the cache,
	 * i.e. has it already been learned. Since protocols are not enforced by
	 * this implementation, this method always returns true.
	 *
	 * @param protocol The protocol to check for
	 * @return always true
	 */
	public boolean hasProtocol(String protocol)
	{
		return true;
	}
	
	/**
	 * Attempt to send the message given
	 *
	 * @param message The message to be sent
	 */
	public void sendMessage(Message msg)
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return;
		}
		
		// Add time-stamp
		msg.getEnvelope().setSlotValue( Service.FIPAOS_DEBUG + "cm-send-time", new Long( System.currentTimeMillis() ) );
		
		try
		{
			// Attempt to send message
			if( _sender != null )
			{
				// See if we can deal with message
				handleMessage( msg );
				
				// Indicate a message was received
				System.out.println( "CM Sending message" );
				
				_sender.sendMessage( msg );
			}
			else
			{
				System.out.println( "No sender assigned to CM - can't send message!");
			}
		}
		catch( Throwable t )
		{
			// TODO: Propogate error back to Agent!
			System.out.println( "Exception occured when CM tried to send the message: " + t);
		}
	}
	
	/**
	 * Set the reference to the MessageSender that should send-outgoing messages
	 */
	public void setMessageSender(MessageSender ms)
	{
		// Save reference
		_sender = ms;
	}
	
	/**
	 * Set the reference to the MessageReceiver that should receive incoming messages
	 */
	public void setMessageReceiver(MessageReceiver mr)
	{
		// Ignore this, since the "next level" above this should be dealing
		// with Conversation objects via the ConversationListener interface
	}
	
	public void shutdown()
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return;
		}
		
		// Indicate we should terminate
		_terminate = true;
		
		// Kill threads
		_monitor.halt();
		_thread.interrupt();
		
		// Remove references ready for garbage collection
		_owner = null;
		_conversation_listener = null;
		_sender = null;
		_agent_id = null;
		_database = null;
		_monitor = null;
		_updates = null;
		_thread = null;
		_active_list.clear();
		_active_list = null;
		_protocol_cache.clear();
		_protocol_cache = null;
		
		System.out.println( "CM has shutdown()" );
	}
	
	public void receiveMessage(fipaos.mts.Message msg)
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return;
		}
		
		// Add time-stamp
		msg.getEnvelope().setSlotValue( Service.FIPAOS_DEBUG + "cm-receive-time",
		new Long( System.currentTimeMillis() ) );
		
		// Pass message to the conversation manager
		try
		{
			// Wait for the Agent to be ready before continuing
			_owner.waitForPushLock();
			
			// Indicate a message was received
			System.out.println( "Message received" );
			
			// See if we can deal with message
			handleMessage( msg );
		}
		catch ( ProtocolErrorException e )
		{
			System.out.println( "Exception occured, sending a not-understood");
			
			// Protocol error!
			_owner.sendNotUnderstood( msg.getACL() );
		}
		catch ( ConversationFinishedException e )
		{
			System.out.println( "Conversation has finished, ignoring message" );
			
			// Conversation has ended!
			_owner.sendNotUnderstood( msg.getACL() );
		}
		catch ( ConversationManagerException e )
		{
			System.out.println( "Unexpected ConversationMangerException");
		}
		catch ( Throwable e )
		{
			System.out.println( "Error, replying with not-understood" );
		}
	}
	
	/**
	 * When an object implementing interface <code>Runnable</code> is used
	 * to create a thread, starting the thread causes the object's
	 * <code>run</code> method to be called in that separately executing
	 * thread.
	 * <p>
	 * The general contract of the method <code>run</code> is that it may
	 * take any action whatsoever.
	 *
	 * @see     java.lang.Thread#run()
	 */
	public void run()
	{
		while(!_terminate)
		{
			try
			{
				_thread.sleep(5000);
				/*
				if the Thread sleep uninterrupted it will then cause the
				planner scheduler to check the state of conversations.
				 */
			}
			catch(InterruptedException e)
			{
			}
			
			checkState();
		}
	}
	
	/**
	 * Check the state of the given conversation, and forward if necessary
	 */
	private void checkState(Conversation next)
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return;
		}
		
		synchronized( _active_list )
		{
			synchronized( next )
			{
				long time_out = next.getTimeOut();
				
				// Get the sender of the message
				AgentID sender = next.getSenderAID( next.getLatestMessageIndex() );
				
				// Ensure ended conversations are dealt with appropriately &
				// then forward them
				switch( next.getState() )
				{
					case 	Conversation.CONVERSATION_END:
						try
						{
							removeConversation( next.getConversationID() );
						}
						catch (InvalidConversationIDException icie)
						{
							System.out.println("Couldn't remove conversation: " + icie);
							break;
						}
						
						default:
							if( !sender.equals( _agent_id ) )
							{
								_conversation_listener.notify( next );
							}
							break;
				}
			}
		}
	}
	
	/**
	 * Check the state of all active conversations, and place into queue to be
	 * dealt with if necessary
	 */
	private void checkState()
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return;
		}
		
		synchronized( _active_list )
		{
			// Attempt to produce an iterator for the _active_list
			Iterator conversations = _active_list.keySet().iterator();
			
			// Iterate through all conversations dealing with their state
			while(conversations.hasNext())
			{
				// Get the next Conversation ID
				String conversation_id = (String) conversations.next();
				
				// Get the associated Conversation object
				LightweightConversationImpl next = (LightweightConversationImpl)_active_list.get(conversation_id);
				
				synchronized( next )
				{
					long time_out = next.getTimeOut();
					
					if( !next.isUpdated() &&
					( ( time_out == -1 ) ||
					( time_out > System.currentTimeMillis() ) ) )
					{
						continue;
					}
					
					switch( next.getState() )
					{
						case 	Conversation.AGENT_ACTION_REQ:
						case 	Conversation.CONVERSATION_END:
							try
							{
								_updates.addItem( next.cloneAndResetUpdate() );
							}
							catch (NullPointerException npe)
							{
							}
							break;
							
						case 	Conversation.NO_AGENT_ACTION_REQ:
							next.resetUpdate();
							break;
							
							default:
								break;
					}
				}
			}
		}
	}
	
	/**
	 * This method creates a new conversation object
	 */
	private LightweightConversationImpl createConversation(String protocol,AgentID currentName) throws UnknownProtocolException
	{
		LightweightConversationImpl conversation = new LightweightConversationImpl(protocol, currentName);
		conversation.setConversationID(createConversationID());
		
		return conversation;
	}
	
	/**
	 * This method creates a new conversation id
	 */
	private synchronized String createConversationID()
	{
		//initialise local variables
		String conversationID = null;
		String stringOfTime = null;
		
		
		//intrement count instance variable by 1
		_count++;
		
		//build conversationID by concatenating agentguid to current time plus count
		long currentTime = System.currentTimeMillis();
		stringOfTime = "" + currentTime;
		
		//concat current_time
		conversationID = _agent_name + stringOfTime + _count;
		return conversationID;
	}
	
	/**
	 * Looks at a Message and checks to see if it part
	 * of an exisiting conversation. if yes adds it to the conversation object
	 * in the conversation list.  Otherwise, creates a new Conversation object
	 */
	public void handleMessage(Message msg) throws UnknownProtocolException, InvalidConversationIDException, ProtocolErrorException, ConversationFinishedException
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return;
		}
		
		ACL message = msg.getACL();
		
		//The case where the message has no id
		if (message.getConversationID() == null)
		{
			// Create conversation object give it an id
			LightweightConversationImpl conversation = createConversation(message.getProtocol(), _agent_id );
			message.setConversationID( conversation.getConversationID() );
			
			synchronized( _active_list )
			{
				// Add to _active_list
				_active_list.put( conversation.getConversationID(), conversation );
				
				synchronized( conversation )
				{
					// Add the message to it, and duplicate
					conversation.addMessage( msg );
					conversation = conversation.cloneAndResetUpdate();
				}
				
				// Deal with the conversation
				dealWithConversation( conversation );
			}
		}
		//the case where the message does have a conversation id
		else
		{
			// Retrieve the Conversation from the _active_list
			synchronized( _active_list )
			{
				LightweightConversationImpl conv = (LightweightConversationImpl) _active_list.get( message.getConversationID() );
				
				// Check if Conversation exists
				if( conv == null )
				{
					// Create and add to _active_list
					conv = createConversation( message.getProtocol(), _agent_name, message.getConversationID() );
					_active_list.put( conv.getConversationID(), conv );
				}
				
				synchronized( conv )
				{
					// Add the message to it, and duplicate
					conv.addMessage( msg );
					conv = conv.cloneAndResetUpdate();
				}
				
				// Deal with the conversation
				dealWithConversation( conv );
			}
		}
	}
	
/**
 * This method creates a new conversation object
 */
	private LightweightConversationImpl createConversation(String protocol,String currentName,String conv_id) throws UnknownProtocolException
	{
		// create conversation Object and give it an id.
		LightweightConversationImpl conversation = createConversation(protocol, _agent_id );
		conversation.setConversationID( conv_id );
		
		return conversation;
	}
	
	/**
	 * Dependant on the state of the conversation, puts the message into the
	 * Conversation queue
	 */
	private void dealWithConversation(LightweightConversationImpl conversation)
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return;
		}
		
		// If conversation requires attention of ConversationListener, add to PS queue
		if( conversation.getState() != Conversation.NO_AGENT_ACTION_REQ &&
		conversation.getState() != Conversation.WAIT_BEFORE_END )
		{
			_updates.addItem( conversation );
		}
	}
	
	public static void main(String[] args)
	{
	}
}
