//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved.
//
//******************************************************************************

package fipaos.agent.conversation.standard;

//----------------------------------------------------------------------

import java.util.List;
import java.util.LinkedList;
import java.util.Map;
import java.util.TreeMap;
import java.util.Iterator;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;

import fipaos.ont.fipa.ACL;
import fipaos.ont.fipa.fipaman.AgentID;
import fipaos.ont.fipa.fipaman.Envelope;
import fipaos.parser.acl.ACLMessage;
import fipaos.mts.Message;
import fipaos.util.*;

import fipaos.ont.fipa.FIPACONSTANTS;
import fipaos.ont.BASETOKENS;

import java.io.Serializable;
import fipaos.skill.db.DatabaseObject;
import fipaos.parser.ParserException;
import fipaos.agent.conversation.Conversation;
import fipaos.agent.conversation.UnknownProtocolException;
import fipaos.agent.conversation.ConversationManager;
import fipaos.agent.conversation.ProtocolErrorException;
import fipaos.agent.conversation.InvalidConversationIDException;
import fipaos.agent.conversation.ConversationFinishedException;

//----------------------------------------------------------------------

  /**
   * <p>This class represents the notion of a Conversation which is comprised of
   * several Message's, the order of which follow a certain protocol
   * (e.g. FIPA-Request).  The methods it provides allow retrieval of information
   * about the conversation, as well as validation of the protocol the conversation
   * is following.  This class cannot be instantiated directly, the static method
   * getConversation() should be used to obtain a Conversation instance which
   * will follow the required protocol.  Protocol definitions are defined in concrete
   * sub-classes, which map from protocol name to class name as such:</p>
   * <bl>
   * <li> "fipa-request" --> protocol.FIPARequest </li>
   * <li> "fipa-iterated-contract-net" --> protocol.FIPAIteratedContractNet </li>
   * </bl>
   * <p> The class name is relative to the package in which this object resides.  The
   * sub-class must also define a public static Object[] called __protocol which
   * contains details of the protocol to be followed.  The list should contain a list
   * of tuples which follow this format:</p>
   * <code>
   * { String performative [, Object conv_state], Integer participant [, Object[] sub_nodes] }<br>
	 * </code>
   * <p> The performative is the type of message recieved at that part of the protocol,
   * the conv_state is the conversation state when that performative is recieved (one of
   * NO_AGENT_ACTION_REQ or AGENT_ACTION_REQ, and if none is given CONVERSATION_END is assumed).  
   * The
   * participant number identifies which participant in the conversation should be sending
   * that message (the numbers used should be consistent throughout the messages according
   * to the protocol requirements).  The sub_nodes array can contain a list of the same
   * format to define what messages should be expected after the one defined by the current
   * list has been received (hence the whole protocol can be defined recursively).
   * <br>
   * <br>
   */
public abstract class ConversationImpl implements Cloneable, Serializable, DatabaseObject, Conversation
{
//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------

	/**
	 * Conversation has ended
	 */
	public final static int CONVERSATION_END = -1;


	/**
	 * Conversation requires no Agent interaction
	 */
	public final static int NO_AGENT_ACTION_REQ = -2;


	/**
	 * Conversation requires Agent interaction
	 */
	public final static int AGENT_ACTION_REQ = -3;

	
	/**
	 * Wait for response before ending conversation
	 */
	public final static int WAIT_BEFORE_END = -4;

//----------------------------------------------------------------------
// STATIC VARIABLES
//----------------------------------------------------------------------

	/**
	 * Not-understood generic node
	 */
	private static ConversationProtocol __not_understood;

	static
	{
		__not_understood = new ConversationProtocol();
		__not_understood._performative = FIPACONSTANTS.NOT_UNDERSTOOD;
		__not_understood._conv_state = CONVERSATION_END;
		__not_understood._following_nodes = new LinkedList();
	}		

//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

	/**
	 * Protocol that conversations of this type should follow (tree like
	 * structure for quick access)
	 */
	protected ConversationProtocol _protocol_tree;
	
	
	/**
	 * Current position in the protocol (null if conversation is empty)
	 */
	private ConversationProtocol _current_node;


	/**
	 * Helps to indicate if this is a deserialised object
	 */
	transient Object _serialised = new Object();
	
	
	/**
	 * Contains the list of MessageDetails
	 */
	private List _messages;
	
	
	/**
	 * Indicates if this conversation has been updated since the last time it was
	 * examined by the planner scheduler
	 */
	private boolean _updated;
	 

	/**
	 * Agent-id of the Agent to whom this belongs
	 */
	protected AgentID _agent_id;
	
	
	/**
	 * Other Agent in the conversations
	 */
	private AgentID _other_agent_aid;
	
	
	/**
	 * Is the Agent this belongs to acting as initiator (or participant)?
	 */
	private boolean _initiator = true;
	
	
	/**
	 * Indicates if the Conversation knows that the _initiator value has
	 * been decided (i.e. if no messages have been put into the Conversation,
	 * then we don't know if the Agent this Conversation is contained within is
	 * the participant or initiator)
	 */
	private boolean _initiator_known = false;

	
	/**
	 * Conversation ID of this Conversation
	 */
	private String _conv_id;
	

	/**
	 * Name of the protocol this conversation is following
	 */
	protected String _protocol;
	
	
	/**
	 * Timeout for WAIT_BEFORE_END
	 */
	private long _wait_before_end_timeout;	 	 
	
//----------------------------------------------------------------------
// CONSTRUCTOR
//----------------------------------------------------------------------

	/**
	 * Generalised constructor
	 */
	protected ConversationImpl()
	{
		// Initialise the participant/message List
		_messages = new LinkedList();
	}

//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------

	/**
	 * Method to dynamically load a Conversation instance which follows the
	 * given protocol and belongs to the given Agent
	 *
	 * @param protocol FIPA protocol name, null for generic NoProtocol conversation
	 *                 object
	 *
	 * @param agent_name AgentID of the Agent on the local machine who is involved
	 *                   in this conversation
	 *
	 * @return The appropriate instance of Conversation for the protocol given
	 *
	 * @exception UnknownProtocolException Thrown if a Conversation class cannot be 
	 *                                     loaded for the given class
	 */
	public static ConversationImpl getConversation( String protocol, String agent_name ) 
		throws UnknownProtocolException, ParserException
	{
		return getConversation( protocol, new AgentID( agent_name ) );
	}
	
	
	/**
	 * Method to dynamically load a Conversation instance which follows the
	 * given protocol and belongs to the given Agent
	 *
	 * @param protocol FIPA protocol name, null for generic NoProtocol conversation
	 *                 object
	 *
	 * @param agent_name AgentID of the Agent on the local machine who is involved
	 *                   in this conversation
	 *
	 * @return The appropriate instance of Conversation for the protocol given
	 *
	 * @exception UnknownProtocolException Thrown if a Conversation class cannot be 
	 *                                     loaded for the given class
	 */
	public static ConversationImpl getConversation(String protocol, AgentID agent_name) 
		throws UnknownProtocolException
	{		
		
		String class_name = ConversationManagerImpl.getProtocolClassName( protocol );
		
		try
		{
			// Attempt to dynamically load the class
			Class conv = Class.forName( class_name );
			Field p_field = conv.getField( "__protocol" );
			
			// Check that it is a sub-class of Conversation
			if( ! ConversationImpl.class.isAssignableFrom( conv ) )
			{
				throw new UnknownProtocolException( "Error loading class " + class_name + " - it doesn't " +
																						"extend ConversationImpl class" );
			}
			
			// Create the new instance, and set the ownership
			ConversationImpl instance = (ConversationImpl) conv.newInstance();
			instance._agent_id = agent_name;
			instance._protocol = protocol;
						
			// Get the protocol from the static variable __protocol
			instance._protocol_tree = setupProtocol( (Object[]) p_field.get( null ) );
			
			// Setup dummy conversation ID - NOT UNIQUE!
			instance._conv_id = System.currentTimeMillis() + "";
			
			return instance;
		}
		catch ( ClassNotFoundException e )
		{
			// Class wasn't found!
			throw new UnknownProtocolException( "Error creating Conversation class to deal with \"" +
																					protocol + "\", " + class_name + " was not found" );
		}
		catch ( InstantiationException e )
		{
			// Class couldn't be instantiated!
			throw new RuntimeException( "Error creating Conversation class to deal with \"" +
																					protocol + "\", " + class_name + " could not be " +
																					"instantiaited" );
		}
		catch ( IllegalAccessException e )
		{
			// Security manager won't let us create the class!
			throw new RuntimeException( "Error creating Conversation class to deal with \"" +
																					protocol + "\", SecurityManager prevented " +
																					class_name + " from being instantiated" );
		}
		catch ( NoSuchFieldException e )
		{
			// Sub-class doesn't contain a __protocol object
			throw new RuntimeException( "Error creating Conversation object from " + class_name + 
																	", it doesn't contain a __protocol definition" );
		}
	}

	
	/** Pass back the index of the latest message in the conversation
	 * @return Index of the latest message, or -1 if there are no messages
	 *        in this conversation
	 */
	public int getLatestMessageIndex()
	{
		return ( _messages.size()-1 );
	}


	/**
	 * Pass-back the latest message in this conversation
	 */
	public ACL getACL( int index )
	{
		ACL acl = null;

		// Attempt to get the message, but catch an exception if the index is out of range		
		try
		{
			// Get the Message object that this is refering to
			Message m = ( (MessageDetails) _messages.get( index ) )._msg;
			
			// Get the ACL object contained within the Message
			acl = m.getACL();
		}
		catch ( Throwable t )
		{
		}

		return acl;
	}
	
	
	/**
	 * Get the Envelope of the latest message received
	 */
	public Envelope getEnvelope( int index )
	{
		Envelope env = null;

		// Attempt to get the message, but catch an exception if the index is out of range		
		try
		{
			// Get the Message object that this is refering to
			Message m = ( (MessageDetails) _messages.get( index ) )._msg;
			
			// Get the ACL object contained within the Message
			env = m.getEnvelope();
		}
		catch ( Throwable t )
		{
		}

		return env;		
	}
	

	/** Method to retrieve the ACLMessage with the given index in the conversation
	 * (e.g. first message has an index of 0, second 1 etc...).  The messages are ordered
	 * based upon the order they are added to the Conversation.
	 * @param index The index of the message to retrieve
	 * @return The ACLMessage at the given index, or null if a message with that index
	 *        doesn't exist
	 * @deprecated The ACLMessage object is no longer supported - please use
	 * the getACL() and getEnvelope() methods instead
	 */
	public ACLMessage getMessage( int index )
	{
		ACLMessage msg = null;

		// Attempt to get the message, but catch an exception if the index is out of range		
		try
		{
			// Get the Message object that this is refering to
			Message m = ( (MessageDetails) _messages.get( index ) )._msg;
			
			// Get the ACL object contained within the Message
			ACL acl = m.getACL();
			
			// Create an ACLMessage object to wrap the ACL message
			msg = new ACLMessage( acl ); 
		}
		catch ( Throwable t )
		{
		}

		return msg;
	}
	

	/**
	 * Gets the sender of the message at the specified index
	 *
	 * @param index The index of the message to retrieve
	 *
	 * @return The name of the sender of the message at the given index, or
	 *         null if no such message exists
	 */
	public String getSender( int index )
	{
		String sender = null;

		// Attempt to get the sender, but catch an exception if the index is out of range		
		try
		{
			MessageDetails md = (MessageDetails) _messages.get( index );
			sender = md._msg.getACL().getSenderAID().toString();
		}
		catch ( Throwable t )
		{
		}

		return sender;
	}


	/**
	 * Gets the sender of the message at the specified index
	 *
	 * @param index The index of the message to retrieve
	 *
	 * @return The id of the sender of the message at the given index, or
	 *         null if no such message exists
	 */
	public AgentID getSenderAID( int index )
	{
		AgentID aid = null;
		
		try
		{
			MessageDetails md = (MessageDetails) _messages.get( index );
			aid = md._msg.getACL().getSenderAID();
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Error when getting AID of sender of message " + index +
														" in conversation " + _conv_id, t, this );
		}
		
		return aid;
	}

	
	/**
	 * Method to retrieve the time stamp of a given message
	 *
	 * @param index The index of the message for which to retrieve the timestamp
	 *
	 * @return The time the message was added in milliseconds since 1/1/1970 0:00 GMT,
	 *         or -1 if no such message exists
	 */
	public long getTimeStamp( int index )
	{
		long time = -1;

		// Attempt to get the message, but catch an exception if the index is out of range		
		try
		{
			time = ( (MessageDetails) _messages.get( index ) )._timestamp;
		}
		catch ( ArrayIndexOutOfBoundsException e )
		{
		}

		return time;
	}


	/**
	 * Method to get the time at which the last message will time out
	 *
	 * @return The time the message will time out in milliseconds since 1/1/1970 0:00 GMT,
	 *         or -1 if there is no timeout associated with the last message
	 */
	public long getTimeOut()
	{
		long time = -1;

		// Attempt to get the message, but catch an exception if the index is out of range		
		try
		{
			time = ( (MessageDetails) _messages.get( _messages.size() - 1 ) )._timeout;
		}
		catch ( Throwable e )
		{
		}

		return time;
	}
	
	
	/**
	 * Method to discover what performatives are allowed next in the protocol.
	 *
	 * @return A list of performatives, or an empty list if no messages can be sent next
	 **/
	public List getNextPerformatives()
	{
		
		List perfs = new LinkedList();		
		
		// Check if we're at the start of the conversation
		if( _current_node == null && _protocol_tree != null )
		{
			DIAGNOSTICS.println( "Protocol not started - returning first performative", this );
			
			// Just add starting protocol
			perfs.add( _protocol_tree._performative );
		}
		else if( _current_node != null )
		{
			// In the middle of a conversation
			
			// Loop around all of the possibilites
			Iterator nodes = _current_node._following_nodes.iterator();			
			while( nodes.hasNext() )
			{
				// Get the next element
				ConversationProtocol curr = (ConversationProtocol) nodes.next();
				
				// Check if initiator flags are equal (indicates that the performative
				// for this node should be sent by the Agent this Conversation belongs to) or
				// if the initiator is unknown
				if( curr._initiator == _initiator || !_initiator_known )
				{
					perfs.add( curr._performative );
				}
				
			DIAGNOSTICS.println( "Protocol started (" + _current_node._performative + 
														") - returning next performatives: " + perfs, this );						
			}				
		}
		else
		{
			DIAGNOSTICS.println( "Couldn't get next performatives!", this );
		}
		
		return perfs;
	}
	
	
	/**
	 * Method to discover what types of messages can be sent next
	 *
	 * @return A list of possible ACL's which could be sent next, or
	 *         return an empty list if none can
	 */
	public List getNextMessages()
	{
		// Create a list to store the results
		List msgs = new LinkedList();

		// Check if we're at the start of the conversation
		if( _current_node == null && _protocol_tree != null )
		{
			// Create a new message, fill in the details and add it to the list
			ACL msg = getFilledInACL();
			msg.setPerformative( _protocol_tree._performative );
			msgs.add( msg );
		}
		else if( _current_node != null )
		{
			// Loop around all of the possibilites
			Iterator nodes = _current_node._following_nodes.iterator();
			while( nodes.hasNext() )
			{
				// Get the next element
				ConversationProtocol curr = (ConversationProtocol) nodes.next();
				
				// Check if we are part of the participant list
				if( curr._initiator == _initiator || !_initiator_known )
				{
					// Create a new filled out ACL and add performative
					ACL msg = getFilledInACL();
					msg.setPerformative( curr._performative );
					msg.setReceiverAID( _other_agent_aid );

					// Add the message to the list
					msgs.add( msg );
				}
			}
		}			

		return msgs;
	}
	
	
	/** Method to produce a filled in ACL - e.g. sender, conversation ID and
	 * protocol fields will be set appropriately for the conversation
	 * @return The filled in ACL object
	 */
	public ACL getFilledInACL()
	{
		ACL msg = new ACL();
		
		try
		{
			msg.setConversationID( _conv_id );
			msg.setSenderAID( _agent_id );
			msg.setProtocol( _protocol );
		}
		catch( Throwable t )
		{
		}
		
		return msg;		
	}
	
	
	/**
	 * Method to get the protocol this Conversation is following
	 */
	public String getProtocol()
	{
		return _protocol;
	}
	
	/**
	 * Gets the conversation ID of the conversation
	 *
	 * @return The String representation of the Conversation's
	 *         ID, or null if one has not been assigned
	 */
	public String getConversationID()
	{
		return _conv_id;
	}


	/**
	 * Method to retrieve the state of the conversation (e.g. does the conversation
	 * require agent input)
	 *
	 * @return One of NO_AGENT_ACTION_REQ, AGENT_ACTION_REQ or CONVERSATION_END
	 */
	public synchronized int getState()
	{
		// Check that this Conversation hasn't been serialised (in which case the
		// Conversation is over!) - _serialised is a transient reference, so this will
		// only be non-null after it is initially created, until it is serialised
		if( _serialised == null )
		{
			DIAGNOSTICS.println( "In getState() (" + _conv_id + ") - _serialised == null" , this );
			return CONVERSATION_END;
		}
		
		// Check that we are in a conversation
		if( _current_node == null && _protocol_tree != null )
		{
			DIAGNOSTICS.println( "In getState() (" + _conv_id + ") - conv not started" , this );
			return NO_AGENT_ACTION_REQ;
		}

		// Check that we are in a conversation which is following a protocol
		if( _protocol_tree == null )
		{ 
			if(	_agent_id.equals( getSenderAID( getLatestMessageIndex() ) ) )
			{
				DIAGNOSTICS.println( "In getState() (" + _conv_id + ") - no-protocol conv in progress, no action req." , this );
				return NO_AGENT_ACTION_REQ;
			}
			else
			{
				DIAGNOSTICS.println( "In getState() (" + _conv_id + ") - no-protocol conv in progress, action req." , this );
				return AGENT_ACTION_REQ;
			}
		} 

		// Check if the end of the conversation has been reached
		if( _current_node._following_nodes.isEmpty() )
		{
			DIAGNOSTICS.println( "In getState() (" + _conv_id + ") - no more paths in ConversationProtocol tree at performative \""
														+ _current_node._performative + "\"" , this );
			return CONVERSATION_END;
		}

		// Check if the current node belongs to our user
		if( _initiator == _current_node._initiator && _initiator_known )
		{
			// Check if the following node also belongs to our use - if so they need to do something
			Iterator nodes = _current_node._following_nodes.iterator();
			while( nodes.hasNext() )
			{
				ConversationProtocol node = (ConversationProtocol) nodes.next();
				if( _initiator == node._initiator )
				{
					DIAGNOSTICS.println( "In getState() (" + _conv_id + ") - action req." , this );
					return AGENT_ACTION_REQ;
				}
			}
			
			DIAGNOSTICS.println( "In getState() (" + _conv_id + ") - no action req." , this );
			return NO_AGENT_ACTION_REQ;			
		}

		
		// Check if we are waiting for a response before ending the conversation
		if( _current_node._conv_state == WAIT_BEFORE_END )
		{
			if( _wait_before_end_timeout > System.currentTimeMillis() )
			{
				DIAGNOSTICS.println( "In getState() (" + _conv_id + ") - waiting for timeout, no action req." , this );
				return NO_AGENT_ACTION_REQ;
			}
			else
			{
				DIAGNOSTICS.println( "In getState() (" + _conv_id + ") - timeout reached" , this );
				return CONVERSATION_END;
			}
		}
		
		// Otherwise return the appropriate value
		DIAGNOSTICS.println( "In getState() (" + _conv_id + ") - just returning node state" , this );
		return _current_node._conv_state;
	}
	
	
	/**
	 * Method to get unique identifier of this DatabaseObject
	 */
	public String getObjectID()
	{
		return _conv_id;
	}
	
	
	/**
	 * Overrides default toString()
	 */
	public String toString()
	{
		
		StringBuffer out = new StringBuffer();
		
		synchronized( this )
		{
			
			out.append( "-----------------------------\n" );
			
			int max = _messages.size();
			
			for( int i=0; i<max; i++ )
			{
				
				out.append( _messages.get( i ) );
				out.append( "\n" );
				
			}
			
			out.append( "State: " );
			
			if( getState() == AGENT_ACTION_REQ )
			{
				
				out.append( "Agent action required\n" );
				
			}
			else if( getState() == NO_AGENT_ACTION_REQ )
			{
				
				out.append( "No action required\n" );
				
			}
			else if( getState() == CONVERSATION_END )
			{
				
				out.append( "End of conversation\n" );
				
			}
			
		}
					
		return out.toString();
		
	}
	

//----------------------------------------------------------------------
// PROTECTED METHODS
//----------------------------------------------------------------------
	
	/**
	 * Clone method to make duplicates of this Conversation object
	 *
	 * @return An independant copy of this Conversation object
	 *
	 * @exception CloneNotSupportedException Thrown if the object cannot be cloned
	 */
	public synchronized Object clone()
	{
		try
		{
			// Duplicate Conversation object
			ConversationImpl conv = (ConversationImpl) super.clone();

			// Clone parent AID
			conv._agent_id = (AgentID) _agent_id.clone();
			
			// Clone other agent AID
			if( _other_agent_aid != null )
			{
				conv._other_agent_aid = (AgentID) _other_agent_aid.clone();
			}
			
			// Clone converstion protocols
			cloneProtocols( conv );

			// Clone messages
			if( _messages != null )
			{
				conv._messages = new LinkedList();
				Iterator i = _messages.iterator();
				while( i.hasNext() )
				{
					conv._messages.add( ( (MessageDetails) i.next() ).clone() );
				}
			}
			
			// Ensure serialisation state information is maintaied for copy
			conv._serialised = _serialised;

			return conv;
		}
		catch( CloneNotSupportedException e )
		{
			return null;
		}
	}
		
	
	/**
	 * Method to clone ConversationProtocol objects into given ConversationImpl
	 */
	private void cloneProtocols( ConversationImpl c )
	{
		// Lists to contain ConversationProtocol objects
		
		// Contains "cloned" ConversationProtocol objects
		List originals = new LinkedList();
		// Contains (on a positional-basis) clones of the ConversationProtocol
		// objects in the originals list
		List clones = new LinkedList();
		
		// Start at the top of the protocol tree, and work downwards, tracking
		// all nodes that have been converted
		if( _protocol_tree != null )
		{
			c._protocol_tree = (ConversationProtocol) _protocol_tree.clone( originals, clones );
		}
		
		// Now find ConversationProtocol's matching other references
		if( _current_node != null )
		{
			if( _current_node == __not_understood )
			{
				c._current_node = __not_understood;
			}
			else
			{
				int index_curr = originals.indexOf( _current_node );
				c._current_node = (ConversationProtocol) clones.get( index_curr );
			}
		}
	}
	
	
//----------------------------------------------------------------------
// package METHODS
//----------------------------------------------------------------------

	/**
	 * Sets all references to null to aid garbage collection - ONLY USE WHEN
	 * OBJECT HAS BEEN FINISHED WITH!
	 */
	synchronized void removeReferences()
	{
		if( _messages != null )
		{
			_messages.clear();
			_messages = null;
		}
		
		_agent_id = null;
		_other_agent_aid = null;
		_conv_id = null;
		_protocol = null;
		
		if( _protocol_tree != null )
		{
			_protocol_tree.removeReferences();
			_protocol_tree = null;
		}
		_current_node = null;
	}
	
	/**
	 * Sets the conversation ID of the Conversation - should only be
	 * called directly after creation of a Conversation object from
	 * the ConversationManagerImpl
	 *
	 * @param conv_id The Conversation ID that this Conversation object should use
	 */
	void setConversationID( String conv_id )
	{
		_conv_id = conv_id;
	}
	
	
	/**
	 * Method to add a message to the conversation - sets conversation state
	 * to updated as well.
	 *
	 * @param mess The Message to be added
	 *
	 * @exception ProtocolErrorException Thrown when the message to be added
	 *                                doesn't follow the correct protocol
	 * @exception InvalidConversationIDException Thrown when the conversation ID
	 *         																of the ACLMessage is different
	 *																				to that of the conversation
	 * @exception ConversationFinishedException Thrown when the conversation has been finished
	 */
	public synchronized void addMessage( Message msg )
		throws ProtocolErrorException, InvalidConversationIDException, ConversationFinishedException
	{
		// Check that the ACL Message has a conversation ID and that
		// it matches this conversation's ID
		String msg_conv_id = msg.getACL().getConversationID();
		
		if( msg_conv_id == null || !msg_conv_id.equals( _conv_id ) )
		{
			throw new InvalidConversationIDException( "ConversationID is invalid or not specified" );	
		}

		// Copy the Message in order to decouple versions that are flying about
		msg = (Message) msg.clone();
		
		// Extract message parameters
		ACL mess = msg.getACL();
		String performative = mess.getPerformative();
		AgentID sender_aid = mess.getSenderAID();
		AgentID receiver_aid = mess.getReceiverAID();
		boolean sender_is_initiator = false;
		
		// Handle special case of multi-cast - this will only be the case if recieving a
		// message
		if( mess.getReceiverAIDs().size() > 1 )
		{
			DIAGNOSTICS.println( "WARNING: Recieved multicast message - recipients = " +
						mess.getReceiverAIDs(), this );
			receiver_aid = (AgentID) _agent_id;
		}
		else
		{
			DIAGNOSTICS.println( "WARNING: Recieved non-multicast message!", this );
			receiver_aid = (AgentID) mess.getReceiverAIDs().get(0);
		}

		
		// Is this the first message?
		if( !_initiator_known )
		{
			// Extract ACL from Message
			ACL acl = msg.getACL();
			
			// Check the sender - if it is our Agent, then it is the initiator
			_initiator = acl.getSenderAID().equals( _agent_id );
			
			// Indicate that the initiator is now known
			_initiator_known = true;
			
			// Store the AID of the other Agent in the conversation
			if( _initiator )
			{
				_other_agent_aid = acl.getReceiverAID();
			}
			else
			{
				_other_agent_aid = acl.getSenderAID();
			}
		}
		
		
		// Determine if the sender is the initiator of the conversation
		
		// Sender is initiator if sender is the Agent this belongs to & it is the initiator...
		sender_is_initiator = ( _initiator && sender_aid.equals( _agent_id ) );
		// or sender is not the Agent this belongs to and it isn't the initiator
		sender_is_initiator |= ( !_initiator && !sender_aid.equals( _agent_id ) );
		
		
		//
		// Now deal with message protocol
		//
		
		
		// Check if the message is a not-understood
		if( performative.equalsIgnoreCase( FIPACONSTANTS.NOT_UNDERSTOOD ))
		{
			// Check if we are already in the not-understood state
			if( _current_node == __not_understood )
			{
				return;
			}
			
			// If message is not from a known participant, just ignore it
			if( !sender_aid.equals( _agent_id ) || 
					!(_other_agent_aid != null && sender_aid.equals( _other_agent_aid ) ) )
			{
				return;
			}

			// Add the message to the message list
			addMessageDetails( msg );

			_current_node = __not_understood;
			_updated = true;
			
			return;
		}


		// Check if a protocol has been specified - if not just add the message
		if( _protocol_tree == null )
		{
			// Add the message to the list of messages
			addMessageDetails( msg );
			_updated = true;
			return;
		}

		
		// Check if the conversation has been started
		if( _current_node == null )
		{
			// Conversation not started - just add message
			
			// Check if the performative matches
			if( _protocol_tree._performative.equalsIgnoreCase( performative ) )
			{
				// Performative matches that expected at the start of conversation
				
				// Set the current node to be the start node
				_current_node = _protocol_tree;
				
				// Add message
				addMessageDetails( msg );
			}
			else
			{
				// Performative doesn't match that expected at the start of conversation
				
				// Add the message for record purposes
				if( sender_aid.equals( _agent_id ) )
				{
					addMessageDetails( msg );
				}
				
				// Throw an exception because this message doesn't belong as part of this
				// protocol!
				throw new ProtocolErrorException( "Unexpected message received of type " + performative );
			}
		}
		else
		{
			// Conversation already started - check its OK to add message
			
			// Check if the end of the conversation has been reached
			if( _current_node._following_nodes.isEmpty() )
			{
				throw new ConversationFinishedException( "Message received in conversation which has already ended!" );
			}
				
			// Setup a list to contain possible nodes to move to
			List nodes = new LinkedList();
			
			// Find all of the branches from the current node in the protocol tree
			// which match
			Iterator sub_nodes = _current_node._following_nodes.iterator();
			while( sub_nodes.hasNext() )
			{
				// Get the next node to check
				ConversationProtocol node = (ConversationProtocol) sub_nodes.next();
				
				// If its performative matches, add it to the list
				if( node._performative.equalsIgnoreCase( performative ) )
				{
					nodes.add( node );	
				}
			}
			
			
			// Check if the list is empty
			if( nodes.isEmpty() )
			{
				// Add the message to the message list
				if( sender_aid.equals( _agent_id ) )
				{
					addMessageDetails( msg );
				}

				// Throw an appropriate exception
				throw new ProtocolErrorException( "Unexpected message received of type " + performative );
			}
						

			// Iterate the possible nodes, and checking if the sender is allowed to send
			// that message at this point
			ConversationProtocol match = null;
			sub_nodes = nodes.iterator();
			while( sub_nodes.hasNext() && match == null )
			{
				// Get the next node to check
				ConversationProtocol node = (ConversationProtocol) sub_nodes.next();
				
				// Check that the sender of the message matches the expected sender of the message
				if( sender_is_initiator == node._initiator )
				{
					match = node;
				}
			}

			// Check if a match was made
			if( match == null )
			{
				// Add the message
				if( !sender_aid.equals( _agent_id ) )
				{
					addMessageDetails( msg );
				}
				else
				{
					// No match! Throw an exception.
					throw new ProtocolErrorException( "Couldn't match expected message types and participant" );
				}
			}
			
			// Set the current node to be the node that matches our message
			_current_node = match;
		}

		// Check the state of the node - if it is a WAIT_BEFORE_END, add a timeout
		// to the current message
		if( _current_node._conv_state == WAIT_BEFORE_END )
		{
			if( _current_node._initiator == _initiator )
			{
				mess.setReplyByUTC( ( new UTCDuration( 0, 0, 0, 0, 0, 10, 0 ) ) );
				_wait_before_end_timeout = System.currentTimeMillis() + 10000;
			}
			else
			{
				_wait_before_end_timeout = 0;
			}
		}
		
		// Set the updated flag
		_updated = true;

		// Add the msg to the message list
		addMessageDetails( msg );
	}
	
		
	/**
	 * Checks if the conversation has been updated.
	 *
	 * @return True if the conversation has been updated
	 */
	synchronized boolean isUpdated()
	{
		return _updated;
	}
	

	/**
	 * Resets the update flag to false.
	 */
	synchronized void resetUpdate()
	{
		_updated = false;
	}

	
	/**
	 * Creates a copy of this Conversation object, and sets the update state
	 * to false
	 *
	 * @return A copy of the current Conversation, with update state set to false
	 */
	synchronized ConversationImpl cloneAndResetUpdate()
	{
		ConversationImpl conv;
		
		// Clone this
		conv = (ConversationImpl) this.clone();
		
		// Change the state of the updated flag
		_updated = false;
		
		return conv;
	}
	
//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------

	/**
	 * Method to setup the _protocol_tree
	 *
	 * @param protocol An object[] with details of the tree
	 */
	private static ConversationProtocol setupProtocol( Object[] protocol )
	{
		List p_tree = null;
	
		try
		{
			// Create the protocol tree
			p_tree = getProtocolTree( protocol, new TreeMap( fipaos.util.Comparators.OBJECT_COMPARE ) );
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}

		// If the protocol tree couldn't be constructed, thrown an exception
		if( p_tree == null || p_tree.isEmpty() )
		{
			return null;
		}
		else
		{
			return (ConversationProtocol) p_tree.get(0);
		}
	}
	
	
	/**
	 * Produces a ConversationProtocol tree based upon the given Object[] array
	 *
	 * @param list An array containing the tree structure
	 * @param prev_nodes A map of previous nodes for loop detection
	 *
	 * @return The List containing the top-level ConversationProtocol's
	 */
	private static List getProtocolTree( Object[] list, Map prev_nodes )
	{
		// Check we haven't been given a null pointer
		if( list == null )
		{
			return (new LinkedList());
		}
		
		// Create a List to store the results
		List out = new LinkedList();

		// Parse the list of objects and create a protocol tree
		int i=0;
		while( i < list.length )
		{
			// Create a ConversationProtocol for the next node
			ConversationProtocol current = new ConversationProtocol();
			
			
			// Store the info in the node
			current._performative = (String) list[i++];

			// Check what information is given next (is it a conversation state)
			if( ( list[i] instanceof java.lang.Integer ) && ( (Integer) list[i] ).intValue()<0 )
			{
				current._conv_state = ( (Integer) list[i++]).intValue();
			}
			else
			{
				current._conv_state = CONVERSATION_END;
			}

			// Add the initiator indicator
			current._initiator = ( (Integer) list[i++] ).intValue() == 0;

			// Check if the next node is a list of follow on messages
			if( ( i < list.length ) && !( list[i] instanceof java.lang.String ) )
			{
				// Check if the child nodes have already been created
				// (Allows for loops in protocol)
				if( prev_nodes.containsKey( list[i] ) )
				{
					current._following_nodes = (List) prev_nodes.get( list[i++] );
				}
				else
				{
					List sub_tree = new LinkedList();
					
					// Add List to the hash (aren't object references wonderful?)
					prev_nodes.put( list, out );
					
					// Produce the sub-tree and store it
					sub_tree = getProtocolTree( (Object[]) list[i++], prev_nodes );
					
					// Add sub-tree to current node
					current._following_nodes = sub_tree;
				}
			}
			else
			{
				// Set the following nodes to be an empty List
				current._following_nodes = new LinkedList();
			}

      // Check that current._following_nodes has been initialised
      if( current._following_nodes == null )
      {
				// Set the following nodes to be an empty List
				current._following_nodes = new LinkedList();          
      }
      
			// Add the current node to the list
			out.add( current );
		}
		
		return out;
	}
	
	

	/**
	 * Add a set of message details to the message list
	 *
	 * @param msg The ACLMessage to add
	 */
	private void addMessageDetails( Message msg )
	{
		// Create a MessageDetails object to store the message in
		MessageDetails details = new MessageDetails();
		
		// Fill in the relevant fields
		details._msg = msg;
		details._timestamp = System.currentTimeMillis();
		details._timeout = -1;
		
		// Get the timeout
		if( msg.getACL().getReplyByUTC() != null )
		{
			// Get the reply-by value
			UTCContainer reply_by = msg.getACL().getReplyByUTC();
			
			// Determine if its a UTCTime or UTCDuration, and take appropriate measures
			if( reply_by instanceof UTCTime )
			{
				long current_time = (new UTCTime() ).getZTime().getTimeInMillis();
				long duration = ( (UTCTime) reply_by ).getZTime().getTimeInMillis() - current_time;
				details._timeout = details._timestamp + duration;
			}
			else if( reply_by instanceof UTCDuration )
			{
				details._timeout = details._timestamp + 
														((UTCDuration) reply_by).getDurationInMillis();
			}
		}
		
		// Add the message to the message list
		_messages.add( details );		
	}
	

//----------------------------------------------------------------------
// INNER-CLASSES
//----------------------------------------------------------------------

	/**
	 * Class to contain information about a message
	 */
	class MessageDetails implements Serializable, Cloneable
	{
		/**
		 * Message that this object is representing
		 */
		public Message _msg;
		
		/**
		 * Local time of message receipt
		 */
		public long _timestamp;

		/**
		 * Duration of the timeout for the last message added 
		 */
		public long _timeout = -1;
		
		/**
		 * Overrides default toString()
		 */
		public String toString()
		{
			
			StringBuffer out = new StringBuffer();
			
			out.append( BASETOKENS.OPENBRACE );
			out.append( BASETOKENS.SPACE );
			out.append( _msg.toString() );
			out.append( BASETOKENS.COMMA );
			out.append( BASETOKENS.SPACE );
			out.append( _timestamp );
			out.append( BASETOKENS.COMMA );
			out.append( BASETOKENS.SPACE );
			out.append( _timeout );			
			out.append( BASETOKENS.SPACE );
			out.append( BASETOKENS.CLOSEBRACE );

			return out.toString();
					
		}
		
		public Object clone()
		{
			try
			{
				MessageDetails md = (MessageDetails) super.clone();
				md._msg = (Message) _msg.clone();
				return md;
			}
			catch( CloneNotSupportedException e )
			{
				return null;
			}
		}
	}
	
	
	/**
	 * Class to contain information about the protocol to be
	 * used.
	 */
	public static class ConversationProtocol implements Serializable, Cloneable
	{
		/**
		 * Performative to be used for this branch
		 */
		public String _performative;
		
		/**
		 * Initiator should send performative at this node?
		 */
		public boolean _initiator;
		
		/**
		 * Reference to following nodes
		 */
		public List _following_nodes;
		
		/**
		 * State of conversation at this node
		 */
		public int _conv_state;
		
		/**
		 * Wait for response, if none is received end the conversation
		 */
		public boolean _wait_for_response;
		
		/**
		 * Overrides default toString()
		 */
		public String toString()
		{
			
			StringBuffer out = new StringBuffer();
			
			out.append( BASETOKENS.OPENBRACE );
			out.append( BASETOKENS.SPACE );
			out.append( _performative );
			out.append( BASETOKENS.COMMA );
			out.append( BASETOKENS.SPACE );
			out.append( "initiator=" );
			out.append( this._initiator );
			
			if( _following_nodes != null )
			{
				
				out.append( BASETOKENS.COMMA );
				out.append( BASETOKENS.SPACE );
				out.append( _following_nodes.toString() );
				
			}
			
			out.append( BASETOKENS.SPACE );
			out.append( BASETOKENS.CLOSEBRACE );

			return out.toString();	

		}
		
		
		/**
		 * Renders object ready for garbage collection
		 */
		public void removeReferences()
		{
			if( _following_nodes != null )
			{
				List copy = new LinkedList( _following_nodes );
				Iterator i = copy.iterator();
				_following_nodes.clear();
				_following_nodes = null;
				while( i.hasNext() )
				{
					ConversationProtocol next = (ConversationProtocol) i.next();
					next.removeReferences();
				}
				copy.clear();
			}
		}
		
		/**
		 * Clone this object
		 */
		public Object clone()
		{
			return clone( new LinkedList(), new LinkedList() );
		}
	
		/**
		 * Clone this object recursively
		 */
		public ConversationProtocol clone( List originals, List clones  )
		{
			try
			{
				// Clone this
				ConversationProtocol cp = (ConversationProtocol) super.clone();
				cp._following_nodes = new LinkedList();
				
				// Update lists
				originals.add( this );
				clones.add( cp );
				
				// Get cloned ConversationProtocol objects for the _following_nodes list
				Iterator i = _following_nodes.iterator();
				while( i.hasNext() )
				{
					// Get next ORIGINAL CP object
					ConversationProtocol cp2 = (ConversationProtocol) i.next();
					
					// Already duplicated?
					if( originals.contains( cp2 ) )
					{
						// If we've already duplicated it, just retrieve its clone
						int index = originals.indexOf( cp2 );
						cp2 = (ConversationProtocol) clones.get( index );
					}
					else
					{
						// If we've not already cloned it, do so now
						cp2 = cp2.clone( originals, clones );
					}
					
					// Add the clone (either new, or previously produced) to the cloned
					// CP _following_nodes List
					cp._following_nodes.add( cp2 );
				}
				
				// Return cloned object :)
				return cp;
			}
			catch( CloneNotSupportedException e )
			{
				return null;
			}
		}
	}
	
//----------------------------------------------------------------------
// TEST METHODS
//----------------------------------------------------------------------

	public static void testAddMessages( String protocol, String[] data, boolean cont ) throws Throwable
	{
		AgentID aid = new AgentID();
		aid.setName( data[1] );
			ConversationImpl conv = getConversation( protocol, aid );

			testAddMessages( conv, data, cont );
	}

	public static void testAddMessages( ConversationImpl conv, String[] data, boolean cont )
	{
		conv.setConversationID( "testid" );
		try
		{
			for( int i=0; i< data.length; i+=2 )
			{
				AgentID a1 = new AgentID();
				AgentID a2 = new AgentID();
				a1.setName( data[ i+1 ] );
				a2.setName( (data[ i+1 ].equalsIgnoreCase(data[1]) ? data[3] : data[1]) );
				
				ACLMessage msg = new ACLMessage();
				msg.setPerformative( data[ i ] );
				msg.setSenderAID( a1 );
				msg.setReceiverAID( a2 ); 
				msg.setConversationID( "testid" );
				
				try
				{
					conv.addMessage( new Message( msg ) );
				}
				catch ( ProtocolErrorException e )
				{
					try
					{
						if( data[ i + 1 ].equals( data[ 1 ] ) )
							throw e;
							
						ACLMessage msg2 = new ACLMessage();
						msg2.setPerformative( FIPACONSTANTS.NOT_UNDERSTOOD );
						msg2.setSenderAID( a1 );
						msg2.setConversationID( "testid" );
						
						conv.addMessage( new Message( msg2 ) );

						if( !cont )
						{
							break;
						}
					}
					catch ( Exception f )
					{
						f.printStackTrace();
						break;
					}
				}
				catch ( Exception e )
				{
					e.printStackTrace();
					break;
				}
				
				fipaos.util.DIAGNOSTICS.println( conv );
			}
		}
		catch (Throwable e)
		{
			e.printStackTrace();
		}

		System.out.println("Final state:");
		System.out.println( conv );
	}
	
	
	/**
	* <br> Returns a Conversation object populated with the passed ACLMessage.
	* <br> Is useful during testing for creating populated Conversations.
	*
	* @param acl The ACLMessage to populate the Conversation with.
	* @param protocol A String identifying the protocol object to create.
	*
	*/
	public static ConversationImpl testPopulate(ACLMessage acl, String protocol)
	{
	   ConversationImpl conversation = null;
		
		 try 
		 {	
			  // create conversation Object and give it an id.
			  conversation = ConversationImpl.getConversation(protocol, acl.getSenderAID());		
		 }
		 catch ( UnknownProtocolException upe )
		 { System.out.println("Error: " + upe); }
	
	   conversation.setConversationID(acl.getConversationID());
		
	   try
		 {
	      conversation.addMessage( new Message( acl ) );
		 }
		 catch(Exception e) { System.out.println("Error adding message: " + e); }
		 
		 return conversation;
		
	} // testPopulate(ACLMessage acl)
		
	
	/**
	*  Used for testing.
	*/
	public static void main(String[] args) throws Throwable
	{
		
		 ACLMessage message_1 = new ACLMessage();
		 message_1.setConversationID("2222");
		 message_1.setSender("agent1@iiop://47.101.224.356:9000/acc");
		 message_1.setPerformative(FIPACONSTANTS.REQUEST);
		 message_1.setOntology("FIPA Agent Management");
		
	   Conversation conversation_1 = testPopulate(message_1, FIPACONSTANTS.FIPA_REQUEST);
	   	
	   System.out.println();
	   System.out.println("Generated Conversation:");
	   System.out.println(conversation_1);
	   	
	   	
	} // main
	
} // Conversation
