//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved.
//
//******************************************************************************

package fipaos.agent.conversation.standard;

//----------------------------------------------------------------------

import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.mts.*;
import fipaos.mts.service.*;
import fipaos.agent.*;
import fipaos.agent.conversation.ConversationListener;
import fipaos.agent.conversation.Conversation;
import fipaos.agent.conversation.ProtocolErrorException;
import fipaos.agent.conversation.ConversationFinishedException;
import fipaos.agent.conversation.ConversationManagerException;
import fipaos.agent.conversation.UnknownProtocolException;
import fipaos.agent.conversation.InvalidConversationIDException;
import fipaos.agent.profile.*;
import fipaos.skill.db.*;
import fipaos.util.*;
import fipaos.parser.acl.ACLMessage;
import fipaos.ont.BASETOKENS;

import java.util.*;
import java.awt.*;
import fipaos.agent.conversation.ConversationManager;


/**
 * This class embodies the core functionality of the ConversationManagerImpl.
 */
public class ConversationManagerImpl implements Runnable, ConversationManager
{
	//----------------------------------------------------------------------
	// CONSTANTS
	//----------------------------------------------------------------------
	
	//----------------------------------------------------------------------
	// STATIC VARIABLES
	//----------------------------------------------------------------------
	
	//----------------------------------------------------------------------
	// INSTANCE VARIABLES
	//----------------------------------------------------------------------
	
	/**
	 * Counts the number of messages sent by this ConversationManagerImpl (ensures
	 * local uniqueness)
	 */
	private int _count = 0;
	
	/**
	 * Agent name
	 */
	private String _agent_name;
	
	/**
	 * Agent id
	 */
	private AgentID _agent_id;
	
	/**
	 * Map of active conversations
	 */
	private Map _active_list;
	
	/**
	 * Cach of protocols
	 */
	private Map _protocol_cache;
	
	/**
	 * Reference to the FIFOQueue containing updates to be dealt with
	 */
	private fipaos.util.Queue _updates;
	
	/**
	 * Reference to MessageSender used to send messages
	 */
	private MessageSender _sender;
	
	/**
	 * Reference to the owner of this
	 */
	private FIPAOSAgent _owner;
	
	/**
	 *  Reference to the object that Conversations should be handed onto.
	 *  Typically this will be the TaskManager.
	 */
	private ConversationListener _conversation_listener;
	
	/**
	 * Database in use to archive finished conversations
	 */
	private Database _database;
	
	/**
	 * Thread running within the CM
	 */
	private Thread _thread;
	
	/**
	 * Monitor running within this CM
	 */
	private Monitor _monitor;
	
	/**
	 * Indicates if the CM has terminated yet
	 */
	private boolean _terminate = false;
	
	/**
	 * Contains the mappings between protocol names and their implementing
	 * classes
	 */
	private static ProtocolProfile __protocol_profile;
	
	//----------------------------------------------------------------------
	// CONSTRUCTOR
	//----------------------------------------------------------------------
	
	/**
	 * Create and initalise the CM.  This will invoked the setMessageSender()
	 * method on the ConversationListener passed as part of the initialisation
	 */
	public ConversationManagerImpl(ConversationListener cl, MessageSender ms,
	FIPAOSAgent agent, DatabaseProfile db_profile, ProtocolProfile protocol_profile)
	{
		// Save references passed
		_owner = agent;
		_conversation_listener = cl;
		_agent_id = agent.getAID();
		_agent_name = agent.getAID().getName();
		_sender = ms;
		
		__protocol_profile = protocol_profile;
		
		// Create Map of active conversations.
		_active_list = new TreeMap( fipaos.util.Comparators.STRING_COMPARE );
		
		// Initialise the protocol cache
		_protocol_cache = new TreeMap( fipaos.util.Comparators.STRING_COMPARE );
		
		// Create a DB for the Agent's conversations
		String database_name = agent.getAID().getName() + "_convdb";
		_database = DatabaseFactory.getDatabase( database_name, db_profile );
		if( _database == null )
		{
			DIAGNOSTICS.println("Couldn't create a new database, " +
			db_profile.getDatabaseType() +
			" couldn't be instantiated", this,
			DIAGNOSTICS.LEVEL_MAX );
			return;
		}
		
		// Open the Database
		try
		{
			_database.openDatabase();
		}
		catch (DatabaseException e)
		{
			DIAGNOSTICS.println( "Error creating PlannerScheduler Database", e, DIAGNOSTICS.LEVEL_MAX );
		}
		
		// Start main handling thread
		_thread = new Thread(this, "CM-" + _agent_name );
		_thread.start();
		
		// Create monitor for updated queue
		_updates = new FIFOQueue();
		_monitor = new Monitor( _updates, new MonitorListener()
		{
			public void monitorUpdate( Object o )
			{
				try
				{
					if( o instanceof Conversation )
					{
						// Standard conversation update
						Conversation conv = (Conversation) o;
						checkState( conv );
					}
					else if( o instanceof Message )
					{
						// Message for ended-conversation
						_conversation_listener.notifyMessageInEndedConversation( (Message) o );
					}
				}
				catch( Throwable t )
				{
					DIAGNOSTICS.println( "Exception when dealing with updated conversation - ignoring", t );
				}
			}
		});
		
		// Indicate that the message sender should send messages via this
		_sender.setMessageReceiver( this );
		
		// Set this as the default MessageSender for the ConversationListener
		_conversation_listener.setMessageSender( this );
	}
	
	//----------------------------------------------------------------------
	// PUBLIC METHODS - MessageSender interface
	//----------------------------------------------------------------------
	
	/**
	 * Shutdown the CM
	 */
	public void shutdown()
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return;
		}
		
		DIAGNOSTICS.println( "CM for " + _owner + " is about to be shutdown()...", DIAGNOSTICS.LEVEL_2 );
		
		// Indicate we should terminate
		_terminate = true;
		
		// Kill threads
		_monitor.halt();
		_thread.interrupt();
		
		// Remove references ready for garbage collection
		_owner = null;
		_conversation_listener = null;
		_sender = null;
		_agent_id = null;
		_database = null;
		_monitor = null;
		_updates = null;
		_thread = null;
		_active_list.clear();
		_active_list = null;
		_protocol_cache.clear();
		_protocol_cache = null;
		
		DIAGNOSTICS.println( "CM for " + _owner + " has shutdown()", DIAGNOSTICS.LEVEL_2 );
	}
	
	
	/**
	 * Attempt to set the receiver of incoming messages - not supported
	 * by this class!
	 */
	public void setMessageReceiver( MessageReceiver p1 )
	{
		// Ignore this, since the "next level" above this should be dealing
		// with Conversation objects via the ConversationListener interface
		DIAGNOSTICS.println( "ConversationManagerImpl.setMessageReceiver() ignored" );
	}
	
	
	/**
	 * Attempt to send the given message
	 */
	public void sendMessage( Message msg )
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return;
		}
		
		// Add time-stamp
		msg.getEnvelope().setSlotValue( Service.FIPAOS_DEBUG + "cm-send-time",
		new Long( System.currentTimeMillis() ) );
		
		try
		{
			// Attempt to send message
			if( _sender != null )
			{
				// See if we can deal with message
				handleMessage( msg );
				
				// Indicate a message will be sent
				DIAGNOSTICS.println( "Sending message: " + msg.getACL(),
				this, DIAGNOSTICS.LEVEL_2 );
				
				_sender.sendMessage( msg );
			}
			else
			{
				DIAGNOSTICS.println( "No sender assigned to CM - can't send message!",
				DIAGNOSTICS.LEVEL_MAX );
			}
		}
		catch( Throwable t )
		{
			// TODO: Propogate error back to Agent!
			DIAGNOSTICS.println( "Exception occured when sending message", t, this, DIAGNOSTICS.LEVEL_MAX );
		}
	}
	
	//----------------------------------------------------------------------
	// PUBLIC METHODS - MessageReceiver interface
	//----------------------------------------------------------------------
	
	/**
	 * Received an incoming message - put it through the CM
	 */
	public void receiveMessage( Message msg )
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return;
		}
		
		// Add time-stamp
		msg.getEnvelope().setSlotValue( Service.FIPAOS_DEBUG + "cm-receive-time",
		new Long( System.currentTimeMillis() ) );
		
		// Pass message to the conversation manager
		try
		{
			// Wait for the Agent to be ready before continuing
			_owner.waitForPushLock();
			
			// Indicate a message was received
			DIAGNOSTICS.println( "Message received by " + _agent_id.getName() + ": " + msg.getACL(),
			this, DIAGNOSTICS.LEVEL_2 );
			
			// See if we can deal with message
			handleMessage( msg );
		}
		catch ( ProtocolErrorException e )
		{
			DIAGNOSTICS.println( "Exception occured, sending a not-understood: " + msg, e ,this,DIAGNOSTICS.LEVEL_MAX );
			
			// Protocol error!
			_owner.sendNotUnderstood( msg.getACL() );
		}
		catch ( ConversationFinishedException e )
		{
			DIAGNOSTICS.println( "Conversation has finished" + msg, e, this );
			
			// Pass event to queue of conversation updates
			_updates.addItem( msg );
		}
		catch ( ConversationManagerException e )
		{
			DIAGNOSTICS.println( "Unexpected exception", e, this, DIAGNOSTICS.LEVEL_MAX );
		}
		catch ( Throwable e )
		{
			DIAGNOSTICS.println( "Error, replying with not-understood to: " + msg, e,this,DIAGNOSTICS.LEVEL_MAX );
		}
	}
	
	public void setMessageSender( MessageSender p1 )
	{
		// Save reference
		_sender = p1;
	}
	
	//----------------------------------------------------------------------
	// PUBLIC METHODS - Runnable interface
	//----------------------------------------------------------------------
	
	/**
	 * Main thread for dealing with non-event based Conversation updates
	 */
	public void run()
	{
		while(!_terminate)
		{
			try
			{
				Thread.sleep(5000);
				checkState();
			}
			catch(Throwable e)
			{
				DIAGNOSTICS.println( "Exception in checkState() - ignoring" + e );
			}
		} // while(true)
	} // run()
	
	
	//----------------------------------------------------------------------
	// PUBLIC METHODS
	//----------------------------------------------------------------------
	
	/**
	 * Method to set the ProtocolProfile for the JVM
	 */
	public static void setProtocolProfile( ProtocolProfile pp )
	{
		__protocol_profile = pp;
	}
	
	
	/**
	 * Looks at a Message and checks to see if it part
	 * of an exisiting conversation. if yes adds it to the conversation object
	 * in the conversation list.  Otherwise, creates a new Conversation object
	 */
	public void handleMessage( Message msg )
	throws UnknownProtocolException, InvalidConversationIDException,
	ProtocolErrorException, ConversationFinishedException
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return;
		}
		
		ACL message = msg.getACL();
		
		//The case where the message has no id
		if (message.getConversationID() == null)
		{
			// Create conversation object give it an id
			ConversationImpl conversation = createConversation(message.getProtocol(), _agent_id );
			message.setConversationID( conversation.getConversationID() );
			
			synchronized( _active_list )
			{
				// Add to _active_list
				_active_list.put( conversation.getConversationID(), conversation );
				
				synchronized( conversation )
				{
					// Add the message to it, and duplicate
					conversation.addMessage( msg );
					conversation = conversation.cloneAndResetUpdate();
				}
				
				// Deal with the conversation
				dealWithConversation( conversation );
			}
		}
		//the case where the message does have a conversation id
		else
		{
			// Retrieve the Conversation from the _active_list
			synchronized( _active_list )
			{
				ConversationImpl conv = (ConversationImpl) _active_list.get( message.getConversationID() );
				
				// Check if Conversation exists
				if( conv == null )
				{
					// Create and add to _active_list
					conv = createConversation( message.getProtocol(), _agent_name, message.getConversationID() );
					_active_list.put( conv.getConversationID(), conv );
				}
				
				synchronized( conv )
				{
					// Add the message to it, and duplicate
					conv.addMessage( msg );
					conv = conv.cloneAndResetUpdate();
				}
				
				// Deal with the conversation
				dealWithConversation( conv );
			}
		}
	}
	
	
	/**
	 * Method to create a new conversation given a protcol, and add it to the active list
	 */
	public Conversation addConversation(String protocol) throws UnknownProtocolException
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return null;
		}
		
		synchronized( _active_list )
		{
			ConversationImpl conv = createConversation( protocol, _agent_id );
			_active_list.put( conv.getConversationID(), conv );
			return (Conversation) conv.clone();
		}
	}
	
	
	/**
	 * Retrieves an active/previous Conversation that an Agent has had
	 */
	public Conversation getConversation(String conversation_id) throws InvalidConversationIDException
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return null;
		}
		
		Conversation conversation = null;
		
		synchronized( _active_list )
		{
			ConversationImpl conv_impl = (ConversationImpl) _active_list.get(conversation_id);
			if( conv_impl != null )
			{
				conversation = (Conversation) conv_impl.clone();
			}
		}
		
		// Check if the conversation was found
		if( conversation == null )
		{
			try
			{
				// Open database for read-only transaction
				_database.startTransaction( true );
				
				conversation = (Conversation) _database.findObject( conversation_id );
			}
			catch( Throwable t )
			{
				DIAGNOSTICS.println( "Error when accessing Database", t, DIAGNOSTICS.LEVEL_MAX );
			}
			finally
			{
				try
				{
					_database.endTransaction();
				}
				catch( Throwable t )
				{
				}
			}
		}
		
		// If conversation still hasn't been found, indicate that the conversation ID is unknown
		if( conversation == null )
		{
			throw new InvalidConversationIDException("Conversation " + conversation_id + " isn't active");
		}
		
		return conversation;
	}
	
	
	/**
	 * Adds a new protocol to the CM - the CM will attempt to learn the protocol and will
	 * cache it for future use.
	 *
	 * @param protocol_name The name of the protocol to store (e.g. fipa-new-protocol)
	 * @param protocol_defn The RDF definition of the protocol
	 * @return True if the protocol was learned and stored correctly, false otherwise
	 **/
	public boolean addProtocol( String protocol_name, String protocol_defn )
	{
		return false;
	}
	
	
	/**
	 * Indicates whether the specified protocol is already in the cache, i.e. has it already been learned.
	 *
	 * @param protocol The protocol to check for
	 * @return True if the protocol has already been learned, false otherwise
	 **/
	public boolean hasProtocol( String protocol )
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return false;
		}
		
		if ( _protocol_cache.containsKey( protocol ) )
		{
			return true;
		}
		// Now try and dynamically load the class from the hard coded protocol files.
		// If we can load it then we know it.
		else
		{
			try
			{
				// Attempt to dynamically load the class
				Class conv = Class.forName( getProtocolClassName( protocol ) );
				
				return true;
			}
			catch ( Exception e )
			{
			}
		}
		
		return false;
	}
	
	
	//----------------------------------------------------------------------
	// PROTECTED METHODS
	//----------------------------------------------------------------------
	
	//----------------------------------------------------------------------
	// PRIVATE METHODS
	//----------------------------------------------------------------------
	
	/**
	 * Returns the protocol class name for the protocol passed in to the
	 * method. If no protocol class can be found then "NoProtocol" is returned.
	 * This will also be returned if you use the "no-protocol" protocol.
	 *
	 * @param protocol The protocol that you want the class name for
	 * @return The class name for the protocol class, or NoProtocol
	 */
	static String getProtocolClassName( String protocol ) throws UnknownProtocolException
	{
		
		String class_name = null;
		
		// Decide if a protocol name has been given & a protocol profile has been
		// specified previously
		if( protocol != null &&
		!protocol.equals("") &&
		__protocol_profile != null)
		{
			Vector mappings = (Vector)__protocol_profile.getProtocolMappings();
			
			if (mappings != null)
			{
				Iterator iter = mappings.iterator();
				
				ProtocolMapping mapping;
				
				while (iter.hasNext())
				{
					
					mapping = (ProtocolMapping)iter.next();
					
					if (mapping.getProtocolName().equalsIgnoreCase(protocol))
					{
						class_name = mapping.getProtocolClass();
					}
					
				} // end while
				
			} // end if mappings != null
			
		}
		else
		{
			// No protocol given, use NoProtocol Conversation object
			DIAGNOSTICS.println( "Couldn't locate protocol specified - using NoProtocol" );
			class_name = "fipaos.agent.conversation.standard.protocol.NoProtocol";
		}
		
		if (class_name == null)
		{
			throw new UnknownProtocolException("CM: No class mapped to protocol " + protocol);
		}
		else
		{
			return class_name;
		}
		
	} // end getProtocolClassName()
	
	
	/**
	 * This method creates a new conversation id
	 */
	private synchronized String createConversationID()
	{
		//initialise local variables
		String conversationID = null;
		String stringOfTime = null;
		
		
		//intrement count instance variable by 1
		_count++;
		
		//build conversationID by concatenating agentguid to current time plus count
		long currentTime = System.currentTimeMillis();
		stringOfTime = "" + currentTime;
		
		//concat current_time
		conversationID = _agent_name + stringOfTime + _count;
		return conversationID;
	}
	
	
	/**
	 * This method creates a new conversation object
	 */
	private ConversationImpl createConversation(String protocol, AgentID currentName)
	throws UnknownProtocolException
	{
		ConversationImpl conversation = null;
		
		// create conversation Object and give it an id.
		conversation = ConversationImpl.getConversation(protocol, currentName);
		conversation.setConversationID(createConversationID());
		
		return conversation;
	}
	
	
	/**
	 * This method creates a new conversation object
	 */
	private ConversationImpl createConversation(String protocol, String currentName, String conv_id )
	throws UnknownProtocolException
	{
		// create conversation Object and give it an id.
		ConversationImpl conversation = createConversation(protocol, _agent_id );
		conversation.setConversationID( conv_id );
		
		return conversation;
	}
	
	
	/**
	 * Dependant on the state of the conversation, puts the message into the
	 * Conversation queue
	 */
	private void dealWithConversation( ConversationImpl conversation )
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return;
		}
		
		// If conversation requires attention of ConversationListener, add to PS queue
		if( conversation.getState() != Conversation.NO_AGENT_ACTION_REQ &&
		conversation.getState() != Conversation.WAIT_BEFORE_END )
		{
			_updates.addItem( conversation );
		}
	}
	
	
	
	/**
	 * Moves a conversation from the Active to the conversation archive
	 */
	public void removeConversation(String conversation_id)
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return;
		}
		
		DIAGNOSTICS.println( "removeConversation( " + conversation_id + " ) invoked!" );
		
		ConversationImpl conv = null;
		
		synchronized( _active_list )
		{
			// Remove the conversation
			conv = (ConversationImpl) _active_list.remove(conversation_id);
		}
		
		// Archive it
		if( conv != null )
		{
			archiveConversation( conv );
		}
	} // conversationInactive(String conversation_id)
	
	
	/**
	 * Place inactive Conversations into the Database associated with this CM
	 */
	private void archiveConversation( ConversationImpl conv )
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return;
		}
		
		try
		{
			_database.startTransaction( false );
			
			// add conversation into the archive
			conv._serialised = null;
			_database.addObject( (DatabaseObject) conv.clone() );
		}
		catch (DatabaseException ex)
		{
			DIAGNOSTICS.println( "Error when archiving Conversations", ex, DIAGNOSTICS.LEVEL_MAX );
		}
		finally
		{
			try
			{
				_database.endTransaction();
			}
			catch( Throwable t )
			{
			}
			
			conv.removeReferences();
		}
	} // archiveConversations()
	
	
	/**
	 * Check the state of all active conversations, and place into queue to be
	 * dealt with if necessary
	 */
	private void checkState()
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return;
		}
		
		synchronized( _active_list )
		{
			// Attempt to produce an iterator for the _active_list
			Iterator conversations = _active_list.keySet().iterator();
			
			// Iterate through all conversations dealing with their state
			while(conversations.hasNext())
			{
				// Get the next Conversation ID
				String conversation_id = (String) conversations.next();
				
				// Get the associated Conversation object
				ConversationImpl next = (ConversationImpl)_active_list.get(conversation_id);
				
				if( next != null )
				{
					synchronized( next )
					{
						long time_out = next.getTimeOut();
						
						if( !next.isUpdated() &&
						( ( time_out == -1 ) ||
						( time_out > System.currentTimeMillis() ) ) )
						{
							continue;
						}
						
						switch( next.getState() )
						{
							case 	Conversation.AGENT_ACTION_REQ:
							case 	Conversation.CONVERSATION_END:
								_updates.addItem( next.cloneAndResetUpdate() );
								break;
								
							case 	Conversation.NO_AGENT_ACTION_REQ:
								next.resetUpdate();
								break;
								
								default:
									break;
						} // switch(_active_conversations.conversationState(conversation_id))
					}
				}
			} // while(conversations.hasMoreElements())
		}
	} // checkState()
	
	
	/**
	 * Check the state of the given conversation, and forward if necessary
	 */
	private void checkState( Conversation next )
	{
		// Check if CM has been shutdown
		if( _terminate )
		{
			return;
		}
		
		// If this is true, pass conversation onwards
		boolean forward = false;
		
		synchronized( _active_list )
		{
			synchronized( next )
			{
				long time_out = next.getTimeOut();
				
				// Check there is a latest message to check!
				if( next.getLatestMessageIndex() >= 0 )
				{
					// Get the sender of the message
					AgentID sender = next.getSenderAID( next.getLatestMessageIndex() );
					
					// Ensure ended conversations are dealt with appropriately &
					// then forward them
					switch( next.getState() )
					{
						case 	Conversation.CONVERSATION_END:
							removeConversation( next.getConversationID() );
							
							default:
								if( !sender.equals( _agent_id ) )
								{
									forward = true;
								}
								break;
					} // switch(_active_conversations.conversationState(conversation_id))
				}
			}
		}
		
		// No problem forwarding Conversation :)
		if( forward )
		{
			_conversation_listener.notify( next );
		}
	} // checkState()
	
	
	//------------------------------------------------------------------------------
	// Test harness
	//------------------------------------------------------------------------------
	
	public static void main( String[] args )
	{
		final String me = "me@iiop://somewhere.com:8080/acc";
		final String other = "other@iiop://somewhere.com:8080/acc";
		final int test1_count = 1000;
		
		DatabaseProfile db_profile = null;
		
		ProtocolProfile protocol_profile = new ProtocolProfileImpl();
		
		ProtocolMapping mapping1 = new ProtocolMappingImpl();
		
		mapping1.setProtocolName("fipa-request");
		mapping1.setProtocolClass("fipaos.agent.conversation.standard.protocol.FIPARequest");
		
		protocol_profile.addProtocolMapping(mapping1);
		
		// Create new ConversationManagerImpl & target
		TestConvListener cl = new TestConvListener();
		ConversationManagerImpl cm = new ConversationManagerImpl( (ConversationListener) cl,
		(MessageSender) new TestMessageSender(),
		(FIPAOSAgent) null, db_profile, protocol_profile );
		
		//--------------------------------------------------------------------------
		// Test 1 - Does ConversationListener get notfication of ALL updates?
		//--------------------------------------------------------------------------
		System.out.println( "Test 1: Counting conversation updates......" );
		
		try
		{
			ACLMessage test_msg = new ACLMessage();
			test_msg.setSender( other );
			test_msg.setReceiver( me );
			test_msg.setPerformative( "inform" );
			test_msg.setConversationID( "test-msg-id" );
			test_msg.setProtocol( "" );
			test_msg.setContent( "" );
			
			for( int i = 0; i < test1_count; i++ )
			{
				cm.handleMessage( new Message( test_msg ) );
			}
		}
		catch( Throwable t )
		{
			System.out.println( "Error during test" );
			t.printStackTrace();
		}
	}
	
	private static class TestMessageSender implements MessageSender
	{
		public void sendMessage( Message msg )
		{
		}
		
		public void setMessageReceiver( MessageReceiver mr )
		{
		}
		
		public void shutdown()
		{
		}
	}
	
	private static class TestConvListener extends Frame implements ConversationListener
	{
		public int _notified;
		private Label _label;
		
		public TestConvListener()
		{
			super( "Test 1" );
			setLayout( new BorderLayout() );
			add( new Label( "Received Conversation Updates:" ), BorderLayout.NORTH );
			_label = new Label("0");
			add( _label, BorderLayout.CENTER );
			pack();
			setVisible( true );
		}
		
		/** May be called to notify of a single conversation that is in need of intervention.
		 */
		public void notify(Conversation conversation)
		{
			// Count number of notify messages
			_notified++;
			
			_label.setText( "" + _notified );
			
			try
			{
				Thread.sleep( 25 );
			}
			catch( InterruptedException e )
			{
			}
		}
		
		/** May be called to notify of a number of conversations that are in need of intervention.
		 */
		public void notify(Enumeration conversations)
		{
			while( conversations.hasMoreElements() )
			{
				notify( (Conversation) conversations.nextElement() );
			}
		}
		
		public void setMessageSender( MessageSender ms )
		{
		}
		
	/**
	 * Indicates that the given Message arrived for a completed Conversation
	 */
		public void notifyMessageInEndedConversation( Message msg )
		{
		}
		
	/**
	 * Indicates that the given Message was not deliverable
	 */
		public void notifyMessageUndeliverable( Message msg )
		{
		}
	}
}

