//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.agent.task;

//----------------------------------------------------------------------

import fipaos.ont.fipa.FIPACONSTANTS;
import fipaos.ont.BASETOKENS;
import fipaos.util.DIAGNOSTICS;
import fipaos.agent.task.Task;
import fipaos.agent.conversation.Conversation;
import fipaos.agent.FIPAOSAgent;
import fipaos.agent.AgentList;
import fipaos.ont.fipa.fipaman.*;
import fipaos.parser.acl.ACLMessage;
import fipaos.ont.fipa.*;
import fipaos.util.*;
import fipaos.parser.*;

import java.util.*;

//----------------------------------------------------------------------

/**
 * This task class handles agent searches of DFs.
 * <br>
 * <br>
 */
public class DFSearchTask extends Task
{

	//----------------------------------------------------------------------
	// CONSTANTS
	//----------------------------------------------------------------------

	// DF search depth constraint
	private static final int SEARCH_DEPTH = 2;


	//----------------------------------------------------------------------
	// INSTANCE VARIABLES
	//----------------------------------------------------------------------

	private AgentList _store;
	private String _address;
	//  private Task _task;
	private DFAgentDescription _df_desc;
	private String _df;
	private SearchConstraints _constraints;
	private boolean _federated;
	private int _selected = -1;


	//----------------------------------------------------------------------
	// CONSTRUCTOR
	//----------------------------------------------------------------------


	/** Constructs the DF search task object which will perform a search using the
	 * given description
	 * @param desc Description of the Agent to search for
	 */
	public DFSearchTask( DFAgentDescription desc )
	{
		_store = new AgentList();
		_df_desc = desc;
		_selected = 0;
	}


	/** Searches the local platform DF for any agents matching the specified DF
	 * description using the specified search constraint.
	 * This method will use the following default search parameters:
	 * <br><ul>
	 * <li>Federated search: yes
	 * </ul>
	 * @param df_desc The DF description to search for
	 * @param constraint The search constraint
	 */
	public DFSearchTask( DFAgentDescription df_desc, SearchConstraints constraints )
	{
		_store = new AgentList();
		_df_desc = df_desc;
		
		_constraints = constraints;
		
		_selected = 1;
	}


	/** Searches the local platform DF for any agents matching the specified DF
	 * description. The search is optionally federated across other DFs.
	 * This method will use the following default search parameters:
	 * <br><ul>
	 * <li>Search constraint: depth 2
	 * </ul>
	 * @param df_desc The DF description to search for
	 * @param federated Flag to signal whether this search should be federated or not
	 */
	public DFSearchTask( DFAgentDescription df_desc, boolean federated )
	{
		_store = new AgentList();
		_df_desc = df_desc;
		_federated = federated;
		_selected = 2;
	}


	/** Searches the specified DF for any agents matching the specified DF
	 * description. The search is optionally federated across other DFs.
	 * This method will use the following default search parameters:
	 * <br><ul>
	 * <li>Search constraint: depth 2
	 * </ul>
	 * @param df The DF to search
	 * @param df_desc The DF description to search for
	 * @param federated Flag to signal whether this search should be federated or not
	 */
	public DFSearchTask( String df, DFAgentDescription df_desc, boolean federated )
	{
		_store = new AgentList();
		_df = df;
		_df_desc = df_desc;
		_federated = federated;
		_selected = 3;
	}


	/** Searches the specified DF for any agents matching the specified DF
	 * description using the specified search contraint.
	 * The search is optionally federated across other DFs.
	 * @param df The DF to search
	 * @param df_desc The DF description to search for
	 * @param constraint The search constraint
	 * @param federated Flag to signal whether this search should be federated or not
	 */
	public DFSearchTask( String df, DFAgentDescription df_desc, SearchConstraints constraints,
	boolean federated )
	{
		_store = new AgentList();
		_df = df;
		_df_desc = df_desc;
		_constraints = constraints;
		_federated = federated;
		_selected = 4;
	}


	/**
	 * Invoked when the task has started
	 */
	protected void startTask()
	{
		_store = new AgentList();
		_address = _owner.getLocalDF();

		// Ensure a SearchConstraints object is provided
		if( _constraints == null )
		{
			_constraints = new SearchConstraints();
		}
		
		switch( _selected )
		{
			case 0:		searchDF( _df_desc );
			break;
			case 1:		searchDF( _df_desc, _constraints );
			break;
			case 2:		searchDF( _df_desc, _federated );
			break;
			case 3:		searchDF( _df, _df_desc, _federated );
			break;
			case 4:		searchDF( _df, _df_desc, _constraints, _federated );
			break;
		}
	}

	//----------------------------------------------------------------------
	// PRIVATE METHODS
	//----------------------------------------------------------------------

	private DFAgentDescription[] fetchDFResults( String content ) throws ParserException
	{
//		String resultContent= StringParser.trimBrackets( content );

		AgentManagementResult result=new AgentManagementResult( content );

		Set resultSet=result.getResultSet();

		DFAgentDescription[] descs=new DFAgentDescription[resultSet.size()];

		Iterator iterator=resultSet.iterator();

		int i=0;

		while (iterator.hasNext())
		{
			descs[i]=(DFAgentDescription) iterator.next();
			i++;
		}

		return descs;
	}

	//----------------------------------------------------------------------
	// PUBLIC METHODS
	//----------------------------------------------------------------------

	// fipa-request protocol used for all comms with other agents

	/**
	 * Handles incoming not-understood messages. An empty array of DF results
	 * is returned to the calling task.
	 *
	 * @param conv The conversation containint the incoming message
	 */
	public void handleNotUnderstood( Conversation conv )
	{
		// Return an empty result array
		String sender = conv.getACL( conv.getLatestMessageIndex() ).getSenderAID().getName();
		_store.remove( sender );
		done( new DFAgentDescription[ 0 ] );
	}

	// -------------------------------------------------------------------------

	/**
	 * Handles incoming refuse messages. An empty array of DF results
	 * is returned to the calling task.
	 *
	 * @param conv The conversation containint the incoming message
	 */
	public void handleRefuse( Conversation conv )
	{
		// Return an empty result array
		String sender = conv.getACL( conv.getLatestMessageIndex() ).getSenderAID().getName();
		_store.remove( sender );
		done( new DFAgentDescription[ 0 ] );
	}

	// -------------------------------------------------------------------------

	/**
	 * Handles incoming agree messages. These messages are ignored.
	 *
	 * @param conv The conversation containint the incoming message
	 */
	public void handleAgree( Conversation conv )
	{

		// Ignore

	}

	// -------------------------------------------------------------------------

	/**
	 * Handles incoming failure messages. An empty array of DF results
	 * is returned to the calling task.
	 *
	 * @param conv The conversation containint the incoming message
	 */
	public void handleFailure( Conversation conv )
	{
		// Return an empty result array
		String sender = conv.getACL( conv.getLatestMessageIndex() ).getSenderAID().getName();
		_store.remove( sender );
		done( new DFAgentDescription[ 0 ] );
	}

	// -------------------------------------------------------------------------

	/**
	 * Handles incoming inform messages. The content of the inform message is
	 * parsed for DF search results, which are then passed to the calling task via
	 * a callback method.
	 *
	 * @param conv The conversation containint the incoming message
	 */
	public void handleInform( Conversation conv )
	{
		// Results back from the DF
		ACL acl = conv.getACL( conv.getLatestMessageIndex() );
		String sender = acl.getSenderAID().getName();

		if ( _store.contains( sender ) )
		{
			_store.remove( sender );
			try
			{
				Object content = acl.getContentObject();
				if( content instanceof String )
				{
					done( fetchDFResults( (String) content ) );
				}
				else if( content instanceof AgentManagementResult )
				{
					AgentManagementResult amr = (AgentManagementResult) content;
					done( amr.getResultSet().toArray( new DFAgentDescription[0] ) );
				}
			}
			catch (Throwable e)
			{
					DIAGNOSTICS.println("Invalid results received",e,this);
			}

		}
	}


	//----------------------------------------------------------------------
	// PUBLIC METHODS - DF SEARCH API
	//----------------------------------------------------------------------

	/** Searches the local platform DF for any agents matching the specified DF
	 * description. This method will use the following default search parameters:
	 * <br><ul>
	 * <li>Search constraint: depth 2
	 * <li>Federated search: yes
	 * </ul>
	 * @param df_desc The DF description to search for
	 * @return False if the search could not be started, true otherwise - a return value
	 * of true does NOT mean that results were found
	 * @deprecated Please use this task as a sub-task, rather than directly
	 * invoking methods on it.
	 */
	public boolean searchDF( DFAgentDescription df_desc )
	{

		String df = _address;
		SearchConstraints cons = new SearchConstraints( SEARCH_DEPTH );

		return searchDF( df, df_desc, cons, true );

	}

	// -------------------------------------------------------------------------

	/** Searches the local platform DF for any agents matching the specified DF
	 * description using the specified search constraint.
	 * This method will use the following default search parameters:
	 * <br><ul>
	 * <li>Federated search: yes
	 * </ul>
	 * @param df_desc The DF description to search for
	 * @param constraint The search constraint
	 * @return False if the search could not be started, true otherwise - a return value
	 * of true does NOT mean that results were found
	 * @deprecated Please use this task as a sub-task, rather than directly
	 * invoking methods on it.
	 */
	public boolean searchDF( DFAgentDescription df_desc, SearchConstraints constraint )
	{

		String df = _address;

		return searchDF( df, df_desc, constraint, true );

	}

	// -------------------------------------------------------------------------

	/** Searches the local platform DF for any agents matching the specified DF
	 * description. The search is optionally federated across other DFs.
	 * This method will use the following default search parameters:
	 * <br><ul>
	 * <li>Search constraint: depth 2
	 * </ul>
	 * @param df_desc The DF description to search for
	 * @param federated Flag to signal whether this search should be federated or not
	 * @return False if the search could not be started, true otherwise - a return value
	 * of true does NOT mean that results were found
	 * @deprecated Please use this task as a sub-task, rather than directly
	 * invoking methods on it.
	 */
	public boolean searchDF( DFAgentDescription df_desc, boolean federated )
	{

		String df = _address;
		SearchConstraints cons = new SearchConstraints( SEARCH_DEPTH );

		return searchDF( df, df_desc, cons, federated );

	}

	// -------------------------------------------------------------------------

	/** Searches the specified DF for any agents matching the specified DF
	 * description. The search is optionally federated across other DFs.
	 * This method will use the following default search parameters:
	 * <br><ul>
	 * <li>Search constraint: depth 2
	 * </ul>
	 * @param df The DF to search
	 * @param df_desc The DF description to search for
	 * @param federated Flag to signal whether this search should be federated or not
	 * @return False if the search could not be started, true otherwise - a return value
	 * of true does NOT mean that results were found
	 * @deprecated Please use this task as a sub-task, rather than directly
	 * invoking methods on it.
	 */
	public boolean searchDF( String df, DFAgentDescription df_desc, boolean federated )
	{

		SearchConstraints cons = new SearchConstraints( SEARCH_DEPTH );

		return searchDF( df, df_desc, cons, federated );

	}

	// -------------------------------------------------------------------------

	/** Searches the specified DF for any agents matching the specified DF
	 * description using the specified search contraint.
	 * The search is optionally federated across other DFs.
	 * @param df The DF to search
	 * @param df_desc The DF description to search for
	 * @param constraint The search constraint
	 * @param federated Flag to signal whether this search should be federated or not
	 * @return False if the search could not be started, true otherwise - a return value
	 * of true does NOT mean that results were found
	 * @deprecated Please use this task as a sub-task, rather than directly
	 * invoking methods on it.
	 */
	public boolean searchDF( String df, DFAgentDescription df_desc, SearchConstraints constraints,
	boolean federated )
	{

		ACL out = getNewConversation( FIPACONSTANTS.FIPA_REQUEST );

		if ( out == null )
		{

			return false;

		}

		out.setSenderAID( _owner.getAID() );
		out.setLanguage( FIPACONSTANTS.SL0 );
		out.setOntology( FIPAMANCONSTANTS.FIPA_AGENT_MANAGEMENT );
		out.setPerformative( FIPACONSTANTS.REQUEST );

		try
		{
			out.setReceiverAID( new AgentID( df ) );
			
			AgentManagementAction agent_mgt_action=new AgentManagementAction();
			agent_mgt_action.setAction(FIPAMANCONSTANTS.SEARCH);
			agent_mgt_action.setAgentID(new AgentID(df));
			agent_mgt_action.setDFAgentDescription(df_desc);
			
			if( federated )
			{
				agent_mgt_action.setSearchConstraints(constraints);
			}
			else
			{
				agent_mgt_action.setSearchConstraints( new SearchConstraints() );
			}

			out.setContentObject( agent_mgt_action );
			

			_store.add( new AgentID( df ).getName(), FIPACONSTANTS.FIPA_DF );

			forward( out );

		}
		catch ( fipaos.parser.ParserException p )
		{

			DIAGNOSTICS.println( "Error when parsing DF AgentID - incorrect format!",
			this, DIAGNOSTICS.LEVEL_MAX );

			return false;

		}
		catch ( Throwable mse )
		{

			return false;

		}

		return true;

	}


} // End DFSearchState
