//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Emorphia are Copyright (C) 2001 Emorphia Limited
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.agent.task;

// -------------------------------------------------------------------------

/**
 * This task enables a Task to be started and its result to be captured by
 * making a single method invocation
 */
public class SynchronousTask extends Task
{
	/**
	 * Child task we're monitoring
	 */
	private Task _child;
	
	/**
	 * Timeout for child Rask
	 */
	private long _timeout;
	
	/**
	 * Monitor lock - used to indicate when done
	 */
	private Object _lock = new Object();
	
	/**
	 * Result of child task
	 */
	private Object _result;
	
	/**
	 * Indicates if we're done
	 */
	private boolean _done = false;
	
	/**
	 * Private constructor since we don't want to allow any code other than that
	 * contained within this class to construct one of these
	 */
	private SynchronousTask( Task child, long timeout )
	{
		_child = child;
		_timeout = timeout;
	}
	
	/**
	 * Invoked when the task starts
	 */
	public void startTask()
	{
		// Check if a timeout has been specified
		if( _timeout > -1 )
		{
			// Start the child task with a timeout
			newTask( _child, _timeout );
		}
		else
		{
			// Start the child task
			newTask( _child );
		}
	}
	

	/** Invokes the appropriate doneX() method on this, given the child task
	 * @param child The child task that has completed
	 */
	void notifyDone( Task child )
	{
		// Invoke superclass equivelant
		super.notifyDone( child );
		
		_result = new NoResult();
		indicateDone();
	}
	
	
	/** Invokes the appropriate doneX(Object) method on this, given the child task
	 * @param child The child task that has completed
	 * @param result The result associated with the complete child Task
	 */
	void notifyDone( Task child, Object result )
	{
		// Invoke superclass equivelant
		super.notifyDone( child, result );

		_result = result;
		indicateDone();
	}
	
	
	/** Invoked by child-tasks to indicate that it has had an error
	 * @param child Child Task which had the error
	 * @param error The associated error exception
	 */
	void notifyError( Task child, Throwable error )
	{
		// Invoke superclass equivelant
		super.notifyError( child, error );
		
		_result = new ErrorResult( error );
		indicateDone();
	}
	
	
	/** Invokes the appropriate timeoutX() method on this, given the child task
	 * @param child The child task that has timed-out
	 */
	void notifyTimeout( Task child )
	{
		// Invoke superclass equivelant
		super.notifyTimeout( child );
		
		_result = new TimeoutResult();
		indicateDone();
	}
	
	
	/**
	 * Indicates to monitoring threads that we're done
	 */
	private void indicateDone()
	{
		// Indicate to interested threads that we're done
		synchronized( _lock )
		{
			_done = true;
			_lock.notifyAll();
		}
		
		// Indicate to TM that we're finished with this Task
		done();
	}
	
	
	/**
	 * Executes the given Task, and returns the result - in the event that something
	 * unexpected happens, an object of type UnusualResult is returned
	 *
	 * <p>The Task must not have been initialised
	 *
	 * <p>WARNING: This method shouldn't be invoked from within the code for a Task!
	 */
	public static Object executeTask( TaskManager tm, Task t )
	{
		return executeTask(tm, t, -1 );
	}
	
	
	/**
	 * Executes the given Task, and returns the result - in the event that something
	 * unexpected happens, an object of type UnusualResult is returned
	 *
	 * <p>The Task must not have been initialised
	 *
	 * <p>WARNING: This method shouldn't be invoked from within the code for a Task!
	 */
	public static Object executeTask(TaskManager tm, Task t, long timeout )
	{
		// Create Task to deal with this
		SynchronousTask st = new SynchronousTask( t, timeout );
		tm.newTask( st );
		
		// Wait for result
		synchronized( st._lock )
		{
			while( !st._done )
			{
				try
				{
					st._lock.wait();
				}
				catch( InterruptedException e )
				{
				}
			}
		}
		
		// Return result
		return st._result;		
	}
	
	
	/**
	 * A sub-class is returned by executeTask() when something out of the ordinary happens
	 */
	public static class UnusualResult
	{
	}
	
	/**
	 * Indicates that the Task timed-out
	 */
	public static class TimeoutResult extends UnusualResult
	{
	}
	
	/**
	 * Indicates that the Task didn't return a result
	 */
	public static class NoResult extends UnusualResult
	{
	}
	
	/**
	 * Indicates that an exception occured within the Task
	 */
	public static class ErrorResult extends UnusualResult
	{
		private Throwable t;
		
		private ErrorResult( Throwable t )
		{
			this.t = t;
		}
		
		public Throwable getThrowable()
		{
			return t;
		}
	}
}