//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved.
//
//******************************************************************************

package fipaos.agent.task;

// -------------------------------------------------------------------------

import java.util.*;

import fipaos.agent.FIPAOSAgent;
import fipaos.agent.conversation.Conversation;
import fipaos.agent.task.event.*;
import fipaos.mts.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.ont.BASETOKENS;
import fipaos.util.*;

// -------------------------------------------------------------------------

/** <p>This class is an abstract representation of an agent task. It contains message dispatchers and dummy
 * handling methods. To implement tasks in your agent, subclass this class and override the methods for
 * the message types that you need to handle. For example, if your agent handles two conversations -
 * one of type request and the other of contract net, then you would create two task classes that
 * subclass this class. One that handles fipa-request protocol messages only, and on the handles
 * fipa-contract-net messages only.
 *
 * <p>The Task-lifecycle is as follows:
 * <ul>
 * <li>Created, non-initialised - In this state the Task implementation has been
 * instantiated, but newTask() has not been invoked on it.  In this state, <b>none</b>
 * of the functionality provided by the Task class is gauranteed to function.
 * <li>Initialised - The Task is made aware of the transition to the state by having
 * its startTask() method invoked - in this state <b>all</b> of the functionality
 * provided by the Task class is gauaranteed to work as described in the JavaDocs
 * for this class.  The call to startTask() is executed in a seperate thread than
 * other call-backs - if you want to take full-advantage of the parallel-processing
 * capabilities of the Task paradigm, just add the functionality to this method in
 * your Task implementation
 * <li>Finished - The Task moves to this state if one of the following occurs:
 * <ul>
 * <li>done() or done( Object ) is invoked
 * <li>The Task times-out (if a time-out is imposed on the-Task)
 * <li>An un-caught exception occurs within the Task
 * </ul>
 * In this state the Task will no longer recieve any events
 * </ul>
 */
public abstract class Task
{
	
	//----------------------------------------------------------------------
	// CONSTANTS
	//----------------------------------------------------------------------
	
	private static final String HANDLE_METHOD_PREFIX = "handle";
	private static final String DONE_METHOD_PREFIX = "done";
	private static final String ERROR_METHOD_PREFIX = "error";
	private static final String TIMEOUT_METHOD_PREFIX = "timeout";
	
	//----------------------------------------------------------------------
	// INSTANCE VARIABLES
	//----------------------------------------------------------------------
	
	/**
	 * The agent that owns this task - should only be used for the purpose of
	 * updating/retreiving parent-Agent state information
	 */
	protected FIPAOSAgent _owner;
	
	/**
	 * The task manager that manages this task object - useful for registering new tasks based on
	 * incoming conversations
	 */
	protected TaskManager _tm;
	
	/**
	 * State of this Task
	 */
	TaskState _state = new TaskState();
	
	/**
	 * ID of this task (?)
	 */
	String _id;
	
	/**
	 * List of children of this task.
	 */
	protected Set _children;
	
	/**
	 * Reference to the "parent" Task of this.  If null, then this task
	 * has no parent.
	 */
	Task _parent;
	
	/**
	 * TaskManager ID's that this is associated with
	 */
	Set _tm_ids = new TreeSet( fipaos.util.Comparators.STRING_COMPARE );
	
	/**
	 * Timeout value for this task
	 */
	long _timeout = -1;
	
	/**
	 * Indicates if this task is done, or has timedout
	 */
	boolean _done = false;
	
	/**
	 * Lock - ensures mutual exclusion between threads modifying this Task
	 */
	protected Object _lock = new Object();
	
	/**
	 * Indicates if a handleX() method didn't execute in sub-class
	 */
	private boolean _non_exec = false;
	
	/**
	 * Context info - allows an object to provide context info to parent-Task
	 * which launched this task
	 */
	private Object _context = null;
	
	//----------------------------------------------------------------------
	// CONSTRUCTOR(S)
	//----------------------------------------------------------------------
	
	/** Constructs a Task object - to fully initialise a Task object, use the
	 * Task or TaskManager newTask() method
	 */
	public Task()
	{
		_children = new TreeSet( fipaos.util.Comparators.OBJECT_COMPARE );
		_id = getClassName( this );
	}
	
	
	/** Constructs a Task object. You must construct tasks using a concrete subclass of this class.
	 * @deprecated This constructor is no longer supported.  Use the
	 * no-argument constructor from sub-classes, and
	 * initialise using the newTask() method of the parent Task
	 * or TaskManager.
	 * @param owner The agent that owns this task
	 * @param taskman Reference to parent TaskManager
	 */
	public Task( FIPAOSAgent owner, TaskManager taskman )
	{
		this();
		
		_tm = taskman;
		_owner = owner;
		_id = owner.toString();
	}
	
	//----------------------------------------------------------------------
	
	/** Constructs a task object. You must construct tasks using a concrete subclass of this class.
	 * @param owner The agent that owns this task
	 * @param taskman The task manager managing this task
	 * @param id The unique id of the task
	 * @deprecated This constructor is no longer supported.  Use the
	 * no-argument constructor from sub-classes, and
	 * initialise using the newTask() method of the parent Task
	 * or TaskManager.
	 */
	public Task( FIPAOSAgent owner, TaskManager taskman, String id )
	{
		this();
		
		_tm = taskman;
		_owner = owner;
		_id = id;
		
	}
	
	
	//----------------------------------------------------------------------
	// PRIVATE METHODS
	//----------------------------------------------------------------------
	
	/**
	 * Capitalises the first character of a string
	 *
	 * @param string The string to capitalise
	 */
	private String capitaliseFirstChar( String string )
	{
		return string.substring( 0, 1 ).toUpperCase() + string.substring( 1 );
	}
	
	
	/** Gets the class name of an object without package names, and the $ char in
	 * InnerClass names converted to _
	 * @return The class-name of this class
	 * @param o The Object to construct the classname from
	 */
	private String getClassName( Object o )
	{
		// Get a reference to the Class of this
		String this_class = o.getClass().getName();
		
		// Replace '$' with '_'
		this_class = this_class.replace( '$', '_' );
		
		// Remove package details
		int last_dot = this_class.lastIndexOf( '.' );
		this_class = this_class.substring( last_dot + 1 );
		
		return this_class;
	}
	
	
	//----------------------------------------------------------------------
	// PUBLIC METHODS
	//----------------------------------------------------------------------
	
	/**
	 * Sets a context object for this Task - allows a parent-Task to associate some
	 * arbritary "context" with a child-Task which it can pickup when the child
	 * completes/fails
	 */
	public void setContext( Object o )
	{
		_context = o;
	}
	
	
	/**
	 * Gets a context object for this Task - allows a parent-Task to associate some
	 * arbritary "context" with a child-Task which it can pickup when the child
	 * completes/fails
	 */
	public Object getContext()
	{
		return _context;
	}
		
	
	/** Returns the TaskState associated with this Task
	 * @return The TaskState associated with this
	 */
	public TaskState getState()
	{
		return _state;
	}
	
	/**
	 * Dispatches incoming message for ended conversation to the correct method.
	 *
	 * @param msg Message that has arrived after end of conversation
	 **/
	public void notifyMessageForEndedConversation( Message msg )
	{
		handleLateMessage( msg );
	}
	
	/**
	 * Dispatches incoming messages that have bounced for some reason.
	 *
	 * @param conv Conversation containing incoming message
	 **/
	public void notifyMessageFailure( Message msg )
	{
		handleUndeliverableMessage( msg );
	}
	
	/**
	 * Dispatches incoming messages to the correct message handling method.
	 *
	 * @param conv Conversation containing incoming message
	 **/
	public void notify( Conversation conv )
	{
		ACL acl = conv.getACL( conv.getLatestMessageIndex() );
		String msg_type = acl.getPerformative().toLowerCase();
		String sender = acl.getSenderAID().getName();
		
		// We only want to inform tasks of incoming messages - remember that the
		// PlannerScheduler notifies us of all conversation events - even outgoing msgs
		if ( ! sender.equals( _owner.getAID().getName() ) )
		{
			
			DIAGNOSTICS.println( "Received message: " + acl.toString(), _owner, DIAGNOSTICS.LEVEL_2 );
			
			// Reset "not executed by child" indicator
			_non_exec = false;
			String method_name = null;
			
			try
			{
				// Try to use more efficient method calls rather than reflection first
				if( msg_type.equalsIgnoreCase( FIPACONSTANTS.INFORM ) )
				{
					method_name = "handleInform";
					handleInform( conv );
				}
				else if( msg_type.equalsIgnoreCase( FIPACONSTANTS.REQUEST ) )
				{
					method_name = "handleRequest";
					handleRequest( conv );
				}
				else if( msg_type.equalsIgnoreCase( FIPACONSTANTS.FAILURE ) )
				{
					method_name = "handleFailure";
					handleFailure( conv );
				}
				else if( msg_type.equalsIgnoreCase( FIPACONSTANTS.NOT_UNDERSTOOD ) )
				{
					method_name = "handleNotUnderstood";
					handleNotUnderstood( conv );
				}
				else if( msg_type.equalsIgnoreCase( FIPACONSTANTS.REFUSE ) )
				{
					method_name = "handleRefuse";
					handleRefuse( conv );
				}
				else if( msg_type.equalsIgnoreCase( FIPACONSTANTS.SUBSCRIBE ) )
				{
					method_name = "handleSubscribe";
					handleSubscribe( conv );
				}
				else if( msg_type.equalsIgnoreCase( FIPACONSTANTS.AGREE ) )
				{
					method_name = "handleAgree";
					handleAgree( conv );
				}
				else if( msg_type.equalsIgnoreCase( FIPACONSTANTS.ACCEPT_PROPOSAL ) )
				{
					method_name = "handleAcceptProposal";
					handleAcceptProposal( conv );
				}
				else if( msg_type.equalsIgnoreCase( FIPACONSTANTS.CFP ) )
				{
					method_name = "handleCfp";
					handleCfp( conv );
				}
				else if( msg_type.equalsIgnoreCase( FIPACONSTANTS.CONFIRM ) )
				{
					method_name = "handleConfirm";
					handleConfirm( conv );
				}
				else if( msg_type.equalsIgnoreCase( FIPACONSTANTS.DISCONFIRM ) )
				{
					method_name = "handleDisconfirm";
					handleDisconfirm( conv );
				}
				else if( msg_type.equalsIgnoreCase( FIPACONSTANTS.INFORM_IF ) )
				{
					method_name = "handleInformIf";
					handleInformIf( conv );
				}
				else if( msg_type.equalsIgnoreCase( FIPACONSTANTS.INFORM_REF ) )
				{
					method_name = "handleInformRef";
					handleInformRef( conv );
				}
				else if( msg_type.equalsIgnoreCase( FIPACONSTANTS.PROPAGATE ) )
				{
					method_name = "handlePropagate";
					handlePropagate( conv );
				}
				else if( msg_type.equalsIgnoreCase( FIPACONSTANTS.PROPOSE ) )
				{
					method_name = "handlePropose";
					handlePropose( conv );
				}
				else if( msg_type.equalsIgnoreCase( FIPACONSTANTS.PROXY ) )
				{
					method_name = "handleProxy";
					handleProxy( conv );
				}
				else if( msg_type.equalsIgnoreCase( FIPACONSTANTS.QUERY_IF ) )
				{
					method_name = "handleQueryIf";
					handleQueryIf( conv );
				}
				else if( msg_type.equalsIgnoreCase( FIPACONSTANTS.QUERY_REF ) )
				{
					method_name = "handleQueryRef";
					handleQueryRef( conv );
				}
				else if( msg_type.equalsIgnoreCase( FIPACONSTANTS.REJECT_PROPOSAL ) )
				{
					method_name = "handleRejectProposal";
					handleRejectProposal( conv );
				}
				else if( msg_type.equalsIgnoreCase( FIPACONSTANTS.REQUEST_WHEN ) )
				{
					method_name = "handleRequestWhen";
					handleRequestWhen( conv );
				}
				else if( msg_type.equalsIgnoreCase( FIPACONSTANTS.REQUEST_WHENEVER ) )
				{
					method_name = "handleRequestWhenever";
					handleRequestWhenever( conv );
				}
				else
				{
					_non_exec = true;
				}
			}
			catch( Throwable t )
			{
				// Just display the exception
				DIAGNOSTICS.println( "WARNING: Unhandled exception when dealing with incoming message in " +
				getClass().getName() + "( method " + method_name +
				"( Conversation ) )", t, DIAGNOSTICS.LEVEL_3 );
			}
			
			// If _non_exec == false, then method was executed by a sub-class :)
			if( !_non_exec )
			{
				return;
			}
			
			// Convert message type to handleX method name
			// This involves changing the name to be capitalised and removing '-' characters
			// and capitalising the next letter after the '-'
			// e.g. not-understood --> handleNotUnderstood
			// e.g. accept-proposal --> handleAcceptProposal
			// e.g. request --> handleRequest
			boolean more_dashes = true;
			StringBuffer method = new StringBuffer( HANDLE_METHOD_PREFIX );
			
			while ( more_dashes )
			{
				
				int dash = msg_type.indexOf( BASETOKENS.DASH );
				
				if ( dash > 0 )
				{
					
					method.append( capitaliseFirstChar( msg_type.substring( 0, dash ) ) );
					msg_type = msg_type.substring( dash + 1 );
					
				}
				else
				{
					
					method.append( capitaliseFirstChar( msg_type ) );
					more_dashes = false;
					
				}
				
			}
			
			
			// Reset non-exec flag (it will be set to true if a method implemented in this
			// class is invoked
			_non_exec = false;
			
			// Use reflection to see if this task has that handle method
			try
			{
				Class method_args[] =
				{ Conversation.class };
				Object invoke_args[] =
				{ conv };
				
				DynamicClass dc = new DynamicClass( this );
				dc.invoke( method.toString(), method_args, invoke_args );
			}
			catch ( NoSuchMethodException nsme )
			{
				
				// Method doesn't exist....
				_non_exec = true;
				
			}
			catch( DynamicClassException t )
			{
				// Just display the underlying exception
				DIAGNOSTICS.println( "WARNING: Unhandled exception when dealing with incoming message in" +
				getClass().getName() + "( method " + method +
				"( Conversation ) )", t, DIAGNOSTICS.LEVEL_3 );
			}
		}
		
		// If method didn't exist OR sub-class doesn't override method, invoke
		// handleOther()
		if( _non_exec )
		{
			handleOther( conv );
		}
	}
	
	/**
	 * Dummy method - please override implementation and DO NOT invoke
	 */
	public void handleInform( Conversation c )
	{
		_non_exec = true;
	}
	
	/**
	 * Dummy method - please override implementation and DO NOT invoke
	 */
	public void handleRequest( Conversation c )
	{
		_non_exec = true;
	}
	
	/**
	 * Dummy method - please override implementation and DO NOT invoke
	 */
	public void handleFailure( Conversation c )
	{
		_non_exec = true;
	}
	
	/**
	 * Dummy method - please override implementation and DO NOT invoke
	 */
	public void handleNotUnderstood( Conversation c )
	{
		_non_exec = true;
	}
	
	/**
	 * Dummy method - please override implementation and DO NOT invoke
	 */
	public void handleRefuse( Conversation c )
	{
		_non_exec = true;
	}
	
	/**
	 * Dummy method - please override implementation and DO NOT invoke
	 */
	public void handleSubscribe( Conversation c )
	{
		_non_exec = true;
	}
	
	/**
	 * Dummy method - please override implementation and DO NOT invoke
	 */
	public void handleAgree( Conversation c )
	{
		_non_exec = true;
	}
	
	/**
	 * Dummy method - please override implementation and DO NOT invoke
	 */
	public void handleAcceptProposal( Conversation c )
	{
		_non_exec = true;
	}
	
	/**
	 * Dummy method - please override implementation and DO NOT invoke
	 */
	public void handleCfp( Conversation c )
	{
		_non_exec = true;
	}
	
	/**
	 * Dummy method - please override implementation and DO NOT invoke
	 */
	public void handleConfirm( Conversation c )
	{
		_non_exec = true;
	}
	
	/**
	 * Dummy method - please override implementation and DO NOT invoke
	 */
	public void handleDisconfirm( Conversation c )
	{
		_non_exec = true;
	}
	
	/**
	 * Dummy method - please override implementation and DO NOT invoke
	 */
	public void handleInformIf( Conversation c )
	{
		_non_exec = true;
	}
	
	/**
	 * Dummy method - please override implementation and DO NOT invoke
	 */
	public void handleInformRef( Conversation c )
	{
		_non_exec = true;
	}
	
	/**
	 * Dummy method - please override implementation and DO NOT invoke
	 */
	public void handlePropagate( Conversation c )
	{
		_non_exec = true;
	}
	
	/**
	 * Dummy method - please override implementation and DO NOT invoke
	 */
	public void handlePropose( Conversation c )
	{
		_non_exec = true;
	}
	
	/**
	 * Dummy method - please override implementation and DO NOT invoke
	 */
	public void handleProxy( Conversation c )
	{
		_non_exec = true;
	}
	
	/**
	 * Dummy method - please override implementation and DO NOT invoke
	 */
	public void handleQueryIf( Conversation c )
	{
		_non_exec = true;
	}
	
	/**
	 * Dummy method - please override implementation and DO NOT invoke
	 */
	public void handleQueryRef( Conversation c )
	{
		_non_exec = true;
	}
	
	/**
	 * Dummy method - please override implementation and DO NOT invoke
	 */
	public void handleRejectProposal( Conversation c )
	{
		_non_exec = true;
	}
	
	/**
	 * Dummy method - please override implementation and DO NOT invoke
	 */
	public void handleRequestWhen( Conversation c )
	{
		_non_exec = true;
	}
	
	/**
	 * Dummy method - please override implementation and DO NOT invoke
	 */
	public void handleRequestWhenever( Conversation c )
	{
		_non_exec = true;
	}
	
	/**
	 * Please override implementation to deal with undeliverable messages
	 * <p>NOTE: Not yet implemented
	 */
	public void handleUndeliverableMessage( Message msg )
	{
	}
	
	/**
	 * Please override implementation - by default sends a not-understood
	 */
	public void handleLateMessage( Message msg )
	{
			// Conversation has ended - send a not understood!
			_owner.sendNotUnderstood( msg.getACL() );
	}
	
	//----------------------------------------------------------------------
	// PROTECTED METHODS - DF SEARCH API
	//----------------------------------------------------------------------
	
	/**
	 * Searches the local platform DF for any agents matching the specified DF
	 * description. This method will use the following default search parameters:
	 * <br><ul>
	 * <li>Search constraint: depth 2
	 * <li>Federated search: yes
	 * </ul>
	 *
	 * @param df_desc The DF description to search for
	 * @return False if the search could not be started, true otherwise - a return value
	 * of true does NOT mean that results were found
	 */
	protected boolean searchDF( DFAgentDescription df_desc )
	{
		newTask( new DFSearchTask( df_desc ) );
		
		return true;
	}
	
	// -------------------------------------------------------------------------
	
	/**
	 * Searches the local platform DF for any agents matching the specified DF
	 * description using the specified search constraint.
	 * This method will use the following default search parameters:
	 * <br><ul>
	 * <li>Federated search: yes
	 * </ul>
	 *
	 * @param df_desc The DF description to search for
	 * @param constraint The search constraint
	 * @return False if the search could not be started, true otherwise - a return value
	 * of true does NOT mean that results were found
	 */
	protected boolean searchDF( DFAgentDescription df_desc, SearchConstraints constraint )
	{
		newTask( new DFSearchTask( df_desc, constraint ) );
		
		return true;
	}
	
	// -------------------------------------------------------------------------
	
	/**
	 * Searches the local platform DF for any agents matching the specified DF
	 * description. The search is optionally federated across other DFs.
	 * This method will use the following default search parameters:
	 * <br><ul>
	 * <li>Search constraint: depth 2
	 * </ul>
	 *
	 * @param df_desc The DF description to search for
	 * @param federated Flag to signal whether this search should be federated or not
	 * @return False if the search could not be started, true otherwise - a return value
	 * of true does NOT mean that results were found
	 */
	protected boolean searchDF( DFAgentDescription df_desc, boolean federated )
	{
		newTask( new DFSearchTask( df_desc, federated ) );
		
		return true;
	}
	
	// -------------------------------------------------------------------------
	
	/**
	 * Searches the specified DF for any agents matching the specified DF
	 * description. The search is optionally federated across other DFs.
	 * This method will use the following default search parameters:
	 * <br><ul>
	 * <li>Search constraint: depth 2
	 * </ul>
	 *
	 * @param df The DF to search
	 * @param df_desc The DF description to search for
	 * @param federated Flag to signal whether this search should be federated or not
	 * @return False if the search could not be started, true otherwise - a return value
	 * of true does NOT mean that results were found
	 */
	protected boolean searchDF( String df, DFAgentDescription df_desc, boolean federated )
	{
		newTask( new DFSearchTask( df, df_desc, federated ) );
		
		return true;
	}
	
	// -------------------------------------------------------------------------
	
	/**
	 * Searches the specified DF for any agents matching the specified DF
	 * description using the specified search contraint.
	 * The search is optionally federated across other DFs.
	 *
	 * @param df The DF to search
	 * @param df_desc The DF description to search for
	 * @param constraint The search constraint
	 * @param federated Flag to signal whether this search should be federated or not
	 * @return False if the search could not be started, true otherwise - a return value
	 * of true does NOT mean that results were found
	 */
	protected boolean searchDF( String df, DFAgentDescription df_desc, SearchConstraints constraint,
	boolean federated )
	{
		newTask( new DFSearchTask( df, df_desc, constraint, federated ) );
		
		return true;
	}
	
	// -------------------------------------------------------------------------
	
	/**
	 * Callback method that is used to signal to the task that the previously
	 * requestes DF search has now finished. The results of the search are returned
	 * as an array of DF description objects. This method should be overridden in
	 * any concrete tasks to provide some functionality.
	 *
	 * @param df_descs An array of search results
	 */
	protected void DFSearchResults( DFAgentDescription[] df_descs )
	{
		
		if( ! (this instanceof SynchronousTask) )
		{
			// Do nothing - this will be overriden in the concrete task subclass
			DIAGNOSTICS.println( "WARNING: DFSearchResults() invoked, but Task implementation" +
			" doesn't override this method!", DIAGNOSTICS.LEVEL_3 );
		}
		
	}
	
	
	/** Callback method to indicate that a DFSearchTask has completed.  This
	 * simply invokes DFSearchResults unless it is overridden.
	 * @param result The result of the DF-search
	 */
	public void doneDFSearchTask( Object result )
	{
		if( !_tm._shutdown )
		{
			if( result != null && result instanceof DFAgentDescription[] )
			{
				DFSearchResults( (DFAgentDescription[]) result );
			}
			else
			{
				DFSearchResults( new DFAgentDescription[0] );
			}
		}
	}
	
	
	/**
	 * Callback method to indicate that a DFSearchTask has failed.  This
	 * simply invokes DFSearchResults unless it is overridden.
	 */
	public void errorDFSearchTask( Task t, Throwable exception )
	{
		if( !_tm._shutdown )
		{
			// Indicate we couldn't get any results
			DFSearchResults( new DFAgentDescription[0] );
		}
	}
	
	
	/** Callback method to indicate that a DFSearchTask has failed.  This
	 * simply invokes DFSearchResults unless it is overridden.
	 * @param t The DFSearchTask that has timed-out
	 * @param exception  */
	public void timeoutDFSearchTask( Task t )
	{
		if( !_tm._shutdown )
		{
			// Indicate we couldn't get any results
			DFSearchResults( new DFAgentDescription[0] );
		}
	}
	
	
	//----------------------------------------------------------------------
	// PROTECTED METHODS
	//----------------------------------------------------------------------
	
	/**
	 * Sends an ACL message.
	 *
	 * @param acl The ACL message to send
	 * @throws MessageSenderException if the message could not be sent
	 */
	protected void forward( ACL acl )
	{
		// Get variables
		TaskManager tm = null;
		synchronized( _lock )
		{
			tm = _tm;
		}
		
		if( _tm != null )
		{
			// Split apart message if we have multiple recipients
			if( acl.getReceiverAIDs().size() > 1 && acl.getProtocol() != null )
			{
				Iterator i = acl.getReceiverAIDs().iterator();
				while( i.hasNext() )
				{
					try
					{
						ACL next = (ACL) acl.clone();
						next.setReceiverAID( (AgentID) i.next() );
						_tm.forward( next, this );
					}
					catch( Throwable t )
					{
						DIAGNOSTICS.println( "Couldn't clone() multicast message" );
					}
				}
			}
			else
			{
				_tm.forward( acl, this );
			}
		}
	}
	
	
	/**
	 * Sends a Message.
	 *
	 * @param msg The Message to send
	 * @throws MessageSenderException if the message could not be sent
	 */
	protected void forward( Message msg )
	{
		// Get variables
		TaskManager tm = null;
		synchronized( _lock )
		{
			tm = _tm;
		}
		
		if( tm != null )
		{
			// Split apart message if we have multiple recipients
			if( msg.getACL().getReceiverAIDs().size() > 1 && msg.getACL().getProtocol() != null )
			{
				Iterator i = msg.getACL().getReceiverAIDs().iterator();
				while( i.hasNext() )
				{
					try
					{
						Message next = (Message) msg.clone();
						next.getACL().setReceiverAID( (AgentID) i.next() );
						next.getEnvelope().setReceivers( next.getACL().getReceiverAIDs() );
						_tm.forward( next, this );
					}
					catch( Throwable t )
					{
						DIAGNOSTICS.println( "Couldn't clone() multicast message" );
					}
				}
			}
			else
			{
				_tm.forward( msg, this );
			}
		}
	}
	
	
	/**
	 * Send a not-understood in response to the given ACL
	 */
	protected boolean sendNotUnderstood( ACL orig_acl )
	{
		// Get variables
		TaskManager tm = null;
		FIPAOSAgent owner = null;
		synchronized( _lock )
		{
			tm = _tm;
			owner = _owner;
		}
		
		if( tm._shutdown && owner != null )
		{
			return owner.sendNotUnderstood( orig_acl );
		}
		else
		{
			return false;
		}
	}
	
	
	
	// -------------------------------------------------------------------------
	
	/**
	 * Returns a new ACL message object that is pre-filled for the specified protocol.
	 * This is the preferred way of constructing new messages (that are not part of
	 * an existing conversation) as this will generate a unique conversation ID.
	 *
	 * @param protocol The agent interaction protocol to use
	 * @return A new ACL message object
	 */
	protected ACL getNewConversation( String protocol )
	{
		return _tm.getNewConversation( protocol );
	}
	
	
	// -------------------------------------------------------------------------
	
	
	/** Initialises the given task, and invokes its start() method.  The task name
	 * will become the name of the class by default.
	 * @param task The new Task to be initialised and added to the system
	 */
	protected void newTask( Task task )
	{
		newTask( getClassName(task), task );
	}
	
	
	/** Initialises the given task, and invokes its start() method.  The task name
	 * will become the name of the class by default.
	 * @param task The new Task to be initialised and added to the system
	 * @param conv Conversation to be associated with this Task
	 */
	protected void newTask( Task task, Conversation conv )
	{
		newTask( getClassName(task), task, conv );
	}
	
	
	/** Initialises the given task, and invokes its start() method
	 * @param task The new Task to be initialised and added to the system
	 * @param name The name to be used for this task
	 */
	protected void newTask( String name, Task task )
	{
		if( _tm != null )
		{

			_tm.newTask(name, task, this );
		}
	}
	
	
	/** Initialises the given task, and invokes its start() method
	 * @param task The new Task to be initialised and added to the system
	 * @param conv Conversation to be associated with this Task
	 * @param name The name to be associated with this task
	 */
	protected void newTask( String name, Task task, Conversation conv )
	{
		if( _tm != null )
		{

			_tm.newTask(name, task, conv, this );
		}
	}
	
	
	
	/** Initialises the given task, and invokes its start() method.  The task name
	 * will become the name of the class by default.  If the task doesn't invoke
	 * done() in the given amount of time, a timeoutX() method will be invoked on
	 * the parent-Task, where X is the id of the task
	 *
	 * @param timeout The time in milliseconds to wait before timing out
	 * @param task The new Task to be initialised and added to the system
	 */
	protected void newTask( Task task,long timeout )
	{
		if( _tm != null )
		{
			_tm.newTask( task, timeout, this );
		}
	}
	
	
	/** Initialises the given task, and invokes its start() method.  The task name
	 * will become the name of the class by default. If the task doesn't invoke
	 * done() in the given amount of time, a timeoutX() method will be invoked on
	 * the parent-Task, where X is the id of the task
	 *
	 * @param timeout The time in milliseconds to wait before timing out
	 * @param task The new Task to be initialised and added to the system
	 * @param conv Conversation to be associated with this Task
	 */
	protected void newTask( Task task, Conversation conv, long timeout )
	{
		if( _tm != null )
		{
			_tm.newTask( task, conv, timeout, this );
		}
	}
	
	
	/** Initialises the given task, and invokes its start() method.  If the task doesn't invoke
	 * done() in the given amount of time, a timeoutX() method will be invoked on
	 * the parent-Task, where X is the id of the task
	 *
	 * @param timeout The time in milliseconds to wait before timing out
	 * @param task The new Task to be initialised and added to the system
	 * @param name The name to be used for this task
	 */
	protected void newTask( String name, Task task, long timeout )
	{
		task._id = name;
		
		if( _tm != null )
		{
			_tm.newTask( task, timeout, this );
		}
	}
	
	
	/** Initialises the given task, and invokes its start() method.  If the task doesn't invoke
	 * done() in the given amount of time, a timeoutX() method will be invoked on
	 * the parent-Task, where X is the id of the task
	 *
	 * @param timeout The time in milliseconds to wait before timing out
	 * @param task The new Task to be initialised and added to the system
	 * @param conv Conversation to be associated with this Task
	 * @param name The name to be associated with this task
	 */
	protected void newTask( String name, Task task, Conversation conv, long timeout )
	{
		task._id = name;
		
		if( _tm != null )
		{
			_tm.newTask( task, conv, timeout, this );
		}
	}
	
	
	// -------------------------------------------------------------------------
	
	/** When invoked, parent task will have the appropriate doneX() method invoked
	 */
	protected void done()
	{
		DIAGNOSTICS.println( "done() invoked!", this );
		
		// Get variables
		TaskManager tm = null;
		Task parent = null;
		synchronized( _lock )
		{
			tm = _tm;
			parent = _parent;
		}
		
		
		// Check if this is the first thread to indicate done
		if( thisThreadFirstToIndicateDone() )
		{
			// Add a suitable event to the parent task, if one is specified
			if( parent != null && tm != null )
			{
				// Create TaskEvent
				TaskEvent te = new ChildDoneEvent( this );
				
				// Add to parent
				tm.addEvent( parent, te );
			}
			// Remove task if it is done & is a top-level-Task
			else if( parent == null && tm != null )
			{
				tm.removeTask( this );
			}
		}
	}
	
	
	/** When invoked, parent task will have the appropriate doneX(Object) method invoked
	 * @param result The Object to pass to parent task
	 */
	protected void done(Object result)
	{
		DIAGNOSTICS.println( "done( " + result + " ) invoked!", this );
		
		// Get variables
		TaskManager tm = null;
		Task parent = null;
		synchronized( _lock )
		{
			tm = _tm;
			parent = _parent;
		}
		
		
		// Check if this is the first thread to indicate done
		if( thisThreadFirstToIndicateDone() )
		{
			// Add a suitable event to the parent task, if one is specified
			if( parent != null && tm != null )
			{
				// Create TaskEvent
				TaskEvent te = new ChildDoneWithResultEvent( this, result );
				
				// Add to parent
				tm.addEvent( parent, te );
			}
			// Remove task if it is done & is a top-level-Task
			else if( _parent == null && tm != null )
			{
				tm.removeTask( this );
			}
		}
		
	}
	
	
	// -------------------------------------------------------------------------
	
	/** Invoked immediately after the task has been initialised by newTask()
	 */
	protected void startTask()
	{
	}
	
	
	// -------------------------------------------------------------------------
	
	/**
	 * Cleans up the reference that this task has to enable quicker garbage collection
	 */
	void removeReferences()
	{
		synchronized( _lock )
		{
			_owner = null;
			_parent = null;
			_tm = null;
			
			if( _state != null )
			{
				_state.removeReferences();
				_state = null;
			}
			
			if( _tm_ids != null )
			{
				_tm_ids.clear();
				_tm_ids = null;
			}
			
			if( _children != null )
			{
				_children.clear();
				_children = null;
			}
		}
	}
	
	/**
	 * Starts the given Task executing
	 */
	void notifyStart()
	{
		// Save reference to this
		final Task task = this;
		
		try
		{
			// Start Task and change its state
			startTask();
			
			synchronized( _lock )
			{
				if( !_done && _state != null )
				{
					_state.setState( TaskState.TASK_IDLE );
				}
			}
		}
		catch( Throwable th )
		{
			// Check if this is the first thread to indicate done
			if( thisThreadFirstToIndicateDone() )
			{
				TaskManager tm = null;
				Task parent = null;
				
				synchronized( _lock )
				{
					tm = _tm;
					parent = _parent;
				}
				
				// Got reference to TM OK?
				if( tm != null )
				{
					// Check if task has parent
					if( parent != null )
					{
						// Create TaskEvent
						TaskEvent te = new ChildFailureEvent( task, th );
						
						// Add to parent task
						tm.addEvent( parent, te );
					}
					else
					{
						DIAGNOSTICS.println( "WARNING: Exception in top-level task " +
						task.getClass().getName() + ": ", th, this,
						DIAGNOSTICS.LEVEL_MAX );
						tm.removeTask( task );
					}
				}
			}
		}
	}
	
	
	/**
	 * Invoked when the timeout for this Task has been reached
	 */
	void timeoutReached()
	{
		// Check if this is the first thread to indicate done
		if( thisThreadFirstToIndicateDone() )
		{
			TaskManager tm = null;
			Task parent = null;
			
			synchronized( _lock )
			{
				tm = _tm;
				parent = _parent;
			}
			
			// Got reference to TM OK?
			if( tm != null )
			{
				// Check if task has parent
				if( parent != null )
				{
					// Create TaskEvent
					TaskEvent te = new ChildTimeoutEvent( this );
					
					// Add to parent task
					tm.addEvent( parent, te );
				}
				else
				{
					DIAGNOSTICS.println( "WARNING: Timeout in top-level task " + this,
					DIAGNOSTICS.LEVEL_2 );
					tm.removeTask( this );
				}
			}
		}
	}
	
	
	/** Invoked by child-tasks to indicate that it has had an error
	 * @param child Child Task which had the error
	 * @param error The associated error exception
	 */
	void notifyError( Task child, Throwable error )
	{
		if( !_tm._shutdown )
		{
			synchronized( _lock )
			{
				// Check child is acutally our child!
				if( _children != null && !_children.contains( child ) )
				{
					DIAGNOSTICS.println( "WARNING: A task (" + child._id + ") has indicated it has " +
					"failed, but it is not registered as a child-Task!",
					DIAGNOSTICS.LEVEL_3 );
					
					return;
				}
			}
			

			// Attempt to find the appropriate method to handle the error and invoke it
			Object callback_target = null;
			if( this instanceof FIPAOSAgent.TopLevelHandlerTask )
			{
				callback_target = _owner;
			}
			else
			{
				callback_target = this;
			}
			
			
			// Invoke appropriate method on object
			if( callback_target instanceof Task )
			{
				Task t = (Task) callback_target;
				t.notifyError( child, ERROR_METHOD_PREFIX + child._id, error );
			}
			else if( callback_target instanceof FIPAOSAgent )
			{
				FIPAOSAgent a = (FIPAOSAgent) callback_target;
				a.notifyError( child, ERROR_METHOD_PREFIX + child._id, error );
			}
			
			
			// Remove child-task from Set of children
			synchronized( _lock )
			{
				if( _children != null )
				{
					// Always remove the child-Task from the _children set
					_children.remove( child );
				}
			}
		}
	}
	
	
	/**
	 * Default method invoked to dispatch error callbacks - override this with
	 * a method to dispatch the callback statically to improve efficiency :)
	 * @param child Child-Task that has failed
	 * @param method Name of the method to invoke
	 */
	protected void notifyError( Task child, String method, Throwable error )
	{
		DynamicClass dc = new DynamicClass( this );
		
		// Attempt to invoke that method
		try
		{
			dc.invoke( method, new Class[]
			{ Task.class, Throwable.class },
			new Object[]
			{ child, error } );
		}
		catch( Throwable t )
		{
			if( ! (this instanceof SynchronousTask) )
			{
				// Just display the error
				DIAGNOSTICS.println( "WARNING: Unhandled exception from child-Task (" + child._id +
				") ", error, DIAGNOSTICS.LEVEL_3 );
			}
		}
	}
	
	
	/** Invokes the appropriate doneX() method on this, given the child task
	 * @param child The child task that has completed
	 */
	void notifyDone( Task child )
	{
		if( !_tm._shutdown )
		{
			synchronized( _lock )
			{
				// Check child is acutally our child!
				if( _children != null && !_children.contains( child ) )
				{
					DIAGNOSTICS.println( "WARNING: A task (" + child._id + ") indicated it has " +
					"completed, but it is not registered as a child-Task!",
					DIAGNOSTICS.LEVEL_3 );
					
					return;
				}
			}
			
			// Attempt to find the appropriate method to handle this completion and invoke it
			Object callback_target = null;
			if( this instanceof FIPAOSAgent.TopLevelHandlerTask )
			{
				callback_target = _owner;
			}
			else
			{
				callback_target = this;
			}
			
			
			// Invoke appropriate method on object
			if( callback_target instanceof Task )
			{
				Task t = (Task) callback_target;
				t.notifyDone( child, DONE_METHOD_PREFIX + child._id, null );
			}
			else if( callback_target instanceof FIPAOSAgent )
			{
				FIPAOSAgent a = (FIPAOSAgent) callback_target;
				a.notifyDone( child, DONE_METHOD_PREFIX + child._id, null );
			}
			
			
			// Remove child-task from Set of children
			synchronized( _lock )
			{
				if( _children != null )
				{
					// Always remove the child-Task from the _children set
					_children.remove( child );
				}
			}
		}
	}
	
	
	/** Invokes the appropriate doneX(Object) method on this, given the child task
	 * @param child The child task that has completed
	 * @param result The result associated with the complete child Task
	 */
	void notifyDone( Task child, Object result )
	{
		if( !_tm._shutdown )
		{
			synchronized( _lock )
			{
				// Check child is acutally our child!
				if( _children != null && !_children.contains( child ) )
				{
					DIAGNOSTICS.println( "WARNING: A task (" + child._id + ") indicated it has " +
					"completed, but it is not registered as a child-Task!",
					DIAGNOSTICS.LEVEL_3 );
					
					return;
				}
			}
			
			// Attempt to find the appropriate method to handle this completion and invoke it
			Object callback_target = null;
			if( this instanceof FIPAOSAgent.TopLevelHandlerTask )
			{
				callback_target = _owner;
			}
			else
			{
				callback_target = this;
			}
			
			
			// Invoke appropriate method on object
			if( callback_target instanceof Task )
			{
				Task t = (Task) callback_target;
				t.notifyDone( child, DONE_METHOD_PREFIX + child._id, result );
			}
			else if( callback_target instanceof FIPAOSAgent )
			{
				FIPAOSAgent a = (FIPAOSAgent) callback_target;
				a.notifyDone( child, DONE_METHOD_PREFIX + child._id, result );
			}
			
			
			// Remove child-task from Set of children
			synchronized( _lock )
			{
				if( _children != null )
				{
					// Always remove the child-Task from the _children set
					_children.remove( child );
				}
			}
		}
	}
	
	
	/**
	 * Default method invoked to dispatch done callbacks - override this with
	 * a method to dispatch the callback statically to improve efficiency :)
	 * @param child Child-Task that has completed
	 * @param method Name of the method to invoke
	 * @param result Result associated with the completion (null if no result given)
	 */
	protected void notifyDone( Task child, String method, Object result )
	{
		DynamicClass dc = new DynamicClass( this );
		
		// Attempt to invoke that method
		try
		{
			if( result == null )
			{
				dc.invoke( method, new Class[]
				{ Task.class }, new Object[]
				{ child } );
			}
			else
			{
				// Try doneX( Task, Object ), then doneX( Object ) if that fails
				try
				{
					dc.invoke( method, new Class[]
					{ Task.class, Object.class }, new Object[]
					{ child, result } );
				}
				catch( NoSuchMethodException t )
				{
					dc.invoke( method, new Class[]
					{ Object.class }, new Object[]
					{ result } );
				}
			}
		}
		catch( NoSuchMethodException t )
		{
			if( ! (this instanceof SynchronousTask) )
			{
				//if (!(child instanceof SynchronousTask))
				//{
				// Just display the error
				DIAGNOSTICS.println( "WARNING: Unhandled completion from child-Task (" +
				child._id + ") - parent " + "Task = " +
				getClass().getName() + "( method " + method +
				"( Task ) not found )", DIAGNOSTICS.LEVEL_3 );
				//}
			}
		}
		catch( DynamicClassException t )
		{
			if( ! (this instanceof SynchronousTask) )
			{
				// Just display the underlying exception
				DIAGNOSTICS.println( "WARNING: Unhandled exception when dealing with completion from child-Task (" +
				child._id + ") in parent " + "Task = " +
				getClass().getName() + "( method " + method +
				"( Task ) )", t, DIAGNOSTICS.LEVEL_3 );
			}
		}
	}
	
	
	/** Invokes the appropriate timeoutX() method on this, given the child task
	 * @param child The child task that has timed-out
	 */
	void notifyTimeout( Task child )
	{
		if( !_tm._shutdown )
		{
			synchronized( _lock )
			{
				// Check child is acutally our child!
				if( _children != null && !_children.contains( child ) )
				{
					DIAGNOSTICS.println( "WARNING: A task (" + child._id + ") indicated it has " +
					"timed-out, but it is not registered as a child-Task!",
					DIAGNOSTICS.LEVEL_3 );
					
					return;
				}
			}
			
			// Attempt to find the appropriate method to handle this completion and invoke it
			Object callback_target = null;
			if( this instanceof FIPAOSAgent.TopLevelHandlerTask )
			{
				callback_target = _owner;
			}
			else
			{
				callback_target = this;
			}

			
			// Invoke appropriate method on object
			if( callback_target instanceof Task )
			{
				Task t = (Task) callback_target;
				t.notifyTimeout( child, TIMEOUT_METHOD_PREFIX + child._id );
			}
			else if( callback_target instanceof FIPAOSAgent )
			{
				FIPAOSAgent a = (FIPAOSAgent) callback_target;
				a.notifyTimeout( child, TIMEOUT_METHOD_PREFIX + child._id );
			}
			
			
			// Remove child-task from Set of children
			synchronized( _lock )
			{
				if( _children != null )
				{
					// Always remove the child-Task from the _children set
					_children.remove( child );
				}
			}
		}
	}
	
	
	/**
	 * Default method invoked to dispatch timeout callbacks - override this with
	 * a method to dispatch the callback statically to improve efficiency :)
	 * @param child Child-Task that has timed-out
	 * @param method Name of the method to invoke
	 */
	protected void notifyTimeout( Task child, String method )
	{
		DynamicClass dc = new DynamicClass( this );
		
		// Attempt to invoke that method
		try
		{
			dc.invoke( method, new Class[]
			{ Task.class }, new Object[]
			{ child } );
		}
		catch( Throwable t )
		{
			if( ! (this instanceof SynchronousTask) )
			{
				if (!(child instanceof SynchronousTask))
				{
					// Just display the error
					DIAGNOSTICS.println( "WARNING: Unhandled timeout from child-Task (" +
					child._id + ") - parent " + "Task = " +
					getClass().getName(), DIAGNOSTICS.LEVEL_3 );
				}
			}
		}
	}
	
	
 /**
	* Ensure that we indicate that the Task is completed, without requiring
	* a lock on the TM first (can lead to dealock since we need to check that
	* _done is false BEFORE sync'ing on _tm - if _done is true it is possible
	* that _tm is null, therefore we cannot safely sync. on it).
	*/
	boolean thisThreadFirstToIndicateDone()
	{
		boolean done_here = false;
		synchronized( _lock )
		{
			if( !_done )
			{
				// Ensure this task is indicated as being done
				_done = true;
				
				// Indicate the Task was "done" within this Thread
				done_here = true;
			}
		}
		
		return done_here;
	}
	
	
	
	// -------------------------------------------------------------------------
	
	/**
	 * This method can be implemented to provide a mechanism for dealing with incoming
	 * messages for which there is no performative handler available in the task subclass.
	 * The default implementation is to send a not-understood message. Override this method
	 * if you wish to add different handling such as learning how to handle the performative.
	 *
	 * @param conv The conversation containing the message to handle
	 **/
	protected void handleOther( Conversation conv )
	{
		if( !_tm._shutdown )
		{
			ACL acl = conv.getACL( conv.getLatestMessageIndex() );
			
			DIAGNOSTICS.println( "WARNING: Unhandled message type " + acl.getPerformative() +
			"  Task = " + getClass().getName(), DIAGNOSTICS.LEVEL_3 );
			
			if ( ! acl.getPerformative().equals( FIPACONSTANTS.NOT_UNDERSTOOD ) ) // don't reply to not-understood
			{
				try
				{
					_owner.sendNotUnderstood( acl, "internal-error unhandled-performative" );
				}
				catch( Throwable t )
				{
				}
			}
		}
	}
} // End Task
