//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved.
//
//******************************************************************************

package fipaos.agent.task;

//----------------------------------------------------------------------

import fipaos.util.DIAGNOSTICS;

import fipaos.agent.FIPAOSAgent; // TEMP: TODO: Remove when platform v2 is used
import fipaos.agent.conversation.*;
import fipaos.parser.acl.ACLMessage;
import fipaos.mts.*;
import fipaos.mts.service.*;
import fipaos.agent.profile.*;
import fipaos.agent.task.*;
import fipaos.agent.task.event.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.util.*;

import fipaos.agent.conversation.Conversation;

import java.util.*;

//----------------------------------------------------------------------

/**
 * The TaskManager class stores task objects with their associated conversation IDs for
 * an agent. The agent must add it's tasks during initialisation.
 * <br>Each task can be used to handle the messages for a particular conversation that an
 * agent participates in. For example. if an agent registers with a DF on startup, one task
 * object can be used to handle this conversation without having to store any state for
 * other conversations - even if they have the same protocol.
 *
 *
 **/
public class TaskManager implements ConversationListener
{
	//----------------------------------------------------------------------
	// INSTANCE VARIABLES
	//----------------------------------------------------------------------
	
	private static final String DF_SEARCH    = "df-search-task";
	private static final String IDLE = "IDLE";
	
	//----------------------------------------------------------------------
	// CLASS VARIABLES
	//----------------------------------------------------------------------
	
	/**
	 * Set of all TaskManagers
	 */
	private static Set __all_task_managers = new HashSet();
	
	//----------------------------------------------------------------------
	// INSTANCE VARIABLES
	//----------------------------------------------------------------------
	
	/**
	 * Task index table.  The table contains a map of conversation-id, index to TaskState objects
	 */
	private Map _ids_map;
	
	/**
	 * The set of all TaskState objects
	 */
	private Set _tasks_set;
	
	/**
	 * Set of timeouts being monitored
	 */
	private Set _timeouts_set;
	
	/**
	 * A List of all conversation ids' for which a Task has been assigned, but the
	 * Conversation hasn't ended - allows Tasks to terminate before the end of a Conversation
	 * without worrying about carrying out the rest of the Conversation (i.e. in a
	 * failure case).
	 */
	protected Set _conversation_ids = new TreeSet(  fipaos.util.Comparators.STRING_COMPARE );
	
	/**
	 * Owner of this
	 */
	private FIPAOSAgent _owner;
	
	/**
	 * Message sender of this
	 */
	private MessageSender _ms;
	
	/**
	 * Indicates if the TM has been shutdown
	 */
	boolean _shutdown = false;
	
	/**
	 * Order in which we should execute the next set of ready Task's
	 */
	private Queue _execution_order = new FIFOQueue();
	
	/**
	 * TaskManagerListener associated with this
	 */
	private TaskManagerListener _tm_listener;
	
	/**
	 * Task to substitue for parent-Task for top-level Tasks
	 */
	private Task _top_level_parent;
	
	/**
	 * ThreadPool used by this TM to dispatch events
	 */
	private ThreadPool _thread_pool;
	
	/**
	 * ThreadPool used by this TM to monitor Tasks
	 */
	private ThreadPool _thread_pool2;
	
	/**
	 * Profile used by this TaskManager
	 */
	private TaskManagerProfile _profile;
	
	//----------------------------------------------------------------------
	// CONSTRUCTOR(S)
	//----------------------------------------------------------------------
	
	/**
	 * Constructs an empty task manager. Tasks can be added via the newTask() methods.
	 *
	 * @param owner Reference to parent Agent
	 * @param ms MessageSender to send messages - this may be null (can be set
	 *						later via the setMessageSender() method).
	 **/
	public TaskManager( FIPAOSAgent owner, MessageSender ms, Task top_level_parent, TaskManagerProfile tmp )
	{
		// Save our profile
		_profile = tmp;
		
		// Set up the Map and the Set
		// Both are used to provide a level of indirection between the index and the tasks, This way
		// many keys can point to the same task object (e.g. multiple conversations being handled by the same
		// task object).
		_ids_map = new TreeMap(  fipaos.util.Comparators.STRING_COMPARE );
		_tasks_set = new HashSet();
		_timeouts_set = new HashSet();
		
		_ms = ms;
		_owner = owner;
		
		// Check if we need to start a thread pool or use the global thread-pool
		if( _profile.getUseGlobalThreads() )
		{
			_thread_pool = ThreadPool.getGlobalThreadPool();
		}
		else
		{
			_thread_pool = new ThreadPool( _profile.getLocalThreads(), owner.getAID().getName() + "-TM-ThreadPool" );
		}
		
		// Initialise Top-level parent
		_top_level_parent = top_level_parent;
		newTask( _top_level_parent );
		
		// Add check-state and check-timeout Jobs to mini thread-pool

		if (owner == null)
		{
			_thread_pool2 = new ThreadPool( 2, "TMControl-ThreadPool" );
		}
		else
		{
			_thread_pool2 = new ThreadPool( 2, owner.getAID().getName() + "-TMControl-ThreadPool" );
		}
		_thread_pool2.addJob( new TMCheckState() );
		_thread_pool2.addJob( new TMCheckTimeout() );
		
		// Add this to set of all TM's
		__all_task_managers.add( this );
	}
	
	
	//----------------------------------------------------------------------
	// PUBLIC METHODS
	//----------------------------------------------------------------------
	
	/**
	 * Shuts-down this TaskManager - prevents any new Tasks from starting
	 */
	public synchronized void shutdown()
	{
		DIAGNOSTICS.println( "shutdown()", this );
		
		__all_task_managers.remove( this );
		
		_shutdown = true;
		
		// Clear all references
		if( _thread_pool != ThreadPool.getGlobalThreadPool() )
		{
			_thread_pool.shutdown();
		}
		_thread_pool = null;
		_thread_pool2.shutdown();
		_thread_pool2 = null;
		_execution_order.releaseBlock();
		_execution_order = null;
		_owner = null;
		_ms = null;
		_ids_map.clear();
		_ids_map= null;
		_tasks_set.clear();
		_tasks_set = null;
		_timeouts_set.clear();
		_timeouts_set = null;
	}
	
	
	/**
	 * Initializes a new "top-level" task and adds it to the TaskManager.
	 */
	public void newTask( Task task )
	{
		DIAGNOSTICS.println( "newTask()", this );
		
		// Check that we haven't shutdown!
		if( !_shutdown )
		{
			// If we're not dealing with a Task with a parent,
			// and it isn't the _top_level_parent Task,
			// and it isn't a SynchronousTask then make _top_level_parent
			// the parent of this Task
			if( task._parent == null && task != _top_level_parent && !(task instanceof SynchronousTask))
			{
				task._parent = _top_level_parent;
				
				// Add the task to our list of child tasks, if not already added
				synchronized( _top_level_parent._lock )
				{
					if( !_top_level_parent._children.contains( task ) )
					{
						_top_level_parent._children.add( task );
					}
				}
			}
			
			synchronized( task._lock )
			{
				// set the task manager for the Task
				task._tm = this;
				
				// set the fipaos agent for the task
				task._owner = _owner;
				
				// Add to set of all Tasks
				_tasks_set.add( task );
			}
			
			if( task != _top_level_parent )
			{
				// Indicate task is ready
				addEvent( task, new InitialisedEvent() );
			}
			else
			{
				// Indicate top-level-Task is ready!
				task.getState().setState( TaskState.TASK_IDLE );
			}
		}
	}
	
	
	/**
	 * Creates a new Task with the given parent
	 */
	void newTask( Task task, Task parent )
	{
		if( !_shutdown )
		{
			synchronized( task._lock )
			{
				// Update internal references
				task._parent = parent;
				task._tm = this;
				task._owner = _owner;
			}
			
			newTask( task );
			
			// Add the task to our list of child tasks, if not already added
			synchronized( parent._lock )
			{
				if( !parent._children.contains( task ) )
				{
					parent._children.add( task );
				}
			}
		}
	}
	
	/**
	 * Creates a new Task with the given parent and the given alternate Task name
	 */
	void newTask(String name, Task task, Task parent )
	{
		if( !_shutdown )
		{
			synchronized( task._lock )
			{
				// Update internal references
				task._parent = parent;
				task._tm = this;
				task._owner = _owner;
				task._id = name;
			}
			
			newTask( task );
			
			// Add the task to our list of child tasks, if not already added
			synchronized( parent._lock )
			{
				if( !parent._children.contains( task ) )
				{
					parent._children.add( task );
				}
			}
		}
	}	
	
	/**
	 * Creates a new Task with the top level handler task as parent and the given alternate Task name
	 */
	public void newTask(String name, Task task)
	{
		if( !_shutdown )
		{
			synchronized( task._lock )
			{
				// Update internal references
				task._parent = _top_level_parent;
				task._tm = this;
				task._owner = _owner;
				task._id = name;
			}
			
			newTask( task );
			
			// Add the task to our list of child tasks, if not already added
			synchronized( _top_level_parent._lock )
			{
				if( !_top_level_parent._children.contains( task ) )
				{
					_top_level_parent._children.add( task );
				}
			}
		}
	}		
	
	/**
	 * Initializes a new "top-level" task and adds it to the TaskManager.  The task is associated with
	 * the given String.
	 */
	public void newTask( Task t, String id )
	{
	
		if( !_shutdown )
		{
			synchronized( t._lock )
			{
				// Add Task to TM
				_ids_map.put( id, t );
				
				// Add id to the list of conversation id's we're monitoring (assuming
				// id is not IDLE)
				if( !id.equalsIgnoreCase( IDLE ) )
				{
					_conversation_ids.add( id );
				}
				
				// Add TM id to Task
				t._tm_ids.add( id );
			}
			
			// Initialise Task
			newTask( t );
		}
	}
	
	
	/**
	 * Initializes a new "top-level" task and adds it to the TaskManager.  The task is associated with
	 * the given conversation.
	 */
	public void newTask( Task t, Conversation conv )
	{
		newTask( t, conv.getConversationID() );
	}
	
	
	/**
	 * Initializes a new child-task and adds it to the TaskManager.  The task is associated with
	 * the given conversation.
	 */
	public void newTask( Task task, Conversation conv, Task parent )
	{
		if( !_shutdown )
		{
			synchronized( task._lock )
			{
				// Update internal references
				task._parent = parent;
				task._tm = this;
				task._owner = _owner;
			}
			
			newTask( task, conv.getConversationID() );
			
			// Add the task to our list of child tasks, if not already added
			synchronized( parent._lock )
			{
				if( !parent._children.contains( task ) )
				{
					parent._children.add( task );
				}
			}
		}
	}
	
	/**
	 * Initializes a new child-task and adds it to the TaskManager.  The task is associated with
	 * the given conversation.
	 *
	 * The name parameter sets the alternate callback method
	 */
	public void newTask( String name, Task task, Conversation conv, Task parent )
	{
		if( !_shutdown )
		{
			synchronized( task._lock )
			{
				// Update internal references
				task._parent = parent;
				task._tm = this;
				task._owner = _owner;
				task._id = name;
			}
			
			newTask( task, conv.getConversationID() );
			
			// Add the task to our list of child tasks, if not already added
			synchronized( parent._lock )
			{
				if( !parent._children.contains( task ) )
				{
					parent._children.add( task );
				}
			}
		}
	}	
	
	
	/** Initialises the given task, and invokes its start() method.  If the task doesn't invoke
	 * done() in the given amount of time, a timeoutX() method will be invoked on
	 * the parent-Task, where X is the id of the task
	 *
	 * @param timeout The time in milliseconds to wait before timing out
	 * @param task The new Task to be initialised and added to the system
	 * @param conv Conversation to be associated with this Task
	 */
	public void newTask( Task task, long timeout )
	{
		if( !_shutdown )
		{
			synchronized( task._lock )
			{
				task._timeout = timeout;
			}
			
			newTask( task );
		}
	}
	
	
	/** Initialises the given task, and invokes its start() method.  If the task doesn't invoke
	 * done() in the given amount of time, a timeoutX() method will be invoked on
	 * the parent-Task, where X is the id of the task
	 *
	 * @param timeout The time in milliseconds to wait before timing out
	 * @param task The new Task to be initialised and added to the system
	 * @param conv Conversation to be associated with this Task
	 */
	void newTask( Task task, long timeout, Task parent )
	{
		if( !_shutdown )
		{
			// Update internal references
			synchronized( task._lock )
			{
				task._parent = parent;
				task._tm = this;
				task._owner = _owner;
			}
			
			newTask( task, timeout );
			
			// Add the task to our list of child tasks, if not already added
			synchronized( parent._lock )
			{
				if( !parent._children.contains( task ) )
				{
					parent._children.add( task );
				}
			}
		}
	}
	
	
	/** Initialises the given task, and invokes its start() method.  If the task doesn't invoke
	 * done() in the given amount of time, a timeoutX() method will be invoked on
	 * the parent-Task, where X is the id of the task
	 *
	 * @param timeout The time in milliseconds to wait before timing out
	 * @param task The new Task to be initialised and added to the system
	 * @param conv Conversation to be associated with this Task
	 */
	public void newTask( Task task, Conversation conv, long timeout )
	{
		if( !_shutdown )
		{
			synchronized( task._lock )
			{
				task._timeout = timeout;
			}
			
			newTask( task, conv );
		}
	}
	
	
	/**
	 * Initializes a new "top-level" task and adds it to the TaskManager.  The task is associated with
	 * the passed in conversation.
	 */
	public void newTask( Task task, Conversation conv, long timeout, Task parent )
	{
		if( !_shutdown )
		{
			synchronized( task._lock )
			{
				// Update internal references
				task._parent = parent;
				task._tm = this;
				task._owner = _owner;
			}
			
			newTask( task, conv, timeout );
			
			// Add the task to our list of child tasks, if not already added
			synchronized( parent._lock )
			{
				if( !parent._children.contains( task ) )
				{
					parent._children.add( task );
				}
			}
		}
	}
	
	
	/**
	 * Removes a task object associated with the specified conversation id from the manager.
	 *
	 * @param key The conversation id of the task to be removed.
	 * @return True if the conversation id and task were removed, false otherwise
	 **/
	public boolean removeTask( String key )
	{
		if( !_shutdown )
		{
			// Task to be removed
			Task task = null;
			
			// Get a reference to the Task to be removed
			synchronized( this )
			{
				// Check such a Task exists
				if (_ids_map.containsKey(key))
				{
					task = (Task) _ids_map.get( key );
				}
			}
			
			// Remove Task if we've found it
			if( task != null )
			{
				return removeTask( task );
			}
			else
			{
				return true;
			}
		}
		
		return false;
	}
	
	
	/**
	 * Removes a task object from the manager.
	 * Any conversation bound to the task will be removed as well.
	 *
	 * @param task Task to be removed
	 * @return True if the task was removed, false otherwise
	 **/
	public boolean removeTask( Task task )
	{
		DIAGNOSTICS.println( "removeTask( " + task + " )", this );
		
		if( !_shutdown )
		{
			// Reference to a duplicate of the set in the given Task
			Set duplicate_tm_ids = null;
			
			// Duplicate _tm_ids in Task
			synchronized( task._lock )
			{
				if( task._tm_ids != null )
				{
					duplicate_tm_ids = new TreeSet(  fipaos.util.Comparators.STRING_COMPARE );
					duplicate_tm_ids.addAll( task._tm_ids );
				}
				else
				{
					duplicate_tm_ids = new TreeSet( fipaos.util.Comparators.STRING_COMPARE );
				}
			}
			
			// Try to remove all mapping of conversation ids and task.
			boolean result = true;
			
			synchronized( this )
			{
				// Iterate through the known ID's of the task
				Iterator i = duplicate_tm_ids.iterator();
				while( i.hasNext() )
				{
					// Get next id
					String id = (String) i.next();
					
					// Remove it from our _ids_map
					result &= (_ids_map.remove( id ) != null );
				}
			}
			
			// Clean up references in Task
			task.removeReferences();
			
			synchronized( this )
			{
				// try to remove the task from the list of all tasks.
				return _tasks_set.remove(task) && result;
			}
		}
		
		return false;
	}
	
	
	/**
	 * Get info. about number of active tasks managed by TM
	 */
	public synchronized int getTaskCount()
	{
		return _tasks_set.size();
	}
	
	
	/**
	 * Get info. about number of timeouts currently being monitored
	 */
	public int getTimeoutCount()
	{
		synchronized( _timeouts_set )
		{
			return _timeouts_set.size();
		}
	}
	
	
	/**
	 * Get info. about the number of pending TaskEvents
	 */
	public synchronized int getPendingTaskEvents()
	{
		return _execution_order.queueLength();
	}
	
	
	/**
	 * Get the name of the owning Agent
	 */
	public String toString()
	{
		return _owner.getAID().getName();
	}
	
	//----------------------------------------------------------------------
	// PUBLIC CLASS METHODS
	//----------------------------------------------------------------------
	
	/**
	 * Get set of all TM's in VM
	 */
	public static Set getVMTaskManagers()
	{
		return new HashSet( __all_task_managers );
	}
	
	//----------------------------------------------------------------------
	// Inter-Task methods
	//----------------------------------------------------------------------
	
	/**
	 * Adds the given TaskEvent to the given Task, and informs the registered
	 * listener
	 */
	void addEvent( Task task, TaskEvent event )
	{
		DIAGNOSTICS.println( "addEvent()", this );
		
		boolean added_event = false;
		
		synchronized( task._lock )
		{
			if( !task._done )
			{
				// Add the event to the task
				task.getState().addEvent( event );
				added_event = true;
			}
			else
			{
				DIAGNOSTICS.println( "Couldn't add " + event.getClass().getName() + " to " +
				task.getClass().getName() + ", task has completed",
				this, DIAGNOSTICS.LEVEL_3 );
			}
		}
		
		synchronized( this )
		{
			// Check if there is a TM listener
			if( added_event && _tm_listener != null )
			{
				_tm_listener.taskEvent( task, event );
			}
		}
	}
	
	//----------------------------------------------------------------------
	// FIPAOSAgent access methods
	//----------------------------------------------------------------------
	
	/** <p>Sends an ACL message on behalf of the given Task.
	 *
	 * <p>If the message is part of a new conversation, a Conversation object will
	 * be created within the ConversationManager and the Task will be bound to that
	 * Conversation's ID - hence the task will subsequently receive any messages in response
	 * automatically.
	 * Note: a call to this method will update the table of all conversation ids and related tasks so
	 * that the conversation id is always associated with the last task that calls this method.
	 * @param acl The ACL message to send
	 * @param task
	 */
	void forward( ACL acl, Task task )
	{
		forward( new Message( acl ), task );
	}
	
	/** <p>Sends a Message on behalf of the given Task.
	 *
	 * <p>If the message is part of a new conversation, a Conversation object will
	 * be created within the ConversationManager and the Task will be bound to that
	 * Conversation's ID - hence the task will subsequently receive any messages in response
	 * automatically.
	 * Note: a call to this method will update the table of all conversation ids and related tasks so
	 * that the conversation id is always associated with the last task that calls this method.
	 * @param msg The Message to send
	 * @param task
	 */
	void forward( Message msg, Task task )
	{
		DIAGNOSTICS.println( "forward()", this );
		MessageSender ms = null;
		FIPAOSAgent owner = null;
		ACL acl = msg.getACL();
		
		String conv_id = acl.getConversationID();
		String protocol = acl.getProtocol();
		
		// If id is null, then create new conversation and  add the (id, task) to the list of ids.
		if (conv_id == null)
		{
			// Here the ACLMessage is the first of a conversation.  Then we nee to:
			
			// 1 - create a new conversation in the conversation manager and get the conversation id.
			conv_id = (createNewConversation(protocol)).getConversationID();
			
			// 2 - associate the new ACLMessage with the newly created conversation.
			acl.setConversationID(conv_id);
		}
		
		// Add time-stamp
		msg.getEnvelope().setSlotValue( Service.FIPAOS_DEBUG + "tm-send-time",
		new Long( System.currentTimeMillis() ) );
		
		
		// Update our internal variables
		synchronized( this )
		{
			// Always update the mapping of conversation ids and bound tasks.
			_ids_map.put( conv_id, task );
			task._tm_ids.add( conv_id );
			
			// Add to the set of Conversation ID's being monitored
			_conversation_ids.add( conv_id );
			
			// Store reference to be used to send a message
			ms = _ms;
			owner = _owner;
		}
		
		
		// Forward the message
		if( ms != null )
		{
			ms.sendMessage( msg );
		}
		else if( owner != null )
		{
			owner.forward( msg );
		}
		else
		{
			DIAGNOSTICS.println( "Couldn't send message for " + task + " - no way to send!", this );
		}
		
		
		// Check if the conversation has completed
		checkConversationEnded( conv_id );
	}
	
	
	/**
	 * Sets up a new Conversation using the given protocol
	 */
	ACL getNewConversation( String protocol )
	{
		DIAGNOSTICS.println( "getNewConversation()", this );
		Conversation conv = createNewConversation( protocol );
		
		if( conv != null )
		{
			return conv.getFilledInACL();
		}
		
		return null;
	}
	
	
	/**
	 * Gets an ACL message that is pre-filled in for a new conversation of the specified
	 * protocol.
	 *
	 * @return The new ACL message
	 **/
	private Conversation createNewConversation( String protocol )
	{
		DIAGNOSTICS.println( "createNewConversation()", this );
		
		try
		{
			return _owner.getNewConversation( protocol );
		}
		catch ( fipaos.agent.conversation.UnknownProtocolException upe )
		{
			DIAGNOSTICS.println( "Exception occured when creating new Conversation:\n"+ upe,
			this, DIAGNOSTICS.LEVEL_MAX );
		}
		
		return null;
	}
	
	
	//----------------------------------------------------------------------
	// PUBLIC METHODS
	//----------------------------------------------------------------------
	
	/**
	 * Append a Task to the execution order (Tasks are executed in the
	 * order they are added to the execution order).
	 */
	public synchronized void appendToExecutionOrder( Task t )
	{
		DIAGNOSTICS.println( "appendToExecutionOrder()", this );
		if( !_shutdown )
		{
			_execution_order.addItem( t );
		}
	}
	
	
	/**
	 * List the active Tasks
	 */
	public synchronized Set getActiveTasks()
	{
		DIAGNOSTICS.println( "getActiveTasks()", this );
		if( !_shutdown )
		{
			Set s = new TreeSet( fipaos.util.Comparators.STRING_COMPARE );
			s.addAll( _tasks_set );
			return s;
		}
		else
		{
			return new TreeSet();
		}
	}
	
	
	/**
	 * Setup the TaskManageristener
	 */
	public synchronized void setTaskManagerListener( TaskManagerListener tml )
	{
		DIAGNOSTICS.println( "setTaskManagerListener()", this );
		if( !_shutdown )
		{
			_tm_listener = tml;
		}
	}
	
	
	//----------------------------------------------------------------------
	// PRIVATE METHODS
	//----------------------------------------------------------------------
	
	/**
	 * Checks the state of the active Task's within the TM, and invoked them
	 * where appropriate
	 */
	private void checkState()
	{
		// Check if we've shutdown
		if( _shutdown )
		{
			return;
		}
		
		//	DIAGNOSTICS.println( "checkState()", this );
		// Iterate through the tasks to be activated the TM, and activate them
		Task next = (Task) _execution_order.getItem();
		Task last = null;
		
		while( next != null && !_shutdown )
		{
			// Next event to pass to Task
			TaskEvent next_event = null;
			
			// Indicates if its OK to do so
			boolean ok_to_deal = false;
			boolean back_off = false;
			
			synchronized( this )
			{
				synchronized( next._lock )
				{
					// Get next Event to pass to Task
					next_event = next.getState().getNextEvent();
					
					// Check that it is still a valid choice
					if( !next._done && _tasks_set.contains( next ) )
					{
						if( next_event != null )
						{
							// Can deal with next event now :)
							ok_to_deal = true;
						}
						else
						{
							// Can't deal with it now for
							// some reason!?!?
							DIAGNOSTICS.println( "NOT dispatching event now, Task " +
							"is busy(?) - will retry later", this );
							
							back_off = true;
						}
					}
					else
					{
						// Task isn't alive anymore?!?!?
						DIAGNOSTICS.println( "NOT dispatching event " +
						next_event + " to " + next + " - " +
						"next._done = " + next._done +
						"; _tasks_set.contains( next ) = " +
						_tasks_set.contains( next ), this );
					}
				}
			}
			
			
			// Back off if necessary
			if( back_off )
			{
				// Back off for a little while
				try
				{
					Thread.sleep( 250 );
				}
				catch( InterruptedException e )
				{
				}

				// Repeat while loop iteration again
				continue;
			}
			
			// OK to carry out event!
			if( ok_to_deal  && !_shutdown )
			{
				// Invoke a method based upon the next event type
				dealWithTask( next, next_event );
			}
			
			
			if( !_shutdown )
			{
				last = next;
				
				// Get next Task to execute
				next = (Task) _execution_order.getItem();
			}
		}
	}
	
	
	/**
	 * Invoked when the current set of timeouts should be checked
	 */
	private void checkTimeout()
	{
		// Check if we've shutdown
		if( _shutdown )
		{
			return;
		}
		
		// Iterate through contents of Set, acting as necessary
		List tasks = new LinkedList();
		synchronized( _timeouts_set )
		{
			DIAGNOSTICS.println( "Checking " + _timeouts_set.size() + " timeouts..." );
			
			Iterator i = _timeouts_set.iterator();
			while( i.hasNext() )
			{
				// Get next Timeout
				Timeout t = (Timeout) i.next();
				
				// Check its state
				if( t.isDone() )
				{
					// We can remove it from the set
					i.remove();
				}
				else if( t.hasTimedout() )
				{
					// Record that we need to notify the task it has timed out
					tasks.add( t.getTask() );
					
					// We can remove it from the set
					i.remove();
				}
			}
		}
		
		DIAGNOSTICS.println( "Found " + tasks + " Task's that have timed-out" );
		
		// Now notify all tasks that have timed out
		Iterator j = tasks.iterator();
		while( j.hasNext() )
		{
			Task t = (Task) j.next();
			t.timeoutReached();
		}
	}
	
	
	/**
	 * Deals with actually invoking methods appropriate to the given event in
	 * the given Task
	 */
	private void dealWithTask( final Task task, final TaskEvent event )
	{
		DIAGNOSTICS.println( "dealWithTask()", this );
		DIAGNOSTICS.println( "Dispatching event " + event + " to " + task, this );
		
		// Indicate task is executing
		synchronized( task )
		{
			task.getState().setState( TaskState.TASK_EXECUTING );
		}
		
		// Work out which event type, and take appropriate action
		if( event instanceof ChildDoneWithResultEvent )
		{
			_thread_pool.addJob( new ThreadPool.Job()
			{
				public void go()
				{
					// Cast to relevant type
					ChildDoneWithResultEvent e = (ChildDoneWithResultEvent) event;
					
					// Notify task
					task.notifyDone( e.getChildTask(), e.getResult() );
					
					// Now remove child-Task
					removeTask( e.getChildTask() );
					
					// Execute postEventDistach
					postEventDispatch( task, event );
				}
			});
		}
		else if( event instanceof ChildDoneEvent )
		{
			_thread_pool.addJob( new ThreadPool.Job()
			{
				public void go()
				{
					if( !_shutdown )
					{
						// Cast to relevant type
						ChildDoneEvent e = (ChildDoneEvent) event;
						
						// Notify task
						task.notifyDone( e.getChildTask() );
						
						// Now remove child-Task
						removeTask( e.getChildTask() );
						
						// Execute postEventDistach
						postEventDispatch( task, event );
					}
				}
			});
		}
		else if( event instanceof ChildFailureEvent )
		{
			_thread_pool.addJob( new ThreadPool.Job()
			{
				public void go()
				{
					if( !_shutdown )
					{
						// Cast to relevant type
						ChildFailureEvent e = (ChildFailureEvent) event;
						
						// Notify task
						task.notifyError( e.getChildTask(), e.getFailureReason() );
						
						// Now remove child-Task
						removeTask( e.getChildTask() );
						
						// Execute postEventDistach
						postEventDispatch( task, event );
					}
				}
			});
		}
		else if( event instanceof ChildTimeoutEvent )
		{
			_thread_pool.addJob( new ThreadPool.Job()
			{
				public void go()
				{
					if( !_shutdown )
					{
						// Cast to relevant type
						ChildTimeoutEvent e = (ChildTimeoutEvent) event;
						
						// Notify task
						task.notifyTimeout( e.getChildTask() );
						
						// Now remove child-Task
						removeTask( e.getChildTask() );
						
						// Execute postEventDistach
						postEventDispatch( task, event );
					}
				}
			});
		}
		else if( event instanceof ConversationUpdateEvent )
		{
			_thread_pool.addJob( new ThreadPool.Job()
			{
				public void go()
				{
					if( !_shutdown )
					{
						// Cast to relevant type
						ConversationUpdateEvent e = (ConversationUpdateEvent) event;
						
						// Notify task
						task.notify( e.getUpdatedConversation() );
						
						// Check if the Conversation ended
						checkConversationEnded( e.getUpdatedConversation() );
						
						// Execute postEventDistach
						postEventDispatch( task, event );
					}
				}
			});
		}
		else if( event instanceof MessageForEndedConversationEvent )
		{
			_thread_pool.addJob( new ThreadPool.Job()
			{
				public void go()
				{
					if( !_shutdown )
					{
						// Cast to relevant type
						MessageForEndedConversationEvent e = (MessageForEndedConversationEvent) event;
						
						// Notify task
						task.notifyMessageForEndedConversation( e.getMessage() );
						
						// Execute postEventDistach
						postEventDispatch( task, event );
					}
				}
			});
		}
		else if( event instanceof MessageDeliveryFailureEvent )
		{
			_thread_pool.addJob( new ThreadPool.Job()
			{
				public void go()
				{
					if( !_shutdown )
					{
						// Cast to relevant type
						MessageDeliveryFailureEvent e = (MessageDeliveryFailureEvent) event;
						
						// Notify task
						task.notifyMessageFailure( e.getMessage() );
						
						// Execute postEventDistach
						postEventDispatch( task, event );
					}
				}
			});
		}
		else if( event instanceof InitialisedEvent )
		{
			_thread_pool.addJob( new ThreadPool.Job()
			{
				public void go()
				{
					if( !_shutdown )
					{
						// Start the task
						task.notifyStart();
						
						// Execute postEventDistach
						postEventDispatch( task, event );
						
					}
				}
			});
			
			synchronized( task._lock )
			{
				// Start timeout if required
				if( !task._done && task._timeout > 0 )
				{
					synchronized( _timeouts_set )
					{
						_timeouts_set.add( new Timeout( task._timeout, task ) );
					}
				}
			}
		}
		else
		{
			DIAGNOSTICS.println( "Unknown event type " + event.getClass().getName() + ", ignoring" );
		}
	}
	
	/**
	 * Post-event dispatch handler
	 */
	private void postEventDispatch( Task task, TaskEvent event )
	{
		// Update the state of the Task if necessary
		if( event != null )
		{
			if( !( event instanceof InitialisedEvent ) )
			{
				synchronized( task._lock )
				{
					if( !task._done && task._state != null )
					{
						task._state.setState( TaskState.TASK_IDLE );
					}
				}
			}
			
			// Clean up TaskEvent
			event.removeReferences();
		}
	}
	
	/**
	 * Check if Conversation is finished - if so we should remove it from
	 * _conversation_ids
	 */
	protected void checkConversationEnded( Conversation conv )
	{
		try
		{
			// If we reach here and conv != null everything went OK - check if we've
			// reached the end of the conversation
			if( conv != null )
			{
				if( conv.getState() == Conversation.CONVERSATION_END )
				{
					// Remove conversation since it is over now
					DIAGNOSTICS.println( "Conversation " + conv.getConversationID() + " appears to have ended, removing reference", this );
					_conversation_ids.remove( conv.getConversationID() );
				}
			}
			else
			{
				// Couldn't find conversation?  It must have completed - remove it!
				DIAGNOSTICS.println( "Conversation " + conv.getConversationID() + " appears to have ended, removing reference", this );
				_conversation_ids.remove( conv.getConversationID() );
			}
		}
		catch( Throwable t )
		{
			// Couldn't find conversation?  It must have completed - remove it!
			DIAGNOSTICS.println( "Conversation " + conv.getConversationID() + " appears to have ended, removing reference", t, this );
			_conversation_ids.remove( conv.getConversationID() );
		}
	}
	
	
	/**
	 * Check if Conversation is finished - if so we should remove it from
	 * _conversation_ids
	 */
	private void checkConversationEnded( String conv_id )
	{
		DIAGNOSTICS.println( "checkConversationEnded()", this );
		FIPAOSAgent owner = null;
		
		// Ensure we can get a reference to our owner, without maintaining a lock
		// on this
		synchronized( this )
		{
			owner = _owner;
		}
		
		if( owner != null )
		{
			try
			{
				Conversation conv = owner.getCurrentConversation( conv_id );
				
				checkConversationEnded( conv );
			}
			catch( Throwable t )
			{
				// Couldn't find conversation?  It must have completed - remove it!
				DIAGNOSTICS.println( "Conversation " + conv_id + " appears to have ended, removing reference", t, this );
				_conversation_ids.remove( conv_id );
			}
		}
	}
	
	
	//----------------------------------------------------------------------
	// PUBLIC METHODS - ConversationListener interface
	//----------------------------------------------------------------------
	
	/**
	 * Conversation listener method to deal with conversations being sent by the
	 * planner scheduler.  These should be overriden as default by the sub-class
	 * if necessary
	 *
	 * @param conv The updated conversation to deal with
	 */
	public void notify( Conversation conv )
	{
		DIAGNOSTICS.println( "notify()", this );
		// Not-understood reason - if non-null, a not-understood will be sent
		// with the given reason
		String not_understood_reason = null;
		
		// Target Task for the Conversation update
		Task task = null;
		
		// Get the conv.id of the Conversation
		String id = conv.getConversationID();
		
		// Check that we haven't shutdown!
		if( !_shutdown )
		{
			// Check that its OK to push (blocks if it isn't)
			_owner.waitForPushLock();
		}
		
		synchronized( this )
		{
			if( !_shutdown )
			{
				// Task already associated with Conversation?
				if( _ids_map.containsKey( id ) )
				{
					task = (Task) _ids_map.get( id );
				}
				// Have we previously dealt with this Conversation?
				else if( _conversation_ids.contains( id ) )
				{
					not_understood_reason = "internal-error Task-no-longer-associated-with-conversation";
				}
				// Never seen this Conversation before?
				// Have we got an IDLE task to deal with it?
				else if( _ids_map.containsKey( IDLE ) )
				{
					task = (Task) _ids_map.get( IDLE );
				}
			}
			else
			{
				DIAGNOSTICS.println( "Received updated conversation, but TM has been shutdown!",
				this, DIAGNOSTICS.LEVEL_MAX );
				return;
			}
		}
		
		
		// Do we have a Task to deal with this event?
		if( task != null )
		{
			if( !task._done )
			{
				// Create TaskEvent + add to appropriate Task
				TaskEvent te = new ConversationUpdateEvent( conv );
				addEvent( task, te );
			}
			else
			{
				// Can't pass update to Task, its finished - better send a not-understood
				not_understood_reason = "internal-error task-completed";
			}
		}
		// No Task & no failure reason? - Better send a not-understood
		else if( not_understood_reason == null )
		{
			DIAGNOSTICS.println( "Received conversation update - no IDLE task " +
			"registered in agent - cannot handle conversation",
			this, DIAGNOSTICS.LEVEL_MAX );
			
			not_understood_reason = "internal-error no-idle-task";
		}
		
		// Send not-understood?
		if( not_understood_reason != null &&
		!conv.getACL( conv.getLatestMessageIndex() ).getPerformative().equals( FIPACONSTANTS.NOT_UNDERSTOOD ) )
		{
			// Reply with a not-understood, with an appropriate error-reason
			_owner.sendNotUnderstood( conv.getACL( conv.getLatestMessageIndex() ),
			not_understood_reason );
			
			// Check to see if the conversation compeleted
			checkConversationEnded( id );
		}
	}
	
	
	/**
	 * Deals with multiple conversations at once - just calls notify( Conversation )
	 * for each one.  Override in sub class to replace this behavior. ( NOTE: When
	 * overriding this method, it should be noted that care should be taken because
	 * this method could be invoked at any time after the FIPAOSAgent()  constructor
	 * had been used.  Normally this method doesn't do anything until after startPushing()
	 * is invoked).
	 *
	 * @param enum An enumeration of updated conversations to deal with
	 */
	public void notify( Enumeration enum )
	{
		if( !_shutdown )
		{
			// Pass to other notify() method
			while( enum.hasMoreElements() )
			{
				notify( (Conversation) enum.nextElement() );
			}
		}
	}
	
	
	/**
	 * Sets the MessageSender that this should use
	 */
	public void setMessageSender( MessageSender ms )
	{
		_ms = ms;
	}
	
	/**
	 * Indicates that the given Message arrived for a completed Conversation - pass it to IDLE task
 */
	public void notifyMessageInEndedConversation(Message msg)
	{
		DIAGNOSTICS.println( "notifyMessageInEndedConversation()", this );
		
		// Target Task for the event
		Task task = null;
		
		// Check that we haven't shutdown!
		if( !_shutdown )
		{
			// Check that its OK to push (blocks if it isn't)
			_owner.waitForPushLock();
		}
		
		synchronized( this )
		{
			if( !_shutdown )
			{
				// Have we got an IDLE task to deal with it?
				if( _ids_map.containsKey( IDLE ) )
				{
					task = (Task) _ids_map.get( IDLE );
				}
				else
				{
					DIAGNOSTICS.println( "No Task to deal with message in ended conversation - ignoring" );
					return;
				}
			}
			else
			{
				DIAGNOSTICS.println( "Received message in ended conversation, but TM has been shutdown!",
				this, DIAGNOSTICS.LEVEL_MAX );
				return;
			}
		}
		
		
		// Do we have a Task to deal with this event?
		if( task != null )
		{
			if( !task._done )
			{
				// Create TaskEvent + add to appropriate Task
				TaskEvent te = new MessageForEndedConversationEvent( msg );
				addEvent( task, te );
			}
			else
			{
				DIAGNOSTICS.println( "Task to deal with message delivery failure has ended - ignoring" );
				return;
			}
		}
	}
	
	/**
	 * Indicates that the given Message was not deliverable
	 * <p>NOTE: This is a new feature which hasn't been fully tested
 	 */
	public void notifyMessageUndeliverable(Message msg)
	{
		DIAGNOSTICS.println( "notifyMessageUndeliverable()", this );
		
		// Target Task for the event
		Task task = null;
		
		// Get the conv.id of the Conversation
		String id = msg.getACL().getConversationID();
		
		// Check that we haven't shutdown!
		if( !_shutdown )
		{
			// Check that its OK to push (blocks if it isn't)
			_owner.waitForPushLock();
		}
		
		synchronized( this )
		{
			if( !_shutdown )
			{
				// Task already associated with Conversation?
				if( _ids_map.containsKey( id ) )
				{
					task = (Task) _ids_map.get( id );
				}
				// Never seen this Conversation before?
				// Have we got an IDLE task to deal with it?
				else if( _ids_map.containsKey( IDLE ) )
				{
					task = (Task) _ids_map.get( IDLE );
				}
				else
				{
					DIAGNOSTICS.println( "No Task to deal with message delivery failure - ignoring" );
					return;
				}
			}
			else
			{
				DIAGNOSTICS.println( "Received delivery failure, but TM has been shutdown!",
				this, DIAGNOSTICS.LEVEL_MAX );
				return;
			}
		}
		
		
		// Do we have a Task to deal with this event?
		if( task != null )
		{
			if( !task._done )
			{
				// Create TaskEvent + add to appropriate Task
				TaskEvent te = new MessageDeliveryFailureEvent( msg );
				addEvent( task, te );
			}
			else
			{
				DIAGNOSTICS.println( "Task to deal with message delivery failure has ended - ignoring" );
				return;
			}
		}
	}
	
	//----------------------------------------------------------------------
	// INNER CLASSES
	//----------------------------------------------------------------------
	
	/**
	 * Default implementation of the TM Listener - just adds events back to TM
	 * in the order they arrive
	 */
	public static class DefaultTaskManagerListener implements TaskManagerListener
	{
		private TaskManager _tm;
		
		public DefaultTaskManagerListener( TaskManager tm )
		{
			_tm = tm;
			_tm.setTaskManagerListener( this );
		}
		
		public void taskEvent( Task t, TaskEvent e )
		{
			DIAGNOSTICS.println( "taskEvent()", this );
			//			DIAGNOSTICS.println( "Event of type " + e.getClass().getName() + " arrived for " +
			//														t, this );
			_tm.appendToExecutionOrder( t );
		}
	}
	
	/**
	 * Inner-class to represent a timeout
	 */
	private class Timeout
	{
		private long _timeout_time;
		private Task _task;
		
		public Timeout( long timeout, Task t )
		{
			_timeout_time = System.currentTimeMillis() + timeout;
			_task = t;
		}
		
		public boolean hasTimedout()
		{
			return ( System.currentTimeMillis() >= _timeout_time );
		}
		
		public boolean isDone()
		{
			return _task._done;
		}
		
		public Task getTask()
		{
			return _task;
		}
	}
	
	
	/**
	 * Job to be started at TM start
	 */
	private class TMCheckState implements ThreadPool.Job
	{
		/**
		 * Method invoked when the TM starts
		 */
		public void go()
		{
			// While TM isn't shutdown
			while( !_shutdown )
			{
				// Let the thread sleep
				try
				{
					// Check state of Tasks
					checkState();
					Thread.sleep( 1000 );
				}
				catch( Throwable t )
				{
					DIAGNOSTICS.println( "Unexpected exception in go()", t, this, DIAGNOSTICS.LEVEL_MAX );
				}
			}
		}
	}
	
	
	/**
	 * Job to check timeouts
	 */
	private class TMCheckTimeout implements ThreadPool.Job
	{
		/**
		 * Method invoked when the TM starts
		 */
		public void go()
		{
			while( !_shutdown )
			{
				// Let the thread sleep
				try
				{
					// Check state of Tasks
					checkTimeout();
					Thread.sleep( 250 );
				}
				catch( Throwable t )
				{
					DIAGNOSTICS.println( "Unexpected exception in go()", t, this, DIAGNOSTICS.LEVEL_MAX );
				}
			}
		}
	}
} // End TaskManager

