//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files ("the
//     Programs") are Copyright (c) 1999 - 2000 Nortel Networks Corporation, 
//     8200 Dixie Road, Suite 100, Brampton, Ontario, Canada L6R 5P6.  
//     All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out in
//     the Nortel Networks FIPA-OS Public License, a copy of which can be found
//     in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version can
//     also be found at http://fipa-os.sourceforge.net/. By downloading the
//     files containing the Programs you accept the terms and  conditions of the
//     Public License. You do not have to accept these terms and conditions, but
//     unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.agent.task;

//----------------------------------------------------------------------

import java.util.*;
import fipaos.agent.task.event.*;

//----------------------------------------------------------------------

/**
 * Represents the current state of a particular Task within the TaskManager
 */
public class TaskState
{
	/**
	 * Indicates that a Task is not executing and has no pending events
	 */
 	public static final int TASK_IDLE = 0;

	/**
	 * Indicates that a Task is currently executing
	 */
	public static final int TASK_EXECUTING = 1;
	
	/**
	 * Indicates that a Task is ready to start initial execution
	 */
	public static final int TASK_READY_TO_START = 2;
	
	/**
	 * Indicates that a Task is ready to continue execution (events pending)
	 */
	public static final int TASK_READY_TO_CONTINUE = 3;
	
	/**
	 * Pending events for this Task
	 */
	private List _events = new LinkedList();

	/**
	 * Current state
	 */
	private int _state = TASK_READY_TO_START;
	
	/**
	 * Get the state of the Task
	 */
	public synchronized int getState()
	{
		if( _state == TASK_READY_TO_START ||
				_state == TASK_EXECUTING )
		{
			return _state;
		}
		else if( _events.size() > 0 )
		{
			return TASK_READY_TO_CONTINUE;
		}
		else
		{
			return TASK_IDLE;
		}
	}
	
	/**
	 * Set the current state of the Task this belongs to
	 */
	synchronized void setState( int state )
	{
		_state = state;
	}
	
	/**
	 * <p>Peek at the next TaskEvent that the Task this belongs to should receive
	 * @return Return null if getState() != TASK_READY_TO_CONTINUE
	 */
	public synchronized TaskEvent peekNextEvent()
	{
		if( getState() != TASK_READY_TO_CONTINUE )
		{
			return null;
		}
		else
		{
			return (TaskEvent) _events.get( 0 );
		}
	}
	
	
	/**
	 * <p>Get the next TaskEvent that the Task this belongs to should receive - this
	 * removes the event from the pending list
	 * @return Return null if getState() != TASK_READY_TO_CONTINUE
	 */
	public synchronized TaskEvent getNextEvent()
	{
		if( getState() != TASK_READY_TO_CONTINUE &&
				getState() != TASK_READY_TO_START )
		{
			return null;
		}
		else
		{
			return (TaskEvent) _events.remove( 0 );
		}
	}
	
	/**
	 * <p>Adds a TaskEvent that the Task this belongs to should receive
	 */
	synchronized void addEvent( TaskEvent event )
	{
		_events.add( event );
	}
	
	/**
	 * Removes all references to other objects to aid garbage collection
	 */
	void removeReferences()
	{
		if( _events != null )
		{
			Iterator i = _events.iterator();
			while( i.hasNext() )
			{
				TaskEvent event = (TaskEvent) i.next();
				event.removeReferences();
			}
			
			_events.clear();
			_events = null;
		}
	}
} // End class
