//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.mts;

//----------------------------------------------------------------------

import fipaos.mts.service.Service;
import fipaos.mts.service.PreParserService;
import fipaos.ont.fipa.fipaman.Envelope;
import fipaos.ont.fipa.fipaman.AgentID;
import fipaos.util.URL;
import fipaos.util.DIAGNOSTICS;

//----------------------------------------------------------------------

/**
 * Provides a basic implementation of a MTP, based upon the MTPBase
 * object and implementing the InternalMTP interface
 * <br>
 * <br>
 */
public abstract class InternalMTPBase extends MTPBase implements InternalMTP
{
//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// STATIC VARIABLES
//----------------------------------------------------------------------
	
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

	/**
	 * Previous PreParserService in the stack
	 */
	private PreParserService _previous;
	
//----------------------------------------------------------------------
// CONSTRUCTOR(S)
//----------------------------------------------------------------------

	/**
	 * Initialises the MTP variables (i.e. _hap, __name & _ns_url), and
	 * then invokes bind()
	 */
	protected InternalMTPBase( String name, URL ns ) throws InitialisationException
	{
		super( name, ns );
	}

//----------------------------------------------------------------------
// PUBLIC METHODS - InternalMTP implementation
//----------------------------------------------------------------------	
	
	/** Invoked when the undelying implementation should initialise itself.  Allows
	 * references to other PreParserService's before and after it in the stack
	 * to be passed
	 * @param previous Previous PreParserService instance (Incoming messages should be passed to this)
	 * @param next Next PreParserService instance (Outgoing messages should be passed to this)
	 */
	public void initialise( PreParserService previous, PreParserService next )
	{
		// Save reference to previous in stack
		_previous = previous;
	}
	
	
	/** Outgoing messages are passed to <CODE>this</CODE> via this method
	 * @param msg Outgoing message
	 */
	public final void outgoing( Message msg )
	{
		DIAGNOSTICS.println( "Dealing with outgoing message", this );
		
		// Pass to super-class to deal with
		super.handleOutgoing( msg.getEnvelope(), msg );
	}
	
	
	/** Incoming messages are passed to <CODE>this</CODE> via this method
	 * @param msg Incoming message
	 */
	public final void incoming( Message msg )
	{
		if( msg.getEnvelope().getErrorCode() == NONE )
		{
			DIAGNOSTICS.println( "Dealt with incoming message", this );
		}
		else
		{
			DIAGNOSTICS.println( "Dealt with problem outoing message", this );			
		}
		
		
		if( _previous != null )
		{
			// Forward message via the stack
			_previous.incoming( msg );
		}
		else
		{
			DIAGNOSTICS.println( "TEST: Would have passed message back up stack!" );
		}
	}
	
	
//----------------------------------------------------------------------
// PACKAGE METHODS - implementation of MTSBase abstract methods
//----------------------------------------------------------------------

	/**
	 * Invoked when the MTPBase class is passing an incoming message back
	 */
	final void receive( Envelope p1, Object p2 )
	{
		// Forward message via stack
		incoming( (Message) p2 );
	}

	
	/**
	 * Send the given message to the given target - the target is the result of
	 * a call to one of the lookup() methods
	 */
	final void send( Object target, Envelope p1, Object p2 )
		throws UnknownTargetException, UnknownProtocolException, UnableToSendException
	{
		// Forward message via underlying comms
		send( target, (Message) p2 );
	}	
	
	
//----------------------------------------------------------------------
// PROTECTED METHODS - methods for use by MTP implementation
//----------------------------------------------------------------------

	/**
	 * Should be invoked by underlying MTP implementation when a message is
	 * received
	 */
	protected final void handleIncoming( Message msg )
	{
		// Forward to superclass
		super.handleIncoming( msg.getEnvelope(), msg );
	}
	
	
//----------------------------------------------------------------------
// ABSTRACT METHODS
//----------------------------------------------------------------------

	/**
	 * Method to retreive the protocol dependant addresses of an object that can be used
	 * to receive messages.
	 *
	 * @return A List containing the required addresses, or null if this Comms
	 *					doesn't publish such an object
	 */
	public abstract java.util.List getAddresses();
		
	
	/**
	 * Retreives the protocols known by this comms implementation
	 *
	 * @return A List of String's 
	 */
	public abstract java.util.List getProtocols();
	
	
	/** This should be invoked when the service should be safely shutdown
	 */
	public abstract void shutdown();
	
	
	/**
	 * Invoked to send the given Message to the given target
	 */
	protected abstract void send( Object target, Message msg ) throws UnknownTargetException, 
																UnknownProtocolException, UnableToSendException;
	
//----------------------------------------------------------------------
// MAIN/TESTING METHOD
//----------------------------------------------------------------------
	
	
	private static class TestInternalMTPBase extends InternalMTPBase
	{
		public TestInternalMTPBase( String name, URL ns_url ) throws InitialisationException
		{
			super( name, ns_url );
		}
		
		/**
		 * Bind the MTP to the NS (if it exists) and perform other start-up operations
		 */
		protected void bind() throws InitialisationException
		{
			System.out.println( "bind() invoked - _name = " + _name + " _hap = " + _hap +
													" _ns_url = " + _ns_url );
		}

		
		/**
		 * Un-bind the MTP to the NS (if it exists)
		 */
		protected void unbind()
		{
			System.out.println( "unbind() invoked - _name = " + _name + " _hap = " + _hap +
													" _ns_url = " + _ns_url );
		}


		/**
		 * Lookup a reference to the MTS at the given URL
		 */
		protected Object lookup( URL name ) throws UnknownTargetException
		{
			System.out.println( "lookup( \"" + name + "\" ) invoked " );
			return name;
		}

		
		/**
		 * Attempt to lookup a reference to the MTS belonging to the given Agent
		 */
		protected Object lookup( String name ) throws UnknownTargetException
		{
			System.out.println( "lookup( \"" + name + "\" ) invoked " );
			return name;
		}

		
		/**
		 * Method to retreive the protocol dependant address of an object that can be used
		 * to receive messages.
		 *
		 * @return A URL containing the required address, or null if this Comms
		 *					doesn't publish such an object
		 */
		public java.util.List getAddresses()
		{
			java.util.List l = new java.util.LinkedList();
			l.add( new URL( "http://localhost:9000/agenta" ) );
			return l;
		}


		/**
		 * Retreives the protocols known by this comms implementation
		 *
		 * @return A List of String's 
		 */
		public java.util.List getProtocols()
		{
			return new java.util.LinkedList();
		}


		/** This should be invoked when the service should be safely shutdown
		 */
		public void shutdown()
		{
			unbind();
		}


		/**
		 * Invoked to send the given Message to the given target
		 */
		protected void send( Object target, Message msg ) throws UnknownTargetException, 
																	UnknownProtocolException, UnableToSendException
		{
			System.out.println( "send( " + target + ", " + msg + " ) invoked" );
		}
	}
	
	
	public static void main( String args[] ) throws Throwable
	{
		DIAGNOSTICS.setMemoryLevel( DIAGNOSTICS.LEVEL_MIN );
		
		String aid = "(agent-identifier :name agent-a " +
											":addresses (sequence fipaos-rmi://localhost:1099/ iiop://somehost.com:8888/acc) " +
											":resolvers (sequence (agent-identifier :name ams) (agent-identifier :name acc) ) )";
		
		String bid = "(agent-identifier :name agent-b " +
											":addresses (sequence fipaos-rmi://localhost:1099/ iiop://somehost.com:8989/acc) " +
											":resolvers (sequence (agent-identifier :name ams) (agent-identifier :name acc) ) )";
		
		
		fipaos.parser.acl.ACLMessage acl = new fipaos.parser.acl.ACLMessage();
		acl.setPerformative( "inform" );
		acl.setOntology( "my-ontology" );
		acl.setLanguage( "my-language" );
		acl.setProtocol( "my-protocol" );
		acl.setSender( aid );
		acl.setReceiver( bid );
		acl.setContent( "( Test Message )" );
		acl.setConversationID( "test-conv-id-2348723648" );
		
		Message msg = new Message( acl );
		msg.getEnvelope().setIntendedReceiver( acl.getReceiverAIDs() );
		
		InternalMTPBase mtpb = new TestInternalMTPBase( "agenta@hap", new URL( "fipaos-rmi://localhost:9000/" ) );

		System.out.println( "=========== Testing handleOutgoing() " );
		
		mtpb.outgoing( msg );
		
		System.out.println( "\n\n=========== Testing handleIncoming() " );

		mtpb.handleIncoming( msg.getEnvelope(), msg );
	}
} // End class
