//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.mts;

//----------------------------------------------------------------------

import fipaos.mts.event.*;
import fipaos.mts.service.Service;
import fipaos.ont.fipa.fipaman.AgentID;
import fipaos.ont.fipa.fipaman.Envelope;
import fipaos.util.*;

import java.util.*;

//----------------------------------------------------------------------

/**
 * Contains the abstract functionality that all MTP's require in order to
 * deal with messages - increases code-reuse in MTP implementations
 * <br>
 * <br>
 * $Revision: 1.10 $ $Date: 2001/04/05 14:54:50 $ $Author: alantreadway $
 */
public abstract class MTPBase
{
     
//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------

	/**
	 * Default name that the ACC binds to the NS using
	 */
	public static String ACC_NAME = "acc";


//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

	/**
	 * URL of this MTP's naming service
	 */
	protected URL _ns_url;

	/**
	 * Reference to our name
	 */
	protected String _name;

	/**
	 * HAP address for the platform we're registered with
	 */
	protected String _hap;

	/**
	 * Indicates if this MTP has been shutdown
	 */
	public boolean _shutdown = false;
	
	/**
	 * Incoming queue for messages
	 */
	private Queue _incoming_queue = new FIFOQueue();

	/**
	 * Monitor for incoming messages
	 */
	private Monitor _incoming_queue_monitor;
	
	/**
	 * Map of previously resolved targets
	 */
	private Map _lookup_table;
	
//----------------------------------------------------------------------
// CONSTRUCTOR(S)
//----------------------------------------------------------------------

	/**
	 * Initialises the MTP variables (i.e. _hap, __name & _ns_url), and
	 * then invokes bind()
	 */
	protected MTPBase( String name, URL ns ) throws InitialisationException
	{
		// Save URL for the NS we're using
		_ns_url = ns;
		
		// Initialise lookup table
		_lookup_table = new TreeMap( fipaos.util.Comparators.STRING_COMPARE );
		
		// Store our name for use with the NS
		int index = name.indexOf( '@' );
		_name = name.substring( 0, index );

		// Store HAP address
		_hap = name.substring( index + 1 );
		
		if( index == -1 )
		{
			throw new InitialisationException( "HAP not spcified in Agent name?!?!?" );
		}
		
		try
		{
			// Attempt to bind to NS
			bind();
		}
		catch( InitialisationException e )
		{
			throw e;
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Exception occured whilst bind()'ing: ", t, this, DIAGNOSTICS.LEVEL_3 );
			throw new InitialisationException( "Some error occured when binding to the NS" );
		}
		
		// Set up the incoming Monitor object
		_incoming_queue_monitor = new Monitor( _incoming_queue, new MonitorListener()
		{
			public void monitorUpdate( Object next )
			{
				// Retreive parameters
				Envelope envelope = (Envelope) ((Object[]) next)[0];
				Object msg = ((Object[]) next)[1];

				try
				{
					// Pass message to code to deal with incoming message
					receive( envelope, msg );
				}
				catch( Throwable t )
				{
					DIAGNOSTICS.println( "Couldn't deal with incoming message for some reason",
																											t, DIAGNOSTICS.LEVEL_MAX );
				}
			}
		} );
	}
	
//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------	

//----------------------------------------------------------------------
// PROTECTED METHODS
//----------------------------------------------------------------------

	/**
	 * Should be invoked by the underlying MTP when an outgoing message is received
	 */
	protected void handleOutgoing( Envelope env, Object msg )
	{
		handleOutgoing( env, msg, true );
	}
	
	
	/**
	 * Should be invoked by the underlying MTP when an outgoing message is received
	 */
	private void handleOutgoing( Envelope env, Object msg, boolean retry )
	{
		// Deal with the message in the appropriate manner
		// Target of message - to be initialised
		Object target = null;
		
		Throwable error_reason = null;
		int error = Service.NONE;
		URL url = null;
		
		try
		{
			// Check if we're sending direct to the target Agent, or via the ACC
			if( env.getViaACC() )
			{
				// Via the ACC
				try
				{
					// Construct URL for ACC
					target = lookup( ACC_NAME );

					DIAGNOSTICS.println( "Routing via ACC", this );
				}
				catch( Throwable t )
				{
					// Set appropriate error code
					error = Service.TARGET_UNKNOWN;
					error_reason = t;
				}
			}
			else
			{
				// Direct to the target Agent

				// Get the AgentID of the target Agent & first URL
				AgentID intended_receiver = (AgentID) env.getIntendedReceiver().get( 0 );

				DIAGNOSTICS.println( "Target agent is: " + intended_receiver, this );

				// Check if addresses are available
				if( intended_receiver.getAddresses().size() > 0 )
				{
					// Attempt to get the URL from the intended receiver field
					try
					{
						target = cachedLookup( intended_receiver );
						url = (URL) intended_receiver.getAddresses().get( 0 );
					}
					catch( Throwable t )
					{
						error_reason = t;
					}
				}
				else
				{
				}
				
				// Not got target and intended receiver is a member of the local AP?
				if( target == null && 
						intended_receiver.getName().endsWith( "@" + _hap ) )
				{					
					String agent_name = intended_receiver.getName(); 
					int index = agent_name.lastIndexOf( "@" + _hap );
					agent_name = agent_name.substring( 0, index );
					
					DIAGNOSTICS.println( "Target agent couldn't be lookup()'ed - trying just \"" 
																													+ agent_name + "\"", this );
					// No addresses - attempt to find Agent locally
					target = lookup( agent_name );
				}
			}
			
			// Attempt to send the message
			if( target != null )
			{
				send( target, env, msg );
			}
			else
			{
				if( error_reason != null )
				{
					throw error_reason;
				}
				else
				{
					throw new UnknownTargetException( "" );
				}
			}
		}
		catch( UnknownTargetException e )
		{
			DIAGNOSTICS.println( "Couldn't resolve target", e, this );
			
			// Target unknown error!
			error = Service.TARGET_UNKNOWN;
		}
		catch( UnknownProtocolException e )
		{
			DIAGNOSTICS.println( "Protocol specified not supported", e, this );

			// Unknown protocol!
			error = Service.PROTOCOL_UNKNOWN;
		}
		catch( UnableToSendException e )
		{
			DIAGNOSTICS.println( "Couldn't send to target for some reason", e, this );

			// Unknown error!
			error = Service.ERROR_UNKNOWN;
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Something unexpected happened...", t, this );
			
			// Unknown error!
			error = Service.ERROR_UNKNOWN;
		}
		
		// Check if an error occured
		if( error != Service.NONE )
		{
			// If reference is potentially cached, remove reference to object and
			// try again
			if( ( url != null ) && retry )
			{
				_lookup_table.remove( url );
				handleOutgoing( env, msg, false );
			}
			else
			{
				// Set error code & pass back up stack
				env.setErrorCode( error );
				handleIncoming( env, msg );
			}
		}
		else
		{
			// Indicate message has left local MTS stack
			Iterator i = env.getMessageListeners().iterator();
			while( i.hasNext() )
			{
				MessageListener ml = (MessageListener) i.next();
				ml.notifyDelivery( env );
			}
		}
	}
	
	/**
	 * Do a cached lookup
	 */
	private Object cachedLookup( AgentID intended_receiver ) throws Exception
	{
		Object target = null;
		
		// Get target URL
		URL url = (URL) intended_receiver.getAddresses().get( 0 );

		// Check MTP can deal with the transport type specified by the URL
		if( getProtocols().contains( url.getProtocol() ) )
		{
			DIAGNOSTICS.println( "MTP supports " + url.getProtocol() + ", attempting lookup()",
														this, DIAGNOSTICS.LEVEL_2 );
			// Lookup target
			if( _lookup_table.containsKey( url.toString() ) )
			{
				DIAGNOSTICS.println( "Using cached reference to: " + url  );
				target = _lookup_table.get( url.toString() );
			}
			else
			{
				target = lookup( url );
				
				// Add target to cache if found
				if( target != null )
				{
					_lookup_table.put( intended_receiver.getAddresses().get( 0 ), target );
				}
			}
		}
		else
		{
			DIAGNOSTICS.println( "MTP doesn't support " + url.getProtocol() + ", not attempting lookup()",
														this, DIAGNOSTICS.LEVEL_2 );

			throw new UnknownProtocolException( url.getProtocol() + " not supported by MTP" );
		}
		
		return target;
	}
	
	/**
	 * Should be invoked by the underlying MTP when an incoming message is received
	 */
	protected void handleIncoming( Envelope envelope, Object msg )
	{
		if( !_shutdown )
		{
			// Add message to queue
			_incoming_queue.addItem( new Object[] { envelope, msg } );
		}
	}
	
	/**
	 * Should be invoked by the undelying implementation when shutting down
	 */
	protected void shutdownMTPBase()
	{
		// Indicate that this has shutdown
		_shutdown = true;
		
		// Dereference objects
		_incoming_queue_monitor.halt();
		_incoming_queue_monitor = null;
		_incoming_queue.flush();
		_incoming_queue = null;
		
		// Unbind from the NS
		unbind();
	}
	
//----------------------------------------------------------------------
// ABSTRACT PROTECTED METHODS
//----------------------------------------------------------------------

	/**
	 * Bind the MTP to the NS (if it exists) and perform other start-up operations
	 */
	protected abstract void bind() throws InitialisationException;

	/**
	 * Un-bind the MTP to the NS (if it exists)
	 */
	protected abstract void unbind();
	
	/**
	 * Lookup a reference to the MTS at the given URL
	 */
	protected abstract Object lookup( URL name ) throws UnknownTargetException;

	/**
	 * Attempt to lookup a reference to the MTS belonging to the given Agent
	 */
	protected abstract Object lookup( String name ) throws UnknownTargetException;	
	
	/**
	 * Send the given message to the given target - the target is the result of
	 * a call to one of the lookup() methods
	 */
	abstract void send( Object target, Envelope envelope, Object msg ) 
		throws UnknownTargetException, UnknownProtocolException, UnableToSendException;
	
	/**
	 * Invoked when a message is received - subclasses should implement this
	 * to deal with incoming messages (i.e. propogate back up the comms stack)
	 */
	abstract void receive( Envelope envelope, Object msg );

	/**
	 * Retreives the protocols known by this comms implementation
	 *
	 * @return A List of String's 
	 */
	public abstract java.util.List getProtocols();
	
//----------------------------------------------------------------------
// MAIN/TESTING METHOD
//----------------------------------------------------------------------
	
	private static class TestMTPBase extends MTPBase
	{
		public TestMTPBase( String name, URL ns_url ) throws InitialisationException
		{
			super( name, ns_url );
		}
		
		public java.util.List getProtocols()
		{
			return new java.util.LinkedList();
		}
		
		/**
		 * Bind the MTP to the NS (if it exists) and perform other start-up operations
		 */
		protected void bind() throws InitialisationException
		{
			System.out.println( "bind() invoked - _name = " + _name + " _hap = " + _hap +
													" _ns_url = " + _ns_url );
		}

		
		/**
		 * Un-bind the MTP to the NS (if it exists)
		 */
		protected void unbind()
		{
			System.out.println( "unbind() invoked - _name = " + _name + " _hap = " + _hap +
													" _ns_url = " + _ns_url );
		}


		/**
		 * Lookup a reference to the MTS at the given URL
		 */
		protected Object lookup( URL name ) throws UnknownTargetException
		{
			System.out.println( "lookup( \"" + name + "\" ) invoked " );
			return name;
		}

		
		/**
		 * Attempt to lookup a reference to the MTS belonging to the given Agent
		 */
		protected Object lookup( String name ) throws UnknownTargetException
		{
			System.out.println( "lookup( \"" + name + "\" ) invoked " );
			return name;
		}

		
		/**
		 * Send the given message to the given target - the target is the result of
		 * a call to one of the lookup() methods
		 */
		void send( Object target, Envelope envelope, Object msg ) 
		{
			System.out.println( "send( " + target + ", " + envelope + ", " +
													msg + " ) invoked" );
		}
		
		/**
		 * Invoked when a message is received - subclasses should implement this
		 * to deal with incoming messages (i.e. propogate back up the comms stack)
		 */
		void receive( Envelope envelope, Object msg )
		{
			System.out.println( "receive( " + envelope + ", " +
													msg + " ) invoked" );			
		}
	}
	
	
	
	
	public static void main( String args[] ) throws Throwable 
	{
		DIAGNOSTICS.setMemoryLevel( DIAGNOSTICS.LEVEL_MIN );
		
		String aid = "(agent-identifier :name agent-a " +
											":addresses (sequence fipaos-rmi://localhost:1099/ iiop://somehost.com:8888/acc) " +
											":resolvers (sequence (agent-identifier :name ams) (agent-identifier :name acc) ) )";
		
		String bid = "(agent-identifier :name agent-b " +
											":addresses (sequence fipaos-rmi://localhost:1099/ iiop://somehost.com:8989/acc) " +
											":resolvers (sequence (agent-identifier :name ams) (agent-identifier :name acc) ) )";
		
		
		fipaos.parser.acl.ACLMessage acl = new fipaos.parser.acl.ACLMessage();
		acl.setPerformative( "inform" );
		acl.setOntology( "my-ontology" );
		acl.setLanguage( "my-language" );
		acl.setProtocol( "my-protocol" );
		acl.setSender( aid );
		acl.setReceiver( bid );
		acl.setContent( "( Test Message )" );
		acl.setConversationID( "test-conv-id-2348723648" );
		
		Message msg = new Message( acl );
		msg.getEnvelope().setIntendedReceiver( acl.getReceiverAIDs() );
		
		MTPBase mtpb = new TestMTPBase( "agenta@hap", new URL( "fipaos-rmi://localhost:9000/" ) );

		System.out.println( "=========== Testing handleOutgoing() " );
		
		mtpb.handleOutgoing( msg.getEnvelope(), msg );
		
		System.out.println( "\n\n=========== Testing handleIncoming() " );

		mtpb.handleIncoming( msg.getEnvelope(), msg );
	}
	
} // End class
