//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.mts;

//----------------------------------------------------------------------

import fipaos.agent.profile.InternalAddress;
import fipaos.mts.service.*;
import fipaos.mts.service.preparse.*;
import fipaos.ont.fipa.*;
import fipaos.util.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.ont.BASETOKENS;
//import fipaos.agent.monitor.*;

import java.util.*;

//----------------------------------------------------------------------

/**
 * This class provides a convenient mechanism for interacting with the
 * comms-level transports
 * <br>
 * <br>
 */
public class MTS implements MessageSender
{
     
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

	/**
	 * URL's of this MTS
	 */ 
	private List _addresses;
	
	/**
	 * MessageReceiver to pass incoming messages to
	 */
	private MessageReceiver _mr;

	/**
	 * Internal stack
	 */
	private InternalStack _internal_stack;

	/**
	 * Indicates if this MTS has been shutdown
	 */
	private boolean _shutdown = false;
	
	/**
	 * Enables message count to be kept in order to serially tag outgoing messages
	 */
	private long _tag_count = 0;
	
//----------------------------------------------------------------------
// CONSTRUCTOR(S)
//----------------------------------------------------------------------

	/**
	 * Constructor for sub-classes
	 */
	protected MTS()
	{
	}
	
	/**
	 * Construct a new MTS object + initialise MTP's
	 */
	public MTS( URL ams, List mtp_ns, AgentID owner )
	{
		// Initialise variables
		_addresses = new LinkedList();
		
		// Setup MTPs
		setupMTPs( owner.getName(), mtp_ns );
		
		// Add transport addresses to AID
		owner.setAddresses( new LinkedList( _addresses ) );
	}

	
//----------------------------------------------------------------------
// PUBLIC METHODS - MessageSender interface
//----------------------------------------------------------------------	
	
	/**
	 * Send the given message
	 */
	public void sendMessage( Message p1 )
	{
		if( !_shutdown )
		{
			// Ensure FIFO ordering of the stack by tagging
			tagMessage( p1 );
			
			// Pass the message down the stack
			_internal_stack.outgoing( p1 );
		}
	}
	
	public void setMessageReceiver( MessageReceiver p1 )
	{
		_mr = p1;
	}
	
	public void shutdown()
	{
		// Check if we've shutdown
		if( !_shutdown )
		{
			// Indicate we're now shutdown
			_shutdown = true;
			
			// Shutdown internal stack
			_internal_stack.shutdown();
			
			// Release references to improve garbage collection
			_internal_stack = null;
			_mr = null;
			_addresses.clear();
			_addresses = null;
		}
	}

//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------
	
	/**
	 * Tag the given message so FIFO order of stack can be maintained
	 */
	private void tagMessage( Message msg )
	{
		long number = 0;
		
		synchronized( this )
		{
			number = _tag_count++;
		}
		
		msg.getEnvelope().setSlotValue( Service.FIPAOS_MSG_ID_SLOT, new Long( number ) );

		DIAGNOSTICS.println( "Msg id " + number + " assigned to: " + msg.getACL(), this );
	}
	
	private void setupMTPs( String name, List mtp_ns )
	{
		// CMS to reference all internal transports
		CommMultiplexService cms = new CommMultiplexService();

		// Attempt to create internal transports
		Map internal_transports = new TreeMap( fipaos.util.Comparators.STRING_COMPARE );
		Iterator i = mtp_ns.iterator();
		
		InternalAddress addr;
		
		while( i.hasNext() )
		{
			
			addr = (InternalAddress)i.next();
			
			// Get URL for NS of next transport
			URL ns = new URL(addr.getAddress());
			
			// Attempt to instantiate transport
			InternalMTP internal = TransportProvider.getDefaultInternalMTP( name, ns );
			
			// Check if succesful
			if( internal != null )
			{
				// Add address to our list of addresses
				_addresses.addAll( internal.getAddresses() );
				
				// Initialise transport with regard to CMS
				internal.initialise( cms, null );
				
				// Iterate through protocols and add the MTP against it
				List protocols = internal.getProtocols();
				Iterator j = protocols.iterator();
				while( j.hasNext() )
				{
					// Add to our Map of internal transports
					internal_transports.put( j.next(), internal );
				}
			}
			else
			{
				// Display warning message
				DIAGNOSTICS.println( "Error starting MTP using NS-URL: " + ns, DIAGNOSTICS.LEVEL_3 );
			}
		}
		
		// Create internal stack
		_internal_stack = new InternalStack();
		ServiceStack stack = new ServiceStack();
		stack.add( _internal_stack );
		stack.add( new BufferService() );
		stack.add( new ACCRouterService( false ) );				
		
		// Add a new Buffer to buffer messages if necessary. Register this with the 
		// Component Monitor
		BufferService message_buffer = new BufferService();
//		ComponentMonitor.registerComponent(message_buffer);
		stack.add(message_buffer);
		
		cms.initialise( internal_transports );
		stack.add( cms );
		stack.initialise();
	}

	
	/**
	 * Class to propogate incoming messages from the internal stack to the
	 * appropriate method in parent class
	 */
	private class InternalStack implements PreParserService
	{
		private PreParserService _next;
		
		/** This should be invoked when the service should be safely shutdown
		 */
		public void shutdown()
		{
			_next.shutdown();
			_next = null;
		}
		
		/** Invoked when the undelying implementation should initialise itself.  Allows
		 * references to other PreParserService's before and after it in the stack
		 * to be passed
		 * @param previous Previous PreParserService instance (Incoming messages should be passed to this)
		 * @param next Next PreParserService instance (Outgoing messages should be passed to this)
		 */
		public void initialise(PreParserService previous,PreParserService next)
		{
			_next = next;
		}
		
		/** Outgoing messages are passed to <CODE>this</CODE> via this method
		 * @param msg Outgoing message
		 */
		public void outgoing(Message msg)
		{
			// If the MTS has shutdown, just ignore
			if( !_shutdown )
			{
				_next.outgoing( msg );
			}
		}
		
		/** Incoming messages are passed to <CODE>this</CODE> via this method
		 * @param msg Incoming message
		 */
		public void incoming(Message msg)
		{
			// If the MTS has shutdown, just ignore
			if( _shutdown )
			{
				return;
			}
			
			int error = msg.getEnvelope().getErrorCode();
			String failure_reason = (String) msg.getEnvelope().getSlotValue( Service.MSG_DELIVERY_FAILURE_REASON );
			
			// Check if an error occured
			if( error == Service.NONE && failure_reason == null )
			{
				msg.getEnvelope().setSlotValue( Service.FIPAOS_DEBUG + "mts-send-time", 
																						new Long( System.currentTimeMillis() ) );
		
				// If not, attempt to pass message to Agent
				if( _mr != null )
				{
					DIAGNOSTICS.println( "Received message, passing to MessageReceiver: " + msg.getACL(), this );
					_mr.receiveMessage( msg );
				}
				else
				{
					DIAGNOSTICS.println( "Received message, but MTS is not connected to anything:\n" +
																msg, this, DIAGNOSTICS.LEVEL_3 );
				}
			}
			else if( error != Service.NONE )
			{
				// TODO: Add error handling support
				
				DIAGNOSTICS.println( "Error when sending message (error reason = " + 
															Service.ERROR_STRING[ error ] +
															"): " + msg, this, DIAGNOSTICS.LEVEL_MAX );
			}
			else if( failure_reason != null )
			{
				DIAGNOSTICS.println( "Error when sending message (error reason = " + 
															failure_reason +
															"): " + msg, this, DIAGNOSTICS.LEVEL_MAX );				
			}
		}
	}

//----------------------------------------------------------------------
// MAIN/TESTING METHOD
//----------------------------------------------------------------------
	
} // End class

