//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis,
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Queen Mary, University of London.
//
//  Portions created by Queen Mary University of London are Copyright (C) 2001
//  Queen Mary, University of London licensed under the Emorphia Public License,
//  available at http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (Stefan Channing)
//
//******************************************************************************
package fipaos.mts.httpMTP;

import java.io.*;
import java.net.*;
import java.util.*;

import fipaos.ont.fipa.fipaman.Envelope;
import fipaos.parser.envelope.xml.*;
import fipaos.util.DIAGNOSTICS;
import fipaos.parser.ParserException;

/**
 * ConnectionThread carries out all communication for a specific Client connection.
 */
public class ConnectionThread extends ThreadingAbility
{
	
	//----------------------------------------------------------------------
	// CONSTANTS
	//----------------------------------------------------------------------
	
	private static final String TEST_PAGE="<HTML><HEAD><TITLE>FIPA-OS Server Test Page</TITLE></HEAD>"+
	"<BODY><H1>FIPA-OS HTTP server</H1>"+
	"This page demonstrates that the server is running<HR>"+
	"<I>Stefan Channing June 2001</I></BODY></HTML>",
	ERROR_CODE="400",
	ERROR_MESSAGE="BAD REQUEST",
	OK_CODE="200",
	OK_MESSAGE="OK";
	
   /**
    * Contants used during the parsing of the MIME content
    */
	private static final int ENVELOPE_START=0,
	ACL_MESSAGE_START=1,
	COMMS_END=2;
	
	//----------------------------------------------------------------------
	// INSTANCE VARIABLES
	//----------------------------------------------------------------------
	
	private Socket clientSocket;
	private PrintStream out;
	private BufferedReader in;
	private CallBack backRef;
	
	//----------------------------------------------------------------------
	// CONSTRUCTOR
	//----------------------------------------------------------------------
	
	public ConnectionThread(Socket connection, CallBack backRef)
	{
		super();
		clientSocket=connection;
		this.backRef=backRef;
		internal.start();
	}
	
	//----------------------------------------------------------------------
	// PROTECTED METHODS
	//----------------------------------------------------------------------
	
   /**
    * A FIPA server does not need to implement GET. This has been added to
    * allow a simple test of the server from a browser. Knowing it is a GET
    * request we ignore anything else that has been sent and return the
    * <code>TEST_PAGE</code>
    */
	protected void doGet() throws IOException
	{
		String line;                //contains each line as it is read from the socket
		DIAGNOSTICS.println("Dealing with GET",this,DIAGNOSTICS.LEVEL_MAX);
		while ((line=in.readLine())!=null && !line.equals(""))
		{}
		out.print("HTTP/1.1 200 OK\r\n");
		out.print("Date: "+new java.util.Date()+"\r\n");
		out.print("Content-length: "+TEST_PAGE.length()+"\r\n");
		out.print("Content-type: text/html\r\n\r\n");
		out.print(TEST_PAGE);
	}
	
   /**
    * Responds to POST request. Parses the HTTP header and extracts various details.
    * One of these is the <code>boundary</code> that is present in the MIME content
    * separating the <code>Envelope</code> and the <code>ACL message</code>. It also
    * marks the end of the content. If this is valid then the MIME content is parsed
    * and the resulting <code>Envelope</code> and <code>ACL message</code> are
    * passed into the FIPA-OS comms stack. A HTTP header is returned to the sender
    * indicating the result.
    */
	protected void doPost() throws IOException
	{
		String line=null,           //contains each line as it is read from the socket
		boundaryString=null, //stores the boundary string used to parse the content
		sender=null,
		envMineType=null,    //stores the MIME type of the envelope (not used)
		msgMineType=null,    //stores the MIME type of the ACL message (not used)
		code=null,           //used in the HTTP response
		message=null;        //used in the HTTP response
		boolean sucess=false;
		Envelope envelope=null;
		byte[] aclMessage=null;
		int contentLength=-1;       //stores the length of the MIME comtent (not used)
		DIAGNOSTICS.println("Dealing with POST",this,DIAGNOSTICS.LEVEL_MAX);
		try
		{
			//read the rest of the header (read up to and including the carriage return)
			while ((line=in.readLine())!=null && !line.equals(""))
			{
				//extract the length of the content from the header (not used)
				if (line.toLowerCase().indexOf("content-length:")!=-1)
					contentLength=Integer.parseInt(line.substring(line.indexOf(":")+1).trim());
				//extract the boundary (used to delimit the MIME content)
				if (line.toLowerCase().indexOf("boundary")!=-1)
				{
					int startMarker=line.indexOf("=\"")+2;//+2 to remove ="
					int endMarker=line.indexOf("\"",startMarker);
					if (endMarker==-1) endMarker=line.length();
					boundaryString=line.substring(startMarker,endMarker);
				}
				//extract the host details (not used)
				if (line.toLowerCase().indexOf("host")!=-1)
					sender=line.substring(line.indexOf(":")+1).trim();
			}//end of while
			System.out.println("****Test information "+contentLength+", "+boundaryString+ ", "+sender+". ****");
			//read the MIME content but only if the boundary has been obtained
			if (boundaryString !=null)
			{
				int boundaryCounter=-1,readContentLength=0;
				StringBuffer envelopeBuf=new StringBuffer(),bodyBuf=new StringBuffer();
				while (boundaryCounter<COMMS_END &&(line=in.readLine())!=null)
				{
					if (line.indexOf(boundaryString)!=-1)
						boundaryCounter++;
					else
					{
						if (boundaryCounter==ENVELOPE_START)
						{
							DIAGNOSTICS.println("Dealing with Envelope",this,DIAGNOSTICS.LEVEL_MAX);
							if (line.toLowerCase().indexOf("content-type:")!=-1)
								envMineType=line.substring(line.indexOf(":")+1);
							else
								envelopeBuf.append(line+"\n");
						}
						else if (boundaryCounter==ACL_MESSAGE_START)
						{
							DIAGNOSTICS.println("Dealing with ACL Message",this,DIAGNOSTICS.LEVEL_MAX);
							if (line.toLowerCase().indexOf("content-type:")!=-1)
								msgMineType=line.substring(line.indexOf(":")+1);
							else
								bodyBuf.append(line+"\n");
						}
					}
					readContentLength+=line.length()+1;//+1 for every carriage return that would have been stripped
				}//end of while
				//attempt to parse the envelope if this fails then indicate BAD_REQUEST
				try
				{
					envelope=Parser.parse(envelopeBuf.toString().trim().getBytes());
					aclMessage=bodyBuf.toString().trim().getBytes();
					sucess=true;
				}
				catch (ParserException PEx)
				{
					sucess=false;
					DIAGNOSTICS.println("Envelope failed to parse correctly",this,DIAGNOSTICS.LEVEL_MAX);
				}
			}
			else //this is the else to if(boundary!=null)
			{
				sucess=false;
				DIAGNOSTICS.println("Message received is not FIPA compliant - boundary is incorrect",this,DIAGNOSTICS.LEVEL_MAX);
				System.err.println("NOT FIPA Compliant");
			}
		}
		catch(IOException IOEx)
		{
			//there has been an error in reading from the socket
			DIAGNOSTICS.println("Communication error: "+IOEx.getMessage(),this,DIAGNOSTICS.LEVEL_MAX);
			sucess=false;
		}
		//setup and send the response information
		if (sucess)
		{
			DIAGNOSTICS.println("Sending Response - OK",this,DIAGNOSTICS.LEVEL_MAX);
			code=OK_CODE;
			message=OK_MESSAGE;
		}
		else
		{
			DIAGNOSTICS.println("Sending Response - ERROR",this,DIAGNOSTICS.LEVEL_MAX);
			code=ERROR_CODE;
			message=ERROR_MESSAGE;
		}
		String reply="HTTP/1.1 "+code+" "+message+"\nContent-Type: text/plain\nCache-Control: no-cache\nConnection: close\n\n";
		out.print(reply);
		//pass the received Envelope and ACL meesage back to HTTPComms to be placed into the incoming queue
		if (sucess) backRef.incomingCall(envelope,aclMessage);
	}
	
	//----------------------------------------------------------------------
	// PROTECTED METHODS - ThreadingAbility implementation
	//----------------------------------------------------------------------
	
	protected void runImplementation()
	{
		try
		{
			//open streams
			in=new BufferedReader(new InputStreamReader(clientSocket.getInputStream()));
			out=new PrintStream(clientSocket.getOutputStream());
			//read first line of header to determine whether GET or POST
			String headerLine=in.readLine();
			if (headerLine.toUpperCase().indexOf("POST")!=-1)
				doPost();
			else
				doGet();
		}
		catch(IOException IOEx)
		{
			DIAGNOSTICS.println("Error in opening the streams: "+IOEx.getMessage(),this,DIAGNOSTICS.LEVEL_MAX);
			IOEx.printStackTrace();
		}
		//this code is guranteed to be executed
		finally
		{
			try
			{
				if (in !=null) in.close();
				if (out !=null) out.close();
				if (clientSocket !=null) clientSocket.close();
				in=null;
				out=null;
				clientSocket=null;
			}
			catch(IOException IOEx)
			{
				IOEx.printStackTrace();
			}
		}
	}
} // End class