//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Queen Mary, University of London.
//
//  Portions created by Queen Mary University of London are Copyright (C) 2001
//  Queen Mary, University of London licensed under the Emorphia Public License,
//  available at http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (Stefan Channing)
//
//******************************************************************************
package fipaos.mts.httpMTP;

import java.util.*;
import java.io.*;

import fipaos.ont.fipa.fipaman.Envelope;
import fipaos.util.URL;

import fipaos.util.DIAGNOSTICS;
import fipaos.mts.*; //this includes all of the exceptions that I require

public class HTTPComms extends ExternalMTPBase implements CallBack
{
    
//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------

   /**
    * Transport specific protocol type
    */
    private static final String[] TRANSPORT_PROTOCOL = { "http" };

//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

    private HTTPTransportService httpTransport;

//----------------------------------------------------------------------
// CONSTRUCTOR
//----------------------------------------------------------------------	

   /**
    * Comms initialiser
    * @param our_name The name of the Agent
    * @param ns_url URL for the transport
    */
    public HTTPComms(String our_name,fipaos.util.URL ns_url) throws InitialisationException
    {
        super(our_name, ns_url);
    }

//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------

   /**
    *
    */
    public List getAddresses()
    {
		List addr_list = new LinkedList();
        addr_list.add(_ns_url);

        return addr_list;
    }

   /**
    * Provides information regarding the Protocol(s) that this MTP can support.
    * @return <code>List</code> containing the Protocols supported by this transport
    */
    public List getProtocols()
    {
        return Arrays.asList(TRANSPORT_PROTOCOL);
    }

   /**
    * Invoked to safely shut down the Comms and to release any resources
    */
    public void shutdown()
    {
        shutdownMTPBase();      //this will call unbind()
        httpTransport=null;
    }

//----------------------------------------------------------------------
// INTERFACE IMPLEMENTATION - CallBack
//----------------------------------------------------------------------	

    public void incomingCall(Envelope env,byte[] msg)
    {
        handleIncoming(env,msg);
    }
    
//----------------------------------------------------------------------
// PROTECTED METHODS - MTPBase implementation
//----------------------------------------------------------------------	

   /**
    * Starts the HTTPTransportService using the port information gained from the
    * acc_profile. 
    */
    protected void bind() throws InitialisationException
    {
        try
        {
	    DIAGNOSTICS.println("Starting HTTPTransportService",this,DIAGNOSTICS.LEVEL_MAX);
	    httpTransport=new HTTPTransportService(_ns_url.getPort(),this);
        }
        catch (IOException IOEx) 
        {
            DIAGNOSTICS.println("Failed to start HTTPTransportService: "+IOEx.toString(),this,DIAGNOSTICS.LEVEL_MAX);
            throw new InitialisationException("Could not start HTTP Server "+IOEx.getMessage());
        }
    }

   /**
    * Informs the <code>httpTransport</code) that is should shut down.
    */
    protected void unbind()
    {
        httpTransport.setActive(false);
    }

   /**
    * Invoked to send the given Message to the given target. This is called by
    * <code>handleOutgoing()</code> from <code>MTPBase</code>. Creates a
    * <code>SenderThread</code> instance to deal with encoding the contents into the 
    * FIPA compliant HTTP header and MIME content. <code>SenderThread</code> also
    * deals with the responding message. Later implementations may do something with
    * this, i.e. log outgoing and incoming information.
    */
    protected void send(Object target,Envelope env,byte[] msg) throws UnknownTargetException, 
                                                                      UnknownProtocolException, 
                                                                      UnableToSendException
    {
        try 
	{
            DIAGNOSTICS.println("Sending ACL Message",this,DIAGNOSTICS.LEVEL_MAX);

			new SenderThread((fipaos.util.URL)target,env,msg);
	}
        //todo work out why the information was not sent to create the most
        //appropriate exception
	catch(Throwable t)
        {
            DIAGNOSTICS.println("Unable to send message: "+t.toString(),this,DIAGNOSTICS.LEVEL_MAX);
            throw new UnableToSendException();
        }			
    }

   /**
    * Lookup a reference to the MTS at the given URL
    * todo this could create an InetAddress object to check the validity of the URL
    */
    protected Object lookup(fipaos.util.URL name) throws UnknownTargetException
    {
        return name;
    }

   /**
    * Attempt to lookup a reference to the MTS belonging to the given Agent
    */
    protected Object lookup(String name) throws UnknownTargetException
    {
        return name;
    }
} // End class