//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved.
//
//******************************************************************************

//***************************************************************************
//
// TODO: Add support for user-defined properties in AgentID
//
//**************************************************************************/

package fipaos.mts.iiop.sun.external;

//----------------------------------------------------------------------
import org.omg.CORBA.ORB;
import org.omg.CosNaming.*;
import org.omg.CosNaming.NamingContextPackage.*;
import org.omg.CORBA.SystemException;
import org.omg.CORBA.Any;
import java.util.Arrays;
import java.util.Set;
import java.util.Properties;
import java.util.List;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.StringTokenizer;

import fipaos.mts.*;
import fipaos.ont.fipa.ACL;
import fipaos.ont.fipa.fipaman.Envelope;
import fipaos.ont.fipa.fipaman.AgentID;
import fipaos.ont.fipa.fipaman.Received;
import fipaos.ont.fipa.fipaman.TransportBehaviour;
import fipaos.util.URL;
import fipaos.util.UTCTime;
import fipaos.util.DIAGNOSTICS;

import fipaos.mts.iiop.sun.external.FIPA.*;



/**
 * SUNIIOPComms uses jdk1.2's IDL classes to provide a FIPA99 Structured IIOP
 * interface for Agent communications.
 * <p>
 * <p>
 * $Revision: 1.15 $ $Date: 2001/04/04 23:21:35 $ $Author: alantreadway $
 */
public class SUNIIOPComms extends ExternalMTPBase
{
	//----------------------------------------------------------------------
	// CONSTANTS
	//----------------------------------------------------------------------
	
	/**
	 * Transport specific protocol type
	 */
	private static final String TRANSPORT_PROTOCOLS[] =
	{ "corbaname", "iiop", "IOR" };
	
	private final static String NS_TYPE = "fipaos-suniiop-ext";
	private final static String NS_KIND = fipaos.mts.iiop.sun.external.FIPA.MTSHelper.id();
	
	private final static String CMD_LINE_SWITCH = "fipaos.sunidl.use_non_existent";
	//----------------------------------------------------------------------
	// INSTANCE VARIABLES
	//----------------------------------------------------------------------
	
	/**
	 * Indicates if the _non_existent() call should be used
	 */
	private static boolean __use_non_existent = System.getProperty( CMD_LINE_SWITCH, "true" ).equals( "true" );
	
	/**
	 * URL for this transport
	 */
	private List _url;
	
	/**
	 * Reference to the ORB
	 */
	private org.omg.CORBA.ORB _orb_ref;
	
	/**
	 * Root naming context
	 */
	private NamingContext _root_context;
	
	/**
	 * Reference to our proxy
	 */
	private fipaos.mts.iiop.sun.external.FIPA.MTS _proxy;
	
	
	//----------------------------------------------------------------------
	// PUBLIC METHODS - Comms implementation
	//----------------------------------------------------------------------
	
	/**
	 * Comms initialiser
	 *
	 * @param our_name The name of the Agent
	 * @param ns_url URL for the transport
	 */
	public SUNIIOPComms( String our_name, URL ns_url ) throws InitialisationException
	{
		super( our_name, ns_url );
	}
	
	
	//----------------------------------------------------------------------
	// PROTECTED METHODS - MTPBase implementation
	//----------------------------------------------------------------------
	
	/**
	 * Bind the MTP to the NS (if it exists) and perform other start-up operations
	 */
	protected void bind() throws InitialisationException
	{
		// Check if we have initialised the transport yet
		if( _orb_ref == null )
		{
			// Initialise the ORB arguments
			Properties props = new Properties();
			props.put("org.omg.CORBA.ORBInitialPort", "" + _ns_url.getPort() );
			props.put("org.omg.CORBA.ORBInitialHost", _ns_url.getHost() );
			
			// Create and initialize the ORB
			String args[] = new String[0];
			_orb_ref = ORB.init( args, props );
			
			// Resolve the root context & Agent context
			_root_context =  resolveRootContext();
			if( _root_context == null || !resolveContext() )
			{
				throw new InitialisationException("Error initialising root context / sub-context in Naming Service");
			}
			
			// Create our proxy (to receive incoming messages)
			_proxy = new MTSListener();
			
			// Connect to the ORB
			try
			{
				_orb_ref.connect( _proxy );
			}
			catch ( SystemException e )
			{
				// Couldn't connect()?
				throw new InitialisationException("Error connecting to ORB - " + e.toString() );
			}
		}
		
		// (Re)Bind this to our Agent name in the NamingService
		if( !rebind( _name ) )
		{
			// Couldn't (re)bind?
			throw new InitialisationException("Couldn't rebind to Naming Service");
		}
		
		// Setup list of URL's for this
		if( _url == null )
		{
			_url = new LinkedList();
		}
		
		// Setup our IOR
		_url.add( new URL( _orb_ref.object_to_string( _proxy ) ) );
		
		try
		{
			// Add old IIOP URL
			URL url = (URL) _ns_url.clone();
			url.setTarget( _name );
			url.setProtocol( "iiop" );
			_url.add( url );
			
			// Add CORBANAME URL
			URL url2 = new URL( "corbaname::" + _ns_url.getHost() + ":" +
													_ns_url.getPort() + "/NameService#" + _name );
			_url.add( url2 );
		}
		catch( Throwable t )
		{
		}
	}
	
	
	/**
	 * Un-bind the MTP from the NS (if it exists)
	 */
	protected void unbind()
	{
		try
		{
			boolean unbind = false;
			
			// Attempt to compare currently bound reference with our proxy
			fipaos.mts.iiop.sun.external.FIPA.MTS obj =
			(fipaos.mts.iiop.sun.external.FIPA.MTS) lookup( _name );
			unbind = obj._is_equivalent( _proxy );
			
			// Unbind the reference if it refers to us!
			if( unbind )
			{
				deregister( _name );
			}
			
			// Disconnect the MTS object from the ORB
			_orb_ref.disconnect( _proxy );
		}
		catch( Throwable e )
		{
		}
	}
	
	
	/**
	 * Lookup a reference to the MTS at the given URL
	 */
	protected Object lookup( URL name ) throws UnknownTargetException
	{
		DIAGNOSTICS.println( "Looking up " + name, this );
		
		fipaos.mts.iiop.sun.external.FIPA.MTS mts_target = null;
		String error_reason = null;
		
		// Attempt to resolve the name
		try
		{
			org.omg.CORBA.Object target = null;
			
			// Lookup based upon the type of addresses being used
			if( name.getProtocol().equalsIgnoreCase( "IOR" ) )
			{
				DIAGNOSTICS.println( "Trying string_to_object() with IOR", this );
				
				// Resolve the URL & check its OK
				target = _orb_ref.string_to_object ( name.toString() );
				
				// Give appropriate error reason
				error_reason = "ORB.string_to_object() returned null!";
			}
			else if( name.getProtocol().equalsIgnoreCase( "IIOP" ) )
			{
				try
				{
					DIAGNOSTICS.println( "WARNING: Trying direct NS lookup with IIOP", this );
					
					// Attempt to use CosNaming to resolve reference
					target = nsResolve( name.getHost(), name.getPort(), name.getTarget() );
				}
				catch( Throwable t )
				{
					DIAGNOSTICS.println( "Problem trying direct NS lookup with IIOP", t, this );
				}
				
				error_reason = "Problem resolving Object in remote NS";
			}
			else if( name.getProtocol().equalsIgnoreCase( "CORBANAME" ) )
			{
				DIAGNOSTICS.println( "WARNING: Trying direct NS lookup with IIOP (CORBANAME)", this );
					
				target = corbaNameResolve( name );
				
				error_reason = "Problem resolving Object in remote NS";
			}
			
			
			// Check we've found a target AND it is alive
			if( target != null )
			{
				DIAGNOSTICS.println( "Found a target, testing existence", this );
				
				if( __use_non_existent )
				{
					if( !target._non_existent() )
					{
						DIAGNOSTICS.println( "Target exists", this );
						
						// Narrow reference
						mts_target = MTSHelper.narrow( target );
					}
					else
					{
						DIAGNOSTICS.println( "WARNING: Target doesn't exist according to Object._non_existent() - this feature is " +
						"not widely supported by CORBA implemenations / SUNIDL, assuming it is OK to attempt " +
						"message delivery.  To disable this test, use -D" + CMD_LINE_SWITCH + "=false as a " +
						"command-line option to the JVM", this, DIAGNOSTICS.LEVEL_MAX );
						
						// Narrow reference
						mts_target = MTSHelper.narrow( target );
					}
				}
				else
				{
					DIAGNOSTICS.println( "Assuming target exists....", this );
					
					// Narrow reference
					mts_target = MTSHelper.narrow( target );
				}
			}
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Error when resolving \"" + name + "\": ", t, this,
			DIAGNOSTICS.LEVEL_3 );
		}
		
		if( mts_target == null )
		{
			if( error_reason == null )
			{
				throw new UnknownTargetException( "Can't resolve \"" + name + "\"" );
			}
			else
			{
				throw new UnknownTargetException( "Can't resolve \"" + name + "\": " + error_reason );
			}
		}
		
		return mts_target;
	}
	
	
	/**
	 * Attempt to lookup a reference to the MTS belonging to the given Agent
	 */
	protected Object lookup( String name ) throws UnknownTargetException
	{
		DIAGNOSTICS.println( "Looking up " + name, this );
		
		// Default result
		Object result = null;
		
		try
		{
			// Attempt to locate Agent
			org.omg.CORBA.Object res = locateAgent( name );
			
			// Check object is alive
			if( !res._non_existent() )
			{
				result = res;
			}
		}
		catch( Throwable t )
		{
		}
		
		// Check that the lookup was completed OK
		if( result == null )
		{
			throw new UnknownTargetException( "Couldn't resolve target: " + name );
		}
		
		// Return result
		return result;
	}
	
	
	//----------------------------------------------------------------------
	// PUBLIC METHODS - ExternalMTPBase implementation
	//----------------------------------------------------------------------
	
	/**
	 * This should be invoked when the transport should be safely shutdown
	 */
	public void shutdown()
	{
		DIAGNOSTICS.println( "Shuting down SUNIIOPComms" );
		
		// Ensure super-class releases its resources
		super.shutdownMTPBase();
		
		// Dereference other objects to improve garbage collection
		_proxy = null;
		_url = null;
		_orb_ref = null;
		_root_context = null;
	}
	
	
	/**
	 * Method to retreive the protocol dependant address of an object that can be used
	 * to receive messages.
	 *
	 * @return A String containing the required address, or null if this Comms
	 *					doesn't publish such an object
	 */
	public List getAddresses()
	{
		List l = new LinkedList( _url );
		//		List l = new LinkedList();
		//		l.add( _url );
		return l;
	}
	
	
	/**
	 * Retreives the protocols known by this comms implementation
	 *
	 * @return A List of String's
	 */
	public List getProtocols()
	{
		return Arrays.asList( TRANSPORT_PROTOCOLS );
	}
	
	
	/**
	 * Invoked to send the given Message to the given target
	 */
	protected void send( Object target, Envelope env, byte msg[] )
	throws UnknownTargetException, UnknownProtocolException, UnableToSendException
	{
		DIAGNOSTICS.println( "Outgoing ACL message is: " + new String(msg), this );
		
		// Check that the target is the appropriate type
		if( target instanceof fipaos.mts.iiop.sun.external.FIPA.MTS )
		{
			// Typecast target
			fipaos.mts.iiop.sun.external.FIPA.MTS mts_target =
			(fipaos.mts.iiop.sun.external.FIPA.MTS) target;
			
			try
			{
				// Convert envelope & send
				FipaMessage fipa_msg = new FipaMessage( convertToIDLEnvelope( env ), msg );
				mts_target.message( fipa_msg );
			}
			catch( Throwable t )
			{
				DIAGNOSTICS.println( "Exception occured whilst sending message: ", t, DIAGNOSTICS.LEVEL_3 );
				throw new UnableToSendException( t.toString() );
			}
		}
		else
		{
			// Throw exception
			throw new UnableToSendException( "Target couldn't be resolved?!?!? - " +
			"incompatible target object type specified" );
		}
	}
	
	//----------------------------------------------------------------------
	// PUBLIC METHODS - MTS implementation
	//----------------------------------------------------------------------
	
	/**
	 * Handle incoming messages
	 */
	public void handleMessage( fipaos.mts.iiop.sun.external.FIPA.FipaMessage msg )
	{
		DIAGNOSTICS.println( "Received message: " + new String( msg.messageBody ), this );
		
		try
		{
			super.handleIncoming( convertToEnvelope(msg.messageEnvelopes ), msg.messageBody );
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Couldn't deal with incoming message for some reason:",
			t, DIAGNOSTICS.LEVEL_MAX );
		}
	}
	
	
	//----------------------------------------------------------------------
	// PRIVATE METHODS
	//----------------------------------------------------------------------
	
	/**
	 * Convert the given IDL envelopes into an Envelope object
	 */
	private static Envelope convertToEnvelope( fipaos.mts.iiop.sun.external.FIPA.Envelope env[] )
	{
		// Construct a new Envelope to contain the information we are about
		// to extract
		Envelope envelope = new Envelope();
		
		// Iterate through the IDL-Envelope objects
		for( int i = 0; i < env.length; i++ )
		{
			// If this is not the "first" enevelope, indicate that we are about
			// to add more info.
			if( i > 0 )
			{
				envelope.newSubEnvelope();
			}
			
			// Set the various parts of the Envelope
			
			// Check if we should add an ACL representation
			if( env[i].aclRepresentation != null && env[i].aclRepresentation.length() > 0 )
			{
				envelope.setACLRepresentation( env[i].aclRepresentation );
			}
			
			// Check if we should add comments
			if( env[i].comments != null && env[i].comments.length() > 0 )
			{
				envelope.setComments( env[i].comments );
			}
			
			// Check if we should add the creation date
			if( env[i].date != null && env[i].date.length == 1 )
			{
				envelope.setDate( convertToUTCTime( env[i].date[0] ) );
			}
			
			// Check if we should add encryption details
			if( env[i].encrypted != null && env[i].encrypted.length != 0 )
			{
				envelope.setEncrypted( Arrays.asList( (Object[]) env[i].encrypted ) );
			}
			
			// Check if we should add from field
			if( env[i].from != null && env[i].from.length == 1 )
			{
				envelope.setSender( convertToAgentID( env[i].from[0] ) );
			}
			
			// Check if we should add intended receiver field
			if( env[i].intendedReceiver != null && env[i].intendedReceiver.length != 0 )
			{
				// Create a list of intended receivers
				List receivers = new LinkedList();
				for( int j = 0; j < env[i].intendedReceiver.length; j ++ )
				{
					// Convert receiver
					receivers.add( convertToAgentID( env[i].intendedReceiver[j] ) );
				}
				
				envelope.setIntendedReceiver( receivers );
			}
			
			// Check if we should add the payload encoding information
			if( env[i].payloadEncoding != null && env[i].payloadEncoding.length() > 0 )
			{
				envelope.setPayloadEncoding( env[i].payloadEncoding );
			}
			
			// Check if we should add the payload length information
			if( env[i].payloadLength > 0 )
			{
				envelope.setPayloadLength( env[i].payloadLength );
			}
			
			if( env[i].received != null && env[i].received.length == 1 )
			{
				// Add the received information
				envelope.setReceived( convertToReceived( env[i].received[0] ) );
			}
			
			// Check if we should add to field
			if( env[i].to != null && env[i].to.length != 0 )
			{
				// Create a list of intended receivers
				List receivers = new LinkedList();
				for( int j = 0; j < env[i].to.length; j ++ )
				{
					// Convert receiver
					receivers.add( convertToAgentID( env[i].to[j] ) );
				}
				
				envelope.setReceivers( receivers );
			}
			
			// Check if we should add the transport behaviour field
			if( env[i].transportBehaviour != null && env[i].transportBehaviour.length == 1 )
			{
				// Create a TransportBehaviour object to represent this part of the envelope
				TransportBehaviour tb = new TransportBehaviour();
				envelope.setTransportBehaviour( tb );
				
				// Iterate through properties
				for( int j = 0; j < env[i].transportBehaviour[0].length; i++ )
				{
					Property p = env[i].transportBehaviour[0][j];
					String keyword = p.keyword;
					String value = p.value.extract_string();
					tb.setProperty( keyword, value );
				}
			}
			
			// Check if we should add the user-defined properties
			if( env[i].userDefinedProperties != null && env[i].userDefinedProperties.length > 0)
			{
				// Iterate through properties
				for( int j = 0; j < env[i].userDefinedProperties.length; j++ )
				{
					Property p = env[i].userDefinedProperties[j];
					String keyword = p.keyword;
					String value = p.value.extract_string();
					envelope.setSlotValue( keyword, value );
				}
			}
		}
		
		return envelope;
	}
	
	
	/**
	 * Convert an IDL AgentID into an AgentID
	 */
	private static AgentID convertToAgentID( fipaos.mts.iiop.sun.external.FIPA.AgentID aid )
	{
		DIAGNOSTICS.println( "Converting AgentID...." );
		
		// Construct new AgentID
		AgentID new_aid = new AgentID();
		
		// Populate!
		
		// Set name
		DIAGNOSTICS.println( "Setting name to \"" + aid.name + "\"" );
		new_aid.setName( aid.name );
		
		// Set addresses
		List addresses = new LinkedList();
		for( int i = 0; aid.addresses != null && i < aid.addresses.length; i++ )
		{
			DIAGNOSTICS.println( "Adding address \"" + aid.addresses[i] + "\"" );
			addresses.add( new URL( aid.addresses[i] ) );
		}
		new_aid.setAddresses( addresses );
		
		// Set resolvers
		List resolvers = new LinkedList();
		for( int i = 0; aid.resolvers != null && i < aid.resolvers.length; i++ )
		{
			DIAGNOSTICS.println( "Adding resolver...." );
			resolvers.add( convertToAgentID( aid.resolvers[i] ) );
		}
		new_aid.setResolvers( resolvers );
		
		// Return new representation
		return new_aid;
	}
	
	
	/**
	 * Convert an IDL DateTime into a UTCTime
	 */
	private static UTCTime convertToUTCTime( fipaos.mts.iiop.sun.external.FIPA.DateTime date )
	{
		DIAGNOSTICS.println( "Converting UTCTime..." );
		DIAGNOSTICS.println( "\"" + date.year + ":" + date.month + ":" + date.day +
		":" + date.hour + ":" + date.hour + ":" + date.seconds +
		":" + date.milliseconds + ":" + date.typeDesignator + "\"" );
		
		
		String type_designator = null;
		
		if( 'a' <= date.typeDesignator && 'z' >= date.typeDesignator )
		{
			type_designator = "" + date.typeDesignator;
			type_designator.toUpperCase();
		}
		else if( 'A' <= date.typeDesignator && 'Z' >= date.typeDesignator )
		{
			type_designator = "" + date.typeDesignator;
		}
		
		return new UTCTime( (int)date.year, (int)date.month, (int)date.day, (int)date.hour,
		(int)date.hour, (int)date.seconds, (int)date.milliseconds,
		type_designator );
	}
	
	
	/**
	 * Convert an IDL ReceivedObject into a Received
	 */
	private static Received convertToReceived( fipaos.mts.iiop.sun.external.FIPA.ReceivedObject rcvd )
	{
		DIAGNOSTICS.println( "Converting Received..." );
		URL from = null;
		String id = null;
		String via = null;
		
		// Only attempt to create from URL if contained
		if( rcvd.from != null && rcvd.from.length() > 0 )
		{
			DIAGNOSTICS.println( "Setting from URL to \"" + rcvd.from + "\"" );
			from = new URL( rcvd.from );
		}
		
		// Populate id & via if necessary
		if( rcvd.id != null && rcvd.id.length() > 0 )
		{
			DIAGNOSTICS.println( "Setting id to \"" + rcvd.id + "\"" );
			id = rcvd.id;
		}
		
		if( rcvd.via != null && rcvd.via.length() > 0 )
		{
			DIAGNOSTICS.println( "Setting via to \"" + rcvd.via + "\"" );
			via = rcvd.via;
		}
		
		DIAGNOSTICS.println( "Received by field = \"" + rcvd.by + "\"" );
		
		// Construct new received object
		Received r = new Received( from, new URL( rcvd.by ), id, via,
		convertToUTCTime( rcvd.date ) );
		
		return r;
	}
	
	
	/**
	 * Convert Envelope into an IDL Envelope
	 */
	private fipaos.mts.iiop.sun.external.FIPA.Envelope[] convertToIDLEnvelope( Envelope envelope )
	{
		// Construct a new Envelope to contain the information we are about
		// to extract
		List envelopes = envelope.getSubEnvelopes();
		fipaos.mts.iiop.sun.external.FIPA.Envelope envs[] =
		new fipaos.mts.iiop.sun.external.FIPA.Envelope[ envelopes.size() ];
		
		// Iterate through the Envelope objects
		Iterator i = envelopes.iterator();
		int count = 0;
		while( i.hasNext() )
		{
			// Get the next Envelope
			Envelope env = (Envelope) i.next();
			
			
			// Create the next envelope
			envs[count] = new fipaos.mts.iiop.sun.external.FIPA.Envelope();
			fipaos.mts.iiop.sun.external.FIPA.Envelope idl_env = envs[count++];
			
			
			// Set the various parts of the Envelope
			
			// Check if we should add an ACL representation
			if( env.getACLRepresentation() != null )
			{
				idl_env.aclRepresentation = env.getACLRepresentation();
			}
			else
			{
				idl_env.aclRepresentation = "";
			}
			
			
			// Check if we should add comments
			if( env.getComments() != null )
			{
				idl_env.comments = env.getComments();
			}
			else
			{
				idl_env.comments = "";
			}
			
			
			// Check if we should add the creation date
			if( env.getDate() != null )
			{
				idl_env.date = new DateTime[]
				{ convertToIDLDateTime( env.getDate() ) };
			}
			else
			{
				idl_env.date = new DateTime[0];
			}
			
			
			// Check if we should add encryption details
			if( env.getEncrypted() != null )
			{
				idl_env.encrypted = (String[]) env.getEncrypted().toArray( new String[0] );
			}
			else
			{
				// Ensure 0 length arrays are assigned for safety
				idl_env.encrypted = new String[0];
			}
			
			
			// Check if we should add from field
			if( env.getSender() != null )
			{
				idl_env.from = new fipaos.mts.iiop.sun.external.FIPA.AgentID[]
				{
					convertToIDLAgentID( env.getSender() ) };
			}
			else
			{
				idl_env.from = new fipaos.mts.iiop.sun.external.FIPA.AgentID[0];
			}
			
			
			// Check if we should add intended receiver field
			if( env.getIntendedReceiver() != null )
			{
				// Create a list of intended receivers
				List receivers = new LinkedList();
				Iterator j = env.getIntendedReceiver().iterator();
				while( j.hasNext() )
				{
					// Convert receiver
					receivers.add( convertToIDLAgentID( (AgentID) j.next() ) );
				}
				
				// Convert list to an array
				idl_env.intendedReceiver = (fipaos.mts.iiop.sun.external.FIPA.AgentID[])
				receivers.toArray( new fipaos.mts.iiop.sun.external.FIPA.AgentID[0] );
			}
			else
			{
				// Ensure 0 length arrays are assigned for safety
				idl_env.intendedReceiver = new fipaos.mts.iiop.sun.external.FIPA.AgentID[0];
			}
			
			// Check if we should add the payload encoding information
			if( env.getPayloadEncoding() != null )
			{
				idl_env.payloadEncoding	= env.getPayloadEncoding();
			}
			else
			{
				idl_env.payloadEncoding = "";
			}
			
			
			// Check if we should add the payload length information
			if( env.getPayloadLength() > 0 )
			{
				idl_env.payloadLength = env.getPayloadLength();
			}
			
			if( env.getReceived() != null )
			{
				// Add the received information
				idl_env.received = new ReceivedObject[]
				{ convertToIDLReceived( env.getReceived() ) };
			}
			else
			{
				idl_env.received = new ReceivedObject[0];
			}
			
			
			// Check if we should add intended receiver field
			if( env.getReceivers() != null )
			{
				// Create a list of intended receivers
				List receivers = new LinkedList();
				Iterator j = env.getReceivers().iterator();
				while( j.hasNext() )
				{
					// Convert receiver
					receivers.add( convertToIDLAgentID( (AgentID) j.next() ) );
				}
				
				// Convert list to an array
				idl_env.to = (fipaos.mts.iiop.sun.external.FIPA.AgentID[])
				receivers.toArray( new fipaos.mts.iiop.sun.external.FIPA.AgentID[0] );
			}
			else
			{
				// Ensure 0 length arrays are assigned for safety
				idl_env.to = new fipaos.mts.iiop.sun.external.FIPA.AgentID[0];
			}
			
			
			// Check if we should add the transport behaviour field
			if( env.getTransportBehaviour() != null )
			{
				// Get TransportBehaviour object
				TransportBehaviour tb = env.getTransportBehaviour();
				
				// Get list of properties
				List properties = tb.getProperties();
				
				// Create correct size property array
				idl_env.transportBehaviour = new fipaos.mts.iiop.sun.external.FIPA.Property[1][ properties.size() ];
				
				// Populate array
				Iterator j = properties.iterator();
				int k = 0;
				while( j.hasNext() )
				{
					String key = (String) j.next();
					String value = tb.getProperty( key );
					Any any = _orb_ref.create_any();
					any.insert_string( value );
					idl_env.transportBehaviour[0][k] = new Property( key, any );
					k++;
				}
			}
			else
			{
				idl_env.transportBehaviour = new fipaos.mts.iiop.sun.external.FIPA.Property[0][0];
			}
			
			
			// Check if we should add the user-defined properties
			if( env.getUserSlots() != null )
			{
				// Get user defined slot names
				Set slots = env.getUserSlots();
				
				// Create Property[] of appropriate size
				idl_env.userDefinedProperties = new fipaos.mts.iiop.sun.external.FIPA.Property[ slots.size() ];
				
				// Populate Property[]
				Iterator j = slots.iterator();
				int k = 0;
				while( j.hasNext() )
				{
					String key = (String) j.next();
					String value = env.getSlotValue( key ).toString();
					Any any = _orb_ref.create_any();
					any.insert_string( value );
					idl_env.userDefinedProperties[k] = new Property( key, any );
					k++;
				}
			}
			else
			{
				idl_env.userDefinedProperties = new fipaos.mts.iiop.sun.external.FIPA.Property[0];
			}
		}
		
		return envs;
	}
	
	
	/**
	 * Convert the given AgentID into the equivelant IDL AgentID
	 */
	private static fipaos.mts.iiop.sun.external.FIPA.AgentID convertToIDLAgentID( AgentID aid )
	{
		// Construct new AgentID
		fipaos.mts.iiop.sun.external.FIPA.AgentID new_aid =
		new fipaos.mts.iiop.sun.external.FIPA.AgentID();
		
		// Populate!
		
		// Set name
		new_aid.name = aid.getName();
		
		// Set addresses
		List addresses = new LinkedList();
		Iterator i = aid.getAddresses().iterator();
		while( i.hasNext() )
		{
			addresses.add( ( (URL) i.next() ).toString() );
		}
		new_aid.addresses = (String[]) addresses.toArray( new String[0] );
		
		// Set resolvers
		List resolvers = new LinkedList();
		i = aid.getResolvers().iterator();
		while( i.hasNext() )
		{
			resolvers.add( convertToIDLAgentID( (AgentID) i.next() ) );
		}
		new_aid.resolvers = (fipaos.mts.iiop.sun.external.FIPA.AgentID[])
		resolvers.toArray( new fipaos.mts.iiop.sun.external.FIPA.AgentID[0] );
		
		// Set user-defined properties to legnth 0
		new_aid.userDefinedProperties = new fipaos.mts.iiop.sun.external.FIPA.Property[0];
		
		// Return new representation
		return new_aid;
	}
	
	
	/**
	 * Convert a UTCTime to an IDL DateTime
	 */
	private static fipaos.mts.iiop.sun.external.FIPA.DateTime convertToIDLDateTime( UTCTime date )
	{
		// Get date-time
		date = date.getZTime();
		
		return new fipaos.mts.iiop.sun.external.FIPA.DateTime( (short)date.get(UTCTime.YEAR),
		(short)date.get(UTCTime.MONTH), (short)date.get(UTCTime.DATE), (short)date.get(UTCTime.HOUR),
		(short)date.get(UTCTime.MINUTE), (short)date.get(UTCTime.SECOND),
		(short)date.get(UTCTime.MILLISECOND), date.getTimeZone().charAt( 0 ) );
	}
	
	
	/**
	 * Convert a Received object to an IDL ReceivedObject
	 */
	private static fipaos.mts.iiop.sun.external.FIPA.ReceivedObject convertToIDLReceived( Received rcvd )
	{
		String from = "";
		String id = "";
		String via = "";
		
		// Only attempt to create from URL if contained
		if( rcvd.getFrom() != null )
		{
			from = rcvd.getFrom().toString();
		}
		
		// Add other parameters if necessary
		if( rcvd.getId() != null )
		{
			id = rcvd.getId();
		}
		
		// Get the via parameter if necessary
		if( rcvd.getVia() != null )
		{
			via = rcvd.getVia();
		}
		
		// Construct new received object
		return new fipaos.mts.iiop.sun.external.FIPA.ReceivedObject( rcvd.getBy().toString(),
		from, convertToIDLDateTime( rcvd.getDateTime() ), id, via );
	}
	
	
	/**
	 * Resolve a reference to the root context of the CORBA Naming
	 * Service
	 *
	 * @return A reference to the root NamingContext object
	 */
	private NamingContext resolveRootContext()
	{
		try
		{
			// Get the initial naming context
			return NamingContextHelper.narrow(_orb_ref.resolve_initial_references("NameService"));
		}
		catch ( SystemException e )
		{
			return null;
		}
		catch ( org.omg.CORBA.ORBPackage.InvalidName e )
		{
			return null;
		}
	}
	
	
	/**
	 * Resolve a reference to the appropriate context within the root context
	 * of the CORBA Naming Service, if the context cannot be resolved
	 * then create a new context
	 *
	 * @return True if everything went OK
	 */
	private boolean resolveContext()
	{
		boolean cold_start = true;
		
		try
		{
			NameComponent[] name = new NameComponent[ 1 ];
			name[ 0 ] = new NameComponent( NS_TYPE, NS_KIND );
			
			try
			{
				_root_context.resolve( name );
				cold_start = false;
			}
			catch ( InvalidName se )
			{
			}
			catch ( CannotProceed se )
			{
			}
			catch ( NotFound se )
			{
			}
			catch ( org.omg.CORBA.SystemException  se )
			{
			}
			
			if( cold_start )
			{
				_root_context.bind_new_context( name );
			}
			
			return true;
		}
		catch ( InvalidName e )
		{
		}
		catch ( CannotProceed e )
		{
		}
		catch ( NotFound e )
		{
		}
		catch ( AlreadyBound e )
		{
		}
		catch ( SystemException e )
		{
		}
		
		return false;
	}
	
	
	/**
	 * Rebind to the CORBA Naming Service
	 *
	 * @param name Name to bind to NS using
	 *
	 * @return True if succesfull
	 */
	private boolean rebind( String name )
	{
		NameComponent[] nc = new NameComponent[ 1 ];
		nc[ 0 ] = new NameComponent( name, NS_KIND );
		
		try
		{
			_root_context.rebind( nc, _proxy );
			return true;
		}
		catch ( InvalidName e )
		{
		}
		catch ( CannotProceed e )
		{
		}
		catch ( NotFound e )
		{
			try
			{
				_root_context.bind( nc, _proxy );
				return true;
			}
			catch( Exception f )
			{
			}
		}
		catch ( SystemException e )
		{
		}
		
		return false;
	}
	
	
	/**
	 * Deregister an agents message handler from the CORBA Naming Service
	 *
	 * @param name Entry to unbind from the NS
	 */
	private void deregister( String name )
	{
		NameComponent[] nc = new NameComponent[ 1 ];
		nc[ 0 ] = new NameComponent(  name, NS_KIND );
		
		try
		{
			_root_context.unbind( nc );
		}
		catch ( InvalidName e )
		{
		}
		catch ( CannotProceed e )
		{
		}
		catch ( NotFound e )
		{
		}
		catch ( SystemException e )
		{
		}
	}
	
	/**
	 * Obtain a CORBA object reference
	 *
	 * @param name Name of the object within the NS
	 *
	 * @return A reference to the object, or null
	 */
	private org.omg.CORBA.Object locateAgent( String name ) throws  NotFound, InvalidName
	{
		org.omg.CORBA.Object object_ref = null;
		
		try
		{
			NameComponent[] nc = new NameComponent[ 2 ];
			nc[ 0 ] = new NameComponent( NS_TYPE, NS_KIND );
			nc[ 1 ] = new NameComponent( name, NS_KIND );
			object_ref = _root_context.resolve( nc );
		}
		catch ( CannotProceed e )
		{
		}
		catch ( SystemException e )
		{
		}
		
		// Check if we have an object ref. and if not, is it because the name given inlcudes
		// a HAP address?
		if( object_ref == null && name.indexOf( '@' ) != -1 )
		{
			object_ref = locateAgent( name.substring( name.indexOf( '@' ) ) );
		}
		
		return object_ref;
	}
	
	
	/**
	 * Attempt to resolve the given CORBANAME url
	 */
	public static org.omg.CORBA.Object corbaNameResolve( URL name )
	{
		try
		{
			// Get part of URL with real info!
			String url_target = name.getHost();

			// Get NS target
			int index_target = url_target.indexOf( '#' );
			String ns_target = url_target.substring( index_target + 1 );

			// Get host, port and service
			String host_port_service = url_target.substring( 1, index_target );

			// Check for service
			int index_service = host_port_service.indexOf( '/' );
			String service = "NameService";
			if( index_service != -1 )
			{
				service = host_port_service.substring( index_service + 1 );
			}
			else
			{
				index_service = host_port_service.length();
			}

			// Check for port
			int index_port = host_port_service.indexOf( ':' );
			String port_num = "900";
			if( index_port != -1 )
			{
				port_num = host_port_service.substring( index_port + 1, index_service );
			}
			else
			{
				index_port = index_service;				
			}
			int port = Integer.parseInt( port_num );

			// Check for host
			String host = host_port_service.substring( 0, index_port );

			// Attempt to use CosNaming to resolve reference
			return MTSHelper.narrow( nsResolve( host, port, ns_target ) );
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Error resolving " + name, t );
		}
		
		return null;
	}
	
	
	/**
	 * Attempt to resolve the target at the given CosNamingService located at the
	 * given ORB server
	 */
	public static org.omg.CORBA.Object nsResolve( String host, int port, String url_target ) throws Exception
	{
		org.omg.CORBA.Object target = null;
		
		// Get reference to remote NS
		Properties props = new Properties();
		props.put("org.omg.CORBA.ORBInitialPort", "" + port );
		props.put("org.omg.CORBA.ORBInitialHost", host );
		ORB remote_ref = ORB.init( new String[0], props );
		
		// Get reference to remote NS
		NamingContext nc = NamingContextHelper.narrow( remote_ref.resolve_initial_references("NameService"));
		
		// Recursively iterate through target name until we reach "IDL:FIPA/MTS:1.0"
		StringTokenizer st = new StringTokenizer( url_target, "/" );
		while( st.hasMoreTokens() )
		{
			// Get next component
			String next = st.nextToken();
			
			// Attempt to get CORBA object
			org.omg.CORBA.Object obj = findEntry( next, nc );
			
			// Attempt to narrow
			try
			{
				MTSHelper.narrow( obj );
				target = obj;
			}
			catch( Throwable t )
			{
				DIAGNOSTICS.println( "Exception when resolving \"" + next + "\"", t );
				nc = NamingContextHelper.narrow( obj );
			}
		}
		
		return target;
	}
	
	
	/**
	 * Find the given entry in the given NamingContext
	 */
	private static org.omg.CORBA.Object findEntry( String name, NamingContext nc ) throws Exception
	{
		// Attempt to iterate through given NC for the given name
		BindingListHolder blh = new BindingListHolder();
		BindingIteratorHolder bih = new BindingIteratorHolder();
		nc.list( 1000, blh, bih );
		Binding bindings[] = blh.value;
		
		// Iterate through objects found, until we reach the end or find a null
		// entry
		for( int i = 0; i < bindings.length && bindings[ i ] != null; i++ )
		{
			NameComponent name_comp[] = bindings[ i ].binding_name;
			if( name_comp[ name_comp.length - 1 ].id.equalsIgnoreCase( name ) )
			{
				return nc.resolve( name_comp );
			}
		}
		
		return null;
	}
	
	
	/**
	 * Listener class for the MTS
	 */
	public class MTSListener extends _MTSImplBase
	{
		/**
		 * Default constructor
		 */
		public MTSListener()
		{
		}
		
		/**
		 * Handles incoming messages by passing them to its parent class
		 */
		public void message( fipaos.mts.iiop.sun.external.FIPA.FipaMessage aFipaMessage )
		{
			// Pass message to parent
			handleMessage( aFipaMessage );
		}
	}
	
	//----------------------------------------------------------------------
	// TEST HARNESS
	//----------------------------------------------------------------------
	
	/**
	 * Test harness for transport
	 *
	 * Use java fipaos.comm.sunidl.SUNIDLComms /? for usage details
	 */
	public static void main( String args[] ) throws Exception
	{
		// Display all output
		DIAGNOSTICS.setMemoryLevel( DIAGNOSTICS.LEVEL_MIN );
		
		URL url_a = new URL( args[0] );
		URL url_b = new URL( args[1] );
		String agent_name_a = url_a.getTarget() + "@localhost";
		String agent_name_b = url_b.getTarget() + "@localhost";
		
		URL url_ns_a = (URL) url_a.clone();
		url_ns_a.setTarget( null );
		
		AgentID aid_a = new AgentID();
		AgentID aid_b = new AgentID();
		
		aid_a.setName( agent_name_a );
		aid_b.setName( agent_name_b );
		aid_a.setAddresses( Arrays.asList( new Object[]
		{ url_a } ) );
		aid_b.setAddresses( Arrays.asList( new Object[]
		{ url_b } ) );
		
		// Create message
		ACL acl_a = new ACL();
		acl_a.setPerformative( "inform" );
		acl_a.setContentObject( "Hello" );
		acl_a.setSenderAID( aid_a );
		acl_a.addReceiverAID( aid_b );
		Message msg_a = new Message( acl_a );
		msg_a.getEnvelope().setIntendedReceiver( Arrays.asList( new Object[]
		{ aid_b } ) );
		
		// Create Comms
		ExternalMTP a = new SUNIIOPComms( agent_name_a, url_ns_a );
		a.initialise( new TestClass(), null );
		
		Message msg_cp = (Message) msg_a.clone();
		
		// Send messages every 5 second
		while( true )
		{
			try
			{
				Thread.sleep( 1000 );
				a.outgoing( msg_a.getEnvelope(), msg_a.getACL().toString().getBytes() );
				msg_a = (Message) msg_cp.clone();
			}
			catch( Throwable t )
			{
				t.printStackTrace();
			}
		}
	}
	
	
	private static class TestClass implements fipaos.mts.service.PostParserService
	{
		int _received = 0;
		
		public void initialise( fipaos.mts.service.PostParserService previous,
		fipaos.mts.service.PostParserService next )
		{
		}
		
		public void outgoing( Envelope env, byte body[] )
		{
		}
		
		public void incoming( Envelope env, byte body[] )
		{
			if( env.getErrorCode() == NONE )
			{
				DIAGNOSTICS.println( "Received message " + (_received++) + ": Envelope = " +
				env + "    Body = " + new String( body ) );
			}
			else
			{
				DIAGNOSTICS.println( "Message \"bounced\"" );
			}
		}
		
		public void shutdown()
		{
		}
	}
}