//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.mts.rmi.internal;

//----------------------------------------------------------------------

import fipaos.mts.*;
import fipaos.mts.service.*;
import fipaos.mts.rmi.internal.ns.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.util.*;
import fipaos.mts.rmi.internal.ns.RMIAgentNamingService;
import fipaos.mts.rmi.internal.ns.RMIAgentNamingServiceImpl;

import java.rmi.*;
import java.rmi.server.*;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.net.MalformedURLException;
import java.net.UnknownHostException;
import java.net.InetAddress;
import java.util.*;

  /**
   * This class is a Comms implementation which offers support
   * for RMI.         
   * <br>
   * <br>
   */
public class RMIComms extends InternalMTPBase
{
//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------

  /**
   * Transport specific protocol type
   */
  private static final String TRANSPORT_PROTOCOL = "fipaos-rmi";
  
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

	/**
	 * Reference to NamingService
	 */
	private RMIAgentNamingService _naming_service;

	/**
	 * Reference to MTS object to export
	 */
	private MTS _proxy;
	
	/**
	 * Reference to the URL of this MTS
	 */
	private URL _url;
	
//----------------------------------------------------------------------
// PUBLIC METHODS - Comms implementation
//---------------------------------------------------------------------- 

	/**
 	 * Comms initialiser
 	 *
 	 * @param our_name The name to use when binding to the transport layers
 	 *								 naming service - should include HAP address, i.e. "agentA@hap"
 	 */
	public RMIComms( String name, URL ns ) throws InitialisationException, RemoteException
	{
		// Invoke super class constructor
		super( name, ns );
		
		// Construct our URL
		_url = (URL) ns.clone();
		_url.setTarget( _name );
		
		// Check if URL host is "localhost"
		if( _url.getHost() != null && _url.getHost().equalsIgnoreCase( "localhost" ) )
		{
			// We need to substitute this hosts' IP address
			try
			{
				String host = InetAddress.getLocalHost().getHostAddress();
				_url.setHost( host );
			}
			catch( Throwable t )
			{
				DIAGNOSTICS.println( "Error when trying to get IP address of this host", t,
															this, DIAGNOSTICS.LEVEL_2 );
			}
		}
	}


//----------------------------------------------------------------------
// PROTECTED METHODS - InternalMTPBase implementation
//---------------------------------------------------------------------- 

	/**
	 * Bind the MTP to the NS (if it exists) and perform other start-up operations
	 */
	protected void bind() throws InitialisationException
	{
		// Initialise comms if necessary
		if( _naming_service == null )
		{
			try
			{
				// Locate NS & create proxy object
				_naming_service = findNS( _ns_url );
				_proxy = getMTSListener( this );
			}
			catch( InitialisationException e )
			{
				throw e;
			}
			catch( Throwable e )
			{
				DIAGNOSTICS.println( "Exception when initialising RMIComms", e, this );
				throw new InitialisationException( "Error when locating / contacting NS: " + 
																						e.toString() );
			}
		}
		
		// Attempt to bind to the NS
		try
		{
			DIAGNOSTICS.println( "Binding \"" + _name + "\" to RMINS" );
			_naming_service.bind( _name, _proxy );
			DIAGNOSTICS.println( "Bound \"" + _name + "\" to RMINS" );
		}
		catch( RemoteException e )
		{
			throw new InitialisationException( "Error when attempting (re)bind to NS: " +
																					e.getMessage() );
		}
	}
	
	
	/**
	 * Un-bind the MTP to the NS (if it exists)
	 */
	protected void unbind()
	{
		try
		{
			boolean unbind = false;
			
			// Attempt to determine if reference in the NS belongs to us
			MTS proxy = resolve( _url );
			unbind = ( proxy.equals( _proxy ) );
			
			// Unbind if reference in NS refers to this!
			if( unbind )
			{
				DIAGNOSTICS.println("Unbinding self from NS");
				_naming_service.unbind( _name );
			}
		}
		catch( RemoteException e )
		{
		}
		catch( UnknownTargetException e )
		{
		}
	}
	
	
	/**
	 * Lookup a reference to the MTS at the given URL
	 */
	protected Object lookup(URL name) throws UnknownTargetException
	{
		// Resolve URL
		MTS target = resolve( name );
		
		// Check target has been resolved
		if( target == null )
		{
			throw new UnknownTargetException( "Couldn't resolve target: " + name );	
		}
		
		return target;
	}
	
	
	/**
	 * Attempt to lookup a reference to the MTS belonging to the given Agent
	 */
	protected Object lookup( String name ) throws UnknownTargetException
	{
		// Attempt to create URL for name
		URL url = getLocalNSAgentURL( name, _ns_url, _hap );
		
		// If we managed to create a URL, return a reference to the object
		if( url != null )
		{
			return lookup( url );
		}
		else
		{
			// Couldn't lookup the object
			throw new UnknownTargetException( "Couldn't re-create address for Agent \"" + name + "\"" );
		}
	}
	

	/**
	 * Invoked to send the given Message to the given target
	 */
	protected void send( Object target, Message msg ) throws UnableToSendException
	{
		// Check the type of the target
		if( target instanceof MTS )
		{
			// Cast and send!
			MTS mts_target = (MTS) target;
			
			// Check other end is alive
			try
			{
				mts_target.ping();
			}
			catch( Throwable t )
			{
				throw new UnableToSendException( "Error when sending message to target - cached reference stale?: " +
																					t.toString() );
			}
			
			try
			{
				// Send the message!
				mts_target.message( msg );
			}
			catch( Throwable t )
			{
				// Something went wrong - best report it!
				DIAGNOSTICS.println( "Exception when sending message", t, this );
				throw new UnableToSendException( "Error when sending message to target:" +
																					t.toString() );
			}
		}
		else
		{
			// Can't send message!
			throw new UnableToSendException( "Target not of correct type!" );
		}
	}

	
	/**
	 * Method to retreive the protocol dependant address of an object that can be used
	 * to receive messages.
	 *
	 * @return A URL containing the required address, or null if this Comms
	 *					doesn't publish such an object
	 */
	public List getAddresses()
	{
		List l = new LinkedList();
		l.add( _url );
		return l;
	}
	
	
	/**
	 * Retreives the protocols known by this comms implementation
	 *
	 * @return A List of String's 
	 */
	public List getProtocols()
	{
		return Arrays.asList( new String[] { TRANSPORT_PROTOCOL } );
	}
	
	
	/**
	 * This should be invoked when the transport should be safely shutdown
	 */
	public void shutdown()
	{
		DIAGNOSTICS.println("Shuting down RMIComms");
		
		// Make sure super-class releases resources!
		super.shutdownMTPBase();
		
		// Release references to other objects
		_proxy = null;
		_naming_service = null;
		_url = null;
	} 

	
	public void handleMessage( Message msg )
	{
		super.handleIncoming( msg );
	}
	
//----------------------------------------------------------------------
// PUBLIC STATIC METHODS - For use within sub-packages of fipaos.mts.rmi.int
//---------------------------------------------------------------------- 

	/**
	 * This method will create a new Registry locally if there isn't already
	 * one running
	 */
	public synchronized static void startupRegistry( int port ) throws InitialisationException
	{
		try
		{
			// Attempt to create a local registry
			LocateRegistry.createRegistry( port );
			
			DIAGNOSTICS.println( "Created RMIRegistry on port " + port );
		}
		catch( RemoteException e )
		{
			// Can't create one? There may be a Registry already, lets attempt to 
			// locate it
			try
			{
				Registry r = LocateRegistry.getRegistry( port );
				
				// See if we can actually do anything!
				r.list();
				
				DIAGNOSTICS.println("Located RMIRegistry on port " + port );
			}
			catch( RemoteException f )
			{
				// We're in trouble if we reach here - we can't create or locate a Registry!
				throw new InitialisationException( "Couldn't locate/create local registry: " + f );
			}
		}
	}


	/**
	 * This method will (re)bind the given entry into the local Registry
	 * NOTE: RMIRegistry's only allow local objects to be bound to them,
	 * hence the need for an additional (lightweight) Naming-Service.
	 *
	 * @param name Name to bind the object into the Registry using
	 * @param r The Remote object reference to bind
	 */
	public static void bindToRegistry( String name, Remote r, int port ) throws InitialisationException
	{
		try
		{
			DIAGNOSTICS.println("Attempting to bind \"" + name + "\" into RMIRegistry" );

			// Get a reference to the local registry
			Registry local_reg = LocateRegistry.getRegistry( port );

			// Attempt to bind the references
			local_reg.rebind( name, r );

			DIAGNOSTICS.println("Bound \"" + name + "\" OK" );
		}
		catch( RemoteException e )
		{
			// Not permited to do that!
			throw new InitialisationException( "Couldn't bind object to local registry: " + 
																					e.getMessage() );
		}
	}

//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------

	/**
	 * Method to resolve the given URL to a MTS RMI "server"
	 *
	 * @param url URL to interogate the RMIAgentNamingService about
	 *
	 * @return A MTS instance if everything went OK, null otherwise
	 */
	private MTS resolve( URL url ) throws UnknownTargetException
	{
		MTS target = null;
		
		try
		{
			DIAGNOSTICS.println( "Resolving " + url );

			// No entry? Use the NS
			RMIAgentNamingService ns = null;
			
			// Check if we should look in our NS or remote NS
			if( getNSURL( url ).equals( getNSURL( _ns_url ) ) )
			{
				DIAGNOSTICS.println( "Target " + url + " is local" );
				
				// Use local NS
				ns = _naming_service;
			}
			else
			{
				DIAGNOSTICS.println( "Target " + url + " is remote" );				

				// Locate remote NS
				ns = findNS( url );
			}
			
			try
			{
				// Attempt to lookup target in NS
				target = ns.lookup( getNSAgentName( url ) );
			}
			catch( Throwable t )
			{
				// Check if we're dealing with local NS
				if( ns == _naming_service )
				{
					// See if we can re-locate the NS
					_naming_service = findNS( _ns_url );
					
					// Try again
					target = _naming_service.lookup( getNSAgentName( url ) );
				}
				else
				{
					// Continue throwing exception
					throw t;
				}
			}
		}
		catch( Throwable e )
		{
			throw new UnknownTargetException("Unable to lookup target name: " + e.getMessage() );
		}
		
		return target;
	}


	/**
	 * Method to locate the RMIAgentNamingService at the given host
	 * If the given host is the localhost & either a RMIRegistry isn't
	 * active, or a RMIAgentNamingService isn't found, they will be started
	 *
	 * @param host Host to interogate for the RMIAgentNamingService
	 *
	 * @return A reference to the RMIAgentNamingService for the given host
	 */
	private static RMIAgentNamingService findNS( URL ns_url ) throws InitialisationException, 
		ConnectException, NoSuchObjectException, NotBoundException
	{
		RMIAgentNamingService ns = null;

		try
		{
			// Workout URL for NamingService
			String url = getNSURL( ns_url );
			
			DIAGNOSTICS.println( "Attempting to resolve NS " + url );				

			// Lookup the object we want
			ns = (RMIAgentNamingService) Naming.lookup( url );
		}
		catch( MalformedURLException e )
		{
			throw new InitialisationException( "Error when locating NS: " + e.getMessage() );
		}
/*		catch( ConnectException e )
		{
			// Attempt to remotely activate the NS
			DIAGNOSTICS.println( "Error connecting to NS", e );
		}
		catch( NoSuchObjectException e )
		{
			// Attempt to remotely activate the NS
			DIAGNOSTICS.println( "Error connecting to NS", e );
		}
		catch( NotBoundException e )
		{
			// Attempt to remotely activate the NS
			DIAGNOSTICS.println( "Error connecting to NS", e );
		}
*/		catch( RemoteException e )
		{
			throw new InitialisationException( "Error when attempting to find NS (RMI-NS started?) " +
																					e.getMessage() );
		}
		
		return ns;
	}
	

	/**
	 * Method to activate the RMIAgentNamingService locally
	 *
	 * @param host The host that the NamingService should be located on
	 * @param port The port that the registry is expected to be found on
	 *
	 * @return A reference to the NS
	 */
	private static RMIAgentNamingService activateNS( URL ns_url ) throws InitialisationException
	{
		try
		{
			InetAddress _ns_host = InetAddress.getByName( ns_url.getHost() );
			InetAddress _local = InetAddress.getLocalHost();
			String host = ns_url.getHost();
		
			// Check if the host given is actually the localhost
			if( host.equals("localhost") || host.equals("127.0.0.1") || _ns_host.equals( _local ) )
			{
				DIAGNOSTICS.println( "Creating RMINS locally" );				

				// Make sure local Registry has been started before continuing
				startupRegistry( ns_url.getPort() );

				// Create a local NS
				return (RMIAgentNamingService) new RMIAgentNamingServiceImpl( ns_url );
			}
			else
			{
				// FIXME: Not implemented yet pending investigation into RMI Activation
				// Remotely activate the NS
				throw new InitialisationException( "Couldn't activate NS because " + host +
																						" is not the localhost.  Remote " +
																						"activation is not yet supported." );
			}
		}
		catch( UnknownHostException e )
		{
			throw new InitialisationException( "Couldn't activate NS: " + e.getMessage() );	
		}
		catch( RemoteException e )
		{
			throw new InitialisationException( "Couldn't activate NS: " + e.getMessage() );	
		}
	}

	
	/**
	 * Get the RMI-compliant URL of a NS, given a URL
	 */
	private static String getNSURL( URL url )
	{
		try
		{
			// Construct the URL
			StringBuffer out = new StringBuffer();
			out.append( "//" );
			out.append( url.getHost() );
			out.append( ":" );
			out.append( url.getPort() );
			out.append( "/" );
			out.append( RMIAgentNamingServiceImpl.RMIREGISTRY_NAME );
			return out.toString();
		}
		catch( Throwable t )
		{
		}
		
		return null;
	}
	
	
	/**
	 * Get the name of a particular Agent, given a URL
	 */
	private static String getNSAgentName( URL url )
	{
		try
		{
			String out = url.getTarget();
			return out;
		}
		catch( Throwable t )
		{
		}
		
		return null;
	}
	
	
	/**
	 * Get the URL for a particular Agent name within a given NS
	 */
	private static URL getLocalNSAgentURL( String name, URL ns, String hap )
	{
		// Check if this is a "local" Agent
		int at_index = name.indexOf( '@' );
		if( at_index != -1 )
		{
			// See if HAP of Agent name is same as local HAP
			String agent_hap = name.substring( at_index + 1 );
			name = name.substring( 0, at_index );
			if( !agent_hap.equalsIgnoreCase( hap ) )
			{
				// If not, we can't create a URL for the Agent
				return null;
			}
		}
		
		// Duplicate NS URL
		URL out = (URL) ns.clone();
		
		// Construct Agent's URL
		out.setTarget( name );
		
		return out;
	}
	
	/**
	 * Method to get an instance of MTSListener
	 */
	protected MTSListener getMTSListener( RMIComms parent ) throws RemoteException
	{
		return new MTSListener( parent );
	}
	
	/**
	 * Class that is actually bound to the NS
	 */
	public class MTSListener extends UnicastRemoteObject implements MTS
	{
		RMIComms _parent;
		
		public MTSListener( RMIComms parent ) throws RemoteException
		{
			_parent = parent;
		}
		
		/**
		 * All messages to an Agent should be passed using this method
		 *
		 * @param msg The Message encapsulating the envelope, ACL and message content
		 */
		public void message( Message p1 ) throws java.rmi.RemoteException
		{
			// Forward into RMIComms
			_parent.handleMessage( p1 );
		}
		
		/**
		 * Simple ping to check if an Agent is still alive
		 */
		public void ping() throws java.rmi.RemoteException
		{
			if( _shutdown )
			{
				throw new java.rmi.RemoteException( "This MTS has shutdown!" );
			}
		}
	}
	
//----------------------------------------------------------------------
// TEST HARNESS
//---------------------------------------------------------------------- 

	/**
	 * Test harness for transport
	 */
	public static void main( String args[] ) throws Exception
	{
		// Display all output
		DIAGNOSTICS.setMemoryLevel( DIAGNOSTICS.LEVEL_MIN );
		
		URL url_a = new URL( args[0] );
		URL url_b = new URL( args[1] );
		String agent_name_a = getNSAgentName( url_a ) + "@localhost";
		String agent_name_b = getNSAgentName( url_b ) + "@localhost";
		
		URL url_ns_a = new URL( TRANSPORT_PROTOCOL + ":" + getNSURL( url_a ) );

		AgentID aid_a = new AgentID();
		AgentID aid_b = new AgentID();
		
		aid_a.setName( agent_name_a );
		aid_b.setName( agent_name_b );
		
		// Create message
		ACL acl_a = new ACL();
		acl_a.setContentObject( "Hello" );
		acl_a.setSenderAID( aid_a );
		acl_a.addReceiverAID( aid_b );
		Message msg_a = new Message( acl_a );
		msg_a.getEnvelope().setIntendedReceiver( Arrays.asList( new Object[] { aid_b } ) );
		
		// Create Comms
		InternalMTP a = new RMIComms( getNSAgentName( url_a ) + "@localhost", url_ns_a );
		a.initialise( new TestClass(), null );

		aid_a.setAddresses( a.getAddresses() );
		aid_b.setAddresses( Arrays.asList( new Object[] { url_b } ) );
		
		Message msg_cp = (Message) msg_a.clone();
		
		// Send messages every 5 second
		while( true )
		{
			try
			{
				long time = System.currentTimeMillis();
				a.outgoing( msg_a );
				time = System.currentTimeMillis() - time;
				System.out.println( "Time to send message = " + time );
				msg_a = (Message) msg_cp.clone();
				Thread.sleep( 1000 );
			}
			catch( Throwable t )
			{
				t.printStackTrace();
			}
		}
	}
	
	
	private static class TestClass implements PreParserService
	{
		int _received = 0;
		
		public void initialise( PreParserService previous, PreParserService next )
		{
		}

		public void outgoing( Message msg )
		{
		}

		public void incoming( Message msg )
		{
			DIAGNOSTICS.println( "Received message " + (_received++) + ": " + msg );
		}
		
		public void shutdown()
		{
		}
	}
}
