//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.mts.rmi.internal.ns;

//----------------------------------------------------------------------

import java.util.*;
import java.rmi.*;
import java.rmi.server.*;
import fipaos.agent.profile.*;
import fipaos.mts.*;
import fipaos.mts.rmi.internal.*;
import fipaos.util.*;
import fipaos.skill.db.*;

/**
 * This class provides a NamingService implementation for RMIComms
 * <br>
 * <br>
 */
public class RMIAgentNamingServiceImpl extends UnicastRemoteObject
																				implements RMIAgentNamingService
{
	
	public static final String RMIREGISTRY_NAME = "RMINS";
	
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

	/**
	 * Database to hold Agent references
	 */
	private Database _references;

//----------------------------------------------------------------------
// CONSTRUCTOR
//----------------------------------------------------------------------

	/**
	 * Constructs the NS
	 *
	 * @param URL to be used to locate RMIRegistry 
	 */
	public RMIAgentNamingServiceImpl( URL url ) throws RemoteException
	{
		this( url.getPort() );
	}
	
	
	/**
	 * Constructs the NS
	 *
	 * @param port Port that the registry is expected to be located on 
	 */
	private RMIAgentNamingServiceImpl( int port ) throws RemoteException
	{
		// Do RMI initialisation stuff + start on any port
		super();	
		
		// Setup the local RMIRegistry if necessary, and bind us to it
		try
		{
			// create a DatabaseProfile that describes the database to load
			DatabaseProfile rmi_db_profile = new DatabaseProfileImpl();
			
			rmi_db_profile.setDatabaseType("SerializationDatabase");
			rmi_db_profile.setDatabaseLocation(System.getProperty("user.dir"));
			
			// Setup the local name cache
			_references = DatabaseFactory.getDatabase( "rmi-ns", rmi_db_profile );
			
			// Attempt to open Database
			_references.openDatabase();
			
			// Sync. to ensure that nothing attempts to access NS whilst we're purging
			// references
			synchronized( this )
			{
				RMIComms.startupRegistry( port );
				RMIComms.bindToRegistry( RMIREGISTRY_NAME, this, port );

				// Check that all remote references are still valid
				checkReferences();
			}

			DIAGNOSTICS.println("RMIAgentNamingService initialised OK");
		}
		catch( Throwable e )
		{
			DIAGNOSTICS.println("Problem initialising RMIAgentNamingService", e, this,
																													DIAGNOSTICS.LEVEL_MAX );
			
			throw new RemoteException( "Problem initialising RMIAgentNamingService", e );
		}
	}
	
//----------------------------------------------------------------------
// PUBLIC METHODS - RMIAgentNamingService interface
//----------------------------------------------------------------------

	/**
	 * Attempts to lookup the given AgentGUID & return the corresponding
	 * MTS RMI "server"
	 *
	 * @param name Name to resolve
	 *
	 * @return MTS object registered against the name, or null
	 */
	public synchronized fipaos.mts.rmi.internal.MTS lookup( String name ) throws RemoteException
	{
		NSReference nsref = null;
		
		try
		{
			// Start DB transaction
			_references.startTransaction( true );
			
			// Attempt to locate reference
			Enumeration enum = _references.getIDs();
			while( enum.hasMoreElements() )
			{
				DatabaseObject object = _references.findObject( (String) enum.nextElement() );
				
				// Check if name is the same
				if( object.getObjectID().equalsIgnoreCase( name ) )
				{
					nsref = (NSReference) object;
				}
			}
				
			fipaos.mts.rmi.internal.MTS out = null;
			
			if( nsref != null )
			{
				out = nsref._ref;
			}
			
			DIAGNOSTICS.println( "Looking up " + name + " - " + 
																				( out != null ? "FOUND" : "NOT FOUND" ) );
			return out;
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Error when performing lookup", t, DIAGNOSTICS.LEVEL_MAX );
			throw new RemoteException( "Error when performing lookup", t );
		}
		finally
		{
			try
			{
				// Close transaction
				_references.endTransaction();
			}
			catch( Throwable t )
			{
				DIAGNOSTICS.println( "Exception when closing DB!", t, DIAGNOSTICS.LEVEL_3 );
			}
		}
	}
	
	/**
	 * Attempts to (re)bind the given name against the given
	 * FIPA_Agent_97 RMI "server"
	 *
	 * @param name Name to bind against
	 * @param server FIPA_Agent_97 to bind
	 */
	public synchronized void bind( String name, fipaos.mts.rmi.internal.MTS server ) throws RemoteException
	{
		DIAGNOSTICS.println( "Binding " + name );
		
		try
		{
			// Start DB transaction
			_references.startTransaction( false );
			
			// Create reference
			NSReference nsref = new NSReference( name, server );
			
			// Add reference
			_references.addObject( nsref );
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Error when performing bind", t, DIAGNOSTICS.LEVEL_MAX );
			throw new RemoteException( "Error when performing bind", t );
		}
		finally
		{
			try
			{
				// Close transaction
				_references.endTransaction();
			}
			catch( Throwable t )
			{
				DIAGNOSTICS.println( "Exception when closing DB!", t, DIAGNOSTICS.LEVEL_3 );
			}
		}
	}
		
	/**
	 * Attempts to unbind the given name from its
	 * MTS RMI "server"
	 *
	 * @param name Name to unbind
	 */
	public synchronized void unbind( String name ) throws RemoteException
	{
		DIAGNOSTICS.println( "Unbinding " + name );

		try
		{
			// Start DB transaction
			_references.startTransaction( false );
			
			// Remove reference
			_references.removeObject( name );
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Error when performing unbind", t, DIAGNOSTICS.LEVEL_MAX );
			throw new RemoteException( "Error when performing unbind", t );
		}
		finally
		{
			try
			{
				// Close transaction
				_references.endTransaction();
			}
			catch( Throwable t )
			{
				DIAGNOSTICS.println( "Exception when closing DB!", t, DIAGNOSTICS.LEVEL_3 );
			}
		}
	}
	

	/**
	 * Checks that all of the references in the NS are still valid
	 */
	private synchronized void checkReferences()
	{
		DIAGNOSTICS.println( "Starting purging of references", DIAGNOSTICS.LEVEL_3 );

		try
		{
			Enumeration enum = null;

			// Get Enumeration of references
			_references.startTransaction( true );
			enum = _references.getIDs();
			_references.endTransaction();


			// Iterate through references
			while( enum.hasMoreElements() )
			{
				String next = (String) enum.nextElement();

				try
				{
					// Get next reference
					fipaos.mts.rmi.internal.MTS ref = lookup( next );

					// Attempt to ping (throws an exception if something goes wrong)
					ref.ping();
				}
				catch( Throwable th )
				{
					DIAGNOSTICS.println( "Removing reference to " + next, this, DIAGNOSTICS.LEVEL_3 );

					try
					{
						// Remove current reference
						unbind( next );
					}
					catch( Throwable s )
					{
						DIAGNOSTICS.println( "Error removing reference to " + next, s, this, DIAGNOSTICS.LEVEL_MAX );								
					}
				}
			}
		}
		catch( Throwable th )
		{
			DIAGNOSTICS.println( "Error when checking validity of NS references", th, this,
														DIAGNOSTICS.LEVEL_MAX );
		}
	}
	
	
	/**
	 * Contains reference to
	 */
	private static class NSReference implements DatabaseObject
	{
		public fipaos.mts.rmi.internal.MTS _ref;
		public String _name;
		
		public NSReference( String name, fipaos.mts.rmi.internal.MTS ref )
		{
			_ref = ref;
			_name = name;
		}
		
		public String getObjectID()
		{
			return _name;
		}
	}
	
//----------------------------------------------------------------------
// TEST HARNESS
//----------------------------------------------------------------------

	public static void main(String args[]) throws Exception
	{
		DIAGNOSTICS.setMemoryLevel( DIAGNOSTICS.LEVEL_MIN );
		
		if( args.length != 1 )
		{
			System.out.println("Usage: java fipaos.comm.rmi.ns.RMIAgentNamingServiceImpl [<ns-url>]");	
			System.exit(-10);
		}
		
		new RMIAgentNamingServiceImpl( Integer.parseInt( args[0] )  );	
	}
}
