//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.mts.service;

//----------------------------------------------------------------------

import fipaos.ont.fipa.fipaman.*;
import fipaos.ont.fipa.*;
import fipaos.mts.*;
import fipaos.util.*;

import java.util.*;

//----------------------------------------------------------------------

/**
 * <br>
 * <br>
 */
public class CommMultiplexService implements PreParserService, PostParserService
{
     
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

	private Map _transports;
	
	private Service _previous;

	/**
	 * Indicates if this Service has been shutdown
	 */
	private boolean _shutdown = false;
	
//----------------------------------------------------------------------
// CONSTRUCTOR(S)
//----------------------------------------------------------------------

	public CommMultiplexService()
	{
		_transports = new TreeMap( fipaos.util.Comparators.STRING_COMPARE );
	}
	
//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------	
	
	public void initialise( Map transports )
	{
		// Save transports for later
		_transports.putAll( transports );
	}
		
//----------------------------------------------------------------------
// PUBLIC METHODS - PreParserService interface
//----------------------------------------------------------------------	
	
	public synchronized void outgoing( Message p1 )
	{
		if( _shutdown )
		{
			DIAGNOSTICS.println( "Couldn't deal with outgoing message, shutdown() has been invoked", this );
			return;
		}
		
		DIAGNOSTICS.println( "Dealing with outgoing message (PreParserService)", this );
		
		// Check the type of message we're dealing with
		if( p1.getEnvelope().getViaACC() )
		{
			// Attempt to send the message via the ACC
			sendViaACC( p1 );
		}
		else
		{
			// Send the message "normally"
			send( p1 );
		}		
	}
	
	public synchronized void incoming( Message p1 )
	{
		if( _shutdown )
		{
			DIAGNOSTICS.println( "Couldn't deal with incoming message, shutdown() has been invoked", this );
			return;
		}
		
		Envelope env = p1.getEnvelope();
		
		// Check if this message has an error-code set
		int error = env.getErrorCode();
		
		if( error != NONE )
		{
			DIAGNOSTICS.println( "Dealing with problem outgoing message (PostParserService)", this );
		
			// Not an "incoming" message, but an outgoing message which was rejected
			// somewhere lower down the stack
			switch( error )
			{
				// Target not known / unreachable - remove first address from intended receiver
				// and try again
				case PROTOCOL_UNKNOWN:
															// Save failed URL, since we don't know if its is correct or not
															DIAGNOSTICS.println( "Adding failed address to list of unknown MTP's" );
															
															AgentID aid1 = (AgentID) env.getIntendedReceiver().get(0);
															List addrs1 = aid1.getAddresses();

															// Check an address has been specified
															if( addrs1.size() > 0 )
															{
																env.getUnknownURLsTried().add( addrs1.get( 0 ) );
															}
															
				case TARGET_UNKNOWN:	DIAGNOSTICS.println( "Removing failed address from AID" );
				
															// If message isn't being routed to the ACC, remove address just tried
															if( !env.getViaACC() )
															{
																AgentID aid = (AgentID) env.getIntendedReceiver().get(0);
																List addrs = aid.getAddresses();
																
																// Check an address has been specified
																if( addrs.size() > 0 )
																{
																	addrs.remove( 0 );
																}
																
																// Reset the error code
																env.setErrorCode( NONE );
															}
															
															outgoing( p1 );
															break;
															
				// Can't deal with error here, pass it back up
				default:							((PreParserService) _previous).incoming( p1 );
			}
		}
		else
		{
			DIAGNOSTICS.println( "Dealing with incoming message  (PreParserService)", this );
		
			// Clear "tried transports" & send message upwards
			p1.getEnvelope().getTransportsTried().clear();
			((PreParserService) _previous).incoming( p1 );
		}
	}
	
	public void initialise( PreParserService p1, PreParserService p2)
	{
		_previous = p1;
	}
	
	
//----------------------------------------------------------------------
// PUBLIC METHODS - PostParserService interface
//----------------------------------------------------------------------	
	
	public synchronized void outgoing( Envelope p1, byte[] p2 )
	{
		if( _shutdown )
		{
			DIAGNOSTICS.println( "Couldn't deal with outgoing message, shutdown() has been invoked", this );
			return;
		}
		
		DIAGNOSTICS.println( "Dealing with outgoing message (PostParserService)", this );
		
		// Check the type of message we're dealing with
		if( p1.getViaACC() )
		{
			// Message needs to be sent via the ACC
			sendViaACC( p1, p2 );
		}
		else
		{
			// Message needs to be sent "normally"
			send( p1, p2 );
		}		
	}
	
	public synchronized void incoming( Envelope p1, byte[] p2 )
	{
		if( _shutdown )
		{
			DIAGNOSTICS.println( "Couldn't deal with incoming message, shutdown() has been invoked", this );
			return;
		}
		
		Envelope env = p1;
		
		// Check if this message has an error-code set
		int error = env.getErrorCode();
		
		if( error != NONE )
		{
			DIAGNOSTICS.println( "Dealing with problem outgoing message (PreParserService)", this );
		
			// Not an "incoming" message, but an outgoing message which was rejected
			// somewhere lower down the stack
			switch( error )
			{
				// Target not known / unreachable - remove first address from intended receiver
				// and try again
				case PROTOCOL_UNKNOWN:
															// Save failed URL, since we don't know if its is correct or not
															DIAGNOSTICS.println( "Adding failed address to list of unknown MTP's" );
															
															AgentID aid1 = (AgentID) env.getIntendedReceiver().get(0);
															List addrs1 = aid1.getAddresses();

															// Check an address has been specified
															if( addrs1.size() > 0 )
															{
																env.getUnknownURLsTried().add( addrs1.get( 0 ) );
															}
															
				case TARGET_UNKNOWN:	DIAGNOSTICS.println( "Removing failed address from AID" );
				
															// If message isn't being routed to the ACC, remove address just tried
															if( !env.getViaACC() )
															{
																AgentID aid = (AgentID) env.getIntendedReceiver().get(0);
																List addrs = aid.getAddresses();
																
																// Check an address has been specified
																if( addrs.size() > 0 )
																{
																	addrs.remove( 0 );
																}
																
																// Reset the error code
																env.setErrorCode( NONE );
															}
															
															outgoing( p1, p2 );
															break;
															
				// Can't deal with error here, pass it back up
				default:							((PostParserService) _previous).incoming( p1, p2 );
			}
		}
		else
		{
			DIAGNOSTICS.println( "Dealing with incoming message (PreParserService)", this );
		
			// Clear "tried transports" & send message upwards
			env.getTransportsTried().clear();
			((PostParserService) _previous).incoming( p1, p2 );
		}
	}
	
	public void initialise( PostParserService p1, PostParserService p2 )
	{
		_previous = p1;
	}

	
//----------------------------------------------------------------------
// PUBLIC METHODS - Service interface
//----------------------------------------------------------------------	
	
	public synchronized void shutdown()
	{
		// Indicate shutdown() has been invoked
		_shutdown = true;
		
		// Get a list of all service immediately following this in stack, and invoke
		// shutdown() on them
		Iterator i = _transports.values().iterator();
		while( i.hasNext() )
		{
			// Get next service
			Service s = (Service) i.next();
			
			// Shutdown service
			s.shutdown();
		}
		
		// Remove references to improve garbage collection
		_previous = null;
		_transports.clear();
		_transports = null;
	}	


//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------
	
	private void sendViaACC( Envelope env, byte body[] )
	{
		// Get tried transports
		List tried = env.getTransportsTried();
		
		// Get available transports
		Set available = new TreeSet( fipaos.util.Comparators.STRING_COMPARE );
		available.addAll( _transports.keySet() );
		
		// Compare which protocols have been used against those available
		available.removeAll( tried );
		
		// Any more protocols available to try?
		if( available.size() > 0 )
		{
			// Get transport name
			String transport = (String) ( new LinkedList( available ) ).get( 0 );
			
			// Get "top" service of stack to transport
			PostParserService pps = (PostParserService) _transports.get( transport );
			
			// Add to list of "tried transports"
			env.getTransportsTried().add( transport );
			
			// Send message
			pps.outgoing( env, body );
		}
		else
		{
			// Send back up stack
			env.setErrorCode( TARGET_UNKNOWN );
			((PostParserService) _previous).incoming( env, body );
		}
	}
	
	private void sendViaACC( Message msg )
	{
		// Get tried transports
		List tried = msg.getEnvelope().getTransportsTried();
		
		// Get available transports
		Set available = new TreeSet( fipaos.util.Comparators.STRING_COMPARE );
		available.addAll( _transports.keySet() );
		
		// Compare which protocols have been used against those available
		available.removeAll( tried );
		
		// Any more protocols available to try?
		if( available.size() > 0 )
		{
			// Get transport name
			String transport = (String) ( new LinkedList( available ) ).get( 0 );
			
			// Get "top" service of stack to transport
			PreParserService pps = (PreParserService) _transports.get( transport );
			
			// Add to list of "tried transports"
			msg.getEnvelope().getTransportsTried().add( transport );
			
			// Send message
			pps.outgoing( msg );
		}
		else
		{
			// Send back up stack
			msg.getEnvelope().setErrorCode( TARGET_UNKNOWN );
			((PreParserService) _previous).incoming( msg );
		}
	}
	
	private void send( Envelope env, byte body[] )
	{
		// Check envelope for "best" address to contact target Agent at
		
		// Get intended-receiver field
		List targets = env.getIntendedReceiver();
		
		// If the intended-receiver field doesn't contain a single entry,
		// create it / split it up as appropriate
		if( targets == null || targets.size() == 0 )
		{
			// Build intended receiver fields and re-send
			List to = env.getReceivers();
			
			// Check receivers have been specified
			if( to == null || to.size() == 0 )
			{
				DIAGNOSTICS.println( "Malformed envelope? No To: field" );
				
				// Set error code & send back up stack
				env.setErrorCode( MALFORMED_ENVELOPE );
				((PostParserService) _previous).incoming( env, body );
			}
			
			DIAGNOSTICS.println( "Set up intended-receiver field" );
			
			// Set the intended receivers
			env.setIntendedReceiver( to );
			
			// Re-send
			outgoing( env, body );
		}
		else if( targets.size() > 1 )
		{
			// Duplicate the message & re-send
			
			DIAGNOSTICS.println( "Splitting up multi-recipient message" );
			
			// Iterate through targets
			Iterator i = targets.iterator();
			while( i.hasNext() )
			{
				// Get next target
				AgentID target = (AgentID) i.next();
				
				// Duplicate original Message, and add target as the intended-receiver
				env.setIntendedReceiver( Arrays.asList( new Object[] { target } ) );
				
				// Recurse back into this
				outgoing( env, body );
			}
		}
		else // targets.size() == 1 - this is where we want to be!
		{
			// Find out where we should be sending this message
			
			// Get AID and addresses for delivery
			AgentID aid = (AgentID) targets.get( 0 );
			List addresses = aid.getAddresses();
			
			DIAGNOSTICS.println( "Transports to try = " + _transports.keySet(), this );
			
			// See if there are any protocols to try that we know how to deal with
			Iterator i = addresses.iterator();
			boolean done = false;
			while( i.hasNext() && !done )
			{
				// Get next URL
				URL url = (URL) i.next();
				String transport = url.getProtocol();
				
				// See if we can deal with protocol
				if( _transports.containsKey( url.getProtocol() ) )
				{
					// Get next service down in stack to invoke
					PostParserService pps = (PostParserService) _transports.get( transport );
					
					// Add transport to "tried transports"
					env.getTransportsTried().add( transport );
					
					// Change intended-receiver slot - move first address we can deal with
					// to first in list
					addresses.remove( url );
					addresses.add( 0, url );
					
					// Send message
					pps.outgoing( env, body );
					
					// We're done
					done = true;
				}
			}
			
			
			// If message wasn't sent, pass to a transport we haven't tried yet
			i = _transports.keySet().iterator();
			List tried_transports = env.getTransportsTried();
			while( i.hasNext() && !done )
			{
				// Get next transport
				String transport = (String) i.next();

				// See if it's already been tried
				if( !tried_transports.contains( transport ) )
				{
					// Not been tried previously - lets give it a go!
					
					// Get next service down in stack to invoke
					PostParserService pps = (PostParserService) _transports.get( transport );
					
					// Add transport to "tried transports"
					tried_transports.add( transport );
					
					// Send message
					pps.outgoing( env, body );
					
					// We're done
					done = true;
				}
			}
			
			
			// Check if message was sent OK
			if( !done )
			{
				DIAGNOSTICS.println( "No more transports to attempt to send message via" );
				
				// Propogate back up the stack, indicating that the protocol is unknown
				env.setErrorCode( TARGET_UNKNOWN );
				((PostParserService) _previous).incoming( env, body );				
			}
		}
	}
	
	private void send( Message msg )
	{
		// Check envelope for "best" address to contact target Agent at
		
		// Get intended-receiver field
		Envelope env = msg.getEnvelope();
		List targets = env.getIntendedReceiver();
		
		// If the intended-receiver field doesn't contain a single entry,
		// create it / split it up as appropriate
		if( targets == null || targets.size() == 0 )
		{
			// Build intended receiver fields and re-send
			List to = env.getReceivers();
			
			// Check receivers have been specified
			if( to == null || to.size() == 0 )
			{
				DIAGNOSTICS.println( "Malformed envelope? No To: field" );
				
				// Set error code & send back up stack
				env.setErrorCode( MALFORMED_ENVELOPE );
				((PreParserService) _previous).incoming( msg );
			}
			
			DIAGNOSTICS.println( "Set up intended-receiver field" );
			
			// Set the intended receivers
			env.setIntendedReceiver( to );
			
			// Re-send
			outgoing( msg );
		}
		else if( targets.size() > 1 )
		{
			// Duplicate the message & re-send
			
			DIAGNOSTICS.println( "Splitting up multi-recipient message" );
			
			// Iterate through targets
			Iterator i = targets.iterator();
			while( i.hasNext() )
			{
				// Get next target
				AgentID target = (AgentID) i.next();
				
				// Duplicate original Message, and add target as the intended-receiver
				msg = new Message( msg );
				msg.getEnvelope().setIntendedReceiver( Arrays.asList( new Object[] { target } ) );
				
				// Recurse back into this
				outgoing( msg );
			}
		}
		else // targets.size() == 1 - this is where we want to be!
		{
			// Find out where we should be sending this message
			
			// Get AID and addresses for delivery
			AgentID aid = (AgentID) targets.get( 0 );
			List addresses = aid.getAddresses();
			
			DIAGNOSTICS.println( "Transports to try = " + _transports.keySet(), this );
			
			// See if there are any protocols to try that we know how to deal with
			Iterator i = addresses.iterator();
			boolean done = false;
			while( i.hasNext() && !done )
			{
				// Get next URL
				URL url = (URL) i.next();
				String transport = url.getProtocol();
				
				// See if we can deal with protocol
				if( _transports.containsKey( transport ) )
				{
					// Get next service down in stack to invoke
					PreParserService pps = (PreParserService) _transports.get( transport );
					
					// Change intended-receiver slot - move first address we can deal with
					// to first in list
					addresses.remove( url );
					addresses.add( 0, url );
					
					// Add transport to "tried transports"
					env.getTransportsTried().add( transport );
					
					// Send message
					pps.outgoing( msg );
					
					// We're done
					done = true;
				}
			}
			
			
			// If message wasn't sent, pass to a transport we haven't tried yet
			i = _transports.keySet().iterator();
			List tried_transports = env.getTransportsTried();
			while( i.hasNext() && !done )
			{
				// Get next transport
				String transport = (String) i.next();

				// See if it's already been tried
				if( !tried_transports.contains( transport ) )
				{
					// Not been tried previously - lets give it a go!
					
					// Get next service down in stack to invoke
					PreParserService pps = (PreParserService) _transports.get( transport );
					
					// Add transport to "tried transports"
					tried_transports.add( transport );
					
					// Send message
					pps.outgoing( msg );
					
					// We're done
					done = true;
				}
			}
			
			
			// Check if message was sent OK
			if( !done )
			{
				DIAGNOSTICS.println( "No more transports to attempt to send message via" );
				
				// Propogate back up the stack, indicating that the protocol is unknown
				env.setErrorCode( TARGET_UNKNOWN );
				((PreParserService) _previous).incoming( msg );				
			}
		}		
	}
	
//----------------------------------------------------------------------
// MAIN/TESTING METHOD
//----------------------------------------------------------------------
	
	private static void main( String args[] ) throws Exception 
	{
		DIAGNOSTICS.setMemoryLevel( DIAGNOSTICS.LEVEL_MIN );
		
		// Load an ACL message from a file and attempt to pass it through our test-stack
		if( args.length != 1 )
		{
			System.out.println( "Usage: java " + CommMultiplexService.class.getName() + " [<acl-file>]" );
			System.exit( -10 );
		}
		
		// Read from file
		String acl = DocumentRetriever.getDocFromFile( args[0] );
		
		if( acl == null )
		{
			System.out.println( "Error reading file!" );
			System.exit( -10 );
		}
		
		System.out.println( "ACL == " + acl );
		
		Message msg1 = new Message( new ACL( acl ) );
		Message msg2 = new Message( new ACL( acl ) );
		
		// Setup test services
		PreParserService a = new UnknownTarget();
		PreParserService b = new UnknownProtocol();
		PreParserService c = new Report();
		PreParserService d = new Report();
		CommMultiplexService e = new CommMultiplexService();
		fipaos.mts.service.preparse.ACCRouterService f = new fipaos.mts.service.preparse.ACCRouterService( false );
		
		// Create test stack
		c.initialise( null, f );
		f.initialise( c, e );
		e.initialise( f, null );
		Map p = new TreeMap( fipaos.util.Comparators.OBJECT_COMPARE );
		p.put( "iiop", a );
		p.put( "http", b );
		p.put( "fipaos-rmi", d );
		e.initialise( p );
		a.initialise( e, null );
		b.initialise( e, null );
		d.initialise( e, null );
		
		// Plonk message into pipe & see what happens
		System.out.println( "OUTGOING =============================================" );
		f.outgoing( msg1 );
		System.out.println( "INCOMING =============================================" );
		f.incoming( msg2 );
	}
	
	
	private static class UnknownTarget implements PreParserService
	{
		PreParserService _prev;
		
		public UnknownTarget()
		{
		}
		
		public void outgoing(final fipaos.mts.Message p1)
		{
			p1.getEnvelope().setErrorCode( TARGET_UNKNOWN );
			_prev.incoming( p1 );
		}

		public void incoming(final fipaos.mts.Message p1)
		{
		}

		public void initialise(final fipaos.mts.service.PreParserService p1,final fipaos.mts.service.PreParserService p2) 
		{
			_prev = p1;
		}
		
		public void shutdown()
		{
		}
	}
	
	private static class UnknownProtocol implements PreParserService
	{
		PreParserService _prev;
		
		public UnknownProtocol()
		{
		}
		
		public void outgoing(final fipaos.mts.Message p1)
		{
			p1.getEnvelope().setErrorCode( PROTOCOL_UNKNOWN );
			_prev.incoming( p1 );
		}

		public void incoming(final fipaos.mts.Message p1)
		{
		}

		public void initialise(final fipaos.mts.service.PreParserService p1,final fipaos.mts.service.PreParserService p2) 
		{
			_prev = p1;
		}
		
		public void shutdown()
		{
		}
	}
	
	private static class Report implements PreParserService
	{
		PreParserService _prev;
		
		public Report()
		{
		}
		
		public void outgoing(final fipaos.mts.Message p1)
		{
			System.out.println( "Message sent: " + p1 );
		}

		public void incoming(final fipaos.mts.Message p1)
		{
			System.out.println( "Message recv'd: " + p1 );

			int error =  p1.getEnvelope().getErrorCode();
			if( error != NONE )
			{
				System.out.println( "Error code: " + error );
			}
		}

		public void initialise(final fipaos.mts.service.PreParserService p1,final fipaos.mts.service.PreParserService p2) 
		{
			_prev = p1;
		}
		
		public void shutdown()
		{
		}
	}
} // End class

	
