//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.mts.service;

//----------------------------------------------------------------------

import fipaos.ont.fipa.fipaman.*;
import fipaos.ont.fipa.*;
import fipaos.util.*;
import fipaos.mts.*;
import fipaos.parser.ParserException;

//----------------------------------------------------------------------

/**
 * This Service attempt to parser/deparse the various parts of the message
 * into the desired encoding (as specified in the Envelope and ACL parts of the message)
 */
public class ParserService implements PreParserService, PostParserService
{
	
//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

	/**
	 * Service "before" this in stack
	 */
	private PreParserService _previous;

	/**
	 * Service "after" this in stack
	 */
	private PostParserService _next;
	
	/**
	 * Indicates if this Service has been shutdown
	 */
	private boolean _shutdown = false;
	
//----------------------------------------------------------------------
// CONSTRUCTOR(S)
//----------------------------------------------------------------------

	public ParserService()
	{
	}
	
//----------------------------------------------------------------------
// PUBLIC METHODS - Service
//----------------------------------------------------------------------	

	/**
	 * Shutdown this service, and propogate to next in stack
	 */
	public synchronized void shutdown()
	{
		_shutdown = true;
		
		// Nothing to shutdown except next in "stack"
		if( _next != null )
		{
			_next.shutdown();
		}
		
		// Dereference other objects to ensure garbage collection
		_next = null;
		_previous = null;
	}

//----------------------------------------------------------------------
// PUBLIC METHODS - PostParserService
//----------------------------------------------------------------------	
	
	/** Outgoing messages are passed to <CODE>this</CODE> via this method
	 * @param p1
	 * @param p2  */
	public synchronized void outgoing( Envelope p1, byte[] p2 )
	{
		DIAGNOSTICS.println( "Dealing with outgoing message", this );
		
		if( _shutdown )
		{
			DIAGNOSTICS.println( "Couldn't deal with outgoing message, shutdown() has been invoked", this );
			return;
		}
		
		_next.outgoing( p1, p2 );		
	}
	
	/** Incoming messages are passed to <CODE>this</CODE> via this method
	 * @param msg Incoming message
	 */
	public synchronized void incoming( Envelope p1, byte[] p2 )
	{
		DIAGNOSTICS.println( "Dealing with incoming message", this );
		
		if( _shutdown )
		{
			DIAGNOSTICS.println( "Couldn't deal with incoming message, shutdown() has been invoked", this );
			return;
		}
		
		// De-parse message & forward to service "below" this
		try
		{
			ACL acl = convertToACL( p1, p2 );
			_previous.incoming( new Message( p1, acl )  );
		}
		catch( Throwable t )
		{
			// TODO: Add exception handling
			DIAGNOSTICS.println( "Error when parsing incoming message: ", t, this,
																												DIAGNOSTICS.LEVEL_MAX );
		}
	}
	
	/** Invoked when the undelying implementation should initialise itself.  Allows
	 * references to other PostParserService's before and after it in the stack
	 * to be passed
	 * @param previous Previous PostParserService instance (Incoming messages should be passed to this)
	 * @param next Next PostParserService instance (Outgoing messages should be passed to this)
	 */
	public void initialise( PostParserService p1, PostParserService p2 )
	{
		// Save reference to next service
		_next = p2;
	}
	
//----------------------------------------------------------------------
// PUBLIC METHODS - PreParserService
//----------------------------------------------------------------------	
	
	/** Outgoing messages are passed to <CODE>this</CODE> via this method
	 * @param msg Outgoing message
	 */
	public synchronized void outgoing( Message p1 )
	{
		DIAGNOSTICS.println( "Dealing with outgoing message", this );
		
		if( _shutdown )
		{
			DIAGNOSTICS.println( "Couldn't deal with outgoing message, shutdown() has been invoked", this );
			return;
		}
		
		// De-parse message & forward to service "below" this
		try
		{
			Envelope env = p1.getEnvelope();
			byte body[] = convertToBytes( env, p1.getACL() );
			_next.outgoing( env, body );
		}
		catch( Throwable t )
		{
			// TODO: Add exception handling
		}
	}
	
	/** Incoming messages are passed to <CODE>this</CODE> via this method
	 * @param msg Incoming message
	 */
	public synchronized void incoming( Message p1 )
	{
		DIAGNOSTICS.println( "Dealing with incoming message", this );
		
		if( _shutdown )
		{
			DIAGNOSTICS.println( "Couldn't deal with incoming message, shutdown() has been invoked", this );
			return;
		}
		
		_previous.incoming( p1 );
	}
	
	/** Invoked when the undelying implementation should initialise itself.  Allows
	 * references to other PreParserService's before and after it in the stack
	 * to be passed
	 * @param previous Previous PreParserService instance (Incoming messages should be passed to this)
	 * @param next Next PreParserService instance (Outgoing messages should be passed to this)
	 */
	public void initialise( PreParserService p1, PreParserService p2 )
	{
		// Save reference to previous service
		_previous = p1;
	}

//----------------------------------------------------------------------
// PROTECTED METHODS
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------

	/**
	 * Invoked when an incoming message must be converted into an ACL object
	 */
	private ACL convertToACL( Envelope env, byte in[] ) throws ParserException
	{
		String acl_encoding = env.getACLRepresentation();
		
		// Check which encoding we need to be produced
		if( acl_encoding == null )
		{
			try
			{
				// Try legacy String Encoding
				return new ACL( new String( in ) );
			}
			catch( Throwable t )
			{
			}
			
			try
			{
				// Try Bit. Efficient Encoding
				return fipaos.parser.acl.bitefficient.Parser.parse( in );
			}
			catch( Throwable t )
			{
			}
			
			try
			{
				// Try serialisation encoding
				return (ACL) Serialize.toObject( in );
			}
			catch( Throwable t )
			{
			}
		}
		else if( acl_encoding.equals( ACL.FIPA00069 ) ) 
		{
			// Bit. Efficient Encoding
			return fipaos.parser.acl.bitefficient.Parser.parse( in );
		}
		else if( acl_encoding.equals( ACL.FIPA00071 ) )
		{
			// XML Encoding
			throw new ParserException( ACL.FIPA00071 + " not currently supported!" );
		}
		else if( acl_encoding.equals( ACL.FIPA00070 ) )
		{
			// Legacy String Encoding
			return new ACL( new String( in ) );
		}
		else if( acl_encoding.equals( ACL.FIPA_OS_SERIALIZATION ) )
		{
			// FIPA-OS Serialisation encoding
			return (ACL) Serialize.toObject( in );
		}
		else
		{
			DIAGNOSTICS.println( "WARNING: ACL encoding not known - assuming " + ACL.FIPA00070,
														this, DIAGNOSTICS.LEVEL_MAX );
			
			// Assume legacy encoding?
			try
			{
				return new ACL( new String( in ) );
			}
			catch( ParserException e )
			{
				DIAGNOSTICS.println( "Exception during ACL parsing", e );
				
				// Assume unknown Encoding
				throw new ParserException( acl_encoding + " not currently supported!" );
			}
		}
		
		return null;
	}
	
	/**
	 * Invoked when an outgoing message must be converted to a byte[]
	 */ 
	private byte[] convertToBytes( Envelope env, ACL in )
	{
		String acl_encoding = env.getACLRepresentation();
		
		// Check which encoding we need to be produced
		if( acl_encoding == null )
		{
			// Assume legacy String encoding
			return in.toString().getBytes();
		}
		else if( acl_encoding.equals( ACL.FIPA00069 ) ) 
		{
			// Bit. Efficient Encoding
			return fipaos.parser.acl.bitefficient.Parser.deparse( in );
		}
		else if( acl_encoding.equals( ACL.FIPA_OS_SERIALIZATION ) )
		{
			// FIPA-OS Serialisation encoding
			return Serialize.toByteArray( in );
		}
/*		else if( acl_encoding.equals( ACL.FIPA00071 ) )
		{
			// XML Encoding
			throw new ParserException( ACL.FIPA00071 + " not currently supported!" );
		}
*/
		else if( acl_encoding.equals( ACL.FIPA00070 ) )
		{
			// Legacy String Encoding
			return in.toString().getBytes();
		}
		else
		{
			DIAGNOSTICS.println( "WARNING: ACL encoding not known - assuming " + ACL.FIPA00070,
														this, DIAGNOSTICS.LEVEL_MAX );
			
			// Set appropriate ACL representation indicator
			env.setACLRepresentation( ACL.FIPA00070 );
			
			// Assume legacy encoding?
			return in.toString().getBytes();
		}
	}
	
//----------------------------------------------------------------------
// MAIN/TESTING METHOD
//----------------------------------------------------------------------
 
	private static void main( String args[] ) throws Exception
	{
		DIAGNOSTICS.setMemoryLevel( DIAGNOSTICS.LEVEL_MIN );
		
		// Load an ACL message from a file and attempt to pass it through our test-stack
		if( args.length != 1 )
		{
			System.out.println( "Usage: java fipaos.mts.service.ParserService [<acl-file>]" );
			System.exit( -10 );
		}
		
		// Read from file
		String acl = DocumentRetriever.getDocFromFile( args[0] );
		
		if( acl == null )
		{
			System.out.println( "Error reading file!" );
			System.exit( -10 );
		}
		
		System.out.println( "ACL == " + acl );
		
		Message msg = new Message( new ACL( acl ) );
		
		// Create stack
		PreParserService pps1 = new Top();
		PostParserService pps2 = new UBend();
		ParserService ps = new ParserService();
		pps1.initialise( null, ps );
		ps.initialise( pps1, null );
		ps.initialise( null, pps2 );
		pps2.initialise( ps, null );
		
		
		// Pass message through stack
		pps1.outgoing( msg );
	}

	
	/**
	 * A "U-bend" for the bottom of the stack
	 */
	private static class UBend implements PostParserService
	{
		private PostParserService _prev;
		
		public UBend()
		{
		}
		
		public void outgoing( Envelope p1, byte[] p2 )
		{
			incoming( p1, p2 );
		}
		
		public void incoming( Envelope p1, byte[] p2 )
		{
			_prev.incoming( p1, p2 );
		}
		
		public void initialise( PostParserService p1 , PostParserService p2 )
		{
			_prev = p1;
		}
		
		public void shutdown()
		{
		}
	}
	
	/**
	 * A test-harness for the top of the stack
	 */
	private static class Top implements PreParserService
	{
		PreParserService _next;
		
		public Top()
		{
		}
		
		public void outgoing( Message p1 )
		{
			DIAGNOSTICS.println( "Sending message: " + p1 );
			_next.outgoing( p1 );
		}
		
		public void incoming( Message p1 )
		{
			DIAGNOSTICS.println( "Received message: " + p1 );
		}

		public void initialise( PreParserService p1, PreParserService p2)
		{
			_next = p2;
		}
		
		public void shutdown()
		{
		}
	}
} // End class

	
