//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files ("the
//     Programs") are Copyright (c) 1999 - 2000 Nortel Networks Corporation, 
//     8200 Dixie Road, Suite 100, Brampton, Ontario, Canada L6R 5P6.  
//     All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out in
//     the Nortel Networks FIPA-OS Public License, a copy of which can be found
//     in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version can
//     also be found at http://fipa-os.sourceforge.net/. By downloading the
//     files containing the Programs you accept the terms and  conditions of the
//     Public License. You do not have to accept these terms and conditions, but
//     unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.mts.service;

//----------------------------------------------------------------------

import fipaos.mts.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.ont.fipa.*;
import fipaos.ont.BASETOKENS;
import fipaos.util.*;

import java.util.*;

//----------------------------------------------------------------------

/** This class provides a convenient mechanism for initialising a set
 * of Services to form a stack
 * <br>
 * <br>
 */
public class ServiceStack
{
     
//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// STATIC VARIABLES
//----------------------------------------------------------------------
	
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

	private List _stack;
	
	private boolean _initialised = false;

//----------------------------------------------------------------------
// CONSTRUCTOR(S)
//----------------------------------------------------------------------

	/** Default constructor
	 */
	public ServiceStack()
	{
		_stack = new LinkedList();
	}

//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------	
	
	/** Adds the given Service to the stack, but doesn't initialise it - this
	 * occurs when initialise() is invoked.  The order of the Services within
	 * the stack is dependant upon the order they are added.
	 * @param s The Service to add to the stack
	 */
	public synchronized void add( Service s )
	{
		// Add service to stack, if one has been specified & stack hasn't been initialised
		if( s != null && !_initialised )
		{
			_stack.add( s );
		}
	}
	
	
	/** Links all Service's within the stack together
	 */
	public synchronized void initialise()
	{
		// Check to see if initialisation has already occured
		if( _initialised )
		{
			return;
		}
		
		// Don't add any more services/initialise again!
		_initialised = true;
		
		// Attempt to link all services that have been added together
		Service prev = null;
		Service curr = null;
		Service next = null;
		boolean postparse = false;
		
		Iterator i = _stack.iterator();
		while( i.hasNext() || curr != null )
		{
			// Check if we have a "next" to insert into the pipe-line
			if( i.hasNext() )
			{
				next = (Service) i.next();
			}
			else
			{
				next = null;
			}
			
			// If current == null, iterate again
			if( curr == null )
			{
				prev = curr;
				curr = next;
				continue;
			}
			
			String prev_string = ( prev == null ? null : prev.getClass().getName() );
			String curr_string = ( curr == null ? null : curr.getClass().getName() );
			String next_string = ( next == null ? null : next.getClass().getName() );

			// If current is a ParserService, we need to initialise it twice
			if( curr instanceof ParserService )
			{
				DIAGNOSTICS.println( "Joining Pre/PostParserService's: " + prev_string + " <=> " +
															curr_string + " <=> " + next_string, this );
			
				( (PreParserService) curr).initialise( ( (PreParserService) prev ), null );
				( (PostParserService) curr).initialise( null, ( (PostParserService) next ) );
				postparse = true;
			}
			// If we are currently within pre-parser services, assume all services implement
			// the PreParserService interface
			else if( !postparse )
			{
				DIAGNOSTICS.println( "Joining PreParserService's: " + prev_string + " <=> " +
															curr_string + " <=> " + next_string, this );
			
				( (PreParserService) curr).initialise( (PreParserService) prev, (PreParserService) next );
			}
			// If we are currently within post-parser services, assume all services implement
			// the PostParserService interface
			else
			{
				DIAGNOSTICS.println( "Joining PostParserService's: " + prev_string + " <=> " +
															curr_string + " <=> " + next_string, this );
			
				( (PostParserService) curr).initialise( (PostParserService) prev, (PostParserService) next );				
			}
			
			// Rotate current Services
			prev = curr;
			curr = next;
		}
	}

//----------------------------------------------------------------------
// PROTECTED METHODS
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// MAIN/TESTING METHOD
//----------------------------------------------------------------------

} // End class
