//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files ("the
//     Programs") are Copyright (c) 1999 - 2000 Nortel Networks Corporation,
//     8200 Dixie Road, Suite 100, Brampton, Ontario, Canada L6R 5P6.
//     All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out in
//     the Nortel Networks FIPA-OS Public License, a copy of which can be found
//     in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version can
//     also be found at http://fipa-os.sourceforge.net/. By downloading the
//     files containing the Programs you accept the terms and  conditions of the
//     Public License. You do not have to accept these terms and conditions, but
//     unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.ont.fipa.fipaman;

import fipaos.ont.BASETOKENS;
import fipaos.agent.conversation.content.Content;
import fipaos.parser.sl.SLParser;
import fipaos.parser.ParserException;
import java.util.List;

/**
 * FIPA defined AMS Agent Description class for the FIPA agent management ontology.
 * See FIPA 2000 6.1.5.
 * e.g. a parseable string into the object is:
 * <br>
 * <b>(ams-agent-description :name (agent-identifier :name CameraProxy@foo.com :addresses (sequence iiop://foo.com/acc))
 *                           :ownership Ann
 *			    :state active)
 * </b>
 *
 *
 **/
public class AMSAgentDescription implements fipaos.skill.db.DatabaseObject, java.io.Serializable
{
	
	//----------------------------------------------------------------------
	// PRIVATE CONSTANTS
	//----------------------------------------------------------------------
	
	//----------------------------------------------------------------------
	// INSTANCE VARIABLES
	//----------------------------------------------------------------------
	
	/**
	 *	The identifier of the agent
	 **/
	private AgentID _agent_id           = null;
	
	/**
	 *	The life cycle of the agent
	 * 	@seeAlso FIPAMANCONSTANTS.initiated
	 *  	@seeAlso FIPAMANCONSTANTS.active
	 *	@seeAlso FIPAMANCONSTANTS.suspended
	 *	@seeAlso FIPAMANCONSTANTS.waiting
	 *	@seeAlso FIPAMANCONSTANTS.transit
	 **/
	private String _state             = null;
	
	/**
	 *	The owner of the agent
	 **/
	
	private String _ownership            = null;
	
	
	//----------------------------------------------------------------------
	// CONSTRUCTOR
	//----------------------------------------------------------------------
	
	/** This constructor takes an AMSAgentDescription String as a parameter and uses the SL Parser to parse it into a Content Object.
	 * @param ams_string the AMS-Agent Description component of an ACL string
	 * @throws ParserException if the SL parser fails to parse the String
	 */
	
	public AMSAgentDescription( String ams_string ) throws ParserException
	{
		try
		{
			parse(SLParser.parseTerm(ams_string));
		}
		catch (fipaos.parser.sl.ParseException e)
		{
			throw new ParserException(e.getMessage());
		}
	}
	
	/** This constructor takes a Content Object as the parameter and parses it into an AMSAgentDescription object.
	 * @param content the content object representing the AMSAgentDescription
	 * @throws ParserException if the String is not valid
	 */
	
	public AMSAgentDescription( Content content ) throws ParserException
	{
		parse(content);
	}
	
	/** Empty constructor which can either be populated using the
	 * parse() method or the set methods.
	 */
	
	public AMSAgentDescription()
	{
	}
	
	//----------------------------------------------------------------------
	// PUBLIC METHODS
	//----------------------------------------------------------------------
	
	/** This method returns an AMS-Agent Description as a string
	 * @return String AMS Agent Description
	 */
	
	public String toString()
	{
		
		StringBuffer buffer=new StringBuffer();
		buffer.append(BASETOKENS.OPENBRACKET);
		buffer.append(FIPAMANCONSTANTS.AMS_AGENT_DESCRIPTION);
		buffer.append(BASETOKENS.NEWLINE);
		
		if (_agent_id !=null )
		{
			addKeywordValuePair(buffer,FIPAMANCONSTANTS.NAME,_agent_id.toString());
		}
		
		if (_ownership !=null )
		{
			addKeywordValuePair(buffer,FIPAMANCONSTANTS.OWNERSHIP,_ownership);
		}
		
		if (_state !=null )
		{
			addKeywordValuePair(buffer,FIPAMANCONSTANTS.STATE,_state);
		}
		
		buffer.append(BASETOKENS.CLOSEBRACKET);
		
		return buffer.toString();
		
	}
	
	private void addKeywordValuePair(StringBuffer buffer, String keyword, String value)
	{
		buffer.append(keyword);
		buffer.append(BASETOKENS.SPACE);
		buffer.append(value);
		buffer.append(BASETOKENS.NEWLINE);
	}
	
	/** This method is required for the DatabaseObject interface for storage.
	 * @return id the key for the database
	 */
	
	public String getObjectID()
	{
		return _agent_id.getName();
	}
	
	
	/** This method returns the agent ID
	 * @return AgentID AgentID stored in the Description
	 */
	
	public AgentID getAgentID()
	{
		return _agent_id;
	}
	
	/** This method sets the agent ID.
	 * @param agent_id the AgentID of the Description
	 */
	
	public void setAgentID(AgentID agent_id)
	{
		_agent_id =agent_id;
	}
	
	/** This method returns the name of the agent ID
	 * @return String name part of the AgentID
	 */
	
	public String getAgentName()
	{
		return _agent_id.getName();
	}
	
	/** This method returns the state of the agent
	 * @return String state of the AMS Agent Description
	 */
	
	public String getState()
	{
		return _state;
	}
	
	/** This method sets the state of the agent
	 * @param state state of the agent
	 */
	
	public void setState(String state)
	{
		_state = state;
	}
	
	/** This method returns the ownership
	 * @return ownership of the description
	 */
	
	public String getOwnership()
	{
		return _ownership;
	}
	
	/** This method sets the ownership
	 * @param ownership Ownership stored in the description
	 */
	
	public void setOwnership(String ownership)
	{
		_ownership = ownership;
	}
	
	/**
	 * Indicates whether this AMS Agent Description is suitable for use in the register action.
	 *
	 * @return True if the mandatory fields are present for a register action, false otherwise
	 **/
	
	public boolean canRegister()
	{
		
		if ( _agent_id != null )
		{
			return true;
		}
		
		return false;
	}
	
	/**
	 * Indicates whether this AMS Agent description is suitable for use in the modify action.
	 *
	 * @return True if the mandatory fields are present for a modify action, false otherwise
	 **/
	
	public boolean canModify()
	{
		
		if ( _agent_id != null )
		{
			return true;
		}
		
		return false;
	}
	
	/**
	 * Indicates whether this AMS Agent Description is suitable for use in the deregister action.
	 *
	 * @return True if the mandatory fields are present for a deregister action, false otherwise
	 **/
	
	public boolean canDeregister()
	{
		
		if ( _agent_id != null )
		{
			return true;
		}
		
		return false;
	}
	
	/**
	 * Indicates whether the agent name was present in the AMS Agent Description
	 *
	 * @return True if the field is present, false otherwise
	 **/
	
	public boolean hasAgentID()
	{
		
		if ( _agent_id != null )
		{
			return true;
		}
		
		return false;
	}
	
	
	/**
	 * Indicates whether the agent state was present in the AMS Agent Description
	 *
	 * @return True if the field is present, false otherwise
	 **/
	
	public boolean hasState()
	{
		
		if ( _state != null )
		{
			return true;
		}
		
		return false;
	}
	
	
	//----------------------------------------------------------------------
	// TEST HARNESS
	//----------------------------------------------------------------------
	
	/** This method is for testing only.
	 * @param args Arguments array
	 */
	public static void main(String[] args)
	{
		
		String amsd_string = "(ams-agent-description \n" +
		":name (agent-identifier :name df@phard1f0.europe.nortel.com :addresses (sequence \n" +
		"fipaos-rmi://localhost:3000/df ) ) \n" +
		":ownership fipaos \n" +
		":state active \n" +
		")";
		
		/*            	String amsd_string = "(ams-agent-description "+
		":name (agent-identifier :name dummy@foo.com :addresses (sequence iiop://foo.com/acc )) " +
		":state active " +
		":ownership the_owner)";
		 */
		fipaos.util.DIAGNOSTICS.println(amsd_string);
		
		try
		{
			AMSAgentDescription amsd = new AMSAgentDescription(amsd_string);
			
			System.out.println(amsd);
			
			fipaos.util.DIAGNOSTICS.println(amsd);
			
			amsd=new AMSAgentDescription();
			amsd.setAgentID(new AgentID("(agent-identifier :name dummy@foo.com :addresses (sequence iiop://foo.com/acc ))"));
			amsd.setOwnership("ME");
			amsd.setState(FIPAMANCONSTANTS.ACTIVE);
			
			System.out.println(amsd);
			
		}
		catch (ParserException e)
		{
			e.printStackTrace();
		}
		
	}
	
	
	//----------------------------------------------------------------------
	// PRIVATE METHODS
	//----------------------------------------------------------------------
	
	/**
	 * parse an AMS-Agent-description string into the AMSAgentDescription object.
	 * e.g. a string, including brackets of:
	 * (ams-agent-description :name (agent-identifier :name CameraProxy@foo.com :addresses (sequence iiop://foo.com/acc))
	 *                        :ownership Ann
	 *			   :state active)
	 *
	 *
	 */
	
	private void parse(Content ams_agent_content) throws IllegalArgumentException, ParserException
	{
		if (ams_agent_content!=null)
		{
			
			boolean mandatorySatisfied=false;
			
			Content functional_content=(Content) ams_agent_content.getAttribute("functionalTerm");
			
			if (functional_content==null || !functional_content.getAttribute("functionalOperator").equals(FIPAMANCONSTANTS.AMS_AGENT_DESCRIPTION))
			{
				throw new IllegalArgumentException("Content is not a AMS-AGENT-DESCRIPTION");
			}
			
			if (functional_content.getAttribute("parameter") instanceof List)
			{
				
				List parameters=(List) functional_content.getAttribute("parameter");
				
				for (int i=0;i<parameters.size();i++)
				{
					Content content=(Content) parameters.get(i);
					
					if (content.getName()!="parameter")
					{
						throw new IllegalArgumentException("Illegal AMS-AGENT-DESCRIPTION FORMAT");
					}
					
					parse2( content );
				}
			}
			else
			{
				Content parameter=(Content) functional_content.getAttribute("parameter");
				parse2( parameter );
			}
		}
	}
	
	
	/**
	 * Extract a parameter from the given Content object - auxilluary method for parse()
	 */
	private void parse2( Content content ) throws IllegalArgumentException, ParserException
	{
		if( content != null )
		{
			String name=((String) content.getAttribute("parameterName"));

			String value=(String) ((Content) content.getAttribute("parameterValue")).getAttribute("formattedString");

			if (name.equals(FIPAMANCONSTANTS.NAME))
			{
				_agent_id=new AgentID(value);
			}
			else if (name.equals(FIPAMANCONSTANTS.OWNERSHIP))
			{
				_ownership=value;
			}
			else if (name.equals(FIPAMANCONSTANTS.STATE))
			{
				_state=value;
			}
		}
	}
}
