//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.ont.fipa.fipaman;

// -------------------------------------------------------------------------

import java.util.List;
import java.lang.Math;

import fipaos.ont.BASETOKENS;
import fipaos.agent.conversation.content.Content;
import fipaos.parser.sl.SLParser;
import fipaos.parser.acl.ACLMessage;
import fipaos.parser.ParserException;
import fipaos.parser.acl.string.ACLParser;
import fipaos.parser.acl.string.ParseException;
import fipaos.util.DIAGNOSTICS;

// -------------------------------------------------------------------------

/** AgentManagementAction is an agent management action string converted
 * into an object.
 * It contains an AgentID, a DFAgentDescription or an AMSAgentDescription
 * and/or a SearchConstraints object
 * <br>
 * <br>
 *
 */

public class AgentManagementAction implements java.io.Serializable
{

	//----------------------------------------------------------------------
	// STATIC VARIABLES
	//----------------------------------------------------------------------
	
	//----------------------------------------------------------------------
	// INSTANCE VARIABLES
	//----------------------------------------------------------------------

	private String _action;
	private DFAgentDescription  _df_description;
	private AMSAgentDescription _ams_description;
	private AgentID _agent;

	private SearchConstraints       _constraints;
	
	//----------------------------------------------------------------------
	// CONSTRUCTOR
	//----------------------------------------------------------------------

	
	/** This constructor creates an AgentManagementAction object with a Content
	 * object generated by the SLParser.
	 * @param content Content representing AgentManagementAction
	 * @throws ParserException if the Content is not valid
	 */
	public AgentManagementAction( Content content ) throws ParserException
	{
		try
		{
			parse(content);
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Exception when parsing AgentManagementAction: ", t, this,
														DIAGNOSTICS.LEVEL_MAX );
			throw new ParserException( t.toString() );
		}
	}
	
	
	/** This constructor takes a string as the parameter.
	 * @param agent_mgt_string string representing the AgentManagementAction
	 * @throws ParserException if the string is not valid
	 */
	public AgentManagementAction( String agent_mgt_string ) throws ParserException
	{
		try  {
			parse(SLParser.parseActionExpression(agent_mgt_string));
		}
		catch (fipaos.parser.sl.ParseException e) {
			throw new ParserException(e.getMessage());
		}
	}

	/** Empty Constructor. AgentManagementAction is populated using
	 * the set methods.
	 */
	public AgentManagementAction()
	{
	}

	//----------------------------------------------------------------------
	// PUBLIC METHODS
	//----------------------------------------------------------------------

	/** This method returns the action type of the AgentManagementAction.
	 * @return String representing the AgentManagementAction
	 */

	public String getAction()
	{
		return _action;
	}

	//----------------------------------------------------------------------


	/** This method sets the action of the AgentManagementAction object.
	 * @param action_type action of the AgentManagementAction
	 */
	public void setAction(String action_type)
	{
		if (action_type.equals(FIPAMANCONSTANTS.REGISTER) || action_type.equals(FIPAMANCONSTANTS.DEREGISTER) || action_type.equals(FIPAMANCONSTANTS.MODIFY) || action_type.equals(FIPAMANCONSTANTS.SEARCH) || action_type.equals(FIPAMANCONSTANTS.GET_DESCRIPTION))
		{
			_action = action_type;
		}
		else
		{
			throw new IllegalArgumentException("no such action value of: "+ action_type);
		}
	}
	//----------------------------------------------------------------------

	/** This method returns the String representation of the object
	 * @return String string representation of the object
	 */
	public String toString()
	{	
			StringBuffer buffer=new StringBuffer();
			buffer.append(BASETOKENS.OPENBRACKET);
			buffer.append(FIPAMANCONSTANTS.ACTION);
			buffer.append(BASETOKENS.NEWLINE);
			
			if (_agent !=null ) {
				buffer.append(_agent.toString());
				buffer.append(BASETOKENS.NEWLINE);
			}
			
			buffer.append(BASETOKENS.OPENBRACKET);
			buffer.append(_action);
			buffer.append(BASETOKENS.NEWLINE);
			
			if (_df_description !=null ) {
				buffer.append(_df_description.toString());
				buffer.append(BASETOKENS.NEWLINE);
			}
			else if (_ams_description !=null ) {
				buffer.append(_ams_description.toString());
				buffer.append(BASETOKENS.NEWLINE);
			}
			
			if (_constraints!=null) {
				buffer.append(_constraints.toString());
				buffer.append(BASETOKENS.NEWLINE);
			}
			
			buffer.append(BASETOKENS.CLOSEBRACKET);
			
			buffer.append(BASETOKENS.CLOSEBRACKET);
			
			return buffer.toString();			
	}


	//----------------------------------------------------------------------
	// GET/SET METHODS
	//----------------------------------------------------------------------

	/** This method returns the df-description object.
	 * @return DFAgentDescription object
	 */
	public DFAgentDescription getDFAgentDescription()
	{
		return _df_description;
	}

	//----------------------------------------------------------------------

	/** This method sets the df-description object
	 * @param df_desc DF Agent Description object
	 */
	public void setDFAgentDescription(DFAgentDescription df_desc)
	{
		_df_description = df_desc;
	}

	//----------------------------------------------------------------------

	/** This method gets the ams-description object
	 * @return AMS Agent Description object
	 */
	public AMSAgentDescription getAMSAgentDescription()
	{
		return _ams_description;
	}

	//----------------------------------------------------------------------

	/** This method sets the ams-description object.
	 * @param ams_desc AMS Agent Description object
	 */
	public void setAMSAgentDescription(AMSAgentDescription ams_desc)
	{
		_ams_description = ams_desc;
	}

	//----------------------------------------------------------------------

	/** This method returns the search constraints.
	 * @return SearchConstraints for search actions
	 */
	public SearchConstraints getSearchConstraints()
	{
		return _constraints;
	}

	//----------------------------------------------------------------------

	/** This method sets the SearchConstraints criteria.
	 * @param constraints SearchConstraints for search actions
	 */
	public void setSearchConstraints(SearchConstraints constraints)
	{
		_constraints = constraints;
	}


	/** This methods sets the AgentID.
	 * @param agent AgentID of the action
	 */
	public void setAgentID(AgentID agent) 
	{
		_agent=agent;
	}
	

	//----------------------------------------------------------------------
	// PRIVATE METHODS
	//----------------------------------------------------------------------

	private void addKeywordValuePair(StringBuffer buffer, String keyword, String value) 
	{
		buffer.append(keyword);
		buffer.append(BASETOKENS.SPACE);
		buffer.append(value);
		buffer.append(BASETOKENS.NEWLINE);
	}
	
	private String getFunctionalOperator(Content term) {

		Content functionalTerm= (Content) term.getAttribute("functionalTerm");

		if (functionalTerm!=null) {
			return (String) functionalTerm.getAttribute("functionalOperator");
		}

		return null;
	}

	private Object getTerms(Content term) {

		Content functionalTerm= (Content) term.getAttribute("functionalTerm");

		if (functionalTerm!=null) {
			return functionalTerm.getAttribute("term");
		}

		return null;
	}

	/**
	 * expected input of the form (action (content))
	 * @param agent_mgt_action_string the string to be parsed into the object.
	 * An example of a parsable string is:
	 * "(register	(ams-agent-description
	 *               (:name (agent-identifier :name CameraProxy@foo.com
	 *			       		 :addresses (sequence iiop://foo.com/acc)))
	 *               (:state active)
	 *               ))"
	 **/

	private void parse(Content action_mgt_content) throws IllegalArgumentException, fipaos.parser.ParserException
	{
		if (action_mgt_content!=null) {

			Content agent=(Content) action_mgt_content.getAttribute("agent");

			Content term=(Content) action_mgt_content.getAttribute("term");

			if  (action_mgt_content.getAttribute("actionExpressionOperator").equals(FIPAMANCONSTANTS.ACTION) && agent!=null && term!=null) {

				_agent=new AgentID((String) agent.getAttribute("formattedString"));

				String functionSymbol=getFunctionalOperator(term);

				Object terms=getTerms(term);

				if (functionSymbol!=null)
				{
					_action=functionSymbol;

					if (functionSymbol.equals(FIPAMANCONSTANTS.REGISTER) || 
							functionSymbol.equals(FIPAMANCONSTANTS.DEREGISTER) ||
							functionSymbol.equals(FIPAMANCONSTANTS.MODIFY) ) {

						String head=getFunctionalOperator((Content) terms);

						if (head.equals(FIPAMANCONSTANTS.AMS_AGENT_DESCRIPTION)) {
//							DIAGNOSTICS.println(terms);
							_ams_description=new AMSAgentDescription((Content) terms);
						}
						else if (head.equals(FIPAMANCONSTANTS.DF_AGENT_DESCRIPTION)) {
							_df_description=new DFAgentDescription((Content) terms);
						}
					}
					else if (functionSymbol.equals(FIPAMANCONSTANTS.SEARCH)) {
						
						if (terms instanceof List) {

							Content desc=(Content) ((List) terms).get(0);

							String head=getFunctionalOperator(desc);
							
//							DIAGNOSTICS.println(head);
							
							if (head.equals(FIPAMANCONSTANTS.AMS_AGENT_DESCRIPTION)) {
								_ams_description=new AMSAgentDescription(desc);
							}
							else if (head.equals(FIPAMANCONSTANTS.DF_AGENT_DESCRIPTION)) {
//								DIAGNOSTICS.println(desc);
								_df_description=new DFAgentDescription(desc);
							}

							Content constraints=(Content) ((List) terms).get(1);

							_constraints=new SearchConstraints(constraints);
						}
						else if (terms instanceof Content) {

							String head=getFunctionalOperator((Content) terms);

							if (head.equals(FIPAMANCONSTANTS.AMS_AGENT_DESCRIPTION)) {
								_ams_description=new AMSAgentDescription((Content) terms);
							}
							else if (head.equals(FIPAMANCONSTANTS.DF_AGENT_DESCRIPTION)) {
								_df_description=new DFAgentDescription((Content) terms);
							}
						}
					}
					else if (functionSymbol.equals(FIPAMANCONSTANTS.GET_DESCRIPTION)) {
						/*
						Object description=functionalTerm.getAttribute("term");

						if (description instanceof Content) {
						Content content=((Content) description)
						}
						 */
					}
				}
				else {
					throw new IllegalArgumentException("Content is not a ACTION-EXPRESSION-DESCRIPTION");
				}
			}
			else {
				throw new IllegalArgumentException("Content is not a ACTION-EXPRESSION-DESCRIPTION");
			}
		}
	}

	//----------------------------------------------------------------------
	// TEST METHODS
	//----------------------------------------------------------------------

	/** Test method
	 * @param args Arguments
	 */
	public static void main(String[] args)
	{
		
		String action="(action (agent-identifier :name CameraProxy@foo.com :addresses (sequence iiop://foo.com/acc)) (register (ams-agent-description :name (agent-identifier :name CameraProxy@foo.com :addresses (sequence iiop://foo.com/acc)) :state active )))";
		
		AgentManagementAction ama;
		
		try {	
			ama=new AgentManagementAction(action);

			System.out.println(ama.getAction());
			System.out.println(ama.getAMSAgentDescription());
			System.out.println(ama.getDFAgentDescription());
			System.out.println(ama.getSearchConstraints());
		}
		catch (ParserException e) {
			e.printStackTrace();
		}
		
		action= "(action (agent-identifier :name df@phard1f0.europe.nortel.com ) "+
					  "(register (df-agent-description :name (agent-identifier :name df2@phard1f0.europe.nortel.com "+ 
						":addresses (sequence fipaos-rmi://localhost:3000/df2 ) ) "+
					  ":services (set (service-description :type fipa-df )) " +
						"))) ";
		
		try {	
			ama=new AgentManagementAction(action);

			System.out.println(ama.getAction());
			System.out.println(ama.getAMSAgentDescription());
			System.out.println(ama.getDFAgentDescription());
			System.out.println(ama.getSearchConstraints());
		}
		catch (ParserException e) {
			e.printStackTrace();
		}
		
		action="(action (agent-identifier :name CameraProxy@foo.com :addresses (sequence iiop://foo.com/acc)) (search (df-agent-description :name (agent-identifier :name CameraProxy@foo.com :addresses (sequence iiop://foo.com/acc)) :ontology (set meeting-scheduler) :language (set FIPA-L0 KIF)) (search-constraints :max-depth 2)))";

		try {
			ama=new AgentManagementAction(action);
		
			System.out.println(ama.getAMSAgentDescription());
			System.out.println(ama.getDFAgentDescription());
			System.out.println(ama.getSearchConstraints());
			
			ama=new AgentManagementAction();
			
			ama.setAction(FIPAMANCONSTANTS.REGISTER);
			ama.setAgentID(new AgentID("(agent-identifier :name CameraProxy@foo.com :addresses (sequence iiop://foo.com/acc))"));
			ama.setAMSAgentDescription(new AMSAgentDescription("(ams-agent-description :name (agent-identifier :name CameraProxy@foo.com :addresses (sequence iiop://foo.com/acc)) :state active )"));
			ama.setSearchConstraints(new SearchConstraints("(search-constraints :max-depth 20 :max-results 100)"));
			
			System.out.println(ama);
		}
		catch (ParserException e) {
			e.printStackTrace();
		}

	}

}

