//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Emorphia are Copyright (C) 2001 Emorphia Limited
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.parser.envelope.xml;

import fipaos.util.XMLDataBinding;
import fipaos.parser.ParserException;

import java.util.List;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.Arrays;

/**
 * Facade to provide FIPA-compliant (XC00085G) Envelope<->XML
 * translation
 */
public class Parser
{
	/**
	 * Convert the given byte[] message into an Envelope
	 * @param msg byte[] representing the Envelope
	 * @return The Envelope object representing the original byte[]
	 */
	public static fipaos.ont.fipa.fipaman.Envelope parse( byte[] env ) throws ParserException
	{
		// Convert XML document into XML databinding objects
		String xml = new String( env );

		try
		{
			Envelope xml_env = (Envelope) XMLDataBinding.zeusUnmarshall( xml, "fipaos.parser.envelope.xml" ); //Parser.class.getPackage().getName() );
			return convert( xml_env );
		}
		catch( Throwable t )
		{
			fipaos.util.DIAGNOSTICS.println( "Error when parsing", t, fipaos.util.DIAGNOSTICS.LEVEL_MAX );
			throw new ParserException( "Exception occured when parsing: " + t );
		}
	}

	/**
	 * Convert the given Envelope into FIPA-defined XML format
	 * @param env The envelope to convert
	 * @return The byte[] representing the Envelope, or null
	 */
	public static byte[] deparse( fipaos.ont.fipa.fipaman.Envelope env ) throws ParserException
	{
		// Convert abstract Envelope into XML databinding envelope
		Envelope xml_env = convert( env );

		try
		{
			String xml = XMLDataBinding.zeusMarshall( xml_env );
			return xml.getBytes();
		}
		catch( Throwable t )
		{
			t.printStackTrace();
			fipaos.util.DIAGNOSTICS.println( "Error when parsing", t );
			throw new ParserException( "Exception occured when parsing: " + t );
		}
	}
	
//------------------------------------------------------------------------------
// Convert to abstract envelope methods
//------------------------------------------------------------------------------
	
	/**
	 * Converts fipaos.parser.envelope.xml.Envelope to fipaos.ont.fipa.fipaman.Envelope
	 */
	private static fipaos.ont.fipa.fipaman.Envelope convert( Envelope env ) throws NumberFormatException
	{
		// Create output object
		fipaos.ont.fipa.fipaman.Envelope out = new fipaos.ont.fipa.fipaman.Envelope();
		
		// Start populating
		
		// Iterate through params and update output Envelope
		List params = env.getParamsList();
		Iterator i = params.iterator();
		while( i.hasNext() )
		{
			// Add current params to envelope
			addParams( out, (Params) i.next() );
			
			// Setup output envelope ready for next set of params
			out.newSubEnvelope();
		}
		
		return out;
	}
	
	/**
	 * Adds parameters from a fipaos.parser.envelope.xml.Params to a
	 * fipaos.ont.fipa.fipaman.Envelope
	 */
	private static void addParams( fipaos.ont.fipa.fipaman.Envelope env, Params params ) throws NumberFormatException
	{
		// Set params values into Envelope
		
		AclRepresentation acl_param = params.getAclRepresentation();
		if( acl_param != null && acl_param.getText() != null && acl_param.getText().length() > 0 )
		{
			env.setACLRepresentation( acl_param.getText() );
		}
		
		
		Comments comments_param = params.getComments();
		if( comments_param != null && comments_param.getText() != null && comments_param.getText().length() > 0 )
		{
			env.setComments( comments_param.getText() );
		}
		
		
		Date date_param = params.getDate();
		if( date_param != null && date_param.getText() != null && date_param.getText().length() > 0 )
		{
			env.setDate( new fipaos.util.UTCTime( date_param.getText() ) );
		}
		
		
		Encrypted enc_param = params.getEncrypted();
		if( enc_param != null && enc_param.getText() != null && enc_param.getText().length() > 0 )
		{
			env.setEncrypted( Arrays.asList( new String[] { enc_param.getText() } ) );
		}
		
		
		From from_param = params.getFrom();
		if( from_param != null && from_param.getAgentIdentifier() != null )
		{
			env.setSender( convert( from_param.getAgentIdentifier() ) );
		}
		
		
		IntendedReceiver ir_param = params.getIntendedReceiver();
		if( ir_param != null && ir_param.getAgentIdentifierList() != null )
		{
			Iterator i = ir_param.getAgentIdentifierList().iterator();
			List out = new LinkedList();
			while( i.hasNext() )
			{
				out.add( convert( (AgentIdentifier) i.next() ) );
			}
			env.setIntendedReceiver( out );
		}
		
		
		PayloadEncoding payload_param = params.getPayloadEncoding();
		if( payload_param != null && payload_param.getText() != null )
		{
			env.setPayloadEncoding( payload_param.getText() );
		}
		
		
		PayloadLength payloadlength_param = params.getPayloadLength();
		if( payloadlength_param != null && payloadlength_param.getText() != null )
		{
			env.setPayloadLength( Integer.parseInt( payloadlength_param.getText() ) );
		}
		
		
		Received received_param = params.getReceived();
		if( received_param != null )
		{
			env.setReceived( convert( received_param ) );
		}
		
		
		To to_param = params.getTo();
		if( to_param != null && to_param.getAgentIdentifierList() != null )
		{
			Iterator i = to_param.getAgentIdentifierList().iterator();
			List out = new LinkedList();
			while( i.hasNext() )
			{
				out.add( convert( (AgentIdentifier) i.next() ) );
			}
			env.setReceivers( out );
		}		
	}
	
	
	/**
	 * Converts fipaos.parser.envelope.xml.AgentIdentifier to fipaos.ont.fipa.fipaman.AgentID
	 */
	private static fipaos.ont.fipa.fipaman.AgentID convert( AgentIdentifier aid )
	{
		fipaos.ont.fipa.fipaman.AgentID out = new fipaos.ont.fipa.fipaman.AgentID();
		List out_addrs = new LinkedList();
		List out_resolvers = new LinkedList();
		
		// Convert individual parts of the AgentIdentifier
		out.setName( aid.getName().getText() );
		Addresses addrs = aid.getAddresses();
		Resolvers resolvers = aid.getResolvers();
		
		// Check addresses are given
		if( addrs != null && addrs.getUrlList() != null )
		{
			// Iterate through addresses and translate
			Iterator i = addrs.getUrlList().iterator();
			while( i.hasNext() )
			{
				Url next = (Url) i.next();
				out_addrs.add( new fipaos.util.URL( next.getText() ) );
			}
		}
		
		// Check resolvers are given
		if( resolvers != null && resolvers.getAgentIdentifierList() != null )
		{
			// Iterate through resolvers and translate
			Iterator i = resolvers.getAgentIdentifierList().iterator();
			while( i.hasNext() )
			{
				AgentIdentifier next = (AgentIdentifier) i.next();
				out_resolvers.add( convert( next ) );
			}
		}
		
		// Put everything into one AID
		out.setAddresses( out_addrs );
		out.setResolvers( out_resolvers );
		
		return out;
	}
	
	
	/**
	 * Converts fipaos.parser.envelope.xml.Received to fipaos.ont.fipa.fipaman.Received
	 */
	private static fipaos.ont.fipa.fipaman.Received convert( Received rcvd )
	{
		fipaos.util.URL from = null;
		fipaos.util.URL by = null;
		fipaos.util.UTCTime time = null;
		String id = null;
		String via = null;
		
		if( rcvd.getReceivedBy() != null && rcvd.getReceivedBy().getValue() != null )
		{
			by = new fipaos.util.URL( rcvd.getReceivedBy().getValue() );
		}
		
		if( rcvd.getReceivedFrom() != null && rcvd.getReceivedFrom().getValue() != null )
		{
			from = new fipaos.util.URL( rcvd.getReceivedFrom().getValue() );
		}
		
		if( rcvd.getReceivedDate() != null && rcvd.getReceivedDate().getValue() != null )
		{
			time = new fipaos.util.UTCTime( rcvd.getReceivedDate().getValue() );
		}
		
		if( rcvd.getReceivedId() != null && rcvd.getReceivedId().getValue() != null )
		{
			id = rcvd.getReceivedId().getValue();
		}
		
		if( rcvd.getReceivedVia() != null && rcvd.getReceivedVia().getValue() != null )
		{
			via = rcvd.getReceivedVia().getValue();
		}
		
		return new fipaos.ont.fipa.fipaman.Received( from, by, id, via, time );
	}
	
//------------------------------------------------------------------------------
// Convert from abstract envelope methods
//------------------------------------------------------------------------------
	
	/**
	 * Converts fipaos.ont.fipa.fipaman.Envelope to fipaos.parser.envelope.xml.Envelope
	 */
	private static Envelope convert( fipaos.ont.fipa.fipaman.Envelope env )
	{
		// Create new Envelope to contain output
		Envelope out = new EnvelopeImpl();
		
		// Get sub-envelopes of given envelope and get Params object for each
		Iterator i = env.getSubEnvelopes().iterator();
		int index = 1;
		while( i.hasNext() )
		{
			// Get next envelope to convert
			fipaos.ont.fipa.fipaman.Envelope next = (fipaos.ont.fipa.fipaman.Envelope) i.next();
			
			// Convert sub-envelope to a Params instance
			Params params = convertParams( next );
			
			// Check a Params was created
			if( params != null )
			{
				// Set index of Params as appropriate
				params.setIndex( Integer.toString( index++ ) );

				// Add Params object to list of Params
				out.addParams( params );
			}
		}
		
		return out;
	}
	
	/**
	 * Converts a (sub) fipaos.ont.fipa.fipaman.Envelope to Params
	 */
	private static Params convertParams( fipaos.ont.fipa.fipaman.Envelope env )
	{
		// Indicates that something has actually been added to Params object to be created
		boolean empty = true;
		
		// Create Params object to contain output
		Params out = new ParamsImpl();
		
		String acl_param = env.getACLRepresentation();
		if( acl_param != null && acl_param.length() > 0 )
		{
			AclRepresentation val = new AclRepresentationImpl();
			val.setText( acl_param );
			out.setAclRepresentation( val );
			empty = false;
		}
		
		
		String comments_param = env.getComments();
		if( comments_param != null && comments_param.length() > 0 )
		{
			Comments val = new CommentsImpl();
			val.setText( comments_param );
			out.setComments( val );
			empty = false;
		}
		
		
		fipaos.util.UTCTime date_param = env.getDate();
		if( date_param != null )
		{
			Date val = new DateImpl();
			val.setText( date_param.getZTime().toString() );
			out.setDate( val );
			empty = false;
		}
		
		
		List enc_param = env.getEncrypted();
		if( enc_param != null && enc_param.size() > 0 )
		{
			Encrypted val = new EncryptedImpl();
			val.setText( enc_param.get(0).toString() );
			out.setEncrypted( val );
			empty = false;
		}
		
		
		fipaos.ont.fipa.fipaman.AgentID from_param = env.getSender();
		if( from_param != null )
		{
			From val = new FromImpl();
			val.setAgentIdentifier( convert( from_param ) );
			out.setFrom( val );
			empty = false;
		}
		
		
		List ir_param = env.getIntendedReceiver();
		if( ir_param != null && ir_param.size() > 0 )
		{
			IntendedReceiver val = new IntendedReceiverImpl();
			List aids = new LinkedList();
			Iterator i = ir_param.iterator();
			while( i.hasNext() )
			{
				fipaos.ont.fipa.fipaman.AgentID aid = (fipaos.ont.fipa.fipaman.AgentID) i.next();
				aids.add( convert( aid ) );
			}
			val.setAgentIdentifierList( aids );
			out.setIntendedReceiver( val );
			empty = false;
		}
		
		
		String payload_param = env.getPayloadEncoding();
		if( payload_param != null && payload_param.length() > 0 )
		{
			PayloadEncoding val = new PayloadEncodingImpl();
			val.setText( payload_param );
			out.setPayloadEncoding( val );
			empty = false;
		}
		
		
		int payloadlength_param = env.getPayloadLength();
		if( payloadlength_param >= 0 )
		{
			PayloadLength val = new PayloadLengthImpl();
			val.setText( Integer.toString( payloadlength_param ) );
			out.setPayloadLength( val );
			empty = false;
		}
		
		
		fipaos.ont.fipa.fipaman.Received received_param = env.getReceived();
		if( received_param != null )
		{
			out.setReceived( convert( received_param ) );
			empty = false;
		}
		
		
		List to_param = env.getReceivers();
		if( to_param != null && to_param.size() > 0 )
		{
			To val = new ToImpl();
			List aids = new LinkedList();
			Iterator i = to_param.iterator();
			while( i.hasNext() )
			{
				fipaos.ont.fipa.fipaman.AgentID aid = (fipaos.ont.fipa.fipaman.AgentID) i.next();
				aids.add( convert( aid ) );
			}
			val.setAgentIdentifierList( aids );
			out.setTo( val );
			empty = false;
		}
		

		// Anything set?
		if( empty )
		{
			// No - just return null
			out = null;
		}
		
		return out;
	}
	
	/**
	 * Converts fipaos.ont.fipa.fipaman.AgentID to fipaos.ont.fipa.fipaman.AgentID
	 */
	private static AgentIdentifier convert( fipaos.ont.fipa.fipaman.AgentID aid )
	{
		AgentIdentifier out = new AgentIdentifierImpl();
		
		// Convert individual parts of the AgentIdentifier
		Name name = new NameImpl();
		name.setText( aid.getName() );
		out.setName( name );
		List addrs = aid.getAddresses();
		List resolvers = aid.getResolvers();
		
		// Check addresses are given
		if( addrs != null && addrs.size() > 0 )
		{
			List out_addrs = new LinkedList();
			
			// Iterate through addresses and translate
			Iterator i = addrs.iterator();
			while( i.hasNext() )
			{
				fipaos.util.URL next = (fipaos.util.URL) i.next();
				Url out_url = new UrlImpl();
				out_url.setText( next.toString() );
				out_addrs.add( out_url );
			}
			
			// Add addresses to converted object
			Addresses out_addrs_obj = new AddressesImpl();
			out_addrs_obj.setUrlList( out_addrs );
			out.setAddresses( out_addrs_obj );
		}
		
		// Check resolvers are given
		if( resolvers != null && resolvers.size() > 0 )
		{
			List out_resolvers = new LinkedList();

			// Iterate through resolvers and translate
			Iterator i = resolvers.iterator();
			while( i.hasNext() )
			{
				fipaos.ont.fipa.fipaman.AgentID next = (fipaos.ont.fipa.fipaman.AgentID) i.next();
				out_resolvers.add( convert( next ) );
			}
			
			// Add addresses to converted object
			Resolvers out_resolvers_obj = new ResolversImpl();
			out_resolvers_obj.setAgentIdentifierList( out_resolvers );
			out.setResolvers( out_resolvers_obj );
		}
		
		return out;
	}
	
	
	/**
	 * Converts fipaos.ont.fipa.fipaman.Received to fipaos.parser.envelope.xml.Received
	 */
	private static Received convert( fipaos.ont.fipa.fipaman.Received rcvd )
	{
		Received out = new ReceivedImpl();
		
		if( rcvd.getBy() != null )
		{
			ReceivedBy by = new ReceivedByImpl();
			by.setValue( rcvd.getBy().toString() );
			out.setReceivedBy( by );
		}
		
		if( rcvd.getFrom() != null )
		{
			ReceivedFrom from = new ReceivedFromImpl();
			from.setValue( rcvd.getFrom().toString() );
			out.setReceivedFrom( from );
		}
		
		if( rcvd.getDateTime() != null )
		{
			ReceivedDate date = new ReceivedDateImpl();
			date.setValue( rcvd.getDateTime().toString() );
			out.setReceivedDate( date );
		}
		
		if( rcvd.getId() != null )
		{
			ReceivedId id = new ReceivedIdImpl();
			id.setValue( rcvd.getId() );
			out.setReceivedId( id );
		}
		
		if( rcvd.getVia() != null )
		{
			ReceivedVia via = new ReceivedViaImpl();
			via.setValue( rcvd.getVia() );
			out.setReceivedVia( via );
		}
		
		return out;
	}
		
	public static void main( String args[] ) throws Throwable
	{
		System.out.println( "Read in XML envelope:" );
		fipaos.ont.fipa.fipaman.Envelope env = parse( fipaos.util.DocumentRetriever.getDoc( args[0] ).getBytes() );
		System.out.println( env );
		
		// Create Misc envelope object
/*		fipaos.ont.fipa.fipaman.Envelope env = new fipaos.ont.fipa.fipaman.Envelope();
		env.setACLRepresentation( fipaos.ont.fipa.ACL.FIPA00069 );
		env.setDate( new fipaos.util.UTCTime() );
		env.setSender( new fipaos.ont.fipa.fipaman.AgentID( "(agent-identifier :name test1@emorphia.com)" ) );
		env.setIntendedReceiver( Arrays.asList( new Object[] {
			new fipaos.ont.fipa.fipaman.AgentID( "(agent-identifier :name test1@emorphia.com)" ) } ) );
		env.setReceivers( env.getIntendedReceiver() );
		env.setPayloadLength( 0 );
		env.setPayloadEncoding( "no-encoding" );
		
		System.out.println( "\n\nPre-prepared envelope:\n" + env );
*/
		System.out.println( "\n\nXML'ised representation of envelope:\n" + new String( deparse( env ) ) );
		
		System.exit( 0 );
	}
}
