//******************************************************************************
//	Open Source Copyright Notice and License: FIPA-OS
//
//	1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//	2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.platform;

//----------------------------------------------------------------------

import fipaos.agent.*;
import fipaos.agent.conversation.*;
import fipaos.agent.task.*;
import fipaos.agent.profile.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.ont.fipa.*;
import fipaos.util.*;
import fipaos.platform.ams.*;
import fipaos.skill.db.*;
import fipaos.ont.BASETOKENS;

//----------------------------------------------------------------------

/**
 * The AMS is the controller of agent activity on the platform. It contains a
 * list of all agent that are currently operating on the platform and has
 * control over which agents are allowed to register on the platform.
 * It also controls agent execution via thread management and controls
 * agent migration. The ACC uses AMS functions to authenticate agents.
 *
 * $Revision: 1.14 $ $Date: 2001/05/25 10:45:52 $ $Author: alantreadway $
 *
 **/
public class AgentManagementSystem extends FIPAOSAgent
{

	//----------------------------------------------------------------------
	// CONSTANTS
	//----------------------------------------------------------------------

	// Other
	private static final String DB_NAME						= "ams_directory";


	//----------------------------------------------------------------------
	// INSTANCE VARIABLES
	//----------------------------------------------------------------------

	// IOR database
	private Database _ams_database;


	//----------------------------------------------------------------------
	// CONSTRUCTOR
	//----------------------------------------------------------------------

	/**
	 * Creates a new AMS. AMS extends FIPAOSAgent and uses the second incarnation of the
	 * Conversation Manager.
	 **/
	public AgentManagementSystem( PlatformProfile platform_profile, AgentProfile agent_profile,
																String name, String owner )
	{
		super( platform_profile, agent_profile, name, owner, true );
		init( name );
	}
	
	/**
	 * Creates a new AMS. AMS extends FIPAOSAgent and uses the second incarnation of the
	 * Conversation Manager.
	 *
	 * @param platform_profile_location The path to the platform profile file
	 * @param name The name of this AMS
	 * @param owner The name of the entity that owns this AMS
	 **/
	public AgentManagementSystem( String platform_profile_location, String name, String owner )
	{
		super( platform_profile_location, name, owner, true );
		init( name );		
	}
	
	/**
	 * Creates a new AMS. AMS extends FIPAOSAgent and uses the second incarnation of the
	 * Conversation Manager.
	 *
	 * @param platform_profile_location The path to the platform profile file
	 * @param name The name of this AMS
	 * @param owner The name of the entity that owns this AMS
	 **/
	protected AgentManagementSystem( PlatformProfile pp, AgentProfile ap, String name,String owner, 
																	FIPAOSAgentComponentFactory factory )
	{
		super( pp, ap, name, owner, true, factory );
		init( name );
	}


	//----------------------------------------------------------------------
	// PRIVATE METHODS
	//----------------------------------------------------------------------

	private void init( String name )
	{
		try
		{
			// Construct Database
			_ams_database = DatabaseFactory.getDatabase( name + "_db", getProfile().getDatabaseProfile() );

			// Open Database
			_ams_database.openDatabase();
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Error when initialising Database: ", t, this,
			DIAGNOSTICS.LEVEL_MAX );
			return;
		}

		// Construct listener task + instantiate
		setListenerTask( new IdleTask() );
	}
	
	//----------------------------------------------------------------------
	// PUBLIC METHODS
	//----------------------------------------------------------------------

	/**
	 * Clears up any outstanding stuff as the object is gc'ed.
	 **/
	public void shutdown()
	{
		// Invoked FIPAOSAgent.shutdown()
		super.shutdown();
		
		// Close down DB
		if( _ams_database != null )
		{
			try
			{
				_ams_database.closeDatabase();
			}
			catch( Throwable t )
			{
			}
		}
	}

	public class IdleTask extends Task
	{
		public IdleTask()
		{
		}

		public void startTask()
		{
			// Don't do anything
		}

		public void handleRequest( Conversation conv )
		{
			// Get the type of request
			ACL acl = conv.getACL( conv.getLatestMessageIndex() );

			// Check ontology
			if( acl.getOntology() == null || 
					!acl.getOntology().equalsIgnoreCase( "fipa-agent-management" ) )
			{
				sendNotUnderstood( acl );
				return;
			}
			
			// Check language
			if( acl.getLanguage() == null || 
					!acl.getLanguage().equalsIgnoreCase( FIPACONSTANTS.SL0 ) )
			{
				sendNotUnderstood( acl );
				return;
			}
			
			try
			{
				// Get content of message
				Object content_obj = acl.getContentObject();
				AgentManagementAction ama = null;
				
				// Decide if we need to parse content
				if( content_obj instanceof AgentManagementAction )
				{
					// Object from another FIPA-OS Agent - no parsing required!
					ama = (AgentManagementAction) content_obj;
				}
				else if( content_obj instanceof String )
				{
					// Assume standard SL0 encoding?
					
					// Parse the content to decide what to do
					String content = (String) content_obj;
//					content = StringParser.trimBrackets( content );
					ama = new AgentManagementAction( content );						
				}
				else
				{
					// Unknown content ?!?!?!
					DIAGNOSTICS.println( "Couldn't extract AgentManagementAction - sending not-understood",
																this, DIAGNOSTICS.LEVEL_MAX);
					sendNotUnderstood( acl );
					return;
				}
				
				
				// Get the action of this
				String function = ama.getAction();

				// Decide which sub-tasks to spawn
				if( function.equals( FIPAMANCONSTANTS.REGISTER ) )
				{
					newTask( new RegisterTask( conv, ama, _ams_database ) );
				}
				else if( function.equals( FIPAMANCONSTANTS.DEREGISTER ) )
				{
					newTask( new DeregisterTask( conv, ama, _ams_database ) );
				}
				else if( function.equals( FIPAMANCONSTANTS.MODIFY ) )
				{
					newTask( new ModifyTask( conv, ama, _ams_database ) );
				}
				else if( function.equals( FIPAMANCONSTANTS.SEARCH ) )
				{
					newTask( new SearchTask( conv, ama, _ams_database ) );
				}
				else if ( function.equals( FIPAMANCONSTANTS.GET_DESCRIPTION ) )
				{
					newTask( new GetDescriptionTask(conv, ama, _ams_database) );
				}
				else
				{
					sendNotUnderstood( acl );
				}
			}
			catch( Throwable t )
			{
				DIAGNOSTICS.println( "Exception occured - sending not-understood", t,
				this, DIAGNOSTICS.LEVEL_MAX);
				sendNotUnderstood( acl );
			}
		}

		public void doneRegisterTask( Task t )
		{
			DIAGNOSTICS.println( "RegisterTask completed OK" );
		}

		public void doneDeregisterTask( Task t )
		{
			DIAGNOSTICS.println( "DeregisterTask completed OK" );
		}

		public void doneModifyTask( Task t )
		{
			DIAGNOSTICS.println( "ModifyTask completed OK" );
		}

		public void doneSearchTask( Task t )
		{
			DIAGNOSTICS.println( "SearchTask completed OK" );
		}

		public void doneGetDescriptionTask( Task t )
		{
			DIAGNOSTICS.println( "GetDescriptionTask completed OK" );
		}

		public void errorRegisterTask( Task task, Throwable t )
		{
			DIAGNOSTICS.println( "RegisterTask didn't completed OK", t, this,
			DIAGNOSTICS.LEVEL_MAX );
		}

		public void errorDeregisterTask( Task task, Throwable t )
		{
			DIAGNOSTICS.println( "DeregisterrTask didn't completed OK", t, this,
			DIAGNOSTICS.LEVEL_MAX );
		}

		public void errorModifyTask( Task task, Throwable t )
		{
			DIAGNOSTICS.println( "ModifyTask didn't completed OK", t, this,
			DIAGNOSTICS.LEVEL_MAX );
		}

		public void errorSearchTask( Task task, Throwable t )
		{
			DIAGNOSTICS.println( "SearchTask didn't completed OK", t, this,
			DIAGNOSTICS.LEVEL_MAX );
		}

		public void errorGetDescriptionTask( Task task, Throwable t )
		{
			DIAGNOSTICS.println( "GetDescriptionTask didn't completed OK", t, this,
			DIAGNOSTICS.LEVEL_MAX );
		}	
	}
} // End AMS
