//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.platform.ams;

//----------------------------------------------------------------------

import fipaos.agent.task.*;
import fipaos.agent.*;
import fipaos.agent.conversation.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.ont.*;


//----------------------------------------------------------------------

/**
 * This class provides the ability for an Agent to register with the AMS.
 * <p>It provides two mechanisms for passing back its results.  Either the
 * Callback interface can be used by specifying an object which implements
 * this interface, or a parent-task can check the results that are passed.
 * <p>The results are of the form:
 * <ul><li>AgentID - Name of Agent succesfully registered with</li>
 * <li>Object[] { AgentID, String } - Name of Agent attempted to register with,
 * and the failure/refusal reason</li>
 * </ul>
 */
public class AMSRegistrationTask extends Task
{
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

	/**
	 * AID of AMS to register with
	 */
	private AgentID _ams;

	/**
	 * Active state of the Agent registering
	 */
	private String _state;
	
//----------------------------------------------------------------------
// CONSTRUCTOR(S)
//----------------------------------------------------------------------

	/**
	 * Create a Task to register with the AMS
	 */
	public AMSRegistrationTask()
	{
		this((AgentID)null );
	}
	
	/**
	 * Create a Task to register with the AMS
	 */
	public AMSRegistrationTask(AgentID ams)
	{
		this(ams, FIPAMANCONSTANTS.ACTIVE );
	}
	
	/**
	 * Create a Task to register with the AMS
	 */
	public AMSRegistrationTask(String state)
	{
		this((AgentID)null, state );
	}
	
	/**
	 * Create a Task to register with the AMS
	 */
	public AMSRegistrationTask(AgentID ams, String state)
	{
		_state = state;
		_ams = ams;
	}
	
//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------	

	/**
	 * Invoked when task starts
	 */
	public void startTask()
	{
		// Check if _ams == null
		if( _ams == null )
		{
			_ams = _owner.getLocalAMSAID();
		}
		
		// Get AID of Agent to register
		AgentID id = _owner.getAID();

		// Create Conversation for interaction with the AMS
		ACL to_ams = getNewConversation( FIPACONSTANTS.FIPA_REQUEST );

		// Set-up message parameters
		to_ams.setReceiverAID( _ams );
		to_ams.setOntology( FIPAMANCONSTANTS.FIPA_AGENT_MANAGEMENT );
		to_ams.setLanguage( FIPACONSTANTS.SL0 );
		to_ams.setPerformative( FIPACONSTANTS.REQUEST );

		// Create AMSAgentDescription
		AMSAgentDescription ams_desc = new AMSAgentDescription();
		ams_desc.setAgentID( id );
		ams_desc.setState( _state );
		ams_desc.setOwnership( _owner.getOwnership() );

		AgentManagementAction ama = new AgentManagementAction();
		ama.setAction( FIPAMANCONSTANTS.REGISTER );
		ama.setAgentID( _ams );
		ama.setAMSAgentDescription( ams_desc );

		to_ams.setContentObject( ama );

		// Send message
		forward( to_ams );
	}
	
	
	/**
	 * Deal with not-understood messages
	 */
	public void handleNotUnderstood( Conversation conv )
	{
		ACL in = conv.getACL( conv.getLatestMessageIndex() );
		AgentID sender = in.getSenderAID();
		
		// Indicate we're done and pass-back results
		done( new Object[] { sender, FIPACONSTANTS.NOT_UNDERSTOOD } );
	}
	
	
	/**
	 * Deal with refuse messages
	 */
	public void handleRefuse( Conversation conv )
	{
		ACL in = conv.getACL( conv.getLatestMessageIndex() );
		AgentID sender = in.getSenderAID();
		String reason = new String();
		Object content = in.getContentObject();
		
		try
		{
			if( content instanceof String )
			{
				AgentManagementException ame = new AgentManagementException( (String)content );

				if( ame.getException() != null )
				{
					reason = ame.getException();
				}
			}
			else if( content instanceof AgentManagementException )
			{
				reason = ( ( AgentManagementException) content ).getException();
			}
		}
		catch( Throwable t )
		{
		}
		
		// Indicate we're done and pass-back results
		done( new Object[] { sender, reason } );
	}
	
	
	/**
	 * Deal with agree messages
	 */
	public void handleAgree( Conversation conv )
	{
		
		// Do nothing
		
	}
	
	
	/**
	 * Deal with failure messages
	 */
	public void handleFailure( Conversation conv )
	{
		ACL in = conv.getACL( conv.getLatestMessageIndex() );
		AgentID sender = in.getSenderAID();
		String reason = new String();
		Object content = in.getContentObject();

		try
		{
			if( content instanceof String )
			{
				AgentManagementException ame = new AgentManagementException( (String)content );

				if( ame.getException() != null )
				{
					reason = ame.getException();
				}
			}
			else if( content instanceof AgentManagementException )
			{
				reason = ( ( AgentManagementException) content ).getException();
			}
		}
		catch( Throwable t )
		{
		}
		
		// Indicate we're done and pass-back results
		done( new Object[] { sender, reason } );
	}
	

	/**
	 * Deal with failure messages
	 */
	public void handleInform( Conversation conv )
	{
		ACL in = conv.getACL( conv.getLatestMessageIndex() );
		AgentID sender = in.getSenderAID();
		
		// Indicate we're done and pass-back results
		done( sender );
	}
	
//----------------------------------------------------------------------
// MAIN/TESTING METHOD
//----------------------------------------------------------------------
	
} // End class
