//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Emorphia are Copyright (C) 2001 Emorphia Limited
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.platform.ams;

//----------------------------------------------------------------------

import fipaos.agent.*;
import fipaos.agent.conversation.*;
import fipaos.agent.task.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.ont.fipa.*;
import fipaos.skill.db.*;
import fipaos.util.*;
import fipaos.ont.BASETOKENS;
import java.util.*;

//----------------------------------------------------------------------

/**
 * This class encapsulates the local search mechanisms of the AMS, and coordinates
 * federated searches where appropriate
 * <br>
 * <br>
 */
public class SearchTask extends AMSTask
{

	//----------------------------------------------------------------------
	// CONSTANTS
	//----------------------------------------------------------------------

	/**
	 * Timeout in milliseconds for federated searches
	 */
	public static final long FED_TIMEOUT = 60000;

//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------
	
	/**
	 * AMSAgentDescription being used in the search
	 */
	private AMSAgentDescription _ams_desc;

	/**
	 * Results of search
	 */
	private Set _results;
	
	
//----------------------------------------------------------------------
// CONSTRUCTOR(S)
//----------------------------------------------------------------------

	/** This is the constructor for the SearchTask.
	 * @param conv Conversation
	 * @param ama Agent Management Action
	 * @param db Database
	 */
	public SearchTask( Conversation conv, AgentManagementAction ama, Database db )
	{
		super(conv,ama,db);
		
		_ams_desc = ama.getAMSAgentDescription();
	}


//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------

	/** This is the initial method invoked by the task.
	 */
	public void startTask()
	{
		// Initialised the results list
		_results = new TreeSet( fipaos.util.Comparators.OBJECT_COMPARE );

		// Agree to the request
		sendAgree();

		// Do a local search
		_results.addAll( localSearch( _ams_desc ) );
		
		// Send back results
		searchComplete();
		
	}

//----------------------------------------------------------------------
// PRIVATE METHODS
//-----------------------------------------------------------------------------

	/**
	 * Search completed - send back results
	 */
	private void searchComplete()
	{
		// Send result back
		sendInformResult( _results );

		// Indicate we're done!
		done();
	}
	
	/**
	 * Do a local search of the database
	 */
	private Set localSearch( AMSAgentDescription ams_desc )
	{
		// Create results set
		Set results = new TreeSet( fipaos.util.Comparators.OBJECT_COMPARE );

		try
		{
			// Start read-only transaction
			_ams_db.startTransaction( true );

			// Retrieve Enumeration of the DFAgentDescription in the DB, and iterate through the
			Enumeration ams_descs = _ams_db.getIDs();
			while( ams_descs.hasMoreElements() )
			{
				// Get next DFAgentDescription
				AMSAgentDescription current = (AMSAgentDescription) _ams_db.findObject( ( (String) ams_descs.nextElement() ) );

				// Check that it matches our requirements
				if ( current != null && compare( ams_desc, current ) )
				{
					results.add( current );
				}
			}
		}
		catch (DatabaseException dbe)
		{
			DIAGNOSTICS.println("Error with getting list of AMS's", dbe, this, DIAGNOSTICS.LEVEL_MAX );
		}
		finally
		{
			try
			{
				_ams_db.endTransaction();
			}
			catch( Throwable t )
			{
			}
		}
		
		return results;
	}


	/**
	 * Compare the second AMSAgentDescription against the first - returns true if it
	 * matches (i.e. first is sub-set of second)
	 */
	private static boolean compare( AMSAgentDescription desc1, AMSAgentDescription desc2 )
	{
		
		// Compare fields
		
		// Compare Agent name
		if( desc1.getAgentID() != null &&
				!desc1.getAgentID().equals( desc2.getAgentID() ) )
		{
			return false;
		}
		
		// Compare ownership
		if( desc1.getOwnership() != null )
		{
			if( desc2.getOwnership() != null )
			{
				String owner1 = desc1.getOwnership();
				String owner2 = desc2.getOwnership();
				
				if( !owner2.equals( owner1 ) )
				{
					return false;
				}
			}
			else
			{
				return false;
			}
		}
		
		// Compare state
		if( desc1.getState() != null )
		{
			if( desc2.getState() != null )
			{
				String state1 = desc1.getState();
				String state2 = desc2.getState();
				
				if( !state2.equals( state1 ) )
				{
					return false;
				}
			}
			else
			{
				return false;
			}
		}
		
		return true;
	}
	
//----------------------------------------------------------------------
// MAIN/TESTING METHOD
//----------------------------------------------------------------------


} // End class
