//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Emorphia are Copyright (C) 2001 Emorphia Limited
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//    Portions created by Michal Dobisek are Copyright (C) 2001 Michal Dobisek
//    licensed under the Emorphia Public License, available at
//    http://www.emorphia.com/EPL/.
//
//******************************************************************************

package fipaos.platform.df;

import fipaos.agent.task.*;
import fipaos.agent.conversation.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.ont.proprietary.subscription.*;
import fipaos.util.*;

/** This Task provides a mechanism for listening to events occuring to a particular DF.
 *
 * <p>Currently this Task (and its counterpart) only support intra-platform communication
 * over a transport which maintains message content as objects (i.e. RMI).
 */
public class DFListenerTask extends Task 
{
	/** Object that is listening for DF events
	 */
	private DFListener _listener;
	
	/** DF that is being monitored
	 */
	private AgentID _df;
	
	/** Conversation ID associated with our subscription messages
	 */
	private String _conv_id;
	
	/** Indicates if subscription has occured succesfully
	 */
	private boolean _subscribed = false;
	
	
	/** Creates a DFListenerTask - once the Task has started, it will attempt to subscribe
	 * to events from the given DF and forward them to the given DFListener
	 * @param listener Object to receive DF event notification
	 * @param df DF to monitor
	 */
	public DFListenerTask( DFListener listener, AgentID df ) 
	{
		// Save listener reference
		_listener = listener;
		
		// Save DF name reference
		_df = df;
	}
	
	/** Invoked when the Task starts - sends subscription message to target DF
	 */
	public void startTask()
	{
		// Attempt to send a subscribe message to the DF we need to monitor
		ACL acl = getNewConversation(FIPACONSTANTS.FIPA_SUBSCRIBE);

		// Save conversation ID
		_conv_id = acl.getConversationID();

		acl.setSenderAID( _owner.getAID() );
		acl.setReceiverAID( _df );
		acl.setPerformative( FIPACONSTANTS.SUBSCRIBE );
		acl.setOntology( FIPAMANCONSTANTS.FIPA_AGENT_MANAGEMENT );
		
		// Create SubscriptionAction object to encapsulate the request
		SubscriptionAction sa = new SubscriptionAction();
		sa.setCondition( new String[] { FIPAMANCONSTANTS.REGISTER, 
																		FIPAMANCONSTANTS.DEREGISTER,
																		FIPAMANCONSTANTS.MODIFY } );
		
		// Add subscribe request to ACL message
		acl.setContentObject( sa );
		
		// Send subscription message
		forward( acl );
	}
	
	/**
	 * Indicates if this DFListenerTask has been succesfull in subscribing
	 */
	public boolean subscribedOK()
	{
		return _subscribed;
	}
	
	/**
	 * Deal with not-understood messages here
	 */
	public void handleNotUnderstood( Conversation conv )
	{
		// End this task
		DIAGNOSTICS.println( "Received a not-understood - terminating", this, DIAGNOSTICS.LEVEL_2 );
		done();
	}
	
	/** 
	 * Invoked when the DF refuses subscription
	 * @param conv
	 */
	public void handleRefuse( Conversation conv )
	{
		DIAGNOSTICS.println( "Subscription to " + _df.getName() + " refused", this, DIAGNOSTICS.LEVEL_3 );
		
		// We couldn't subscribe!
		done();
	}
	
	/** Invoked when an inform is received - either indicates that the subscription
	 * has been setup OK, or indicates an event has occured that meets the requirements
	 * of the original subscription request
	 * @param conv
	 */
	public void handleInform( Conversation conv )
	{
		// Get message
		ACL msg = conv.getACL( conv.getLatestMessageIndex() );
		
		// Check if we're subscribed
		if( !_subscribed && msg.getContentObject() instanceof SubscriptionAction )
		{
			DIAGNOSTICS.println( "Subscription to " + _df.getName() + " confirmed", this,
														DIAGNOSTICS.LEVEL_3 );

			_subscribed = true;
			return;
		}
		
		// Get SubscriptionEvent & DFAD
		AgentManagementAction ama = null;
		DFAgentDescription dfad = null;
		
		if( msg.getContentObject() instanceof AgentManagementAction )
		{
			ama = (AgentManagementAction) msg.getContentObject();
			
		}
		else
		{
			DIAGNOSTICS.println( "Content of message not an AgentManagementAction",
														this, DIAGNOSTICS.LEVEL_MAX );
			return;
		}
		
		// Get DFDescription, if available
		if( ama != null && ama.getDFAgentDescription() != null )
		{
			dfad = ama.getDFAgentDescription();
		}
		
		DIAGNOSTICS.println( "Received event update from DF of type " + ama.getAction(), this, DIAGNOSTICS.LEVEL_3 );
		
		// Deal with notification of listener
		if( ama.getAction().equals( FIPAMANCONSTANTS.REGISTER ) )
		{
			_listener.registered( dfad );
		}
		else if( ama.getAction().equals( FIPAMANCONSTANTS.DEREGISTER ) )
		{
			_listener.deregistered( dfad );
		}
		else if( ama.getAction().equals( FIPAMANCONSTANTS.MODIFY ) )
		{
			_listener.modified( dfad );
		}
		else
		{
			DIAGNOSTICS.println( "Unknown event type received!", this, DIAGNOSTICS.LEVEL_MAX );
		}
	}
	
	
/** When invoked, causes the subscription with the DF to be cancelled
 */	
	public void stopListening()
	{
		// If this Task hasn't completed, and we succesfully subscribed to the
		// DF, indicate that we wish to cancel the subscription and end the Task
		if( _subscribed )
		{
			_subscribed = false;
			sendCancel();
			done();
		}
	}
	
	
	/** Sends a cancel message to the DF currently subscribed to
	 */
	private void sendCancel()
	{
		// Attempt to send a subscribe message to the DF we need to monitor
		ACL acl = new ACL();
		acl.setSenderAID( _owner.getAID() );
		acl.setReceiverAID( _df );
		acl.setPerformative( FIPACONSTANTS.CANCEL );
		acl.setProtocol( FIPACONSTANTS.FIPA_SUBSCRIBE );
		acl.setOntology( FIPAMANCONSTANTS.FIPA_AGENT_MANAGEMENT );
		acl.setConversationID( _conv_id );
		
		// Create SubscriptionAction object to encapsulate the request
		SubscriptionAction sa = new SubscriptionAction();
		sa.setCondition( new String[] { FIPAMANCONSTANTS.REGISTER, 
																		FIPAMANCONSTANTS.DEREGISTER,
																		FIPAMANCONSTANTS.MODIFY } );
		
		// Add subscribe request to ACL message
		acl.setContentObject( sa );
		
		// Send subscription message
		forward( acl );
	}
}

