//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Emorphia are Copyright (C) 2001 Emorphia Limited
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.platform.df;

//----------------------------------------------------------------------

import fipaos.agent.task.*;
import fipaos.agent.*;
import fipaos.agent.conversation.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.ont.*;


//----------------------------------------------------------------------

/**
 * This class provides the ability for an Agent to register with a DF.
 * <p>It provides two mechanisms for passing back its results.  Either the
 * Callback interface can be used by specifying an object which implements
 * this interface, or a parent-task can check the results that are passed.
 * <p>The results are of the form:
 * <ul><li>AgentID - Name of Agent succesfully registered with</li>
 * <li>Object[] { AgentID, String } - Name of Agent attempted to register with,
 * and the failure/refusal reason</li>
 * </ul>
 */
public class DFRegistrationTask extends Task
{
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

	/**
	 * AID of DF to register with
	 */
	private AgentID _df;

	/**
	 * DF description to register
	 */
	private DFAgentDescription _dfad;
	
//----------------------------------------------------------------------
// CONSTRUCTOR(S)
//----------------------------------------------------------------------

	/**
	 * Create a Task to register with the DF, with the given service type & name
	 */
	public DFRegistrationTask(String type )
	{
		this((AgentID)null, type );
	}
	
	/**
	 * Create a Task to register with the DF
	 */
	public DFRegistrationTask(AgentID df, String type)
	{
		this(df, new DFAgentDescription() );

		// Create appropriate ServiceDescription
		ServiceDescription sd = new ServiceDescription();
		sd.setServiceName( type );
		sd.setServiceType( type );
		
		// Put ServiceDescription into DFAgentDescription
		_dfad.setAgentServices( new ServiceDescription[] { sd } );
	}
	
	/**
	 * Create a Task to register with the DF
	 */
	public DFRegistrationTask(DFAgentDescription dfad )
	{
		this((AgentID)null, dfad );
	}
	
	/**
	 * Create a Task to register with the DF
	 */
	public DFRegistrationTask(AgentID df, DFAgentDescription dfad )
	{
		_dfad = dfad;
		_df = df;
	}
	
//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------	

	/**
	 * Invoked when task starts
	 */
	public void startTask()
	{
		// Check if _ams == null
		if( _df == null )
		{
			_df = _owner.getLocalDFAID();
		}
		
		// Get AID of Agent to register
		AgentID id = _owner.getAID();

		// Finish setting up DFAgentDescription
		_dfad.setAgentName( id );
		
		// Create Conversation for interaction with the AMS
		ACL to_df = getNewConversation( FIPACONSTANTS.FIPA_REQUEST );

		// Set-up message parameters
		to_df.setReceiverAID( _df );
		to_df.setOntology( FIPAMANCONSTANTS.FIPA_AGENT_MANAGEMENT );
		to_df.setLanguage( FIPACONSTANTS.SL0 );
		to_df.setPerformative( FIPACONSTANTS.REQUEST );

		// Create AgentManagementAction
		AgentManagementAction ama = new AgentManagementAction();
		ama.setAction( FIPAMANCONSTANTS.REGISTER );
		ama.setAgentID( _df );
		ama.setDFAgentDescription( _dfad );

		to_df.setContentObject( ama );

		// Send message
		forward( to_df );
	}
	
	
	/**
	 * Deal with not-understood messages
	 */
	public void handleNotUnderstood( Conversation conv )
	{
		ACL in = conv.getACL( conv.getLatestMessageIndex() );
		AgentID sender = in.getSenderAID();
		
		// Indicate we're done and pass-back results
		done( new Object[] { sender, FIPACONSTANTS.NOT_UNDERSTOOD } );
	}
	
	
	/**
	 * Deal with refuse messages
	 */
	public void handleRefuse( Conversation conv )
	{
		ACL in = conv.getACL( conv.getLatestMessageIndex() );
		AgentID sender = in.getSenderAID();
		String reason = new String();
		Object content = in.getContentObject();
		
		try
		{
			if( content instanceof String )
			{
				AgentManagementException ame = new AgentManagementException( (String)content );

				if( ame.getException() != null )
				{
					reason = ame.getException();
				}
			}
			else if( content instanceof AgentManagementException )
			{
				reason = ( ( AgentManagementException) content ).getException();
			}
		}
		catch( Throwable t )
		{
		}
		
		// Indicate we're done and pass-back results
		done( new Object[] { sender, reason } );
	}
	
	
	/**
	 * Deal with agree messages
	 */
	public void handleAgree( Conversation conv )
	{
		
		// Do nothing
		
	}
	
	
	/**
	 * Deal with failure messages
	 */
	public void handleFailure( Conversation conv )
	{
		ACL in = conv.getACL( conv.getLatestMessageIndex() );
		AgentID sender = in.getSenderAID();
		String reason = new String();
		Object content = in.getContentObject();
		
		try
		{
			if( content instanceof String )
			{
				AgentManagementException ame = new AgentManagementException( (String)content );

				if( ame.getException() != null )
				{
					reason = ame.getException();
				}
			}
			else if( content instanceof AgentManagementException )
			{
				reason = ( ( AgentManagementException) content ).getException();
			}
		}
		catch( Throwable t )
		{
		}
		
		// Indicate we're done and pass-back results
		done( new Object[] { sender, reason } );
	}
	

	/**
	 * Deal with failure messages
	 */
	public void handleInform( Conversation conv )
	{
		ACL in = conv.getACL( conv.getLatestMessageIndex() );
		AgentID sender = in.getSenderAID();
		
		// Indicate we're done and pass-back results
		done( sender );
	}
	
	
//----------------------------------------------------------------------
// INNER-CLASSES
//----------------------------------------------------------------------	
	
//----------------------------------------------------------------------
// MAIN/TESTING METHOD
//----------------------------------------------------------------------
	
} // End class
