//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.platform.df;

//----------------------------------------------------------------------

import fipaos.agent.*;
import fipaos.agent.conversation.*;
import fipaos.agent.task.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.ont.fipa.*;
import fipaos.skill.db.*;
import fipaos.util.*;
import fipaos.ont.BASETOKENS;
import java.util.*;

//----------------------------------------------------------------------

/**
 * This class encapsulates the local search mechanisms of the DF, and coordinates
 * federated searches where appropriate
 * <br>
 * <br>
 */
public class SearchTask extends DFTask
{
	
	//----------------------------------------------------------------------
	// CONSTANTS
	//----------------------------------------------------------------------
	
	/**
	 * Timeout in milliseconds for federated searches
	 */
	public static final long FED_TIMEOUT = 60000;
	
	//----------------------------------------------------------------------
	// INSTANCE VARIABLES
	//----------------------------------------------------------------------
	
	/**
	 * Search constraints for this search
	 */
	private SearchConstraints _search_constraints;
	
	/**
	 * DFAgentDescription being used in the search
	 */
	private DFAgentDescription _df_desc;
	
	/**
	 * Results of search so far
	 */
	private Set _results;
	
	/**
	 * Maximum number of results
	 */
	
	private int _max_results;
	
	/**
	 * Number of results being waited for in a federated search
	 */
	private int _waiting_for = 0;
	
	//----------------------------------------------------------------------
	// CONSTRUCTOR(S)
	//----------------------------------------------------------------------
	
	public SearchTask( Conversation conv, AgentManagementAction ama, Database db )
	{
		super(conv,ama,db);
		
		// Get the search constriants for this
		_search_constraints = ama.getSearchConstraints();
		
		// Set up the parameter for the maximum results
		
		if (_search_constraints!=null && _search_constraints.getMaxResults()!=null)
		{
			_max_results=_search_constraints.getMaxResults().intValue();
		}
		else
		{
			// indicates no limit
			
			_max_results=-1;
		}
		
		_df_desc = ama.getDFAgentDescription();
	}
	
	
	//----------------------------------------------------------------------
	// PUBLIC METHODS
	//----------------------------------------------------------------------
	
	public void startTask()
	{
		// Initialised the results list
		_results = new TreeSet( fipaos.util.Comparators.OBJECT_COMPARE );
		
		// Agree to the request
		sendAgree();
		
		//Check if the max. result is unspecified or greater than zero if specified
		if ( _max_results == -1 || _max_results > 0 )
		{
			// Do a local search
			_results.addAll( localSearch( _df_desc ) );
			
			// Do a remote search (if necessary)
			if( _search_constraints != null && 
					_search_constraints.getMaxDepth() != null &&
					_search_constraints.getMaxDepth().intValue() > 0 )
			{
				DIAGNOSTICS.println("Constraints indicate a federated search");
				
				// Send federated search messages
				remoteSearch( _df_desc, _search_constraints );
			}
			else
			{
				DIAGNOSTICS.println("Constraints indicate a local search only");
				
				// Send back results
				searchComplete();
			}
		}
		else
		{
			
			// Send back empty results
			searchComplete();
		}
	}
	
	
	public void doneDFSearchTask( Object result )
	{
		DIAGNOSTICS.println( "Received results!" );
		
		if( result != null )
		{
			// Type-cast to appropriate types
			DFAgentDescription results[] = (DFAgentDescription[]) result;
			
			// Add to results
			synchronized( _results )
			{
				for (int i=0;i<results.length;i++)
				{
					
					// Check if the size of the result is bigger than the limit
					
					if (_max_results==-1 || _results.size()<_max_results)
					{
						_results.add( results[i] );
					}
					else
					{
						break;
					}
					
				}
			}
		}
		
		searchComplete();
	}
	
	
	public void errorDFSearchTask( Task task, Throwable t )
	{
		DIAGNOSTICS.println( "DFSearchTask failed: ", t, this, DIAGNOSTICS.LEVEL_MAX );
		
		searchComplete();
	}
	
	public void timeoutDFSearchTask( Task task )
	{
		DIAGNOSTICS.println( "DFSearchTask timedout: ", this );
		
		searchComplete();
	}
	
	//----------------------------------------------------------------------
	// PRIVATE METHODS
	//-----------------------------------------------------------------------------
	
	/**
	 * Search completed - send back results
	 */
	private void searchComplete()
	{
		// Check if search is complete
		synchronized( _results )
		{
			// Decrement counter
			_waiting_for--;
			
			DIAGNOSTICS.println( "Waiting for " + _waiting_for + " more results...", this );
			
			// Check if we're done
			if( _waiting_for <= 0 )
			{
				// Send result back
				sendInformResult( _results );
				
				// Indicate we're done!
				done();
			}
		}
	}
	
	
	/**
	 * Start a federated search
	 */
	private void remoteSearch( DFAgentDescription df_desc, SearchConstraints search_const )
	{
		// Get a list of DF's to search
		List dfs = getDFs();
		_waiting_for = dfs.size();
		
		// Check if we can acutally federate the search
		if( dfs.size() <= 0 )
		{
			DIAGNOSTICS.println( "No DF's to federate search!", this, DIAGNOSTICS.LEVEL_3 );
			searchComplete();
		}
		else
		{
			// Decrement MAX value
			search_const.setMaxDepth( new Integer( search_const.getMaxDepth().intValue() - 1 ) );
			
			// Iterate through DF's and send request
			Iterator i = dfs.iterator();
			while( i.hasNext() )
			{
				AgentID aid = (AgentID) i.next();
				
				newTask( new DFSearchTask( aid.toString(), df_desc, search_const, true ), FED_TIMEOUT );
			}
		}
	}
	
	
	/**
	 * Get a list of the AgentID's of the DF's registered with this
	 */
	private List getDFs()
	{
		// Create DFAgentDescription describing DF's
		DFAgentDescription dfad = new DFAgentDescription();
		ServiceDescription sd = new ServiceDescription();
		sd.setServiceType( FIPAMANCONSTANTS.FIPA_DF );
		dfad.setAgentServices( new ServiceDescription[]
		{ sd } );
		
		// Now search for DF's registered with this
		Set results = localSearch( dfad );
		
		// Put the AgentID's of the given DF's into a list
		List agent_ids = new LinkedList();
		Iterator i = results.iterator();
		while( i.hasNext() )
		{
			// Get next DFAgentDescription
			DFAgentDescription curr = (DFAgentDescription) i.next();
			
			// Pull the AgentID out!
			agent_ids.add( curr.getAgentID() );
		}
		
		return agent_ids;
	}
	
	
	/**
	 * Do a local search of the database
	 */
	private Set localSearch( DFAgentDescription df_desc )
	{
		// Create results set
		Set results = new TreeSet( fipaos.util.Comparators.OBJECT_COMPARE );
		
		try
		{
			// Start read-only transaction
			_df_db.startTransaction( true );
			
			// Retrieve Enumeration of the DFAgentDescription in the DB, and iterate through the
			Enumeration df_descs = _df_db.getIDs();
			while( df_descs.hasMoreElements() )
			{
				// Get next DFAgentDescription
				DFAgentDescription current = (DFAgentDescription) _df_db.findObject( ( (String) df_descs.nextElement() ) );
				
				// Check that it matches our requirements and the size is within the limit
				if ( compare( df_desc, current ) && (_max_results==-1 || results.size()<_max_results) )
				{
					results.add( current );
				}
			}
			
			_df_db.endTransaction();
		}
		catch (DatabaseException dbe)
		{
			DIAGNOSTICS.println("Error with getting list of DF's", dbe, this, DIAGNOSTICS.LEVEL_MAX );
		}
		
		return results;
	}
	
	
	/**
	 * Compare the second DFAgentDescription against the first - returns true if it
	 * matches (i.e. first is sub-set of second)
	 */
	private static boolean compare( DFAgentDescription desc1, DFAgentDescription desc2 )
	{
		// Compare fields
		
		// Compare Agent name
		if( desc1.getAgentID() != null &&
		!desc1.getAgentID().equals( desc2.getAgentID() ) )
		{
			return false;
		}
		
		// Compare languages
		if( desc1.getLanguages() != null )
		{
			if( desc2.getLanguages() != null )
			{
				Set lang1 = desc1.getLanguages();
				Set lang2 = desc2.getLanguages();
				
				if( !lang2.containsAll( lang1 ) )
				{
					return false;
				}
			}
			else
			{
				return false;
			}
		}
		
		// Compare ontologies
		if( desc1.getOntologies() != null )
		{
			if( desc2.getOntologies() != null )
			{
				Set ont1 = desc1.getOntologies();
				Set ont2 = desc2.getOntologies();
				
				if( !ont2.containsAll( ont1 ) )
				{
					return false;
				}
			}
			else
			{
				return false;
			}
		}
		
		// Compare protocols
		if( desc1.getProtocols() != null )
		{
			if( desc2.getProtocols() != null )
			{
				Set prot1 = desc1.getProtocols();
				Set prot2 = desc2.getProtocols();
				
				if( !prot2.containsAll( prot1 ) )
				{
					return false;
				}
			}
			else
			{
				return false;
			}
		}
		
		// Compare protocols
		if( desc1.getAgentServices() != null )
		{
			if( desc2.getAgentServices() != null )
			{
				Set serv1 = desc1.getAgentServices();
				Set serv2 = desc2.getAgentServices();
				
				if( !compare( serv1, serv2 ) )
				{
					return false;
				}
			}
			else
			{
				return false;
			}
		}
		
		return true;
	}
	
	
	/**
	 * Compare the second list of ServiceDescriptions against the first - returns true if it
	 * matches (i.e. first is sub-set of second)
	 */
	private static boolean compare( Set serv1, Set serv2 )
	{
		// Iterate through the first list and check that all services
		// are also specified in the second list
		Iterator i = serv1.iterator();
		boolean found_match = true;
		while( i.hasNext() && found_match )
		{
			ServiceDescription service = (ServiceDescription) i.next();
			
			found_match = false;
			Iterator j = serv2.iterator();
			while( j.hasNext() && !found_match)
			{
				ServiceDescription current = (ServiceDescription) j.next();
				found_match = compare( service, current );
			}
		}
		
		return found_match;
	}
	
	
	/**
	 * Compare the second ServiceDescription against the first - returns true if it
	 * matches (i.e. first is sub-set of second)
	 */
	private static boolean compare( ServiceDescription serv1, ServiceDescription serv2 )
	{
		// Compare service names
		if( serv1.getServiceName() != null )
		{
			if( serv2.getServiceName() != null )
			{
				if( !serv1.getServiceName().equals( serv2.getServiceName() ) )
				{
					return false;
				}
			}
			else
			{
				return false;
			}
		}
		
		
		// Compare service types
		if( serv1.getServiceType() != null )
		{
			if( serv2.getServiceType() != null )
			{
				if( !serv1.getServiceType().equals( serv2.getServiceType() ) )
				{
					return false;
				}
			}
			else
			{
				return false;
			}
		}
		
		
		// Compare ontologies
		if( serv1.getOntologies() != null )
		{
			if( serv2.getOntologies() != null )
			{
				Set ont1 = serv1.getOntologies();
				Set ont2 = serv2.getOntologies();
				
				if( !ont2.containsAll( ont1 ) )
				{
					return false;
				}
			}
			else
			{
				return false;
			}
		}
		
		
		// Compare ownership
		if( serv1.getOwnership() != null )
		{
			if( serv2.getOwnership() != null )
			{
				if( !serv1.getOwnership().equals( serv2.getOwnership() ) )
				{
					return false;
				}
			}
			else
			{
				return false;
			}
		}
		
		
		// Compare properties
		if( serv1.getProperties() != null )
		{
			if( serv2.getProperties() != null )
			{
				Set prop1 = serv1.getProperties();
				Set prop2 = serv2.getProperties();
				
				if( !prop2.containsAll( prop1 ) )
				{
					return false;
				}
			}
			else
			{
				return false;
			}
		}
		
		
		// Compare protocols
		if( serv1.getProtocols() != null )
		{
			if( serv2.getProtocols() != null )
			{
				Set prot1 = serv1.getProtocols();
				Set prot2 = serv2.getProtocols();
				
				if( !prot2.containsAll( prot1 ) )
				{
					return false;
				}
			}
			else
			{
				return false;
			}
		}
		
		return true;
	}
	
	//----------------------------------------------------------------------
	// MAIN/TESTING METHOD
	//----------------------------------------------------------------------
	
	
} // End class
