//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Emorphia are Copyright (C) 2001 Emorphia Limited
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.platform.df;

import fipaos.agent.task.*;
import fipaos.agent.conversation.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.ont.proprietary.subscription.*;
import fipaos.util.*;

/** This Task provides a mechanism for informing subscribed Agents about the something
 * that has changed within the DF - DFListenerTask is the counterpart to this Task.
 *
 * <p>Currently this Task (and its counterpart) only support intra-platform communication
 * over a transport which maintains message content as objects (i.e. RMI).
 */
public class SubscriptionHandlerTask extends Task 
{
	/** Agent that is monitoring the DF
	 */
	private AgentID _agent;
	
	/** Conversation ID associated with our subscription messages
	 */
	private String _conv_id;
	
	/**
	 * The original subscription object submitted by the Agent this Task
	 * is interacting with
	 */
	private SubscriptionAction _subscription;
	
	/**
	 * Indicates if the Agent that this task is dealing with is still
	 * subscribed
	 */
	private boolean _still_subscribed = true;

	/** Creates a SubscribeAction - once the Task has started, it will attempt 
	 * to "agree" and "inform" the subscribing Agent that the subscription has
	 * been started OK
	 * @param conv Conversation associated with this subscription
	 */
	public SubscriptionHandlerTask( Conversation conv ) 
	{
		// Save conversation id
		_conv_id = conv.getConversationID ();
		
		// Get AID of other Agent
		_agent = conv.getSenderAID( 0 );
		
		// Get subscription request
		ACL msg = conv.getACL( 0 );
		if( msg.getContentObject() != null && msg.getContentObject() instanceof SubscriptionAction )
		{
			_subscription = (SubscriptionAction) msg.getContentObject();
		}		
	}
	
	/** Invoked when the Task starts - sends subscription message to target DF
	 */
	public void startTask()
	{
		// Send agree and inform messages in response
		ACL msg = new ACL();
		msg.setPerformative( FIPACONSTANTS.INFORM );
		msg.setSenderAID( _owner.getAID() );
		msg.setReceiverAID( _agent );
		msg.setOntology( FIPAMANCONSTANTS.FIPA_AGENT_MANAGEMENT );
		msg.setConversationID( _conv_id );
		msg.setContentObject( _subscription );

		forward( msg );
	}
	
	/** Invoked when agree messages are received - generally indicates that DF will allow
	 * us to subscribe
	 * @param conv The subscription conversation
 */
	public void handleCancel( Conversation conv )
	{
		DIAGNOSTICS.println( "Subscription from " + _agent.getName() + " cancelled", this, DIAGNOSTICS.LEVEL_3 );
		
		// Indicate that this Task is complete, and that we don't want to deal with
		// updates any more
		_still_subscribed = false;
		done();
	}
	
	/**
	 * Deal with not-understood messages here
	 */
	public void handleNotUnderstood( Conversation conv )
	{
		// End this task
		DIAGNOSTICS.println( "Received a not-understood - terminating", this, DIAGNOSTICS.LEVEL_2 );
		_still_subscribed = false;
		done();
	}
	
	/**
	 * Invoked by the DF when the DF has something to propogate to its subscribed
	 * Agents
	 */
	void DFUpdate( AgentManagementAction ama )
	{
		// Check that this Task is still sending updates to the subscribed Agent
		if( _still_subscribed )
		{
			// Send a message containing the update
			ACL msg = new ACL();
			msg.setPerformative( FIPACONSTANTS.INFORM );
			msg.setSenderAID( _owner.getAID() );
			msg.setReceiverAID( _agent );
			msg.setOntology( FIPAMANCONSTANTS.FIPA_AGENT_MANAGEMENT );
			msg.setConversationID( _conv_id );
			msg.setContentObject( ama );

			forward( msg );			
		}
	}
}
