//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Emorphia are Copyright (C) 2001 Emorphia Limited
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.platform.dfgui;

import fipaos.agent.*;
import fipaos.agent.task.*;
import fipaos.platform.ams.*;
import fipaos.platform.df.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.util.*;
import java.util.*;

/** This a basic shell for DFGUI implementations to extend - all Agent based interactions
 * are handled by this class, enabling developers to concentrate on the UI to present
 * this functionality/information
 */
public abstract class DFGUIAgent extends FIPAOSAgent
{
	private static final int DF_TIMEOUT = 30000;
	
	private Map _df_listener_tasks;
	private Map _df_contents;
	private DFGUIAgent _parent = this;
	
/** Generic constructor - see <CODE>FIPAOSAgent</CODE>
 * @param name Name of the Agent
 * @param platform Location of the platform profile
 * @param owner Owner of this Agent
 */	
	public DFGUIAgent( String name, String platform, String owner)
	{
		super( name, platform, owner );
		
		_df_listener_tasks = new TreeMap( Comparators.STRING_COMPARE );
		_df_contents = new TreeMap( Comparators.STRING_COMPARE );
		
		startPushing();
		
		try
		{
			// Register with AMS
			registerWithAMS();
		}
		catch (AMSRegistrationException amsre)
		{
			String reason = amsre.getExceptionReason();
			
			if (reason != null && reason.equals(FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED))
			{
				DIAGNOSTICS.println( "Error registering DF: " + reason, this, DIAGNOSTICS.LEVEL_MAX );	
			}
			else
			{
				DIAGNOSTICS.println( "Error registering DF: " + reason, this, DIAGNOSTICS.LEVEL_MAX );
				return;				
			}
		}
	}
	
/** Invoked when an Agent has been registered with a DF that has previously
 * been searched
 * @param df DF associated with this event
 * @param df_desc DF description of the entry that has been updated
 */	
	public abstract void registered( AgentID df, DFAgentDescription df_desc );
	
/** Invoked when an Agent has been deregistered with a DF that has previously
 * been searched
 * @param df DF associated with this event
 * @param df_desc DF description of the entry that has been updated
 */	
	public abstract void deregistered( AgentID df, DFAgentDescription df_desc);
	
/** Invoked when an Agent has been modified with a DF that has previously
 * been searched
 * @param df DF associated with this event
 * @param df_desc DF description of the entry that has been updated
 */	
	public abstract void modified( AgentID df, DFAgentDescription df_desc);
	
/** Shutdown the DFGUI Agent "nicely" (attempts to cancel subscriptions to DF's
 * before killing the Agent).
 */	
	public synchronized void shutdown()
	{
		// Iterate though DF's subscribed to, and remove them
		Set s = new TreeSet( fipaos.util.Comparators.STRING_COMPARE );
		s.addAll( _df_listener_tasks.keySet() );
		Iterator i = s.iterator();
		
		AgentID aid = new AgentID();
		while( i.hasNext() )
		{
			aid.setName( (String) i.next() );
			
			DIAGNOSTICS.println( "Attempting to unsubscribe from " + aid.getName(), DIAGNOSTICS.LEVEL_3 );
			
			removeDF( aid );
		}
		
		try
		{
			// Ensure that there is enough time for cancel messages to be sent
			Thread.sleep( 5000 );
		}
		catch( Throwable t )
		{	
		}
		
		// Check if we're registered with the AMS
		if (registeredWithAMS() == true)
		{
			try
			{
				// Deregister with AMS
				deregisterWithAMS();
				DIAGNOSTICS.println("Deregistered with AMS", this, DIAGNOSTICS.LEVEL_MAX );
			}
			catch (AMSRegistrationException amsre)
			{
				// A problem deregistering occured..... we can obtain the reason though!
				String reason = amsre.getExceptionReason();
				
				DIAGNOSTICS.println(amsre + ":" + reason, this, DIAGNOSTICS.LEVEL_MAX );
			}
		}
		
		// Shutdown Agent
		super.shutdown();
	}
	
/** Lists the Agents registered with a particular DF.  This automatically causes the
 * DFGUIAgent to subscribe to updates from this DF, if not already doing so
 * @param df DF to search
 * @return A Map of Agent name -> DFAgentDescription
 */	
	protected Map listAgents( AgentID df )
	{
		Map df_contents = null;
		DFListenerTask dflt	= null;

		synchronized( this )
		{
			// Check if DF is currently subscribed to
			df_contents = (Map) _df_contents.get( df.getName() );
			dflt = (DFListenerTask) _df_listener_tasks.get( df.getName() );
		}
		
		if( df_contents == null && dflt == null )
		{
			// Attempt to add DF
			if( addDF( df ) )
			{
				// DF subscribed to!
				df_contents = (Map) _df_contents.get( df.getName() );
			}
		}
		else if( dflt != null && !dflt.subscribedOK() )
		{
			// Attempt to manually update list
			df_contents = getDFEntries( df, 1 );
		}
		
		return df_contents;
	}
	
	
	/**
	 * Send a federated search to the given DF, and retrieve the results
	 */
	public synchronized Map federatedSearch( AgentID df, int depth )
	{
		return getDFEntries( df, depth );
	}
	
	
/** Indicates that the DFGUI should stop listeneing to the DF indicated (i.e.
 * this cancels the subscription with the DF)
 * @param df The DF to cancel subscriptions with
 * @return True if the subscription was cancelled OK
 */	
	protected synchronized boolean stopListeningToDF( AgentID df )
	{
		return removeDF( df );
	}
	
/** Registers the DFAgentDescription with the given DF Agent
 * @param df The DF to register the DFAgentDescription with
 * @param dfad The Agent's details to register
 * @return True if registration was succesfull
 */	
	protected synchronized boolean registerAgent( AgentID df, DFAgentDescription dfad )
	{
		DIAGNOSTICS.println( "registerAgent() yet to be implemented", this, DIAGNOSTICS.LEVEL_MAX );
		
		return false;
	}

/** Deregisters the DFAgentDescription with the given DF Agent
 * @param df The DF to deregister the DFAgentDescription from
 * @param dfad The Agent's details to deregister
 * @return True if deregistration was succesfull
 */	
	protected synchronized boolean deregisterAgent( AgentID df, DFAgentDescription dfad )
	{
		DIAGNOSTICS.println( "deregisterAgent() yet to be implemented", this, DIAGNOSTICS.LEVEL_MAX );		
		
		return false;
	}
	
/** Modifies the DFAgentDescription within the given DF Agent
 * @param df The DF to modify the DFAgentDescription with
 * @param dfad The Agent's details to modify
 * @return True if modification was succesfull
 */	
	protected synchronized boolean modifyAgent( AgentID df, DFAgentDescription dfad )
	{
		DIAGNOSTICS.println( "modifyAgent() yet to be implemented", this, DIAGNOSTICS.LEVEL_MAX );		
		
		return false;
	}
	
	private boolean addDF( AgentID df )
	{
		// Create new DF entry, if it doesn't already exist
		if( !_df_listener_tasks.containsKey( df.getName() ) )
		{
			// Get DF contents
			Map entries = getDFEntries( df, 1 );
			if( entries == null )
			{
				return false;
			}
			
			synchronized( this )
			{
				// Add entries Map to our DF map
				_df_contents.put( df.getName(), entries );

				// Create DFListenerTask to subscribe to updates from the DF
				DFListenerTask dflt = new DFListenerTask( new ParticularDFListener( df ), df );
				_df_listener_tasks.put( df.getName(), dflt );
				_tm.newTask( dflt );			
			}
		}
		
		return true;
	}
	
	private synchronized boolean removeDF( AgentID df )
	{
		// Check if we are actually monitoring the DF
		if( _df_contents.containsKey( df.getName() ) )
		{
			// Get DFListenerTask, and stop subscription
			DFListenerTask dflt = (DFListenerTask) _df_listener_tasks.remove( df.getName() );
			dflt.stopListening();
			
			// Get contents of DF, and purge
			Map dfads = (Map) _df_contents.remove( df.getName() );
			dfads.clear();
		}
		
		return true;
	}
	
	private Map getDFEntries( AgentID df, int depth )
	{
		// Get a list of the entries in the DF first
		Object search_result = SynchronousTask.executeTask( _tm, 
			new DFSearchTask( df.toString(), new DFAgentDescription(), 
												new SearchConstraints( depth ), depth > 1 ), DF_TIMEOUT );

		// Check if there was a problem with the result
		if( search_result instanceof SynchronousTask.UnusualResult )
		{
			return null;
		}

		// Extract the search results & add them to the list
		Map entries = new TreeMap( Comparators.STRING_COMPARE );

		if( search_result != null && search_result instanceof DFAgentDescription[] )
		{
			// Entries in the DF - index them!
			DFAgentDescription dfad[] = (DFAgentDescription[]) search_result;

			DIAGNOSTICS.println( "Found " + dfad.length + " DF entries", this, DIAGNOSTICS.LEVEL_2 );

			// Iterate through entries and add them to the Map of Agents
			for( int i = 0; i < dfad.length; i++ )
			{
				// Get name, and map it against the DFAgentDescription object
				entries.put( dfad[ i ].getAgentID().getName(), dfad[ i ] );
			}
		}
		else
		{
			DIAGNOSTICS.println( "Didn't find DF entries", this, DIAGNOSTICS.LEVEL_2 );
		}
		
		return entries;
	}
	
	private class ParticularDFListener implements DFListener
	{
		/**
		 * AgentID of the DF this is listening to
		 */
		private AgentID _df;
		
		/** Default constructor
		 * @param df DF this is receiving events for
		 */		
		public ParticularDFListener( AgentID df )
		{
			_df = df;
		}
		
			/** Invoked when an Agent has succesfully registered with the DF
		 * @param dfad DFAgentDescription associated with this event
		 */
		public void registered( DFAgentDescription dfad )
		{
			// Add DFAD into DF's list of Agents
			try
			{
				synchronized( _df_contents )
				{
					Map df_contents = (Map) _df_contents.get( _df.getName() );
					df_contents.put( dfad.getAgentID().getName(), dfad );
				}
			}
			catch( Throwable t )
			{
				DIAGNOSTICS.println( "Error when updating DF's list of Agents", t, this, DIAGNOSTICS.LEVEL_3 );
			}
			
			try
			{
				// Indicate to sub-classes that an event has occured
				_parent.registered( _df, dfad );
			}
			catch( Throwable t )
			{
				DIAGNOSTICS.println( "An exception occured whilst propogating DF event to DFGUIAgent instance", t, DIAGNOSTICS.LEVEL_MAX );
			}
		}

		/** Invoked when an Agent has successfully updated its DFAgentDescription with the DF
		 * @param dfad DFAgentDescription associated with this event
		 */
		public void modified( DFAgentDescription dfad )
		{
			// Add DFAD into DF's list of Agents
			try
			{
				synchronized( _df_contents )
				{
					Map df_contents = (Map) _df_contents.get( _df.getName() );
					df_contents.put( dfad.getAgentID().getName(), dfad );
				}
			}
			catch( Throwable t )
			{
				DIAGNOSTICS.println( "Error when updating DF's list of Agents", t, this, DIAGNOSTICS.LEVEL_3 );
			}
			
			try
			{
				// Indicate to sub-classes that an event has occured
				_parent.modified( _df, dfad );
			}
			catch( Throwable t )
			{
				DIAGNOSTICS.println( "An exception occured whilst propogating DF event to DFGUIAgent instance", t, DIAGNOSTICS.LEVEL_MAX );
			}
		}

		/** Invoked when an Agent has successfully deregistered from the DF
		 * @param dfad DFAgentDescription associated with this event
		 */
		public void deregistered( DFAgentDescription dfad )
		{
			// Add DFAD into DF's list of Agents
			try
			{
				synchronized( _df_contents )
				{
					Map df_contents = (Map) _df_contents.get( _df.getName() );
					df_contents.remove( dfad.getAgentID().getName() );
				}
			}
			catch( Throwable t )
			{
				DIAGNOSTICS.println( "Error when updating DF's list of Agents", t, this, DIAGNOSTICS.LEVEL_3 );
			}
			
			try
			{
				// Indicate to sub-classes that an event has occured
				_parent.deregistered( _df, dfad );
			}
			catch( Throwable t )
			{
				DIAGNOSTICS.println( "An exception occured whilst propogating DF event to DFGUIAgent instance", t, DIAGNOSTICS.LEVEL_MAX );
			}
		}
	}
}
