//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Emorphia are Copyright (C) 2001 Emorphia Limited
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.platform.dfgui.swing;

import fipaos.platform.dfgui.DFGUIAgent;
import fipaos.ont.fipa.fipaman.*;
import fipaos.ont.fipa.*;
import fipaos.agent.task.*;
import fipaos.agent.conversation.*;
import fipaos.util.DIAGNOSTICS;

// Import registration exception classes
import fipaos.platform.ams.AMSRegistrationException;
import fipaos.platform.df.DFRegistrationException;

import java.util.*;

/**
 * This is the DF GUI agent that uses Java Swing GUI to display controls
 * for DF(s).
 */
public class SwingDFGUIAgent extends DFGUIAgent
{
	
	private DFGUIFrame _gui;
	
		/**
		 * Creates new SwingDFGUI.
		 * Instansiates a DFGUIFrame that will start up the actual GUI.
		 * @param platform_profile location of the platform profile
		 * @param name name of the agent
		 * @param owner owner of the agent
		 */
	public SwingDFGUIAgent(String platform_profile, String name, String owner)
	{
		super(platform_profile, name, owner);
		
		AgentID df = null;
		try
		{
			df = new AgentID(getLocalDF());
		}
		catch (Exception e)
		{
		}
		
		_gui = new DFGUIFrame(this, df);
		fipaos.util.DIAGNOSTICS.println("GUI started", fipaos.util.DIAGNOSTICS.LEVEL_4);
	}
	
	/**
	 * Method to cleanly kill off Agent
	 */
	public void shutdown()
	{
		// Close down GUI
		if( _gui != null )
		{
			_gui.dispose();
			_gui = null;
		}
		
		super.shutdown();
	}
	
	/**
	 * Invoked when GUI should be shown
	 */
	public void activate()
	{
		if( _gui != null )
		{
			_gui.setVisible( true );
		}
	}
		
/**
 * Registers an agent with a DF. Not implemented, and never will be. This
 * DF GUI doesn't allow user to register agents, since all the agents that are
 * registered with a DF monitored will always show in the GUI. Therefore there
 * is no need to be able to register "ghost" agents.
 *
 * @param agent name of the agent to be registed
 * @return if the operation was successful
 */
	boolean registerRequest(String agent)
	{
		return false;
	}
	
/**
 * Modifies a certain agent's df-agent-description. Not yet implemented.
 *
 * @param agent name of the agent to be modified
 * @return if the operation was successful
 */
	boolean modifyRequest(String agent)
	{
		return false;
	}
	
/**
 * Deregisters a certain agent from a DF. Not yet implemented.
 *
 * @param agent name of the agent to be deregistered
 * @return if the operation was successful
 */
	boolean deregisterRequest(String agent)
	{
		return false;
	}
	
/**
 * Lists all agents on a certain DF.
 *
 * @param df AID for the DF
 * @return all the agents registered
 */
	Map listAgentsRequest(AgentID df)
	{
		return listAgents(df);
	}
	
/**
 * All agents registered on a certain DF.
 * @param df name of the DF
 * @return all the agents registered
 */
	Map listAgentsRequest(String df)
	{
		return listAgents(getAIDfromName(df));
	}
	
/**
 * Notification from the super class when agent has been registered.
 *
 * @param aid AID of the DF this happens with
 * @param description the description of the agent registered
 */
	public void registered(AgentID aid,DFAgentDescription description)
	{
		_gui.notifyMessage("Agent registered");
		_gui.updateTree();
	}
	
/**
 * Notification from the super class when agent has deregistered.
 * @param aid AID of the DF this agent has deregisted from
 * @param description description of the agent deregistered
 */
	public void deregistered(AgentID aid,DFAgentDescription description)
	{
		_gui.notifyMessage("Agent deregistered");
		_gui.agentDeregistered(aid, description.getAgentID());
	}
	
/**
 * Notification from the super class when agent has been modified.
 *
 * @param aid AID of the DF the agent has been modified on
 * @param description the description of the modified agent
 */
	public void modified(AgentID aid,DFAgentDescription description)
	{
		//no other action is required, modified information
		//is now known to GUI
		_gui.notifyMessage("Agent modified");
	}
	
/**
 * Method envoced by the GUI that a particular DF should no longer be monitored.
 *
 * @param df_name name of the DF to be dropped
 */
	void dropDF(String df_name)
	{
		super.stopListeningToDF(getAIDfromName(df_name));
	}
	
/**
 * Returns an AID from a name of an agent.
 * @param name name of an agent
 * @return AID for the agent
 */
	private AgentID getAIDfromName(String name)
	{
		try
		{
			AgentID aid = new AgentID();
			aid.setName(name);
			return aid;
		}
		catch (Throwable t)
		{
			DIAGNOSTICS.println("No name for the agent: " + t, DIAGNOSTICS.LEVEL_MAX);
			return null;
		}
	}
	
/**
 * Request for the GUI that DF should be registered with another DF. This method
 * is temporary, since DF GUI agent will actually pretend to be the DF requesting
 * the registration - more permanent solution in the next release.
 * @param new_df DF the other DF will be registered on
 * @param old_df DF requesting registration
 * @return if the operation was successful
 */
	boolean registerDFRequest(AgentID new_df,String old_df)
	{
		AgentID old_df_aid = new AgentID();
		ACL acl = new ACL();
		DFAgentDescription desc = new DFAgentDescription();
		ServiceDescription[] service_desc =
		{new ServiceDescription()};
		AgentManagementAction ama = new AgentManagementAction();
		
		try
		{
			old_df_aid.setName(old_df);
			service_desc[0].setServiceName("df");
			service_desc[0].setServiceType(FIPAMANCONSTANTS.FIPA_DF);
			service_desc[0].setOntologies(new String[]
			{FIPAMANCONSTANTS.FIPA_AGENT_MANAGEMENT});
			desc.setAgentName(old_df_aid);
			desc.setAgentServices(service_desc);
			ama.setAction(FIPAMANCONSTANTS.REGISTER);
			ama.setDFAgentDescription(desc);
			ama.setAgentID(old_df_aid);
			acl.setSenderAID(old_df_aid);
			acl.setReceiverAID(new_df);
			acl.setOntology(FIPAMANCONSTANTS.FIPA_AGENT_MANAGEMENT);
			acl.setProtocol(FIPACONSTANTS.FIPA_REQUEST);
			acl.setContentObject(ama);
			acl.setPerformative(FIPACONSTANTS.REQUEST);
			
			DIAGNOSTICS.println("Sending: " + acl.toString(), DIAGNOSTICS.LEVEL_MIN);
			forward(acl);
			
			return true;
		}
		catch (Throwable t)
		{
			_gui.notifyError(t.toString());
			return false;
		}
	}
	
}