//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files ("the
//     Programs") are Copyright (c) 1999 - 2000 Nortel Networks Corporation, 
//     8200 Dixie Road, Suite 100, Brampton, Ontario, Canada L6R 5P6.  
//     All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out in
//     the Nortel Networks FIPA-OS Public License, a copy of which can be found
//     in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version can
//     also be found at http://fipa-os.sourceforge.net/. By downloading the
//     files containing the Programs you accept the terms and  conditions of the
//     Public License. You do not have to accept these terms and conditions, but
//     unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.skill.constraint.ccl.constraint;

//----------------------------------------------------------------------

import fipaos.ont.fipa.FIPACONSTANTS;
import fipaos.ont.BASETOKENS;
import fipaos.ont.fipa.fipaman.FIPAMANCONSTANTS;

import java.util.*;

import fipaos.skill.constraint.ccl.variable.IndexPair;
import fipaos.skill.constraint.ccl.variable.List;
import fipaos.skill.constraint.ccl.variable.Tuple;

import fipaos.agent.conversation.content.Content;

//----------------------------------------------------------------------

/**
 * 
 * <br> This class represents the object "CSP-relation" as described in the document
 * <br> "CCL Language Specification v2.01" section 4.2.3.2, p14.
 * <br>
 * <br> This object represents a relation between two variables. Any varibles named
 * <br> in the Relation-body must appear in the set of Varibles of the relation.
 * <br> The set of index-pairs identifies which slots in a tuple valued variable are
 * <br> covered by the relation. For example, for an equality relation between two 
 * <br> variables with 3 tuples as values (e.g. (x, y, z)), setting the set of indices
 * <br> to {(2,2), (3,3)} indicates that only the 2nd and 3rd slot of the value tuples
 * <br> need ever be equal- the constraint is not violated even if the values in the
 * <br> 1st slots are unequal.
 *
 */
 
public class CSPRelation
{
     
//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------

   // Relation types as described on p15 of the CCL specification.
   
   /**
   * <br> This specifies that all the variables listed in the Variables list of the 
   * <br> relevant CSP-constraint object must take equal values in any instantiation.
   */
   public final static String EQUALITY = "Intensional-equality";
   
   /**
   * <br> This specifies that all the variables list in the Variables list of the 
   * <br> relevant CSP-constraint object must take strictly different values in any
   * <br> instantiation.
   */
   public final static String INEQUALITY = "Intensional-inequality";
   
   /**
   * <br> This specifies that the variables in the Variables list of the relevant 
   * <br> CSP-constraint object are related by a "greater than" relationship ">" such
   * <br> that the order of the tuple defines the order in the relationship- the first
   * <br> variable in the list is strictly greater than the second, which is strictly 
   * <br> greater than the third etc.
   * <br>
   * <br> This relation is only valid for variable types which have an ordering relation
   * <br> defined in the domain ontology.
   */
   public final static String GREATER_THAN = "Intensional-greaterThan";
   
   /**
   * <br> This specifies that the variables in the Variables list of the relevant
   * <br> CSP-constraint object are related by a "less than" relationship "<" such that
   * <br> that order of the tuple defines the order in the relationship- the first variable
   * <br> in the list is strictly less than the second, which is strictly less than the 
   * <br> third, etc. 
   * <br> 
   * <br> This relation is only valid for variable types which have an ordering relation
   * <br> defined in the domain ontology.
   */
   public final static String LESS_THAN = "Intensional-lessThan";
   
   /**
   * <br> Similar to greaterThan described above but using a "greater than or equals" relation.
   */
   public final static String GREATER_THAN_EQUAL = "Intensional-greaterThanEqual";
   
   /**
   * <br> Similar to lessThan described above but using a "less than or equals" relation.
   */
   public final static String LESS_THAN_EQUAL = "Intensional-lessThanEqual";
   
   /**
   * <br> This specifies that there are no allowed combinations of values for these variables.
   */
   public final static String EMPTY = "Intensional-Empty";
   
   
   /**
   *  String that a CSPRelation will be referenced with when represented as a Content object.
   */
   public final static String CSPRELATION = "CSP-relation";
   
   /**
   * <br> String that a Set of Indices will be referenced with when represented as an Attribute
   * <br> of a Content object.
   */
   public final static String INDICES = "Indices";
   
   /**
   * <br> String that the relation type is referenced with when represented as an Attribute of
   * <br> of a Content object.
   */
   public final static String RELATION_TYPE = "Relation-type";
   
   /**
   * <br> String that a collection of tags is referenced with when represented as an Attribute 
   * <br> of a Content object.
   */
   public final static String TAGS = "Tags";
   
   /**
   *  The only Attribute stored in a Content object representing "Tags" information.
   */
   public final static String NAME = "Name";
   
   /**
   * <br> String that a set of variables are referenced with when represented as an Attribute
   * <br> of a Content object.
   */
   public final static String VARIABLES = "Variables";
   

//----------------------------------------------------------------------
// STATIC VARIABLES
//----------------------------------------------------------------------
		
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

   /**
   * Name of first CSP-variable in the relation.
   */
   private String _variable_name_one = "";
   
   /**
   * Name of second CSP-variable in the relation.
   */
   private String _variable_name_two = "";
   
   /**
   * The relation type between these two variables
   */
   private String _relation_type = "";
   
   /**
   * <br> Indices.
   * <br> Specifies what sub-fields of variable values the relation refers to.
   * <br> This is a Vector of IndexPair objects.
   */
   private Vector _indices = null;
   
   /**
   * Contains a list of symbols that allow selective constraints.
   */
   private Vector _tags = null;
   

//----------------------------------------------------------------------
// CONSTRUCTOR
//----------------------------------------------------------------------

   /**
   *  Constructs an empty CSPRelation object.
   */
   public CSPRelation()
	 {
	    _indices = new Vector();
	    _tags = new Vector();		
	 }
	 
	 /**
	 * <br> Constructs a new CSPRelation object and populates it with the data
	 * <br> passed in the content object
	 */
	 public CSPRelation(Content content)
	 {
	    _indices = new Vector();
	    _tags = new Vector();
	    
	    assignContentData(content);
	 }
	  

//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------

   /**
   * <br> Takes the passed Content object and copies the values it contains into
   * <br> the CSPRelation object.
   */
   private void assignContentData(Content content)
	 {
	 	  //Retrieve the set of Attribute names
	 	  Set content_attributes = content.getAttributeNames();
	 	  
	 	  // Retrieve Values for Indices.
	 	  if(content_attributes.contains(INDICES))
	 	  {
	 	     String indices = (String)content.getAttribute(INDICES);
	 	     IndexPair pair = new IndexPair(indices);
	 	     addIndexPair(pair);      
	 	  } 
	 	  
	 	  // Retrieve the relation type.
	 	  if(content_attributes.contains(RELATION_TYPE))
	 	  {
	 	     setRelationType((String)content.getAttribute(RELATION_TYPE));   
	 	  }
	 	  
	 	  // Retrieve the Variables stored.
	 	  if(content_attributes.contains(VARIABLES))
	 	  {
	 	     String var = (String) content.getAttribute(VARIABLES);
	 	     
	 	     // convert String to a Vector of it components
	 	     Vector v = List.parseList(var);
	 	     
	 	     // set the name of the two Variables present
	 	     setVariableNameOne((String)v.firstElement());
	 	     setVariableNameTwo((String)v.lastElement());
	 	     	
	 	  } // if(content_attributes.contains(VARIABLES))
	 	  
	 	  if(content_attributes.contains(TAGS))
	 	  {
	 	     Content tags = (Content) content.getAttribute(TAGS);
	 	     Set tags_attributes = tags.getAttributeNames();
	 	     
	 	     if(tags_attributes.contains(NAME))
	 	  	 {
	 	  	 	  String tag = (String)tags.getAttribute(NAME);
	 	  	 	  
	 	  	 	  // Assign the retrieved Vector of tag values to the variable "_tags".
	 	  	 	  _tags = List.parseList(tag);
	 	  	 }	
	 	     	
	 	  } // if(content_attributes.contains(TAGS))
	 	
	 } // assignContentData(Content content)

//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------

   
   /**
   *  Adds a new Tag.
   */
   public void addTag(String str)
	 {
	 	  _tags.addElement(new String(str));
	 	
	 } // addTag(String str)
	 	
	 
	 /**
	 *  Returns a Vector of the Tags stored in the CSPRelation.
	 */
	 public Vector getTags()
	 {
	    Vector value = new Vector();
	    
	    for(Enumeration e = _tags.elements(); e.hasMoreElements(); )
	 	  {
	 	     value.addElement(new String((String) e.nextElement()));
	 	  }
	 	  
	 	  return value;
	 	
	 } // getTags()
	 	
   
   /**
   *  Adds a new IndexPair.
   */
   public void addIndexPair(IndexPair indices)
	 {
	    _indices.addElement(indices.copy());
	 }
	 
	 
	 /**
	 *  Returns a Vector of the IndexPairs stored in the CSPRelation.
	 */
	 public Vector getIndexPairs()
	 {
	    Vector value = new Vector();
	    
	    for(Enumeration e = _indices.elements(); e.hasMoreElements(); )
	 	  {
	 	  	 IndexPair pair = (IndexPair) e.nextElement();
	 	     value.addElement(pair.copy());
	 	  }
	 	  
	 	  return value;	
	 	
	 } // getIndexPairs()


   /**
   *  Sets a value for the first variable this CSPRelation concerns.
   */
   public void setVariableNameOne(String str)
	 {
	    _variable_name_one = new String(str);
	 }
	 
	 
	 /**
	 *  Returns the name of the first variable this CSPRelation concerns.
	 */
	 public String getVariableNameOne()
	 {
	    return new String(_variable_name_one);
	 }
	 
	 
	 /**
	 *  Sets a value for the second variable this CSPRelation concerns.
	 */
	 public void setVariableNameTwo(String str)
	 {
	    _variable_name_two = new String(str);
	 }
	 
	 
	 /**
	 *  Returns the name of the second variable this CSPRelation concerns.
	 */
	 public String getVariableNameTwo()
	 {
	 	  return new String(_variable_name_two);
	 }	
	 
	 
	 /**
	 * <br> Sets the nature of the relationship between the two variables this 
	 * <br> CSPRelation concerns.
	 * <br> Should be set using the constants this class provides.
	 */
	 public void setRelationType(String str)
	 {
	    _relation_type = new String(str);
	 }
	 
	 
	 /**
	 * <br> Returns a String indicating the relationship between the two variables
	 * <br> this CSPRelation concerns.
	 */
	 public String getRelationType()
	 {
	 	  return new String(_relation_type);
	 }
	 
	 
	 /**
	 *  Returns a copy of the CSPRelation object.
	 */
	 public CSPRelation copy()
	 {
	    CSPRelation value = new CSPRelation();
	    
	    value.setVariableNameOne(_variable_name_one);
	    value.setVariableNameTwo(_variable_name_two);
	    value.setRelationType(_relation_type);
	    
	    for(Enumeration e = _indices.elements(); e.hasMoreElements(); )
	 	  {
	 	  	 IndexPair next = (IndexPair) e.nextElement();
	 	     value.addIndexPair(next.copy());
	 	  	
	 	  } // for
	    	
	    for(Enumeration e = _tags.elements(); e.hasMoreElements(); )
	 	  {
	 	     value.addTag(new String((String) e.nextElement()));
	 	  	
	 	  } // for
	 	
	 	  return value;
	 	  
	 } // copy()
	 
	 
	 /**
	 *  Returns a String representation of the CSPRelation object.
	 */
	 public String toString()
	 {
	 	  String value = "\n\n******************************" +
	 	  	             "\nCSPRelation:" + 
	 	  	             "\nVariable name one: " + _variable_name_one +
	 	  	             "\nVariable name two: " + _variable_name_two +
	 	  	             "\nRelation type: " + _relation_type +
	 	  	             "\nIndices: " + _indices +
	 	  	             "\nTags: " + _tags +
	 	  	             "\n\n******************************\n\n";
	 	  
	 	  return value;
	 	
	 } // toString()
	 	
	 	
	 /**
	 *  Returns a Content object representing this class populated with the Data it contains.
	 */ 
	 public Content toContent()
	 {
	    Content value = new Content(CSPRELATION);	
	    
	    /*
	       handles sets of Indices of the form (1, 1) i.e. a Tuple
	    */
	    if(_indices.size() == 1)
	 	  {
	 	     IndexPair pair = (IndexPair)_indices.firstElement(); 
	 	     String pair_string = "(" + pair.getFirstValue() + ", " + pair.getSecondValue() + ")";
	 	     value.addAttribute(INDICES, pair_string); 
	 	  }
	 	  
	 	  if(_indices.size() > 1)
	 	  {
	 	     System.out.println("The set of indices is of a size not currently handled by CSPRelation");
	 	  }
	 	  
	 	  // sets the relation type i.e. "Intensional-greaterThan"
	 	  if(!_relation_type.equals(""))
	 	  {
	 	     value.addAttribute(RELATION_TYPE, _relation_type);
	 	  }
	 	  
	 	  // Sets the name of the two variables the object is concerned with.
	 	  String var_string = "{ " + _variable_name_one + ", " + _variable_name_two + " }";
	 	  value.addAttribute(VARIABLES, var_string);
	 	  
	 	  Content tag = new Content(TAGS);
	 	  
	 	  // Sets the tags it contains.
	 	  if(_tags.size() == 1)
	 	  {
	 	     tag.addAttribute(NAME, (String) _tags.firstElement());
	 	     value.addAttribute(TAGS, tag);   	
	 	  	
	 	  } // if(_tags.size() > 0)
	 	  	
	 	  // Presents the tags in the form "{R1, R2}"
	 	  if(_tags.size() > 1)
	 	  {
	 	     String tag_str = List.toString(_tags);
	 	     tag.addAttribute(NAME, tag_str);
	 	     	value.addAttribute(TAGS, tag);
	 	  }
	 	
	 	  return value;
	 	
	 } // toContent()
	
//----------------------------------------------------------------------
// MAIN/TESTING METHOD
//----------------------------------------------------------------------

   /**
   * Used for testing.
   */
   public static void main(String args[])
	 {
	    System.out.println();
	    System.out.println("Testing CSPRelation class");
	    System.out.println("Creating new CSPRelation object");
	    
	    CSPRelation relation = new CSPRelation();
	    
	    System.out.println();
	    System.out.println("Adding variable name one");	
	    
	    relation.setVariableNameOne("house");
	    
	    System.out.println();
	    System.out.println("Variable name one: " + relation.getVariableNameOne());
	    
	    System.out.println();
	    System.out.println("Adding variable name two");	
	    
	    relation.setVariableNameTwo("insurance");
	    
	    System.out.println();
	    System.out.println("Variable name two: " + relation.getVariableNameTwo());
	    
	    System.out.println();
	    System.out.println("Adding Relation");	
	    
	    relation.setRelationType(INEQUALITY);
	    
	    System.out.println();
	    System.out.println("Relation: " + relation.getRelationType());
	    
	    System.out.println();
	    System.out.println("Adding Tags");
	    
	    relation.addTag("high");
	    relation.addTag("middle");
	    
	    System.out.println();
	    System.out.println("Tags: " + relation.getTags());
	    
	    System.out.println();
	    System.out.println("Adding indices");
	    
	    IndexPair index = new IndexPair("(2, 2)");
	    relation.addIndexPair(index);
	    index = new IndexPair("(3, 3)");
	    // relation.addIndexPair(index);
	    
	    System.out.println();
	    System.out.println("Indices: " + relation.getIndexPairs());
	    
	    System.out.println();
	    System.out.println("Values held in CSPRelation:");
	    System.out.println(relation);
	    
	    System.out.println();
	    System.out.println("Values held in copy of CSPRelation:");
	    System.out.println(relation.copy());
	    System.out.println();
	    
	    System.out.println("CSPRelation represented as a Content object");
	    System.out.println(relation.toContent());
	    
	    System.out.println();
	    System.out.println("Creating Content object with which to populate CSPRelation");
	    
	    Content content = new Content(CSPRELATION);
	    content.addAttribute(INDICES, "(2, 2)");
	    content.addAttribute(RELATION_TYPE, EQUALITY);
	    content.addAttribute(VARIABLES, "{buyer, constraint}");
	    
	    Content tag = new Content(TAGS);
	    tag.addAttribute(NAME, "R1");
	    content.addAttribute(TAGS, tag);
	 	
	 	  System.out.println();
	    System.out.println("Values held in Content object");
	    System.out.println(content);
	    System.out.println();
	    
	    relation = new CSPRelation(content);
	    
	    System.out.println();
	    System.out.println("Values held in CSPRelation:");
	    System.out.println(relation);
	    System.out.println();
	 	
	 } // main
  
	
} // CSPRelation

	
