//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files ("the
//     Programs") are Copyright (c) 1999 - 2000 Nortel Networks Corporation, 
//     8200 Dixie Road, Suite 100, Brampton, Ontario, Canada L6R 5P6.  
//     All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out in
//     the Nortel Networks FIPA-OS Public License, a copy of which can be found
//     in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version can
//     also be found at http://fipa-os.sourceforge.net/. By downloading the
//     files containing the Programs you accept the terms and  conditions of the
//     Public License. You do not have to accept these terms and conditions, but
//     unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.skill.constraint.ccl.object;

//----------------------------------------------------------------------

import fipaos.ont.fipa.FIPACONSTANTS;
import fipaos.ont.BASETOKENS;
import fipaos.ont.fipa.fipaman.FIPAMANCONSTANTS;

import java.util.*;

import fipaos.agent.conversation.content.Content;

import fipaos.skill.constraint.ccl.variable.CSPVariable;
import fipaos.skill.constraint.ccl.variable.IndexPair;
import fipaos.skill.constraint.ccl.variable.CSPValue;
import fipaos.skill.constraint.ccl.variable.CSPRange;

import fipaos.skill.constraint.ccl.constraint.CSPRelation;

//----------------------------------------------------------------------

/**
 * <br> This class represents a CSP Object as described in the 
 * <br> "CCL Language specification v2.01" section 4.1.1.1, p5.
 * <br> 
 * <br> A CSP object represents a choice problem.
 * <br> For a CSP object to be well defined, the items in the Exclusion and Relations
 * <br> slots must only refer to variables which are present in the Variables slot.
 * <br> 7/3/00
 *
 */
 
public class CSP
{
     
//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------

   /**
   *  String CSP-variable objects are always referenced with when stored in Content objects.
   */
   public final static String CSPVARIABLE = "CSP-variable";
   
   /**
   *  String CSP-relation objects are always referenced with when stored in Content objects.
   */
   public final static String CSPRELATION = "CSP-relation";
   
   /**
   *  String CSP-exclusion objects are always referenced with when stored in Content objects.
   */
   public final static String CSPEXCLUSION = "CSP-exclusion";
   
   /**
   *  String used to reference the variable containing the name of the CSP.
   */
   public final static String CSPREF = "CSP-ref";
   
   /**
   *  String used to reference a CSP object when represented as a Content object.
   */
   public final static String CSP = "CSP";
   
   
   // Used to facilitate the use of switch statements to determine the type
   // of object a Content object represents.
   
   public final static int CONTENT = 1;
   public final static int LIST = 2;
   public final static int LONG = 3;
   public final static int STRING = 4;
   

//----------------------------------------------------------------------
// STATIC VARIABLES
//----------------------------------------------------------------------
		
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

   /**
   * <br> The reference to this CSP object.
   * <br> This represents the information held in a CSP-identifier object.
   * <br> This is the unique name of the CSP.
   */
   private String _identifier = "";

   /**
   * <br> This Vector represents the choices which need to be taken in the choice 
   * <br> problem. The variables listed in this slot must all have unique names.
   * <br> 
   * <br> This Vector contains a set of CSPVariable objects.
   */
   private Vector _variables = null;
   
   /**
   * <br> This Vector represents the relationships between the choices to be made.
   * <br>
   * <br> This Vector contains a set of CSPRelation objects.
   */
   private Vector _relations = null;
   
   /**
   * <br> This Vector represents a list of unary relations on single variables
   * <br> which exclude certain values from variable domains.
   * <br>
   * <br> This Vector contains a set of CSPExclusion objects.
   */
   private Vector _exclusions = null;

//----------------------------------------------------------------------
// CONSTRUCTOR
//----------------------------------------------------------------------

   /**
   * <br> Constructs an empty CSP object.
   * <br> This will then be ready to be populated so it represents a 
   * <br> Constraint Satisfaction Problem.
   */
   public CSP()
	 {
	    _variables = new Vector();
	    _relations = new Vector();
	    _exclusions = new Vector();
	 }
	 
	 
	 /**
	 * <br> Constructs a new CSP object.
	 * <br> This will then be populated using the Data passed in the Content object.
	 */
	 public CSP(Content content)
	 {
	    _variables = new Vector();
	    _relations = new Vector();
	    _exclusions = new Vector();
	    
	    assignContentData(content);
	 }

//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------

   /**
   * Populates the CSP object using the Data held in the passed Content object.
   */
   private void assignContentData(Content content)
	 {
	    Set attribute_names = content.getAttributeNames();
	    
	    /* If more than one CSPVariable is present will return a list.
	       If only one CSPVariable was present will return a Content object.
	    */
	    if(attribute_names.contains(CSPVARIABLE))	
	 	  {
	 	  	 Object obj = content.getAttribute(CSPVARIABLE);
	 	  	
	 	  	 // switch on the type of object "obj" is
	 	     switch(type(obj))	 
	 	  	 {
	 	  	 	  // Object is a Content object representing a single CSP-variable
	 	  	    case CONTENT:
	 	  	 	  {
	 	  	 	  	 // Cast as a Content object and add into the Vector of CSP-variables
	 	  	 	  	 Content variable = (Content)obj;
	 	  	 	  	 _variables.addElement(new CSPVariable(variable));
	 	  	       break;
	 	  	 	  }
	 	  	    	
	 	  	    // Object is a List of Content object each representing a CSP-variable
	 	  	    case LIST:
	 	  	 	  {
	 	  	 	  	 List list = (List)obj;
	 	  	 	  	 Content next_content = null;
	 	  	 	  	 
	 	  	 	  	 for(Iterator i = list.iterator(); i.hasNext(); )
	 	  	 	  	 {
	 	  	 	  	    next_content = (Content)i.next();
	 	  	 	  	    _variables.addElement(new CSPVariable(next_content));  	
	 	  	 	  	 	
	 	  	 	  	 } // for
	 	  	 	    	
	 	  	       break;
	 	  	       
	 	  	 	  } // case LIST:
	 	  	    	
	 	  	    default:
	 	  	 	  {
	 	  	 	  	 System.out.println("Error: Unrecognised object type");
	 	  	 	     break;
	 	  	 	  }
	 	  	 	
	 	  	 } // switch(type(obj))
	 	  	
	 	  } // if(attribute_names.contains(CSPVARIABLE))
	 	  	
	 	  	
	 	  /* If more than one CSPRelation is present will return a list.
	       If only one CSPRelation was present will return a Content object.
	    */
	    if(attribute_names.contains(CSPRELATION))	
	 	  {
	 	  	 Object obj = content.getAttribute(CSPRELATION);
	 	  	
	 	  	 // switch on the type of object "obj" is
	 	     switch(type(obj))	 
	 	  	 {
	 	  	 	  // Object is a Content object representing a single CSP-relation
	 	  	    case CONTENT:
	 	  	 	  {
	 	  	 	  	 // Cast as a Content object and add into the Vector of CSP-relation
	 	  	 	  	 Content relation = (Content)obj;
	 	  	 	  	 _relations.addElement(new CSPRelation(relation));
	 	  	       break;
	 	  	 	  }
	 	  	    	
	 	  	    // Object is a List of Content object each representing a CSP-relation
	 	  	    case LIST:
	 	  	 	  {
	 	  	 	  	 List list = (List)obj;
	 	  	 	  	 Content next_content = null;
	 	  	 	  	 
	 	  	 	  	 for(Iterator i = list.iterator(); i.hasNext(); )
	 	  	 	  	 {
	 	  	 	  	    next_content = (Content)i.next();
	 	  	 	  	    _relations.addElement(new CSPRelation(next_content));  	
	 	  	 	  	 	
	 	  	 	  	 } // for
	 	  	 	    	
	 	  	       break;
	 	  	       
	 	  	 	  } // case LIST:
	 	  	    	
	 	  	    default:
	 	  	 	  {
	 	  	 	  	 System.out.println("Error: Unrecognised object type");
	 	  	 	     break;
	 	  	 	  }
	 	  	 	
	 	  	 } // switch(type(obj))
	 	  	
	 	  } // if(attribute_names.contains(CSPRELATION))
	 	  	
	 	  	
	 	  // Retrieves and assigns the name of this CSP
	 	  if(attribute_names.contains(CSPREF))
	 	  {
	 	     identify((String)content.getAttribute(CSPREF));  
	 	  }
	 	
	 } // assignContentData(Content content)
	 	
	 	
	 /**
   * <br> Method to evaluate whether the passed object is an instance of
   * <br> String, Long, Content or List.
   * <br> This allows a switch statement to be used to act on the type of object.
   *
   * @param 	The object to evalutate the type of.
   * @return 	An integer representing the type of the object.
   */
   private int type(Object obj)
	 {
	 	  // Not one of the Constants specified in this class for identifying 
	 	  // types. Hence should cause switch statements to go to the default option.
	 	  int type = 5;
	 	
	 	  if( obj instanceof Content )
	 	  {
	 	     type = CONTENT;	
	 	  }
	 	  	
	 	  if( obj instanceof List )
	 	  {
	 	     type = LIST;	
	 	  }
	 	  	
	 	  if( obj instanceof Long )
	 	  {
	 	     type = LONG;	
	 	  }
	 	  	
	 	  if( obj instanceof String )
	 	  {
	 	     type = STRING;	
	 	  }
	 	
	 	  return type;
	 	  
	 } // type(Object obj)
	 	

//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------


   /**
   *  Adds a unique name with which to identify this CSP.
   */
   public void identify(String str)
	 {
	    _identifier = new String(str);
	 }
	 
	 
	 /**
	 *  Returns the identifier to this CSP.
	 */
	 public String identity()
	 {
	    return new String(_identifier);
	 }


   /**
   *  Adds a Variable to the CSP.
   */
   public void addVariable(CSPVariable variable)	
	 {
	    _variables.addElement(variable.copy());
	 } 
	 
	 
	 /**
	 *  Retrieves a Vector of all the variables in this CSP.
	 */
	 public Vector getVariables()
	 {
	 	  Vector value = new Vector(_variables.size());
	 	  
	 	  for(Enumeration e = _variables.elements(); e.hasMoreElements(); )
	 	  {
	 	     CSPVariable next = (CSPVariable) e.nextElement();
	 	     value.addElement(next.copy());
	 	  }
	 	  
	 	  return value;
	 	
	 } // getVariables()
	 
	 
	 /**
	 * <br> Adds a new CSPRelation.
	 * <br> This identifies a relationship between variables in this CSP.
	 */
	 public void addRelationship(CSPRelation relation)
	 {
	    _relations.addElement(relation.copy());
	 }
	 
	 
	 /**
	 *  Returns a Vector of all the Relations contained in the CSP.
	 */
	 public Vector getRelations()
	 {
	    Vector value = new Vector(_relations.size());
	 	  
	 	  for(Enumeration e = _relations.elements(); e.hasMoreElements(); )
	 	  {
	 	     CSPRelation next = (CSPRelation) e.nextElement();
	 	     value.addElement(next.copy());
	 	  }
	 	  
	 	  return value;
	 	  
	 } // getRelations()
	 
	 
	 /**
	 * Returns a String representation of the CSP.
	 */
	 public String toString()
	 {
	 	  StringBuffer buff = new StringBuffer();
	 	  
	 	  buff.append("\n\n******************************");
	 	  buff.append("\nCSP: " + _identifier);
	 	
	 	  for(Enumeration e = _variables.elements(); e.hasMoreElements(); )
	 	  {
	 	  	 CSPVariable var = (CSPVariable) e.nextElement();
	 	     buff.append(var.toString());  	
	  	}
	  	
	  	for(Enumeration e = _relations.elements(); e.hasMoreElements(); )
	 	  {
	 	  	 CSPRelation rel = (CSPRelation) e.nextElement();
	 	     buff.append(rel.toString());
	  	}
	 	
	 	  buff.append("\n******************************\n\n");
	 	
	 	  return buff.toString();
	 	
	 } // toString()
	 
	 /**
	 * Returns a copy of this CSP.
	 */
	 public CSP copy()
	 {
	 	  CSP value = new CSP();
	 	  
	 	  for(Enumeration e = _variables.elements(); e.hasMoreElements(); )
	 	  {
	 	     value.addVariable((CSPVariable) e.nextElement());  	
	  	}  
	  		
	  	for(Enumeration e = _relations.elements(); e.hasMoreElements(); )
	 	  {
	 	     value.addRelationship((CSPRelation) e.nextElement());
	  	} 	
	  	
	  	value.identify(_identifier);
	 	  	
	 	  return value;
	 	
	 } // copy()
	 	
	 
	 	
	 /**
	 * <br> Returns a Content object populated with Data representing the information
	 * <br> held in this CSP object.
	 */
   public Content toContent()
	 {
	    Content value = new Content(CSP);
	    
	    // Add all CSPVariables to the Content object
	    for(Enumeration e = _variables.elements(); e.hasMoreElements(); )
	 	  {
	 	     CSPVariable var = (CSPVariable)e.nextElement();
	 	     Content content = var.toContent();
	 	     value.addAttribute(CSPVARIABLE, content);
	 	  
	 	  } // for
	 	  	
	 	  // Add all CSPRelations to the Content object
	 	  for(Enumeration e = _relations.elements(); e.hasMoreElements(); )
	 	  {
	 	     CSPRelation rel = (CSPRelation)e.nextElement();
	 	     Content content = rel.toContent();
	 	     value.addAttribute(CSPRELATION, content);
	 	  	
	 	  } // for   
	 	  	
	 	  // add the identifier of this CSP
	 	  value.addAttribute(CSPREF, _identifier);
	    
	    return value;	
	 	
	 } // toContent()	 
	 
	
//----------------------------------------------------------------------
// MAIN/TESTING METHOD
//----------------------------------------------------------------------

   /**
   * <br> With no variables the main method constructs and tests a CSP object.
   * <br> If a String is specified as a command line argument it will read it in, 
   * <br> parse it and try to form a CSP object using a Content object.
   */
   public static void main(String args[])
	 {
	    if(args.length == 0)
	 	  {
	 	     System.out.println();
	 	     System.out.println("Testing CSP class");	
	 	     System.out.println("Creating new CSP");
	 	     
	 	     CSP csp = new CSP();
	 	     
	 	     System.out.println();
	 	     System.out.println("Populating new CSP");
	 	     System.out.println("Identifying CSP");
	 	     csp.identify("auction4");
	 	     
	 	     System.out.println();
	 	     System.out.println("CSP named: " + csp.identity());
	 	     
	 	     System.out.println();
	       System.out.println("Creating new CSPRelation object"); 
	       CSPRelation relation = new CSPRelation();
	       
	       relation.setVariableNameOne("house");
	       relation.setVariableNameTwo("insurance");
	       relation.setRelationType(CSPRelation.EQUALITY);
	       relation.addTag("high");
	       relation.addTag("middle");
	       IndexPair index = new IndexPair("(2, 2)");
	       relation.addIndexPair(index);
	       index = new IndexPair("(3, 3)");
	       relation.addIndexPair(index);
	    
	       System.out.println();
	       System.out.println("Adding CSPRelation");
	       
	       csp.addRelationship(relation);
	       
	       System.out.println("Creating new CSPRelation object"); 
	       relation = new CSPRelation();
	       
	       relation.setVariableNameOne("boat");
	       relation.setVariableNameTwo("insurance");
	       relation.setRelationType(CSPRelation.INEQUALITY);
	       relation.addTag("extreme");
	       relation.addTag("middle");
	       index = new IndexPair("(4, 2)");
	       relation.addIndexPair(index);
	       index = new IndexPair("(3, 7)");
	       relation.addIndexPair(index);
	    
	       System.out.println();
	       System.out.println("Adding CSPRelation");
	       
	       csp.addRelationship(relation);
	       
	       System.out.println();
	       System.out.println("Retrieving CSPRelations");
	       System.out.println(csp.getRelations());
	       
	       
	       System.out.println();
	       System.out.println("Adding CSPVariables");
	       
	       System.out.println();
	       System.out.println("Creating new CSPVariable");
	       CSPVariable variable = new CSPVariable();
	    
	       CSPRange range = new CSPRange();
	 	     range.addRange("{house, car, boat}");	 	
	 	     variable.setRange(range); 
	 	  
	 	     CSPValue value = new CSPValue();
	       value.addElement("blue");
	       value.addElement("red");
	       value.addElement("green");
	       value.addElement("purple");
	       value.addTag("special");
	       value.addTag("reserved");
	       variable.addValue(value);
	       
	       variable.setName("auction");  
	       variable.addType("hot");
	       variable.addType("cold");
	       variable.setRole("lead");
	       variable.setRole("follow");
	   
	       csp.addVariable(variable);
	       
	       System.out.println();
	       System.out.println("Retrieving CSPVariable");
	       System.out.println(csp.getVariables());
	       
	       System.out.println();
	       System.out.println("Values held in CSP object:");
	       System.out.println(csp);
	       
	       System.out.println();
	       System.out.println();
	       System.out.println("Values held in copy of CSP object:");
	       System.out.println(csp.copy());
	       
	       System.out.println();
	       System.out.println("CSP Represented as a Content object:");
	       
	       Content content = csp.toContent();
	       
	       System.out.println(content);
	       
	       csp = new CSP(content);
	       
	       System.out.println();
	       System.out.println("Constructing new CSP using this content object");
	       System.out.println("Contents of new CSP:");
	       System.out.println(csp);

	  
	 	     
	 	     
	 	     System.out.println();
	 	  	
	 	  } // if(args.length == 0)
	 	  
	 	  if(args.length == 1)
	 	  {
	 	     System.out.println();
	 	     System.out.println("Reading input file: " + args[0]);
	 	     
	 	     /*
	 	        Used to test constructing CSP objects using a read and parsed text file.
	 	     */	
	 	     
	 	     System.out.println("NOT IMPLEMENTED YET");
	 	  	
	 	  } // if(args.length == 1)	
	    	
	 } // main
  
	
} // CSP

	
