//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files ("the
//     Programs") are Copyright (c) 1999 - 2000 Nortel Networks Corporation, 
//     8200 Dixie Road, Suite 100, Brampton, Ontario, Canada L6R 5P6.  
//     All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out in
//     the Nortel Networks FIPA-OS Public License, a copy of which can be found
//     in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version can
//     also be found at http://fipa-os.sourceforge.net/. By downloading the
//     files containing the Programs you accept the terms and  conditions of the
//     Public License. You do not have to accept these terms and conditions, but
//     unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.skill.constraint.ccl.object;

//----------------------------------------------------------------------

import fipaos.ont.fipa.FIPACONSTANTS;
import fipaos.ont.BASETOKENS;
import fipaos.ont.fipa.fipaman.FIPAMANCONSTANTS;

import java.util.*;

import fipaos.skill.constraint.ccl.variable.CSPVariableAssignment;
import fipaos.skill.constraint.ccl.variable.CSPValue;

import fipaos.agent.conversation.content.Content;

//----------------------------------------------------------------------

/**
 * 
 * <br> This class represents as CSP-solution Object as described in the 
 * <br> "CCL Language Specification v2.01" section 4.1.1.2, p5.
 * <br>
 * <br> This object captures the notion of a solution to a choice problem.
 * <br> Here all the choices are assigned an appropriate value and the 
 * <br> assignment violates none of the posted constraints.
 *
 */
 
public class CSPSolution
{
     
//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------

   /**
   *  String used to reference a Content object representing a CSPSolution object
   */
   public static final String CSPSOLUTION = "CSP-solution";
   
   /**
   *  String used to reference a Content object representing a CSPVariableAssignment object.
   */
   public static final String CSPVARIABLEASSIGNMENT = "CSP-variable-assignment";
   
   /**
   * <br> String used to reference the identifier of a CSPSolution object when represented
   * <br> as an Attribute of a Content object.
   */
   public static final String HREF = "href";


   // Used to facilitate the use of switch statements to determine the type
   // of object a Content object represents.
   
   public final static int CONTENT = 1;
   public final static int LIST = 2;
   public final static int LONG = 3;
   public final static int STRING = 4;

//----------------------------------------------------------------------
// STATIC VARIABLES
//----------------------------------------------------------------------
		
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

   /**
   *  This references the CSP object this object is a solution for.
   */
   private String _identifier = "";
   
   /**
   * <br> This Vector contains a list of CSPVariable-Assignment objects.
   * <br> These are the solution to the CSP object that that CSP-identifier
   * <br> referrs to.
   * <br>
   * <br> This list of variable assignments is such that:
   * <br>
   * <br> The list contains one and only one assignment for each and every
   * <br> variable defined in the CSP.
   * <br>
   * <br> The assignment of these values violates none of the constraints
   * <br> posted for the CSP. That is the variable assignment is consistent.
   */
   private Vector _assignments = null;

//----------------------------------------------------------------------
// CONSTRUCTOR
//----------------------------------------------------------------------

   /**
   * Constructs an empty CSPSolution
   */
   public CSPSolution()
	 {
	    _assignments = new Vector();   
	 }
	 
	 
	 /**
   * <br> Constructs a CSPSolution with the given name.
   * <br> This name refers to the CSP this is the solution for.
   */
   public CSPSolution(String name)
	 {
	    _assignments = new Vector();   
	    _identifier = new String(name);
	 }
	 
	 /**
	 *  Constructs a new CSPSolution and populates it with the Data in the passed Content object.
	 */
	 public CSPSolution(Content content)
	 {
	    _assignments = new Vector();
	    assignContentData(content);
	 }
	 

//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------

   /**
   *  Populates the Content object with the Data contained in the passed Content object.
   */
   private void assignContentData(Content content)
	 {
	 	  Set attributes = content.getAttributeNames();
	 	  
	 	  // assign the identifier for this CSPSolution
	 	  if(attributes.contains(HREF))
	 	  {
	 	     identify((String)content.getAttribute(HREF));   
	 	  }
	 	  
	 	  // retrieve the CSP-variable-assignment objects if present
	 	  if(attributes.contains(CSPVARIABLEASSIGNMENT))
	 	  {
	 	     Object obj = content.getAttribute(CSPVARIABLEASSIGNMENT);
	 	     
	 	     switch(type(obj))
	 	  	 {	 	
	 	  	 	  // Insert the single VariableAssignment into Vector
	 	  	    case CONTENT:	
	 	  	 	  {
	 	  	 	  	 Content assignment = (Content)obj;
	 	  	 	  	 _assignments.addElement(new CSPVariableAssignment(assignment));
	 	  	 	     break;
	 	  	 	  }
	 	  	 	
	 	  	 	  // Retrieve the List of VariableAssignments and insert into Vector
	 	  	 	  case LIST:
	 	  	 	  {
	 	  	 	  	 List list = (List)obj;
	 	  	 	  	 Content assignment = null;
	 	  	 	  	 
	 	  	 	  	 for(Iterator i = list.iterator(); i.hasNext(); )
	 	  	 	  	 {
	 	  	 	  	    assignment = (Content)i.next();
	 	  	 	  	    _assignments.addElement(new CSPVariableAssignment(assignment));
	 	  	 	  	 }
	 	  	 	  	
	 	  	 	     break;
	 	  	 	     
	 	  	 	  } // LIST
	 	  	 	  	
	 	  	 	  default:
	 	  	 	  {
	 	  	 	     System.out.println("Error, unknown object type");
	 	  	 	     break;
	 	  	 	  }	
	 	  	 	
	 	  	 } // switch(type(obj))		
	 	  	
	 	  } // if(attributes.contains(CSPVARIABLEASSIGNMENT))
	 	
	 } // assignContentData(Content content)	
	 	
	 	
	 /**
   * <br> Method to evaluate whether the passed object is an instance of
   * <br> String, Long, Content or List.
   * <br> This allows a switch statement to be used to act on the type of object.
   *
   * @param 	The object to evalutate the type of.
   * @return 	An integer representing the type of the object.
   */
   private int type(Object obj)
	 {
	 	  // Not one of the Constants specified in this class for identifying 
	 	  // types. Hence should cause switch statements to go to the default option.
	 	  int type = 5;
	 	
	 	  if( obj instanceof Content )
	 	  {
	 	     type = CONTENT;	
	 	  }
	 	  	
	 	  if( obj instanceof List )
	 	  {
	 	     type = LIST;	
	 	  }
	 	  	
	 	  if( obj instanceof Long )
	 	  {
	 	     type = LONG;	
	 	  }
	 	  	
	 	  if( obj instanceof String )
	 	  {
	 	     type = STRING;	
	 	  }
	 	
	 	  return type;
	 	  
	 } // type(Object obj)
   	

//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------

   /**
   * <br> Identifies the CSPSolution.
   * <br> This name refers to the CSP it is a solution for.
   */
   public void identify(String str)
	 {
	    _identifier = new String(str);
	 }
	 
	 
	 /**
	 * <br> Returns the identifier for the CSPSolution.
	 * <br> This name refers to the CSP it is a solution for.
	 */
	 public String identity()
	 {
	    return new String(_identifier);
	 }
	 
	 /**
	 * <br> Adds a new Variable Assignment to the CSPSolution
	 * <br> This represents another element in the solution.
	 */
	 public void addVariable(CSPVariableAssignment var)	
	 {
	    _assignments.addElement(var.copy()); 
	 }
	 
	 /**
	 * <br>  Returns a Vector containing all the Variable Assignments that
	 * <br>  represent the solution to the CSP problem.
	 */
	 public Vector variables()
	 {
	 	  Vector value = new Vector(_assignments.size());
	 	  
	 	  for(Enumeration e = _assignments.elements(); e.hasMoreElements(); )
	 	  {
	 	     CSPVariableAssignment var = (CSPVariableAssignment) e.nextElement();	
	 	  	 value.addElement(var.copy());
	 	  }
	 	  
	 	  return value;
	 	
	 } // variables()
	 
	 
	 /**
	 *  Returns a copy of the CSPSolution object.
	 */
	 public CSPSolution copy()
	 {
	    CSPSolution value = new CSPSolution(_identifier);
	    
	    for(Enumeration e = _assignments.elements(); e.hasMoreElements(); )
	 	  {
	 	     CSPVariableAssignment var = (CSPVariableAssignment) e.nextElement();	
	 	  	 value.addVariable(var.copy());
	 	  }
	   
	    return value;	
	 	
	 } // copy()
	 
	 
	 /**
	 *  Returns a String representation of the CSPSolution.
	 */
	 public String toString()
	 {
	    StringBuffer buff = new StringBuffer();
	    
	    buff.append("\n\n******************************");
	    buff.append("\nCSPSolution: " + _identifier);
	    
	    for(Enumeration e = _assignments.elements(); e.hasMoreElements(); )
	 	  {
	 	     CSPVariableAssignment var = (CSPVariableAssignment) e.nextElement();	
	 	  	 buff.append(var.toString());
	 	  }
	    
	    buff.append("\n******************************\n\n");
	 	
	 	  return buff.toString();
	 	
	 } // toString()
	 	
	 	
	 /**
	 *  Returns a Content object populated with the Data held in this CSPSolution object.
	 */
	 public Content toContent()
	 {
	 	  Content value = new Content(CSPSOLUTION);
	 	  
	 	  // assign identifier
	 	  value.addAttribute(HREF, _identifier);
	 	  
	 	  // assign VariableAssignment objects
	 	  
	 	  for(Enumeration e = _assignments.elements(); e.hasMoreElements(); )
	 	  {
	 	     CSPVariableAssignment assignment = (CSPVariableAssignment)e.nextElement();
	 	     value.addAttribute(CSPVARIABLEASSIGNMENT, assignment.toContent());
	 	  }
	 	  
	 	  return value;
	 	
	 } // toContent()
	 
	 
	
//----------------------------------------------------------------------
// MAIN/TESTING METHOD
//----------------------------------------------------------------------

   /**
   *  Used for testing the class.
   */
   public static void main(String args[])
	 {
	    System.out.println();
	    System.out.println("Testing class CSPSolution");
	    System.out.println("Creating new CSPSolution");	
	 	
	 	  CSPSolution sol = new CSPSolution();
	 	  
	 	  System.out.println();
	 	  System.out.println("Adding name");
	 	  
	 	  sol.identify("auction4");
	 	  
	 	  System.out.println();
	 	  System.out.println("Name: " + sol.identity());
	 	  System.out.println();
	 	  
	    System.out.println("Creating new CSPVariableAssignment object");
	    CSPVariableAssignment var = new CSPVariableAssignment();
	   
	    var.variableName("bob");
	    
	    CSPValue value = new CSPValue();
	    value.addElement("blue");
	    value.addElement("red");
	    value.addElement("green");
	    value.addElement("purple");
	    value.addTag("special");
	    value.addTag("reserved");
	    
	    var.addCSPValue(value);
	    
	    sol.addVariable(var);
	    
	    System.out.println("Creating new CSPVariableAssignment object");
	    var = new CSPVariableAssignment();
	   
	    var.variableName("joe");
	    
	    value = new CSPValue();
	    value.addElement("blue");
	    value.addElement("red");
	    value.addElement("silver");
	    value.addElement("purple");
	    value.addTag("special");
	    value.addTag("free");
	    
	    var.addCSPValue(value);
	    
	    sol.addVariable(var);
	    
	    System.out.println();
	    System.out.println("Variables held in CSPSolution:");
	    System.out.println(sol.variables());
	    
	    System.out.println();
	    System.out.println("Values held in CSPSolution:");
	    System.out.println(sol);
	  
	    System.out.println();
	    System.out.println("Values held in copy of CSPSolution:");
	    System.out.println(sol.copy());
	    
	    Content content = sol.toContent();
	    
	    System.out.println();
	    System.out.println("CSPSolution represented as a Content object");
	    System.out.println(content);
	    
	    System.out.println();
	    System.out.println("Constructing new CSPSolution using this content object");
	    
	    CSPSolution next = new CSPSolution(content);
	    
	    System.out.println();
	    System.out.println("Values held in CSPSolution:");
	    System.out.println(next);
	 	
	 	
	 } // main
  
	
} // CSPSolution

	
