//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files ("the
//     Programs") are Copyright (c) 1999 - 2000 Nortel Networks Corporation, 
//     8200 Dixie Road, Suite 100, Brampton, Ontario, Canada L6R 5P6.  
//     All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out in
//     the Nortel Networks FIPA-OS Public License, a copy of which can be found
//     in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version can
//     also be found at http://fipa-os.sourceforge.net/. By downloading the
//     files containing the Programs you accept the terms and  conditions of the
//     Public License. You do not have to accept these terms and conditions, but
//     unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.skill.constraint.ccl.variable;

//----------------------------------------------------------------------

import fipaos.ont.fipa.FIPACONSTANTS;
import fipaos.ont.BASETOKENS;
import fipaos.ont.fipa.fipaman.FIPAMANCONSTANTS;

import java.util.*;

import fipaos.skill.constraint.ccl.variable.Tuple;
import fipaos.skill.constraint.ccl.variable.List;

import fipaos.agent.conversation.content.Content;

//----------------------------------------------------------------------

/**
 * 
 * <br> This class describes the object CSP-range as described in the document 
 * <br> "CCL Language Specification v2.01" section 4.2.2.1, p12.
 * <br>
 * <br> This object represents a complete domain, to be used when explicit
 * <br> enumeration of values would be too inefficient. the two items "Range"
 * <br> and "Tuple-range" are optional however one or the other must be present.
 *
 */
 
public class CSPRange
{
     
//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------

   /*
   * <br> Name of an Attribute required to be part of a text representation
   * <br> of a CSPRange. 
   */
   public final static String RANGE = "Range";
     
   /**
   *  Name of the Element where the range of values a CSPRange may take are listed.
   */
   public final static String LIST_RANGE = "List-range";
   
   /**
   *  Name of the Attribute where the values are listed.
   */
   public final static String VALUES = "Values";
   
   /**
   * Name a CSPRange is labelled with when represented as a Content object.
   */
   public final static String CSPRANGE = "CSP-range";
   

//----------------------------------------------------------------------
// STATIC VARIABLES
//----------------------------------------------------------------------
		
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

   /**
   * <br> Range.
   * <br> This variable defines complete domains such as ordered lists of numbers,
   * <br> world airports, etc. These must be part of a common ontology.
   * <br>
   * <br> An example of a range would be {car, boat, plane}
   */
   private Vector _range = null;
   
   
   /**
   *  Contains a Reference to the type of Objects this Range contains.
   */
   private String _range_type = "";
   
   /**
   * <br> Tuple-range
   * <br> This variable defines a combination of all the legal values in a tuple
   * <br> A range is given for each slot in the tuple and this parameter specifies 
   * <br> that all combinations of values from the given ranges in each slot in 
   * <br> the tuple are legal.
   * <br> A tuple specifies sets of allowed combinations.
   * <br>
   * <br> An example of a Tuple range would be:
   * <br> {(Savoy, London), (Hyatt, Singapore), (Waldorf, NewYork)}
   */
   
   // This Vector will contain a set of Tuple objects.
   private Vector _tuple_range = null;
   

//----------------------------------------------------------------------
// CONSTRUCTOR
//----------------------------------------------------------------------

   /**
   * <br> Creates an empty CSPRange object.
   */
   public CSPRange()
	 {
	 }
	 
	 
	 /**
	 * <br> Constructs a new CSPRange object.
	 * <br> Populates it using the Data held in the passed Content object.
	 */
	 public CSPRange(Content content)
	 {
	    assignContentData(content);
	 }
	 

//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------

   /**
   * <br> Copies the values held in the passed Content object 
   * <br> to the Data members in the class.
   */
   private void assignContentData(Content content)
	 {
	 	  // System.out.println();
	 	  // System.out.println("assignContentData called with Content object:");
	 	  // System.out.println(content);
	 	  // System.out.println();
	 	  
	 	  // Extract all Data from passed Content object and assign it.
	 	  
	 	  // Contains names of all the Attributes in the passed Content object.
	 	  Set attribute_names = content.getAttributeNames();
	 	  
	 	  // System.out.println(attribute_names);
	 	  
	 	  // Assigns a Value for the Range
	 	  if(attribute_names.contains(RANGE))
	 	  {
	 	  	 String range = (String)content.getAttribute(RANGE);
	 	  	 // System.out.println("Range: " + range);
	 	     _range_type = new String(range);
	 	  }
	 	  
	 	  
	 	  // Assign values for the Range
	 	  if(attribute_names.contains(LIST_RANGE))
	 	  {
	 	  	 Content list_range = (Content)content.getAttribute(LIST_RANGE);
	 	  	 
	 	  	 // System.out.println(list_range);
	 	  	 
	 	  	 // Contains names of all the Attributes in the "List-range" Content object.
	 	  	 Set range_attributes = list_range.getAttributeNames();
	 	  	 
	 	  	 // System.out.println("Attrs: " + range_attributes);
	 	  	 
	 	  	 if(range_attributes.contains(VALUES))
	 	  	 {
	 	  	    String value = (String)list_range.getAttribute(VALUES);
	 	  	    System.out.println("Value for CSPRange: " + value);
	 	  	    
	 	  	    /*
	 	  	       Tests to see if it is of the form
	 	  	 	     {car, boat}
	 	  	 	     or
	 	  	 	     {(car, blue), (boat, green)}
	 	  	    */
	 	  	    if(value.indexOf("(") == -1)
	 	  	 	  {
	 	  	       addRange(value);	
	 	  	 	  }
	 	  	 	  else
	 	  	 	  {
	 	  	 	     Vector v = List.parseList(value);
	 	  	 	     addRange(v);	
	 	  	 	  }
	 	  	 	  
	 	   	 } // if(range_attributes.contains(VALUES))	
	 	  	
	 	  } // if(attribute_names.contains(LIST_RANGE))
	 	
	 } // assignContentData()
   

//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------

   /**
   * <br> Adds a domain range to the CSPRange.
   * <br> The String representing this range should take the following form:
   * <br> {house, car, boat}
   */	
   public void addRange(String str)
	 {
	 	  // System.out.println("String range: " + str);
	 	
	 	  if(_range == null)
	 	  {
	 	     _range = new Vector();
	  	}
	  
			// Trim off the brackets
			if(str.indexOf("{") != -1)
			{
			   str = str.substring(str.indexOf("{") + 1, str.indexOf("}"));	
			 	 // System.out.println("substring is: " + str);
			 	  	 
			 	 StringTokenizer st = new StringTokenizer(str, ",");
			 	  
		      while (st.hasMoreTokens()) 
		      {
		         String next = st.nextToken();  
		         next = next.trim();  
		         _range.addElement(new String(next));
		      }
	 	  }
	 	  else
		  {
		  	 // Copes if there is only a single value
		     _range.addElement(str.trim());
			}
	 	  
	 } // addRange(String str)
	 	
	 	
	 /**
	 * <br> Adds a domain range to the CSPRange.
	 * <br> The Vector should contain the Strings that comprise the Range.
	 */
	 public void addRange(Vector range)
	 {
	    if(_range == null)
	 	  {
	 	     _range = new Vector();
	  	}
	  	
	  	for(Enumeration e = range.elements(); e.hasMoreElements(); )
	 	  {
	 	     _range.addElement(new String((String) e.nextElement()));
	 	  }	
	 	
	 } // addRange(Vector range)
	 
	 	
	 	
	 /**
	 * <br> Adds a Tuple to the CSPRange.
	 */
	 public void addTuple(Tuple tuple)
	 {
	    if(_tuple_range	== null)
	 	  {
	 	     _tuple_range = new Vector();
	 	  }
	 	  
	 	  _tuple_range.addElement(tuple.copy());
	 	
	 } // addTuple(Tuple tuple)
	 	
	 	
	 /**
	 * <br> Returns a Vector of the Strings that comprise the Range 
	 * <br> Value held in the CSPRange.
	 */	 
	 public Vector getRange()
	 {
	    Vector value = new Vector();
	    
	    for(Enumeration e = _range.elements(); e.hasMoreElements(); )
	 	  {
	 	     value.addElement(new String((String) e.nextElement()));
	 	  }
	    	
	 	  return value;
	 	
	 } // getRange()
	 	
   	 	
	 /**
	 * <br> Returns a Vector of Tuple objects.
	 * <br> This is a copy of the Vector of Tuple objects held in the CSPRange object.
	 */
	 public Vector getTuples()
	 {
	 	  Vector value = new Vector();
	    
	    for(Enumeration e = _tuple_range.elements(); e.hasMoreElements(); )
	 	  {
	 	  	 Tuple next = (Tuple) e.nextElement();
	 	     value.addElement(next.copy());
	 	  }
	    	
	 	  return value;
	 	
	 } // getTuples()
	 	
	 	
	 /**
	 *  Returns a copy of this CSPRange object.
	 */ 
	 public CSPRange copy()
	 {
	    CSPRange value = new CSPRange();
	    
	    if(!_range_type.equals(""))
	 	  {
	 	     value.setRangeType(_range_type);
	 	  }
	    
	    if(_range != null)
	 	  {
	       value.addRange(_range);
	 	  }
	    
	    if(_tuple_range != null)
	 	  {
	       for(Enumeration e = _tuple_range.elements(); e.hasMoreElements(); )
	 	     {
	 	  	    Tuple next = (Tuple) e.nextElement();
	 	        value.addTuple(next.copy());
	 	     }
	 	  }
	    
	    return value;	
	 	
	 } // copy()
	 	
	 	
	 /**
	 * <br> Returns a String representation of a CSPRange.
	 */
	 public String toString()
	 {
	    StringBuffer value = new StringBuffer();
	    
	    value.append("\n******************************");	
	    value.append("\nCSPRange:");
	    
	    // System.out.println("Range: " + _range_type);
	    
	    if(!_range_type.equals(""))
	 	  {
	 	     value.append("\nRangeType: " + _range_type + "\n");
	 	  }
	    
	    // Test to see whether it contains a domain or a Tuple Range.
	    if(_range != null)
	 	  {
	 	  	 value.append("\nValues: ");
	 	  	 value.append("\n{ ");
	 	  	
	 	  	 for(Enumeration e = _range.elements(); e.hasMoreElements(); )
	 	  	 {
	 	  	 	  value.append((String) e.nextElement());
	 	  	 	  
	 	  	 	  if(e.hasMoreElements())
	 	  	 	  {
	 	  	 	     value.append(", ");
	 	  	 	  }
	 	  	 	  
	 	  	 } // for	
	 	  	 	
	 	  	 value.append(" }\n");
	 	  	 	
	 	  } // if(_range != null)
	 	  
	 	  if(_tuple_range != null)
	 	  {
	 	  	 value.append("\nValues: ");
	 	  	 value.append("\n{ ");
	 	  	
	 	     for(Enumeration e = _tuple_range.elements(); e.hasMoreElements(); )
	 	  	 {
	 	  	 	  value.append((Tuple) e.nextElement());
	 	  	 	  
	 	  	 	  if(e.hasMoreElements())
	 	  	 	  {
	 	  	 	     value.append(", ");
	 	  	 	  }
	 	  	 	  
	 	  	 } // for
	 	  	
	 	  	 value.append(" }\n");
	 	  	
	 	  } // if(_tuple_range != null)
	 	  	
	 	  value.append("\n******************************\n");
	    
	    return value.toString();
	    
	 } // toString()
	 	
	 	
	 /**
	 *  Sets the type of values that are held in this CSPRange object.
	 */
	 public void setRangeType(String str)	
	 {
	 	  _range_type = new String(str);
	 }
	 
	 
	 /**
	 *  Return the type of values that may be held in this CSPRange object.
	 */
	 public String getRangeType()
	 {
	    return new String(_range_type);
	 }
	 	
	 	
	 /**
	 * <br> Creates a new Content object populates it with the Values
	 * <br> held in this class then returns it.
	 */
	 public Content toContent()
	 {
	 	  Content value = new Content("CSP-range");
	 	  
	 	  if(!_range_type.equals(""))
	 	  {
	 	     value.addAttribute(RANGE, new String(_range_type));
	 	  }
	 	  
	 	  // Adds the Range values held in the "_range" Vector as the Value of the CSP-range.
	 	  if(_range != null)
	 	  {
	 	  	 Content content = new Content(LIST_RANGE);
	 	  	 content.addAttribute(VALUES, List.toString(_range));
	 	  	
	 	     value.addAttribute(LIST_RANGE, content);
	 	  }
	 	  
	 	  // Adds the Tuple values held in the "_tuple_range" Vector as the Value of the CSP-range.
	 	  if(_tuple_range != null)
	 	  {
	 	  	 Content content = new Content(LIST_RANGE);
	 	  	 
	 	  	 
	 	  	 StringBuffer buff = new StringBuffer();
	 	  
	 	     for(Enumeration e = _tuple_range.elements(); e.hasMoreElements(); )
	 	  	 {
	 	  	 	  buff.append((Tuple) e.nextElement());
	 	  	 	  
	 	  	 	  if(e.hasMoreElements())
	 	  	 	  {
	 	  	 	     buff.append(", ");
	 	  	 	  }
	 	  	 	  
	 	  	 } // for
	 	  	
	 	  	 content.addAttribute(VALUES, "{ " + buff + " }");
	 	     value.addAttribute(LIST_RANGE, content);
	 	  }
	 	  
	 	  return value;
	 	
	 } // toContent()
	 
	 
	
//----------------------------------------------------------------------
// MAIN/TESTING METHOD
//----------------------------------------------------------------------

   /**
   * <br> Used for testing.
   * <br>
   */
   public static void main(String args[])
	 {
	 	  System.out.println();
	 	  System.out.println("Testing CSPRange class.");
	 	  System.out.println();
	 	  System.out.println("Creating new CSPRange");
	 	  System.out.println();
	 	  
	 	  CSPRange range = new CSPRange();
	 	  range.setRangeType("Transport");
	 	  range.addRange("{house, car, boat}");
	 	  
	 	  System.out.println("Values held in CSPRange object");
	 	  System.out.println(range);
	 	  
	 	  System.out.println();
	 	  System.out.println("Creating a new CSPRange object using a series of tuples.");
	 	  
	 	  range = new CSPRange();
	 	  range.addTuple(new Tuple("(Savoy, London)"));
	 	  range.addTuple(new Tuple("(Ritz, Paris)"));
	 	  
	 	  System.out.println("Values held in CSPRange object");
	 	  System.out.println(range);
	 	  
	 	  System.out.println();
	 	  System.out.println("Retrieving Content object");
	 	  System.out.println(range.toContent());
	 	  System.out.println();
	 	  
	 	  System.out.println();
	 	  System.out.println("Values held in a copy of range object");
	 	  System.out.println(range.copy());
	 	  
	 	  System.out.println();
	 	  System.out.println("Retrieving tuples held in CSPRange object");
	 	  System.out.println(range.getTuples());
	 	  
	 	  
	 	  
	 	  System.out.println();
	 	  System.out.println("Creating new CSPRange");
	 	  
	 	  range = new CSPRange();
	 	  
	 	  System.out.println("Populating using a Vector of Strings");
	 	  Vector v = new Vector();
	 	  v.addElement("blue");
	 	  v.addElement("green");
	 	  
	 	  range.addRange(v);
	 	  
	 	  System.out.println("Values held in CSPRange object");
	 	  System.out.println(range);
	 	  System.out.println();
	 	  
	 	  System.out.println("Retrieving Vector of values");
	 	  System.out.println(range.getRange());
	 	  
	 	  System.out.println();
	 	  System.out.println("Values held in a copy of range object");
	 	  System.out.println(range.copy());
	 	  
	 	  
	 	  System.out.println();
	 	  System.out.println("Creating new CSPRange using Content object.");
	 	  System.out.println();
	 	  
	 	  Content range_content = new Content("CSP-range");
	 	  range_content.setAttribute("Range", "Car");
	 	  
	 	  Content value_content = new Content("List-range");
	 	  value_content.setAttribute("Values", "{Renault, Fiat}");
	 	  range_content.setAttribute("List-range", value_content);
	 	  
	 	  range = new CSPRange(range_content);
	 	  
	 	  System.out.println();
	 	  System.out.println("Retrieving Content object");
	 	  System.out.println(range.toContent());
	 	  System.out.println();
	 	  
	 	  range_content = new Content("CSP-range");
	 	  range_content.setAttribute("Range", "Car");
	 	  
	 	  value_content = new Content("List-range");
	 	  value_content.setAttribute("Values", "{(Renault, 1000), (Fiat, 2000)}");
	 	  range_content.setAttribute("List-range", value_content);
	 	  
	 	  range = new CSPRange(range_content);
	 	  
	 	  System.out.println();
	 	  System.out.println("Retrieving Content object");
	 	  System.out.println(range.toContent());
	 	  System.out.println();
	 	  
	 	  System.out.println();
	 	  System.out.println();
	 	  System.out.println();
	 	  
	 	  
	 	  System.out.println("Values held in CSPRange object");
	 	  System.out.println(range);
	 	  System.out.println();
	 	  
	    
	    CSPRange constraint_range = new CSPRange();
	    constraint_range.setRangeType("Constraint");
	    
	    Vector constraint = List.parseList("{(Fiat, 1000), (Renault, 1000)}");
	 	  constraint_range.addRange(constraint);	 	
	 	  
	 	  System.out.println();
	 	  System.out.println("Constraint Range:");
	 	  System.out.println(constraint_range);
	 	
	 } // main
   
  
	
} // CSPRange

	
