//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files ("the
//     Programs") are Copyright (c) 1999 - 2000 Nortel Networks Corporation, 
//     8200 Dixie Road, Suite 100, Brampton, Ontario, Canada L6R 5P6.  
//     All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out in
//     the Nortel Networks FIPA-OS Public License, a copy of which can be found
//     in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version can
//     also be found at http://fipa-os.sourceforge.net/. By downloading the
//     files containing the Programs you accept the terms and  conditions of the
//     Public License. You do not have to accept these terms and conditions, but
//     unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.skill.constraint.ccl.variable;

//----------------------------------------------------------------------

import fipaos.ont.fipa.FIPACONSTANTS;
import fipaos.ont.BASETOKENS;
import fipaos.ont.fipa.fipaman.FIPAMANCONSTANTS;

import java.util.*;

import fipaos.agent.conversation.content.Content;

//----------------------------------------------------------------------

/**
 * 
 * <br> This class describes the object CSP-value as described in the document 
 * <br> "CCL Language Specification v2.01" section 4.2.2.2, p12.
 * <br>
 * <br> This object represents an option. In general this can be a tuple- hence
 * <br> the variable is an ordered list of domain terms.
 * <br>
 * <br> May contain the following information:
 * <br>
 * <br> Npart      integer                1
 * <br> Elements   List {domain-term}     {blue, green, red}
 * <br> Tags       Set {symbol}           {T1, T2}
 *
 */
 
public class CSPValue
{
     
//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------

   /**
   * <br> Used to reference the Element that contains the Value when represented as 
   * <br> a Content object.
   */
   public final static String ELEMENTS = "Elements";
   
   /**
   *  Name CSPValue objects are labelled with when represented as Content objects.
   */
   public final static String CSPVALUE = "CSP-value";
   
   /**
   *  Name of the Attribute the Npart value is stored with when represented in Content objects.
   */
   public final static String NPART = "Npart";
   
   /**
   *  Name of the Attribute the Values contained in this class are stored as in Content objects.
   */
   public final static String VALUE = "Value";
   
   // for handling tags values.
   public final static String TAGS = "Tags";
   public final static String NAME = "Name";

//----------------------------------------------------------------------
// STATIC VARIABLES
//----------------------------------------------------------------------
		
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

   /**
   * <br>  Integer containing the Npart variable.
   * <br>  The Npart variable identifies the number of elements
   * <br>  that a tuple variable may have.
   * <br>  This must be identical to the number of items in the Elements parameter.
   * <br>
   * <br>  This value should be set automatically whenever a new Element is added.
   * <br>  For a CSPValue this value should always be 0 or 1 since it can only
   * <br>  contain a single list of domain terms
   */
   private int _npart = 0;
   
   /**
   * <br> Contains the list of domain terms referred to by Elements.
   */
   private Vector _domain_terms = null;
   
   /**
   * <br>
   */
   private Vector _Tags = null;

//----------------------------------------------------------------------
// CONSTRUCTOR
//----------------------------------------------------------------------

   /**
   * Constructs an Empty CSPValue.
   */
   public CSPValue()
	 {
	    _Tags =  new Vector();
	    _domain_terms = new Vector();
	 }
	 
	 
	 /**
	 * <br> Constructs a new CSPValue object and populates it using the Data held
	 * <br> in the passed Content object.
	 */
	 public CSPValue(Content content)
	 {
	 	  _Tags =  new Vector();
	    _domain_terms = new Vector();
	 	
	    assignContentData(content);
	 }
   

//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------

   /**
   * <br> Extracts Data from the passed Content object and uses it to 
   * <br> populate this class with Data.
   */
   private void assignContentData(Content content)
	 {
	 	  // Retrieve the names of the Attributes held in the content object
	 	  Set attribute_names = content.getAttributeNames();
	 	
	 	  // Set the NPart values
	 	  if(attribute_names.contains(NPART))
	 	  {
	 	  	 Object obj = content.getAttribute(NPART);
	 	  	 
	 	  	 if(obj instanceof String)
	 	  	 {
	 	  	 	  // System.out.println("npart: " + obj);
	 	  	 	  String npart = (String) obj;
	 	  	 	  _npart = new Integer(npart).intValue();
	 	  	 }
	 	  	  
	 	  	 if(obj instanceof Long)
	 	  	 {
	 	  	    Long npart = (Long) obj;
	 	        _npart = npart.intValue();
	 	  	 }
	 	  	
	 	  } // if(attribute_names.contains(NPART))
	 	  
	 	  // Element containing Value data
	 	  if(attribute_names.contains(ELEMENTS))
	 	  {
	 	     Content element_content = (Content)content.getAttribute(ELEMENTS);
	 	     
	 	     Set element_attributes = element_content.getAttributeNames();
	 	     
	 	     if(element_attributes.contains(VALUE))
	 	  	 {
	 	  	 	  String value = (String)element_content.getAttribute(VALUE);
	 	  	 	  // System.out.println("Value: " + value);
	 	  	 	  
	 	  	 	  // Converts String to a vector of values
	 	  	 	  Vector values = List.parseList(value);
	 	  	 	  _domain_terms = values;
	 	  	 }
	 	     
	 	  } // if(attribute_names.contains(ELEMENTS))
	 	  	
	 	  // retrieve tag values
	 	  if(attribute_names.contains(TAGS))
	 	  {
	 	     Content tag_content = (Content)content.getAttribute(TAGS);
	 	    
	 	     Set tag_attributes = tag_content.getAttributeNames();
	 	     
	 	     if(tag_attributes.contains(NAME))
	 	  	 {
	 	        String tag = (String)tag_content.getAttribute(NAME);
	 	        System.out.println("Tag: " + tag);	
	 	  	    _Tags = List.parseList(tag);
	 	  	 }
	 	  	
	 	  } // if(attribute_names.contains(TAGS))
	 	
	 	
	 } // assignContentData(Content content)

//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------

   /**
   * <br> Adds an Element to the CSPValue.
   */	
   public void addElement(String str)
	 {   
	 	  if(_domain_terms == null)
	 	  {
	 	     _domain_terms = new Vector();
	 	  }
	 
	 	  _domain_terms.addElement(new String(str));
	 	  
	 	  // Reset the Npart variable to indicate that there is a value present.
	 	  if(_npart == 0)
	 	  {
	 	     _npart = 1;
	 	  }
	 	  
	 } // addElement(String str)
	 	
	 /**
	 * <br> Returns a Vector of all the Elements held
	 */
	 public Vector getElements()
	 {
	    Vector value = new Vector();
	 	  
	 	  for(Enumeration e = _domain_terms.elements(); e.hasMoreElements(); )
	 	  {
	 	     value.addElement(new String((String) e.nextElement()));
	 	  }
	 	  
	 	  return value;	
	 	
	 	
	 } // getElements()
	 	
	 	
	 /**
	 * <br> Returns a String representation of the collection of Elements
	 */
	 public String getElementString()
	 {
	 	  StringBuffer strbuff = new StringBuffer();
	 	
	    for(Enumeration e = _domain_terms.elements(); e.hasMoreElements(); )
	 	  {
	 	     strbuff.append((String) e.nextElement());
	 	     
	 	     if(e.hasMoreElements())
	 	  	 {
	 	  	    strbuff.append(", ");
	 	  	 }
	 	  }   
	 	  
	 	  return "{ " + strbuff + " }";	
	 	
	 } // getElementString()
	 	
	 	
	 /**
	 * <br> Adds a Tag to the CSPValue.
	 */
	 public void addTag(String str)
	 {
	 	  if(_Tags == null)
	 	  {
	 	     _Tags = new Vector();
	 	  }
	 	  
	 	  _Tags.addElement(new String(str));
	 	
	 } // addTag(String str)
	 	
	 	
	 /**
	 * Retrieves the set of all the tags.
	 */
	 public Vector getTags()
	 {
	 	  Vector value = new Vector();
	 	  
	 	  for(Enumeration e = _Tags.elements(); e.hasMoreElements(); )
	 	  {
	 	     value.addElement(new String((String) e.nextElement()));
	 	  }
	 	  
	 	  return value;
	 	
	 } // getTags()
	 	
	 	
	 	/**
	 * <br> Returns a String representation of the collection of Tags
	 * <br> If there are more than one will return them surrounded by brackets.
	 * <br>
	 */
	 public String getTagString()
	 {
	 	  StringBuffer strbuff = new StringBuffer();
	 	
	    for(Enumeration e = _Tags.elements(); e.hasMoreElements(); )
	 	  {
	 	     strbuff.append((String) e.nextElement());
	 	     
	 	     if(e.hasMoreElements())
	 	  	 {
	 	  	    strbuff.append(", ");
	 	  	 }
	 	  }  
	 	  
	 	  if(_Tags.size() == 1)
	 	  {
	 	     return strbuff.toString();
	 	  } 
	 	  else
	 	  { 
	 	     return "{ " + strbuff + " }";	
	 	  }
	 	
	 } // getTagString()
	 	
	 /**
	 * <br> Returns the number of items in the Elements Parameter.
	 */
	 public int getNparts()
	 {
	 	  return new Integer(_npart).intValue();
	 	  	
	 } // getNparts()
	 	
	 	
	 /**
	 * <br> Returns a copy of this CSPValue object.
	 */
	 public CSPValue copy()
	 {
	    CSPValue value = new CSPValue();
	    
	    for(Enumeration e = _domain_terms.elements(); e.hasMoreElements(); )
	 	  {
	 	     value.addElement((String) e.nextElement());
	 	  }
	    
	    for(Enumeration e = _Tags.elements(); e.hasMoreElements(); )
	 	  {
	 	     value.addTag((String) e.nextElement());
	 	  }
	 	  
	 	  return value;	
	 	
	 } // copy()
	 	
	 	
	 /**
	 * <br> Returns a String representation of this CSPValue object.
	 */
	 public String toString()
	 {
	 	  String value = "\n\n******************************\n" +
	 	  	             "CSPValue:\n" +
	 	  	             "Elements: " + getElementString() + "\n" +
	 	  	             "Tags: " + getTagString() + "\n" +
	 	  	             "Nparts: " + getNparts() + "\n" +
	 	  	             "******************************\n\n";
	 	
	 	  return value;
	 	
	 } // toString()
	 	
	 	
	 /**
	 *  Returns a Content object populated with the Data held in this CSPValue object.
	 */
	 public Content toContent()
	 {
	 	  Content content = new Content(CSPVALUE);
	 	  
	 	  content.addAttribute(NPART, String.valueOf(_npart));
	 	  
	 	  Content elements = new Content(ELEMENTS);
	 	  Content tags = new Content(TAGS);
	 	  
	 	  if(_domain_terms.size() >= 1)
	 	  {
	 	  	 // converts value in a Vector to a String based list.
	 	     String value = List.toString(_domain_terms); 
	 	     elements.addAttribute(VALUE, value); 
	 	     content.addAttribute(ELEMENTS, elements);
	 	  }
	 	  
	 	  if(_Tags.size() >= 1)
	 	  {
	 	     String tag = List.toString(_Tags);
	 	     tags.addAttribute(NAME, tag);
	 	     content.addAttribute(TAGS, tags);
	 	  }
	 	
	 	  return content;
	 	
	 } // toContent()
	 	
	
//----------------------------------------------------------------------
// MAIN/TESTING METHOD
//----------------------------------------------------------------------

   /**
   * <br> Used for testing.
   */
   public static void main(String args[])
	 {
	    System.out.println();
	    System.out.println("Testing CSPValue class");
	    
	    System.out.println("Creating new CSPValue");
	    
	    CSPValue value = new CSPValue();
	    value.addElement("blue");
	    value.addElement("red");
	    value.addElement("green");
	    value.addElement("purple");
	    
	    value.addTag("special");
	    value.addTag("reserved");
	    
	    System.out.println();
	    System.out.println("Tags:");
	    System.out.println(value.getTagString());
	    System.out.println(value.getTags());
	    
	    
	    System.out.println();
	    System.out.println("Elements:");
	    System.out.println(value.getElementString());
	    System.out.println(value.getElements());
	    
	    System.out.println();
	    System.out.println("parts: " + value.getNparts());
	    System.out.println();
	    
      
      System.out.println();
      System.out.println("Copy of the previous CSPValue is:");
      System.out.println();
      System.out.println(value.copy());
      System.out.println();
      
      System.out.println();
      System.out.println("CSPValue object represented as a Content object");
      System.out.println(value.toContent());
      
      Content val = value.toContent();
      
      System.out.println();
      System.out.println("Constructing new CSPValue using content object");
      
      CSPValue test_val = new CSPValue(val);
      
      System.out.println();
      System.out.println("Values held in CSPValue object:");
      System.out.println(test_val);
      
      
      System.out.println();
      System.out.println("Populating CSPValue object using a Content object");
      
      Content content = new Content(CSPVALUE);
      content.addAttribute(NPART, new Long(1));
      
      Content value_content = new Content(ELEMENTS);
      value_content.addAttribute(VALUE, "{Renault}");
      content.addAttribute(ELEMENTS, value_content);
      
      System.out.println("Content object:");
      System.out.println(content);
      
      value = new CSPValue(content);
      
      
      System.out.println();
      System.out.println("Values held in CSPValue object:");
      System.out.println(value);
     
	 	
	 } // main
  
	
} // CSPValue

	
