//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files ("the
//     Programs") are Copyright (c) 1999 - 2000 Nortel Networks Corporation, 
//     8200 Dixie Road, Suite 100, Brampton, Ontario, Canada L6R 5P6.  
//     All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out in
//     the Nortel Networks FIPA-OS Public License, a copy of which can be found
//     in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version can
//     also be found at http://fipa-os.sourceforge.net/. By downloading the
//     files containing the Programs you accept the terms and  conditions of the
//     Public License. You do not have to accept these terms and conditions, but
//     unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.skill.constraint.ccl.variable;

//----------------------------------------------------------------------

import fipaos.ont.fipa.FIPACONSTANTS;
import fipaos.ont.BASETOKENS;
import fipaos.ont.fipa.fipaman.FIPAMANCONSTANTS;

import java.util.*;

import fipaos.skill.constraint.ccl.variable.CSPValue;
import fipaos.skill.constraint.ccl.variable.CSPRange;
import fipaos.skill.constraint.ccl.variable.List;

import fipaos.agent.conversation.content.Content;



//----------------------------------------------------------------------

/**
 * 
 * <br> This class describes the object CSP-variable as described in the document 
 * <br> "CCL Language Specification v2.01" section 4.2.2.4, p13.
 * <br>
 * <br> This object represents a single choice to be made, along with a set of possible
 * <br> options for that choice.
 * <br> The Type and Role slots enable the Variable to be situated within the problem
 * <br> solving context.
 *
 */
 
 /*
 An example of a CSPVariable represented as text. From "ccl3.xml"
 
      <CSP-variable Name="BillEndTime" Type="time">
				<Domain>
					<CSP-value-list Npart="1">
							<List-values Values="{13:00:00,13:30:00,14:00:00}"/>
							<Tags Name="T1"/>
					</CSP-value-list>
				</Domain>
			</CSP-variable>
 */
 
public class CSPVariable
{
     
//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------

   /**
   *  Used to identify a variable within a CSP when represented as a Content object.
   */
   public final static String NAME = "Name";
   
   /**
   * <br> Used to identify the type of variable this CSPVariable may contain when 
   * <br> represented as a Content object.
   */
   public final static String TYPE = "Type";
   
   /**
   * <br> Used to identify a Content object containing the values this variable may
   * <br> take when the CSPVariable is represented as a Content object.
   */
   public final static String DOMAIN = "Domain";
   
   /**
   * <br> Attribute name a CSPVariable is always referred to when represented as a 
   * <br> Content object.
   */
   public final static String CSPVARIABLE = "CSP-variable";
   

//----------------------------------------------------------------------
// STATIC VARIABLES
//----------------------------------------------------------------------
		
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

   /**
   * <br> Gives a unique symbol that is used to make references to the variable
   * <br> within the context of a single CSP
   */
   private String _name = "";
   
   /**
   * Contains information on possible values the variable may take.
   */
   private CSPRange _range = null;
   
   /**
   * <br> Contains information on possible values the variable may take.
   * <br> Will contain CSPValue objects.
   */
   private Vector _values = null;
   
   /**
   * <br> Identifies the position of the variable within the problem solving context.
   * <br> Contains a series of domain-role-terms.
   */
   private Vector _role = null;
   
   /**
   * <br> Specifies the type of values that the variable takes.
   */
   private Vector _type = null;
   

//----------------------------------------------------------------------
// CONSTRUCTOR
//----------------------------------------------------------------------

   /**
   * Constucts an Empty CSPVariable.
   */
   public CSPVariable()
	 {
	 }
	 
	 /**
	 * <br> Constructs a new CSPVariable and populates it using the values passed
	 * <br> in the Content object.
	 */
	 public CSPVariable(Content content)
	 {
	    assignContentData(content);  
	 }
	 

//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------


   /**
   *  Populates the CSPVariable with the Data in the passed Content object.
   */
   private void assignContentData(Content content)
	 {
	    // Find the set of Attributes contained in the passed Content object
	    Set attributes = content.getAttributeNames();
	    
	    // Retrieve "Type" information
	    if(attributes.contains(TYPE))
	 	  {
	 	     String type = (String) content.getAttribute(TYPE);	
	 	     
	 	     // assign type value retrieved
	 	     addType(type);
	 	  }
	 	  
	 	  // Retrieve "Name" information
	 	  if(attributes.contains(NAME))
	 	  {
	 	  	 // assign name value retrieved
	 	     _name = new String((String) content.getAttribute(NAME));	
	 	  }
	 	  
	 	  // Retrieve "Domain" information
	 	  if(attributes.contains(DOMAIN))
	 	  {
	 	     Content domain = (Content) content.getAttribute(DOMAIN);
	 	     
	 	     // retrieve CSPRange object if present
	 	     Set domain_attributes = domain.getAttributeNames();
	 	     
	 	     if(domain_attributes.contains(CSPRange.CSPRANGE))
	 	  	 {
	 	  	 	  Content range = (Content) domain.getAttribute(CSPRange.CSPRANGE); 
	 	  	 	  
	 	  	 	  // assign retrieved CSPRange
	 	  	 	  _range = new CSPRange(range);
	 	  	 }	
	 	  	 
	 	  } // if(attributes.contains(DOMAIN))
	 	
	 } // assignContentData(Content content)
	 

//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------


   /**
   * <br> Returns a Vector containing the possible values this Variable can take.
   * <br>
   */
   public Vector getValues()
	 {
	    Vector value = new Vector();
	    
	    for(Enumeration e = _values.elements(); e.hasMoreElements(); )
	 	  {
	 	  	 CSPValue next = (CSPValue) e.nextElement();
	 	  	
	 	     value.addElement(next.copy());
	 	  }
	 	  
	 	  return value;
	 	
	 } // getValues()


   /**
   * Adds another the value to the list of values this object can take.
   */
   public void addValue(CSPValue value)
	 {
	    if(_values == null)
	 	  {
	 	     _values = new Vector();
	 	  }
	 	  
	 	  _values.addElement(value.copy());
	 	
	 } // addValue()
   
   
   /**
   *  Returns a Vector of the roles that this CSPVariable may take.
   */
   public Vector getRoles()
	 {
	    Vector value = new Vector();
	    
	    for(Enumeration e = _role.elements(); e.hasMoreElements(); )
	 	  {
	 	     value.addElement(new String((String)e.nextElement()));
	 	  }
	 	  
	 	  return value;
	 	
	 } // getRoles()


   /**
   * Adds another role.
   */
   public void setRole(String str)
	 {
	    if(_role == null)
	 	  {
	 	     _role = new Vector();
	 	  }
	 	  
	 	  _role.addElement(new String(str));
	 	
	 } // setRole()


   /**
   * <br> This retrieves a Vector containing the types that this CSPVariable can take.
   */
   public Vector getTypes()
	 {	 
	    Vector value = new Vector();
	    
	    for(Enumeration e = _type.elements(); e.hasMoreElements(); )
	 	  {
	 	     value.addElement(new String((String)e.nextElement()));
	 	  }
	 	  
	 	  return value;	
	 	
	 } // getTypes()
   
   
   /**
   * <br> This specifies another type for the CSPVariable.
   * <br> This specifies the type of values a CSPVariable can take.
   */
   public void addType(String str)
	 {
	 	  if(_type == null)
	 	  {
	 	     _type = new Vector();
	 	  }
	 	
	 	  _type.addElement(new String(str));
	 	
	 } // addType(String str)
   
   
   /**
   * <br> Retrieves the CSPRange value held in this variable.
   * <br> This specifies the possible values this variable may take.
   */
   public CSPRange getRange()
	 {
	 	  return _range.copy();
	 	
	 } // getRange()


   /**
   * <br> Sets a CSPRange object for the CSPVariable.
   * <br> This specifies the possible values this variable may take.
   */
   public void setRange(CSPRange range)
	 {
	 	  _range = range.copy();
	 	
	 } // setRange(CSPRange range)


   /**
   * Returns the name of the CSPVariable
   */
   public String getName()
	 {
	 	  return new String(_name);
	 	  
	 } // getName()

   
   /**
   * Sets of name for the CSPVariable
   */
   public void setName(String str)
	 {
	    _name = new String(str);	
	    
	 } // setName(String str)
	 	  	

   /**
   * Produces a copy of a CSPVariable object.
   */
   public CSPVariable copy()
	 { 
	    CSPVariable value = new CSPVariable();
	    
	    value.setName(_name);
	    
	    if(_type != null)
	 	  {
	 	     for(Enumeration e = _type.elements(); e.hasMoreElements(); )
	 	  	 {
	 	  	 	  value.addType((String)e.nextElement());
	 	  	 	  
	 	  	 } // for
	 	  	 
	 	  } // if(_type != null)
	    	
	    if(_role != null)
	 	  {
	 	     for(Enumeration e = _role.elements(); e.hasMoreElements(); )
	 	  	 {
	 	  	 	  value.setRole((String)e.nextElement());
	 	  	 	  
	 	  	 } // for
	 	  	 	
	 	  } // if(_role != null)
	    	
	    if(_values != null)
	 	  {
	 	     for(Enumeration e = _values.elements(); e.hasMoreElements(); )
	 	  	 {
	 	  	 	  value.addValue((CSPValue)e.nextElement());
	 	  	 	
	 	  	 } // for	
	 	  	
	 	  } // if(_values != null)
	    	
	    if(_range != null)
	 	  {
	 	     value.setRange(_range);
	  	}   
	    
	    return value;
	 	
	 } // copy()
	 	

   /**
   * Produces a String representation of a CSPVariable.
   */
   public String toString()
	 {
	 	  String type = "";
	 	  String role = "";
	 	  String values = "";
	 	  String range = "";
	 	
	 	  if(_type != null)
	 	  {
	 	     type = _type.toString();
	 	  }
	 	  	
	 	  if(_role != null)
	 	  {
	 	     role = _role.toString();
	 	  }
	 	  
	 	  if(_values != null)
	 	  {
	 	     values = _values.toString();
	 	  }
	 	  
	 	  if(_range != null)
	 	  {
	 	     range = _range.toString();
	 	  }
	 	
	 	
	    String value = "\n\n******************************" +
	    	             "\nCSPVariable:" +
	    	             "\nName: " + _name +
	    	             "\nTypes: " + type +
	    	             "\nRoles: " + role +
	    	             "\nValues: " + values +
	    	             "\nRange: " + range +
	    	             "\n\n******************************\n\n";
	    	             	
	    return value;
	 	
	 } // toString()	
	 	
	 	
	 /**
	 *  Returns a Content object populated using the Data held in this CSPVariable.
	 */
	 public Content toContent()
	 {
	    Content content = new Content(CSPVARIABLE);
	    
	    // add type information if present
	    if(_type != null)
	 	  {
	 	  	 String types = "";
	 	  	 
	 	  	 if(_type.size() == 1)
	 	  	 {
	 	  	    types = (String)_type.firstElement();
	 	  	 }
	 	  	 else
	 	  	 {
	 	  	    types = List.toString(_type);
	 	  	 }
	 	  	
	 	     content.addAttribute(TYPE, types);
	 	     
	 	  } // if(_type != null)
	    
	    // adds Name information
	    if(!_name.equals(""))
	 	  {
	 	     content.addAttribute(NAME, _name);
	 	  }
	 	  
	 	  // Creates a new Content object to represent "Domain" information
	 	  Content domain = new Content(DOMAIN);
	 	  
	 	  // Adds domain information.
	 	  // Note: Will have to add code to handle domain information contained
	 	  // within a list of CSPValues.
	 	  if(_range != null)
	 	  {
	 	  	 Content range = _range.toContent();
	 	  	 domain.addAttribute(CSPRange.CSPRANGE, range);
	 	     content.addAttribute(DOMAIN, domain);	
	 	  	
	 	  } // if(_range != null)  
	    
	    return content;	
	 	
	 } // toContent()
	 	
	
//----------------------------------------------------------------------
// MAIN/TESTING METHOD
//----------------------------------------------------------------------
  
   /**
   * Used for testing.
   */
   public static void main(String args[])
	 {
	    System.out.println();
	    System.out.println("Testing CSPVariable class");
	    
	    System.out.println("Creating new CSPVariable");
	    CSPVariable variable = new CSPVariable();
	    
	    System.out.println("Creating new CSPRange");
	    CSPRange range = new CSPRange();
	 	  range.addRange("{house, car, boat}");	 	
	 	  
	 	  variable.setRange(range); 
	 	  System.out.println("Range:");
	 	  System.out.println(variable.getRange()); 
	 	  System.out.println();
	 	  
	 	  System.out.println("Creating new CSPValue");
	 	  CSPValue value = new CSPValue();
	    value.addElement("blue");
	    value.addElement("red");
	    value.addElement("green");
	    value.addElement("purple");
	    
	    value.addTag("special");
	    value.addTag("reserved");
	    
	    variable.addValue(value);
	    System.out.println();
	    System.out.println("Values:");
	    System.out.println(variable.getValues());
	    System.out.println();
	    
	    variable.setName("auction");
	    System.out.println();
	    System.out.println("Name:");
	    System.out.println(variable.getName());
	    	
	    	
	    variable.addType("hot");
	    variable.addType("cold");
	    System.out.println();
	    System.out.println("Types:");
	    System.out.println(variable.getTypes());
	    
	    variable.setRole("lead");
	    variable.setRole("follow");
	    System.out.println();
	    System.out.println("Roles:");
	    System.out.println(variable.getRoles());
	    
	    System.out.println();
	    System.out.println("Contents of CSPVariable:");
	    System.out.println(variable);
	    
	    System.out.println();
	    System.out.println("Contents of copy of CSPVariable:");
	    System.out.println(variable.copy());
	    System.out.println();
	   
	    
	    System.out.println("CSPVariable represented as a Content object:");
	    System.out.println(variable.toContent());
	    System.out.println();
	    
	    System.out.println("Constructing Content object to populate CSPVariable");
	    
	    Content range_content = new Content("CSP-range");
	 	  range_content.setAttribute("Range", "Car");
	 	  
	 	  Content value_content = new Content("List-range");
	 	  value_content.setAttribute("Values", "{(Renault, 1000), (Fiat, 2000)}");
	 	  range_content.setAttribute("List-range", value_content);
	 	  
	 	  Content domain = new Content(DOMAIN);
	 	  domain.addAttribute(CSPRange.CSPRANGE, range_content);
	 	  
	 	  Content var_content = new Content(CSPVARIABLE);
	 	  var_content.addAttribute(TYPE, "Car");
	 	  var_content.addAttribute(NAME, "buyer_request");
	 	  var_content.addAttribute(DOMAIN, domain);
	    
	    System.out.println();
	    System.out.println("Values in Content object:");
	    System.out.println(var_content);
	    System.out.println();
	    
	    System.out.println("Constucting CSPVariable");
	    
	    CSPVariable var = new CSPVariable(var_content);
	    
	    System.out.println();
	    System.out.println("Values contained in new CSPVariable");
	    System.out.println(var);
	    System.out.println();
	    
	    
	    CSPVariable constraint_var = new CSPVariable();
	    
	    CSPRange constraint_range = new CSPRange();
	    constraint_range.setRangeType("Constraint");
	    
	    Vector constraint = List.parseList("{(Fiat, 1000), (Renault, 1000)}");
	 	  constraint_range.addRange(constraint);	 
	 	  constraint_var.setRange(constraint_range);	
	 	  
	 	  System.out.println();
	 	  System.out.println("Constraint Variable:");
	 	  System.out.println(constraint_var);
	 	
	 } // main
	
} // CSPVariable

	
