//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files ("the
//     Programs") are Copyright (c) 1999 - 2000 Nortel Networks Corporation, 
//     8200 Dixie Road, Suite 100, Brampton, Ontario, Canada L6R 5P6.  
//     All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out in
//     the Nortel Networks FIPA-OS Public License, a copy of which can be found
//     in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version can
//     also be found at http://fipa-os.sourceforge.net/. By downloading the
//     files containing the Programs you accept the terms and  conditions of the
//     Public License. You do not have to accept these terms and conditions, but
//     unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.skill.constraint.ccl.variable;

//----------------------------------------------------------------------

import fipaos.ont.fipa.FIPACONSTANTS;
import fipaos.ont.BASETOKENS;
import fipaos.ont.fipa.fipaman.FIPAMANCONSTANTS;

import java.util.*;

import fipaos.skill.constraint.ccl.variable.CSPValue;

import fipaos.agent.conversation.content.Content;

//----------------------------------------------------------------------

/**
 * 
 * <br> This class describes the object CSP-variable-assignment as described in the document 
 * <br> "CCL Language Specification v2.01" section 4.2.2.5, p13.
 * <br>
 * <br> The variable named in the Variable-name slot is assigned the value given in the
 * <br> Variable-value slot. This represents a variable instantiation- a choice being made.
 *
 */
 
public class CSPVariableAssignment
{
     
//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------

   /**
   * <br> Name that a CSPVariableAssignment is referenced with when represented
   * <br> as a Content object.
   */
   public final static String CSPVARIABLEASSIGNMENT = "CSP-variable-assignment";
   
   /**
   *  Name of the attribute a "Name" value is represented as in Content objects.
   */
   public final static String NAME  = "Name";

//----------------------------------------------------------------------
// STATIC VARIABLES
//----------------------------------------------------------------------
		
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

   /**
   *  The name of the variable having a value assigned to it.
   */
   private String _variable_name = "";
   
   /**
   * <br> The value being assigned. The value assigned must match with the type of
   * <br> the variable.
   */
   private CSPValue _value = null;

//----------------------------------------------------------------------
// CONSTRUCTOR
//----------------------------------------------------------------------

   /**
   * Constucts an empty CSPVariableAssignment
   */
   public CSPVariableAssignment()
	 {
	 } // CSPVariableAssignment()
	 	
	 	
	 /**
	 * <br> Constructs a new CSPVariableAssignment and populates it with 
	 * <br> Data held in the passed Content object.
	 */
	 public CSPVariableAssignment(Content content)
	 {
	    assignContentData(content);
	 }

//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------


   /**
   * <br> Takes the Data held in the passed content object and assigns it to the 
   * <br> variables held within this object.
   */
   private void assignContentData(Content content)
	 {
	    // obtain the names of all the attributes held in the passed Content object
	    Set attributes = content.getAttributeNames();
	    
	    // System.out.println("Attributes: " + attributes);
	    
	    // retrieves name value
	    if(attributes.contains(NAME))
	 	  {
	 	     variableName((String) content.getAttribute(NAME));	
	 	  }
	 	  
	 	  // retrieves CSPValue 
	 	  if(attributes.contains(CSPValue.CSPVALUE))
	 	  {
	 	  	 Content value = (Content) content.getAttribute(CSPValue.CSPVALUE);
	 	     addCSPValue(new CSPValue(value));
	  	}
	 	
	 } // assignContentData(Content content)


//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------


   /**
   * <br> This is the value that is to be assigned.
   * <br> The value assigned must match with the type of variable.
   */
   public void addCSPValue(CSPValue value)
	 {
	 	  _value = value.copy();
	 	
	 } // addCSPValue(CSPValue value)
	 	
	 /**
	 * Returns the CSPValue this object contains
	 */
	 public CSPValue getCSPValue()
	 {
	    return _value.copy();
	 }	
	 	
	 	
	 /**
   * <br> Assigns a name to the value.
   */
	 public void variableName(String str)
	 {
	    _variable_name = new String(str);
	 	
	 } // variableName(String str)
	 	
	 	
	 /**
	 * Returns the name of the value.
	 */
	 public String getVariableName()
	 {
	    return new String(_variable_name);
	 }
	 	
	 
	 /**
	 *  Returns a copy of the CSPVariableAssignment object.
	 */ 
	 public CSPVariableAssignment copy()
	 {
	    CSPVariableAssignment value = new CSPVariableAssignment();
	    value.variableName(_variable_name);	
	 	
	 	  if(_value != null)
	 	  {
	 	  	 value.addCSPValue(_value.copy());
	 	  }
	 	  
	 	  return value;
	 	
	 } // copy()
	 	
	 
	 /**
	 * <br> Returns a String representation of this object.
	 */
	 public String toString()
	 {
	 	  String value = "\n\n******************************\n" +
	 	  	              "CSPVariableAssignment\n" +
	 	  	              "Name: " + _variable_name + "\n" +
	 	  	              "Contains CSPValue: " + _value + "\n" +
	 	  	               "******************************\n\n";
	 	
	 	  return value;
	 	
	 } // toString()
	 	
	 	
	 /**
	 * <br> Returns a Content object populated with the Data held in this class
	 * <br> that represents the CSPVariableAssignment.
	 */
	 public Content toContent()
	 {
	    Content value = new Content(CSPVARIABLEASSIGNMENT);
	    
	    if(_value != null)
	 	  {
	 	     Content csp_value = _value.toContent();
	 	     value.addAttribute(CSPValue.CSPVALUE, csp_value);
	 	  }
	 	  
	 	  if(!_variable_name.equals(""))
	 	  {
	 	     value.addAttribute(NAME, _variable_name);
	 	  }
	    
	    return value;	
	 	
	 } // toContent()
	 	
	
//----------------------------------------------------------------------
// MAIN/TESTING METHOD
//----------------------------------------------------------------------

   /**
   * Used for testing.
   */
   public static void main(String args[])
	 {
	    System.out.println();
	    System.out.println("Testing CSPVariableAssignment class.");
	    System.out.println();   
	    
	    System.out.println("Creating new CSPVariableAssignment object");
	    CSPVariableAssignment var = new CSPVariableAssignment();
	    
	    System.out.println("populating");
	    
	    var.variableName("bob");
	    
	    CSPValue value = new CSPValue();
	    value.addElement("blue");
	    value.addElement("red");
	    value.addElement("green");
	    value.addElement("purple");
	    
	    value.addTag("special");
	    value.addTag("reserved");
	    
	    var.addCSPValue(value);
	    
	    System.out.println();
	    System.out.println("Values held in object are:");
	    System.out.println(var);
	    System.out.println();
	    
	    System.out.println("Values held in a copy of this object are:");
	    System.out.println(var.copy());
	    System.out.println();
	    
	    System.out.println("CSPValue represented as a Content object:");
	    System.out.println(var.toContent());
	    System.out.println();
	    
	    System.out.println("Constructing CSPVariableAssignment using Content object");
	    
	    Content content = new Content(CSPValue.CSPVALUE);
      content.addAttribute(CSPValue.NPART, new Long(1));
      
      Content value_content = new Content(CSPValue.ELEMENTS);
      value_content.addAttribute(CSPValue.VALUE, "{Renault}");
      content.addAttribute(CSPValue.ELEMENTS, value_content);
      
      Content csp_value = new Content(CSPVARIABLEASSIGNMENT);
      csp_value.addAttribute(NAME, "transaction_item");
      csp_value.addAttribute(CSPValue.CSPVALUE, content);
      
      System.out.println("Values held in content object:");
      System.out.println(csp_value);
      
      CSPVariableAssignment assign = new CSPVariableAssignment(csp_value);
      
      System.out.println();
      System.out.println("Values held in new CSPVariableAssignment object:");
      System.out.println(assign);
	    
	    
	    System.out.println();
	    
	    
	    
	 	
	 } // main
  
	
} // CSPVariableAssignment

	
