//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files ("the
//     Programs") are Copyright (c) 1999 - 2000 Nortel Networks Corporation, 
//     8200 Dixie Road, Suite 100, Brampton, Ontario, Canada L6R 5P6.  
//     All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out in
//     the Nortel Networks FIPA-OS Public License, a copy of which can be found
//     in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version can
//     also be found at http://fipa-os.sourceforge.net/. By downloading the
//     files containing the Programs you accept the terms and  conditions of the
//     Public License. You do not have to accept these terms and conditions, but
//     unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.skill.constraint.message;

//----------------------------------------------------------------------

// import fipaos.fipa.FIPACONSTANTS;
// import fipaos.ont.BASETOKENS;
// import fipaos.ont.fipaman.FIPAMANCONSTANTS;

import fipaos.agent.conversation.content.Content;

import java.util.*;
import java.io.*;

import fipaos.skill.constraint.ccl.object.CSP;
import fipaos.skill.constraint.ccl.object.CSPSolution;

import fipaos.parser.xml.XMLContentHandler;
import fipaos.parser.*;
import fipaos.util.DIAGNOSTICS;


//----------------------------------------------------------------------

/**
 * 
 * <br> This class acts as an interface between text messages, a tree of parsed
 * <br> Content objects and the CCLMessage class.
 * <br> The purpose of the CCLMessage class is to allow the values held in a 
 * <br> CCLMessage to be easily retrieved.
 *
 */
public class CCLMessage
{
     
//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------

   // Used to facilitate the use of switch statements to determine the type
   // of object a Content object represents.
   
   public final static int CONTENT = 1;
   public final static int LIST = 2;
   public final static int LONG = 3;
   public final static int STRING = 4;
   
   

//----------------------------------------------------------------------
// STATIC VARIABLES
//----------------------------------------------------------------------
		
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

   /**
   * <br> Reference to a CSP the message may contain.
   */
   private CSP _csp = null;
   
   /**
   * <br> Reference to a CSPSolution the message may contain.
   */
   private CSPSolution _solution = null;

//----------------------------------------------------------------------
// CONSTRUCTOR
//----------------------------------------------------------------------

   /**
   * <br> Constructs a new CCLMessage object using the information held in the passed
   * <br> String to populate it.
   * <br> This String is parsed into a Content object which is then used to populate
   * <br> the CCLMessage class with Values.
   */
   public CCLMessage(String str)
	 {
	 	
	 	  Parser xml_parser = new XMLContentHandler();
	    
	    // parse String to form a tree of Content objects using the XMLContentHandler
			try
			{
				Content root = xml_parser.parse(str);
				
				Content expression = (Content)root.getAttribute("Expression");	
				Content object = (Content)expression.getAttribute("Object");

				// A CCLMessage may contain either a CSP, a CSPSolution or a CSPSolutionList.

				Content ccl_data = null;

				// retrieve the set of attributes the object contains
				Set attributes = object.getAttributeNames();

				// Retrieve and assign CSP
				if(attributes.contains(CSP.CSP))
				{
					 Content csp = (Content)object.getAttribute(CSP.CSP);
					 _csp = new CSP(csp);
				}

				// Retrieve and assign CSP-solution
				if(attributes.contains(CSPSolution.CSPSOLUTION))
				{
					 Content solution = (Content)object.getAttribute(CSPSolution.CSPSOLUTION);
					 _solution = new CSPSolution(solution);
				}
			}
			catch( ParserException t )
			{
				DIAGNOSTICS.println( "An exception occured whilst parsing a message", t, DIAGNOSTICS.LEVEL_MAX );
			}
	    
	 } // CCLMessage(String str)
	 	
	

//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------
	 	

//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------


   /**
   * <br> Takes the passed Content object and converts it into a String which
   * <br> it then returns.
   * <br> The passed Content object should be a representation of a CSP, CSPSolution
   * <br> or a CSPSolutionList.
   */	
   public static String toTextMessage(Content content)
	 {
	 	  Parser xml_parser = new XMLContentHandler();
	 	  
	 	  // Element that contains everything else in CCL.
	 	  Content message = new Content("Expression");
	 	  Content object = new Content("Object");
	 	  message.addAttribute("Object", object);
	 	  
	 	  // Test to see whether passed Content object represents a CSP or CSPSolution
	 	  String type = content.getName();
	 	  
	 	  if(type.equals(CSP.CSP))
	 	  {
	 	     object.addAttribute(CSP.CSP, content);
	 	     object.addAttribute("Name", "CSP");  
	 	  }
	 	  
	 	  if(type.equals(CSPSolution.CSPSOLUTION))
	 	  {
	 	     object.addAttribute(CSPSolution.CSPSOLUTION, content);
	 	     object.addAttribute("Name", "CSPSolution");
	 	  }
	 	  
			try
			{
		 	  return xml_parser.deparse(message);
			}
			catch( ParserException t )
			{
				DIAGNOSTICS.println( "An exception occured whilst parsing a message", t, DIAGNOSTICS.LEVEL_MAX );
			}
	 	
			return null;
	 } // toTextMessage(Content content)
	 	
	 	
	 /**
	 *  Returns a CSP the message may contain.
	 */
   public CSP getCSP()
	 {	
	    return _csp.copy();
	 }
   
   
   /**
   *  Returns a CSPSolution the message may contain.
   */
   public CSPSolution getCSPSolution()
	 {
	 	  return _solution.copy();
	 }
	 
	 
	 /**
	 *  Returns true if the message contains a CSP
	 */
	 public boolean CSPPresent()
	 {
	    return (_csp != null);
	 }
	 
	 
	 /**
	 *  Returns true if the message contains a CSPSolution
	 */
	 public boolean CSPSolutionPresent()
	 {
	    return (_solution != null);
	 }
   
   
	
//----------------------------------------------------------------------
// MAIN/TESTING METHOD
//----------------------------------------------------------------------

   public static void main(String args[])
	 {
	 	  BufferedReader fin = null;
	 	  
	 	  System.out.println();
	 	
	 	  if( args.length <= 0 )
		  {
			   System.out.println("Usage: java fipaos.skill.constraint.message.CCLMessage [<filename>]");
		   	 System.exit(-10);
		  }
		  else
	 	  {
		     System.out.println("Reading in the specified file: " + args[0]);
		     // Read in the given file
		     
		     try
		  	 {
		        fin = new BufferedReader( new FileReader( args[0] ) );
		     }
		  	 catch(FileNotFoundException e)
		  	 {
		  	    System.out.println("Error reading file: " + e.toString());
		  	 }
	 	  }
		
		  // String to build up input
		  String in = new String();
		
		  try
	 	  {
		
		     String next = fin.readLine();
		     while( next != null )
		     {
			      in += next + "\n";
			      next = fin.readLine();	
		     }
		     
		     // Close the File
		     fin.close();
		     
		     System.out.println("Completed reading from file.");
		  
	 	  }
	 	  catch(IOException e)
	 	  {
	 	     System.out.println("Error reading from input stream: " + e.toString());
	 	  }
	 	  
	 	  // Create new CCLMessage using the passed String
	 	  CCLMessage message = new CCLMessage(in);
	 	  
	 	  // Handle message if it contains a CSP
	 	  if(message.CSPPresent())
	 	  {
	 	  	 CSP csp = message.getCSP();
	 	  	
	 	     System.out.println();
	 	     System.out.println("CSP:");
	 	     System.out.println(csp);
	 	     System.out.println();
	 	     
	 	     // retrieve Content object and convert to a text message
	 	     String text_message = toTextMessage(csp.toContent());
	 	     
	 	     System.out.println();
	 	     System.out.println("Text representation of CCL Message containing CSP");
	 	     System.out.println(text_message);
	 	     System.out.println();
	 	  }
	 	  
	 	  // Handle message if it contains a CSPSolution
	 	  if(message.CSPSolutionPresent())
	 	  {
	 	  	 CSPSolution sol = message.getCSPSolution();
	 	  	
	 	     System.out.println();
	 	     System.out.println("CSPSolution:");
	 	     System.out.println(sol);
	 	     System.out.println();
	 	     
	 	     // retrieve Content object and convert to a text message
	 	     String text_message = toTextMessage(sol.toContent());
	 	     
	 	     System.out.println();
	 	     System.out.println("Text representation of CCL Message containing CSPSolution");
	 	     System.out.println(text_message);
	 	     System.out.println();
	 	  }
	 	  
	 	   
	 	  
	 	  
	 	  
	 	
	 	
	 } // main
  
	
} // CCLMessage

	
