//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files ("the
//     Programs") are Copyright (c) 1999 - 2000 Nortel Networks Corporation, 
//     8200 Dixie Road, Suite 100, Brampton, Ontario, Canada L6R 5P6.  
//     All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out in
//     the Nortel Networks FIPA-OS Public License, a copy of which can be found
//     in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version can
//     also be found at http://fipa-os.sourceforge.net/. By downloading the
//     files containing the Programs you accept the terms and  conditions of the
//     Public License. You do not have to accept these terms and conditions, but
//     unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.skill.db;

//----------------------------------------------------------------------

import java.util.Enumeration;

  /**
   * This is the interface that all database implementations should implement
   * to provide the ability to change the database used at runtime
   *
   * <br>
   * <br>
   */

public interface Database
{

//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// STATIC VARIABLES
//----------------------------------------------------------------------
		
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// CONSTRUCTOR
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------

	/**
	 * Method to create the database store.  Either this method or openDatabase()
	 * should be called before attempting to interact with the Database
	 */
	public void createDatabase() throws DatabaseException;


	/**
	 * Method to re-open the database store.  Either this method or createDatabase()
	 * should be called before attempting to interact with the Database
	 */
	public void openDatabase() throws DatabaseException;


	/**
	 * Method to close the database store.  Should only be called after openDatabase()
	 * or createDatabase() have been used and after all necessary transactions have been
	 * made.  Transactions cannot take place again until the database is re-opened.
	 */
	public void closeDatabase() throws DatabaseException;


	/**
	 * Method to remove the database store.  The database will be closed if necessary before
	 * the store is removed.  The database can only be used again once a new store has been
	 * created.
	 */
	public void deleteDatabase() throws DatabaseException;


	/**
	 * Method to add the given DatabaseObject to the database, in the given relation.
	 * 
	 * @param obj The DatabaseObject to add
	 */
	public void addObject( DatabaseObject obj ) throws DatabaseException;

	
	/**
	 * Method to locate the object in the given relation with the given UID
	 *
	 * @param unique_id The unique id of the DatabaseObject to find
	 */
	public DatabaseObject findObject( String unique_id ) throws DatabaseException;


	/**
	 * Method to update the given DatabaseObject in the Database
	 *
	 * @param obj The DatabaseObject to update
	 */
	public void updateObject( DatabaseObject obj ) throws DatabaseException;


	/**
	 * Method to remove the given DatabaseObject from the Database
	 *
	 * @param obj The DatabaseObject to remove
	 */
	public void removeObject( DatabaseObject obj ) throws DatabaseException;
	
	
	/**
	 * Method to remove the DatabaseObject from the Database with the given UID
	 *
	 * @param unique_id The unique id of the DatabaseObject to remove
	 */
	public void removeObject( String unique_id ) throws DatabaseException;	

	
	/**
	 * Creates the given relation within the Database
	 *
	 * @param relation The name of the relation to create
	 */
	public void createRelation( String relation ) throws DatabaseException;


	/**
	 * Deletes the given relation from the Database
	 *
	 * @param relation The name of the relation to delete
	 */
	public void deleteRelation( String relation ) throws DatabaseException;


	/**
	 * Indicates that a transaction is about to start.  THIS MUST BE CALLED
	 * BEFORE THE DATABASE IS MANIPULATED.
	 *
	 * @param type If true, the transaction will be read-only
	 */
	public void startTransaction( boolean readonly ) throws DatabaseException;


	/**
	 * Indicates that a transaction has finished.  THIS MUST BE CALLED
	 * AFTER THE DATABASE HAS BEEN MANIPULATED IN ORDER TO COMMIT CHANGES.
	 */
	public void endTransaction() throws DatabaseException;
		
		
	/**
	 * Method to return an enumeration of the keys in the database
	 *
	 * @return An Enumeration of the ID's of the objects in the database
	 */
	public Enumeration getIDs() throws DatabaseException;
	
//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------

}
