//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Emorphia are Copyright (C) 2001 Emorphia Limited
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.skill.db;

//----------------------------------------------------------------------

import java.util.*;
import fipaos.agent.profile.*;
import fipaos.agent.conversation.content.*;

/**
 * Simple testharness to stress test a particular Database implementation
 */
public class DatabaseTestHarness
{
	private static List _writes = new LinkedList();
	private static List _reads = new LinkedList();
	private static List _lists = new LinkedList();
	private static List _deletes = new LinkedList();
	
	public static void main( String args[] )
	{
		// Check args
		if( args.length < 1 )
		{
			System.out.println( "Usage: java " + DatabaseTestHarness.class.getName() + " [<db-name>]" );
			System.exit( -10 );
		}
		
		try
		{
			// Create database
			DatabaseProfile dbp = new DatabaseProfileImpl();
			dbp.setDatabaseLocation( System.getProperty( "user.dir" ) );
			dbp.setDatabaseType( args[0] );
			Database db = DatabaseFactory.getDatabase( "test-db", dbp );
			db.createDatabase();
			
			// Execute tests
			testWrites( db, 1000 );
			System.out.println( "Average write time = " + calcAverage( _writes ) + " ms" );
			testReads( db, 1000 );
			System.out.println( "Average read time = " + calcAverage( _reads ) + " ms" );
			testLists( db, 1000 );
			System.out.println( "Average list time = " + calcAverage( _lists ) + " ms" );
			testDeletes( db, 500 );
			System.out.println( "Average delete time = " + calcAverage( _deletes ) + " ms" );
			System.out.println( "Number of entries left in database = " + calcSize( db ) );
		}
		catch( Throwable t )
		{
			System.exit( -20 );
		}
	}
	
	
	private static void testWrites( Database db, int number ) throws DatabaseException
	{
		System.out.println( "Testing " + number + " writes...." );

		// Prefix for object names
		String pre = "test-object";
		
		for( int i = 0; i < number; i++ )
		{
			// Create DBTestObject
			DatabaseObject dbo = new DBTestObject( pre + i );
			
			// Start timer
			Timer t = new Timer();
			
			db.startTransaction( false );
			db.addObject( dbo );
			db.endTransaction();
			
			// Stop timer
			t.stop();
			
			_writes.add( t );
		}
	}
	
	
	private static void testReads( Database db, int number ) throws DatabaseException
	{
		System.out.println( "Testing " + number + " reads...." );

		// Prefix for object names
		String pre = "test-object";
		
		for( int i = 0; i < number; i++ )
		{
			// Start timer
			Timer t = new Timer();
			
			db.startTransaction( true );
			db.findObject( pre + i );
			db.endTransaction();
			
			// Stop timer
			t.stop();
			
			_reads.add( t );
		}
	}
	
	
	private static void testLists( Database db, int number ) throws DatabaseException
	{
		System.out.println( "Testing " + number + " \"lists\"...." );

		for( int i = 0; i < number; i++ )
		{
			// Start timer
			Timer t = new Timer();
			
			db.startTransaction( true );
			
			
			int count = 0;
			Enumeration enum = db.getIDs();
			while( enum.hasMoreElements() )
			{
				enum.nextElement();
				count++;
			}
			
			if( count != number )
			{
				System.out.println( "Only found " + count + " objects in database" );
			}
			
			
			db.endTransaction();
			
			// Stop timer
			t.stop();
			
			_lists.add( t );
		}
	}
	
	
	private static void testDeletes( Database db, int number ) throws DatabaseException
	{
		System.out.println( "Testing " + number + " deletes...." );

		db.startTransaction( true );

		int count = 0;
		Enumeration enum = db.getIDs();
		while( enum.hasMoreElements() && count < number )
		{
			String id = (String) enum.nextElement();
			
			// Start timer
			Timer t = new Timer();

			// Delete object
			db.removeObject( id );
			
			// Stop timer
			t.stop();

			// Increment counter
			count++;
			
			_deletes.add( t );
		}

		db.endTransaction();
	}
	
	
	private static int calcSize( Database db ) throws DatabaseException
	{
		db.startTransaction( true );


		int count = 0;
		Enumeration enum = db.getIDs();
		while( enum.hasMoreElements() )
		{
			enum.nextElement();
			count++;
		}

		db.endTransaction();
		
		return count;
	}
	
		private static long calcAverage( List timers )
	{
		// Iterate through list, calculating average
		double result = 0;
		long size = timers.size();
		
		Iterator i = timers.iterator();
		while( i.hasNext() )
		{
			Timer curr = (Timer) i.next();
			result += (double) curr.time() / (double) size;
		}
		
		return (long) result;
	}
	
	private static class DBTestObject implements DatabaseObject
	{
		byte[] content;
		String name;
		
		public DBTestObject( String name )
		{
			content = new byte[4096];
			this.name = name;
		}
		
		public String getObjectID()
		{
			return name;
		}
	}
	
	public static class Timer
	{
		private long start;
		private long stop;
		
		public Timer()
		{
			start = System.currentTimeMillis();
		}
		
		public void stop()
		{
			if( stop == 0 )
			{
				stop = System.currentTimeMillis();
			}
		}
		
		public long time()
		{
			if( stop != 0 )
			{
				return stop - start;
			}
			else
			{
				return -1;
			}
		}
		
		public String toString()
		{
			return time() + "ms";
		}
	}
}