//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files ("the
//     Programs") are Copyright (c) 1999 - 2000 Nortel Networks Corporation, 
//     8200 Dixie Road, Suite 100, Brampton, Ontario, Canada L6R 5P6.  
//     All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out in
//     the Nortel Networks FIPA-OS Public License, a copy of which can be found
//     in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version can
//     also be found at http://fipa-os.sourceforge.net/. By downloading the
//     files containing the Programs you accept the terms and  conditions of the
//     Public License. You do not have to accept these terms and conditions, but
//     unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.skill.db.memorydatabase;

//----------------------------------------------------------------------

import java.util.Hashtable;
import java.util.Enumeration;

import fipaos.skill.db.Database;
import fipaos.skill.db.DatabaseObject;
import fipaos.agent.profile.*;
import fipaos.skill.db.*;

/**
 * This Database implementation provides a simple non-persistent database
 *
 * <br>
 * <br>
 */
public class MemoryDatabase implements Database
{
     
//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// STATIC VARIABLES
//----------------------------------------------------------------------
	
	/**
	 * Contains all databases controlled by the MemoryDatabase
	 */
	private static Hashtable __databases = new Hashtable();
	
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

	/**
	 * The Hashtable that this represents as the DB
	 */
	private Hashtable _our_hashtable;

	/**
	 * This Database name
	 */
	private String _db_name;

	/**
	 * If true, we're in a transaction and another cannot begin
	 */
	private boolean _in_transaction = false;

	/**
	 * Lock for our DB
	 */
	private DatabaseInfo _database_info;

	/**
	 * Thread which is in the transaction
	 */
	private Thread _in_control_thread;

	/**
	 * DatabaseProfile to use
	 */
	private DatabaseProfile _db_profile;
	
//----------------------------------------------------------------------
// CONSTRUCTOR
//----------------------------------------------------------------------

	/**
	 * Default constructor
	 */	
	public MemoryDatabase( String name, DatabaseProfile db_profile )
	{
		_db_name = name;
		_db_profile = db_profile;
	}

//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------

	/**
	 * Method to create the database store.  Either this method or openDatabase()
	 * should be called before attempting to interact with the Database
	 */
	public void createDatabase()
	{
		synchronized ( __databases )
		{
			_our_hashtable = new Hashtable();
			_database_info = new DatabaseInfo();
			_database_info._db = _our_hashtable;
			
			__databases.put( _db_name, _database_info );
			_in_transaction = false;
		}
	}


	/**
	 * Method to re-open the database store.  Either this method or createDatabase()
	 * should be called before attempting to interact with the Database
	 */
	public void openDatabase()
	{
		synchronized ( __databases )
		{
			_database_info = (DatabaseInfo) __databases.get( _db_name );

			_in_transaction = false;
		
			if( _database_info == null )
			{
				createDatabase();
			}
			else
			{
				_our_hashtable = _database_info._db;
			}
		}
	}


	/**
	 * Method to close the database store.  Should only be called after openDatabase()
	 * or createDatabase() have been used and after all necessary transactions have been
	 * made.  Transactions cannot take place again until the database is re-opened.
	 */
	public void closeDatabase()
	{
		_our_hashtable = null;
	}


	/**
	 * Method to remove the database store.  The database will be closed if necessary before
	 * the store is removed.  The database can only be used again once a new store has been
	 * created.
	 */
	public void deleteDatabase()
	{
		synchronized ( __databases )
		{
			_our_hashtable = null;
			__databases.remove( _db_name );
		}
	}


	/**
	 * Method to add the given DatabaseObject to the database, in the given relation.
	 * 
	 * @param obj The DatabaseObject to add
	 */
	public void addObject( DatabaseObject obj ) throws DatabaseException
	{
		// Check we're in a transaction
		if( !_in_transaction )
		{
			throw new DatabaseException( "Error - no transaction associated with DB!" );
		}
		
		// Get the UID of the object to add, and use that as the key to insert
		// the object into the hashtable
		_our_hashtable.put( obj.getObjectID(), obj );
	}

	
	/**
	 * Method to locate the object in the given relation with the given UID
	 *
	 * @param unique_id The unique id of the DatabaseObject to find
	 */
	public DatabaseObject findObject( String unique_id ) throws DatabaseException
	{
		// Check we're in a transaction
		if( !_in_transaction )
		{
			throw new DatabaseException( "Error - no transaction associated with DB!" );
		}

		// Get the object with the given UID
		return (DatabaseObject) _our_hashtable.get( unique_id );
	}


	/**
	 * Method to update the given DatabaseObject in the Database
	 *
	 * @param obj The DatabaseObject to update
	 */
	public void updateObject( DatabaseObject obj ) throws DatabaseException
	{
		// Check we're in a transaction
		if( !_in_transaction )
		{
			throw new DatabaseException( "Error - no transaction associated with DB!" );
		}

		if( _our_hashtable.get( obj.getObjectID() ) == null )
		{
			throw new DatabaseException( "Error - cannot update object, it doesn't exist in DB" );
		}

		addObject( obj );
	}


	/**
	 * Method to remove the given DatabaseObject from the Database
	 *
	 * @param obj The DatabaseObject to remove
	 */
	public void removeObject( DatabaseObject obj ) throws DatabaseException
	{
		// Check we're in a transaction
		if( !_in_transaction )
		{
			throw new DatabaseException( "Error - no transaction associated with DB!" );
		}

		_our_hashtable.remove( obj.getObjectID() );
	}
	
	
	/**
	 * Method to remove the DatabaseObject from the Database with the given UID
	 *
	 * @param unique_id The unique id of the DatabaseObject to remove
	 */
	public void removeObject( String unique_id ) throws DatabaseException
	{
		// Check we're in a transaction
		if( !_in_transaction )
		{
			throw new DatabaseException( "Error - no transaction associated with DB!" );
		}

		_our_hashtable.remove( unique_id );
	}	


	/**
	 * Creates the given relation within the Database
	 *
	 * @param relation The name of the relation to create
	 */
	public void createRelation( String relation )
	{
	}


	/**
	 * Deletes the given relation from the Database
	 *
	 * @param relation The name of the relation to delete
	 */
	public void deleteRelation( String relation )
	{
	}


	/**
	 * Indicates that a transaction is about to start.  THIS MUST BE CALLED
	 * BEFORE THE DATABASE IS MANIPULATED.
	 */
	public void startTransaction( boolean readonly )
	{
		synchronized ( _database_info )
		{
			// Wait until we can lock the database
			while( ( _database_info._lock != 0 ) && ( _in_control_thread != Thread.currentThread() ) )
			{
				try
				{
					_database_info.wait();
				}
				catch ( InterruptedException e )
				{
				}	
			}

			fipaos.util.DIAGNOSTICS.println( Thread.currentThread().getName() + " - Started transaction" );

			// Lock DB
			_database_info._lock++;
			_in_transaction = true;

			// Record which thread has access
			_in_control_thread = Thread.currentThread();
		}
	}


	/**
	 * Indicates that a transaction has finished.  THIS MUST BE CALLED
	 * AFTER THE DATABASE HAS BEEN MANIPULATED IN ORDER TO COMMIT CHANGES.
	 */
	public void endTransaction()
	{
		synchronized ( _database_info )
		{
			fipaos.util.DIAGNOSTICS.println( Thread.currentThread().getName() + " - Ending transaction" );

			// Unlock DB
			_database_info._lock--;
			_database_info.notify();

			if( _database_info._lock == 0 )
			{
				_in_transaction = false;
				_in_control_thread = null;
			}
		}
	}

	public class DatabaseInfo
	{
		public Hashtable _db;
		public int _lock = 0;
	}

	/**
	 * Method to return an enumeration of the keys in the database
	 *
	 * @return An Enumeration of the ID's of the objects in the database
	 */
	public Enumeration getIDs() throws DatabaseException
	{
		// Check we're in a transaction
		if( !_in_transaction )
		{
			throw new DatabaseException( "Error - no transaction associated with DB!" );
		}

		return _our_hashtable.keys();
	}
}
