//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files ("the
//     Programs") are Copyright (c) 1999 - 2000 Nortel Networks Corporation, 
//     8200 Dixie Road, Suite 100, Brampton, Ontario, Canada L6R 5P6.  
//     All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out in
//     the Nortel Networks FIPA-OS Public License, a copy of which can be found
//     in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version can
//     also be found at http://fipa-os.sourceforge.net/. By downloading the
//     files containing the Programs you accept the terms and  conditions of the
//     Public License. You do not have to accept these terms and conditions, but
//     unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.skill.db.serializationdatabase;

//----------------------------------------------------------------------

import java.util.*;
import java.io.*;

import fipaos.agent.profile.*;
import fipaos.skill.db.*;

import fipaos.util.DIAGNOSTICS;

/**
 * This Database implementation provides a simple persistent database
 *
 */
public class SerializationDatabase implements Database
{
     
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

	/**
	 * This Databases name
	 */
	private String _db_name;

	/**
	 * Reference to the directory containing the Databases' files
	 */
	private File _db_dir;

	/**
	 * If true, we're in a transaction and another cannot begin
	 */
	private boolean _in_transaction = true;

	/**
	 * Thread which is in the transaction
	 */
	private Thread _in_control_thread;

	/**
	 * Lock count (for recursive transactions)
	 */
	private int _lock = 0;

	/**
	 * Stack containing type of lock required
	 */
	private Stack _locks = new Stack();

	/**
	 * Profile being used by this DB
	 */
	private DatabaseProfile _db_profile;
	
//----------------------------------------------------------------------
// CONSTRUCTOR
//----------------------------------------------------------------------

	/**
	 * Default constructor
	 *
	 * @param name The name of the database.  Can be the fully specified path.
	 */	
	public SerializationDatabase( String name, DatabaseProfile db_profile )
	{
		_db_name = name + ".jsdb";
		_db_profile = db_profile;
	}

//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------

	/**
	 * Method to create the database store.  Either this method or openDatabase()
	 * should be called before attempting to interact with the Database.  If the
	 * Database already exists, it will be "overwritten"
	 */
	public synchronized void createDatabase() throws DatabaseException
	{
		try
		{
			// Create a file object for the database.
 			_db_dir = new File( _db_profile.getDatabaseLocation() );
			_db_dir = new File( _db_dir, _db_name );
 			
			// Check if it exists
			if( _db_dir.exists() )
			{
				if( !removeDir( _db_dir ) )
				{
					throw new IOException( "Couldn't remove previous DB in directory " + _db_dir.getPath() );
				}
			}
			else if( !_db_dir.mkdirs() )
			{
				throw new IOException( "Couldn't create directory " + _db_dir.getPath() );
			}
		}
		catch ( Throwable t )
		{
			DIAGNOSTICS.println( "An error occured whilst creating the Database", t,
																												DIAGNOSTICS.LEVEL_MAX );
			throw new DatabaseException( "An error occured whilst creating the Database" + t );
		}
	}


	/**
	 * Method to re-open the database store.  Either this method or createDatabase()
	 * should be called before attempting to interact with the Database
	 */
	public synchronized void openDatabase() throws DatabaseException
	{
		try
		{
			// Create a file object for the database.
 			_db_dir = new File( _db_profile.getDatabaseLocation() );
			_db_dir = new File( _db_dir, _db_name );
 			
			// Check if it exists
			if( !_db_dir.exists() && !_db_dir.mkdirs() )
			{
				throw new IOException( "Couldn't create directory " + _db_dir.getPath() );
			}
		}
		catch ( Throwable t )
		{
			DIAGNOSTICS.println( "An error occured whilst opening the Database", t,
																												DIAGNOSTICS.LEVEL_MAX );
			throw new DatabaseException( "An error occured whilst creating the Database" + t );
		}
	}


	/**
	 * Method to close the database store.  Should only be called after openDatabase()
	 * or createDatabase() have been used and after all necessary transactions have been
	 * made.  Transactions cannot take place again until the database is re-opened.
	 */
	public synchronized void closeDatabase() throws DatabaseException
	{
		_db_dir = null;
	}


	/**
	 * Method to remove the database store.  The database will be closed if necessary before
	 * the store is removed.  The database can only be used again once a new store has been
	 * created.
	 */
	public synchronized void deleteDatabase() throws DatabaseException
	{
		try
		{
			if( _db_dir != null )
			{
				closeDatabase();
			}

			// Create a file object for the database.
			File db_dir = new File( _db_profile.getDatabaseLocation() );
			db_dir = new File( db_dir, _db_name );

			// Check it exists
			if( db_dir.exists() )
			{
				removeDir( db_dir );
			}
		}
		catch( Throwable t )
		{
			throw new DatabaseException( "Error when attempting to delete database: " + t );
		}
	}


	/**
	 * Method to add the given DatabaseObject to the database, in the given relation.
	 * This overwrites any object that has the same key as the given object 
	 * @param obj The DatabaseObject to add
	 */
	public synchronized void addObject( DatabaseObject obj ) throws DatabaseException
	{
		// Check we're in a transaction
		if( !_in_transaction )
		{
			throw new DatabaseException( "No transaction associated with DB!" );
		}
		
		if( ( (Boolean) _locks.peek() ).booleanValue() )
		{
			throw new DatabaseException( "addObject() called in 'readonly' transaction" );
		}
		
		// Get the UID of the object to add, and use that as the key to insert
		// the object into the hashtable
		writeObjectFile( obj );
	}

	
	/**
	 * Method to locate the object in the given relation with the given UID
	 *
	 * @param unique_id The unique id of the DatabaseObject to find
	 */
	public synchronized DatabaseObject findObject( String unique_id ) throws DatabaseException
	{
		// Check we're in a transaction
		if( !_in_transaction )
		{
			throw new DatabaseException( "Error - no transaction associated with DB!" );
		}
		
		try
		{
			// Get the object with the given UID
			return readObjectFile( unique_id );
		}
		catch( Throwable t )
		{
			return null;
		}
	}


	/**
	 * Method to update the given DatabaseObject in the Database. This throws a
	 * RuntimeException if an object with the same key as the object given is not
	 * found within the database.
	 *
	 * @param obj The DatabaseObject to update
	 */
	public void updateObject( DatabaseObject obj ) throws DatabaseException
	{
		// Check we're in a transaction
		if( !_in_transaction )
		{
			throw new DatabaseException( "No transaction associated with DB!" );
		}

		if( ( (Boolean) _locks.peek() ).booleanValue() )
		{
			throw new DatabaseException( "updateObject() called in 'readonly' transaction" );
		}

		updateObjectFile( obj );
	}


	/**
	 * Method to remove the given DatabaseObject from the Database
	 *
	 * @param obj The DatabaseObject to remove
	 */
	public void removeObject( DatabaseObject obj ) throws DatabaseException
	{
		// Check we're in a transaction
		if( !_in_transaction )
		{
			throw new DatabaseException( "No transaction associated with DB!" );
		}

		if( ( (Boolean) _locks.peek() ).booleanValue() )
		{
			throw new DatabaseException( "removeObject() called in 'readonly' transaction" );
		}
			
		removeObjectFile( obj.getObjectID() );
	}
	
	
	/**
	 * Method to remove the DatabaseObject from the Database with the given UID
	 *
	 * @param unique_id The unique id of the DatabaseObject to remove
	 */
	public void removeObject( String unique_id ) throws DatabaseException
	{
		// Check we're in a transaction
		if( !_in_transaction )
		{
			throw new DatabaseException( "No transaction associated with DB!" );
		}

		if( ( (Boolean) _locks.peek() ).booleanValue() )
		{
			throw new DatabaseException( "removeObject() called in 'readonly' transaction" );
		}
			
		removeObjectFile( unique_id );
	}	
	

	/**
	 * Creates the given relation within the Database
	 *
	 * @param relation The name of the relation to create
	 */
	public void createRelation( String relation )
	{
	}


	/**
	 * Deletes the given relation from the Database
	 *
	 * @param relation The name of the relation to delete
	 */
	public void deleteRelation( String relation )
	{
	}


	/**
	 * Indicates that a transaction is about to start.  THIS MUST BE CALLED
	 * BEFORE THE DATABASE IS MANIPULATED.
	 */
	public void startTransaction( boolean readonly ) throws DatabaseException
	{
		synchronized ( this )
		{
			// Wait until we can lock the database
			while( ( _lock != 0 ) && ( _in_control_thread != Thread.currentThread() ) )
			{
				try
				{
					wait();
				}
				catch ( InterruptedException e )
				{
				}	
			}

			try
			{
				if( _lock > 0 )
				{
					// Check that the type of transaction required doesn't conflict with a
					// parent transaction
					if( ( (Boolean) _locks.peek() ).booleanValue() && !readonly )
					{
						throw new DatabaseException("'update' transaction attempted within a 'readonly' transaction");
					}
				}
				
				_locks.push( new Boolean( readonly ) );
			}
			catch ( Throwable t )
			{
				throw new DatabaseException( "Exception during startTransaction(): " + t );
			}

			DIAGNOSTICS.println( Thread.currentThread().getName() + " - Started transaction" );

			// Lock DB
			_lock++;
			_in_transaction = true;

			// Record which thread has access
			_in_control_thread = Thread.currentThread();
		}
	}


	/**
	 * Indicates that a transaction has finished.  THIS MUST BE CALLED
	 * AFTER THE DATABASE HAS BEEN MANIPULATED IN ORDER TO COMMIT CHANGES.
	 */
	public void endTransaction() throws DatabaseException
	{
		synchronized ( this )
		{
			DIAGNOSTICS.println( Thread.currentThread().getName() + " - Ending transaction" );

			// Unlock DB
			_lock--;

			try
			{
				if( _lock == 0 )
				{
					_in_transaction = false;
					_in_control_thread = null;

					notify();
				}
				
				// Pop the last transaction type off of the stack
				_locks.pop();
			}
			catch ( Throwable t )
			{
				throw new DatabaseException( "Exception during endTransaction(): " + t );
			}
		}
	}


	/**
	 * Method to return an enumeration of the keys in the database
	 *
	 * @return An Enumeration of the ID's of the objects in the database
	 */
	public synchronized Enumeration getIDs() throws DatabaseException
	{
		// Check we're in a transaction
		if( !_in_transaction )
		{
			throw new DatabaseException( "No transaction associated with DB!" );
		}

		return getObjectFileEnumeration();
	}
	
//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------

	/**
	 * Convert an ObjectID into a filename
	 */
	private static String keyToFilename( String key )
	{
		// Create StringBuffer to put name into
		StringBuffer filename = new StringBuffer();
		StringBuffer keyname = new StringBuffer( key );
		
		// Iterate through keyname, looking for out of range characters
		for( int i = 0; i < keyname.length(); i++ )
		{
			char curr =  keyname.charAt( i );
			if( Character.isLetterOrDigit( curr ) )
			{
				// Just append
				filename.append( curr );
			}
			else
			{
				// Conver to base 10 (10 digits)
				byte value = (byte) curr;
				
				// Append escape
				filename.append( '#' );
				if( value < 16 )
				{
					filename.append( '0' );
				}
				filename.append( Integer.toHexString( (int) value ) );
			}
		}
		
		return filename.toString();
	}
	
	
	/**
	 * Convert a filename into an ObjectID
	 */
	private static String filenameToKey( String file ) throws NumberFormatException
	{
		// Create StringBuffer to put name into
		StringBuffer filename = new StringBuffer( file );
		StringBuffer keyname = new StringBuffer();
		
		// Iterate through keyname, looking for out of range characters
		for( int i = 0; i < filename.length(); i++ )
		{
			char curr =  filename.charAt( i );
			if( Character.isLetterOrDigit( curr ) )
			{
				// Just append
				keyname.append( curr );
			}
			else if( curr == '#' )
			{
				// Extract encoded value
				byte value = Byte.valueOf( filename.toString().substring( i + 1, i + 3 ), 16 ).byteValue();

				// Write character into key
				keyname.append( (char) value );
				
				// Increment counter
				i += 2;
			}
		}
		
		return keyname.toString();
	}

	
	/**
	 * Remove the contents of the given dir
	 */
	private boolean removeDir( File dir ) throws DatabaseException
	{
		try
		{
			// Get a list of the files in the given directory, and remove
			String files[] = dir.list();
			for( int i = 0; i < files.length; i++ )
			{
				File file = new File( dir, files[ i ] );

				if( !file.delete() )
				{
					throw new DatabaseException( "Couldn't remove " + file.getPath() );
				}
			}
		}
		catch( DatabaseException e )
		{
			throw e;
		}
		catch( Throwable t )
		{
			throw new DatabaseException("Error removing directory " + dir.getPath() + " : " + t );
		}
		
		return true;
	}
	

	/**
	 * Write the given DatabaseObject to disk
	 */
	private void writeObjectFile( DatabaseObject dbo ) throws DatabaseException
	{
		// Create reference to file containing object
		File file = new File( _db_dir, keyToFilename( dbo.getObjectID() ) );
		
		try
		{
			// Open ObjectOuputStream and write object
			ObjectOutputStream oos = new ObjectOutputStream( new FileOutputStream( file ) );
			
			// Write DB object
			oos.writeObject( dbo );
			
			// Close ObjectOutputStream
			oos.close();
		}
		catch( Throwable t )
		{
			throw new DatabaseException("Error writing file " + file.getPath() + " : " + t );
		}
	}
	

	/**
	 * Read the Object with the given key in
	 */
	private DatabaseObject readObjectFile( String key )  throws DatabaseException
	{
		// Create reference to file containing object
		File file = new File( _db_dir, keyToFilename( key ) );
		
		// Assume couldn't get result
		DatabaseObject result = null;
		
		try
		{
			// Create an ObjectInputStream
			ObjectInputStream ois = new ObjectInputStream( new FileInputStream( file ) );
			
			// Retreive object
			result = (DatabaseObject) ois.readObject();
			
			// Close ObjectInputStream
			ois.close();
		}
		catch( Throwable t )
		{
			throw new DatabaseException( "Error reading file " + file.getPath() + t );
		}
		
		return result;
	}
	

	/**
	 * Remove the object with the given key from disk
	 */
	private void removeObjectFile( String key )  throws DatabaseException
	{
		// Create reference to file containing object
		File file = new File( _db_dir, keyToFilename( key ) );
		
		if( !file.delete() )
		{
			throw new DatabaseException("Error removing file " + file.getPath() + " !" );
		}
	}
	

	/**
	 * Update the given object
	 */
	private void updateObjectFile( DatabaseObject dbo ) throws DatabaseException
	{
		removeObjectFile( dbo.getObjectID() );
		writeObjectFile( dbo );
	}
	

	/**
	 * Get an Enumeration of all objects in the database
	 */
	private Enumeration getObjectFileEnumeration() throws DatabaseException
	{
		Vector out = new Vector();
		
		try
		{
			// Get a list of files in the DB directory
			String files[] = _db_dir.list();
			
			// Iterate through and convert to key names
			for( int i = 0; i < files.length; i++ )
			{
				out.addElement( filenameToKey( files[i] ) );
			}
		}
		catch( Throwable t )
		{
			throw new DatabaseException("Problem when creating list of objects in DB: " + t );
		}
		
		return out.elements();
	}
}
