//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Emorphia are Copyright (C) 2001 Emorphia Limited
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.skill.wrapper;

//----------------------------------------------------------------------

import fipaos.skill.wrapper.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.ont.BASETOKENS;
import fipaos.agent.task.*;
import fipaos.agent.conversation.*;
import fipaos.agent.FIPAOSAgent;

//----------------------------------------------------------------------

/**
 * This is a default implementation of a FIPA Wrapper. It will respond to
 * all valid incoming FIPA-Wrapper messages with a default refuse for that message type.
 * It will also respond to other messages with a not-understood. To implement a working
 * FIPA Wrapper agent simply subclass this Task in the agent state that will
 * contain the wrapper functionality in your agent. Your agent can either subclass
 * this state for it's IDLE listener state, or it can pass incoming wrapper requests
 * on to a dedicated wrapper state that subclasses this object.<p>
 * Implementation notes:<br><ul>
 * <li>Query-if and query-ref messages are currently not handled further than saying
 * not-understood - they need support for iota queries,
 * predicates and assertions in the relevant parts of FIPA-OS.
 * </ul>
 * <br>
 * <br>
 */
public abstract class FIPAWrapperTask extends Task
{
  
//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------

	private static final String METHOD_PREFIX = "handle";
	
	
//----------------------------------------------------------------------
// STATIC VARIABLES
//----------------------------------------------------------------------
		
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

  /**
   * Incremental counter of service-instance-ids
   */
  private int _service_id_counter;
   
  
//----------------------------------------------------------------------
// CONSTRUCTOR
//----------------------------------------------------------------------

  /**
   * FIPA Wrapper constructor creates the wrapper state and makes the calls
   * to the API to start everything going. Your state's constructor should
   * mirror the parameters of this method and just make a call to <code>super</code>.
   *
   * @param owner The owner agent of this state
   * @param sm The Task Manager that will control this state
   */
  public FIPAWrapperTask( FIPAOSAgent owner, TaskManager sm )
  {
   
    super( owner, sm );
       
    _service_id_counter = 0;

  }
  

//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------

  /**
   * Handler ACL messages with a <code>request</code> performative.
   * These are the starting messages of the fipa-request protocol.
   *
   * @param conv The conversation containing the new message
   */
  public void handleRequest( Conversation conv )
  {
  	
  	ACL acl = conv.getACL( conv.getLatestMessageIndex() );
  	
  	dispatch( getWrapperCommand( (String) acl.getContentObject() ), conv );
  	
  }
  
//----------------------------------------------------------------------  
  
  /**
   * Handler ACL messages with a <code>query-ref</code> performative.
   * These are the starting messages of the fipa-query protocol.
   *
   * @param conv The conversation containing the new message
   */
  public void handleQueryRef( Conversation conv )
  {
  	
  	ACL acl = conv.getACL( conv.getLatestMessageIndex() );
  	
  	dispatch( getWrapperCommand( (String) acl.getContentObject() ), conv );  	
  	
  }
  
//----------------------------------------------------------------------  
  
  /**
   * Handler ACL messages with a <code>query-if</code> performative.
   * These are the starting messages of the fipa-query protocol.
   *
   * @param conv The conversation containing the new message
   */
  public void handleQueryIf( Conversation conv )
  {
  	
  	ACL acl = conv.getACL( conv.getLatestMessageIndex() );
  	
  	dispatch( getWrapperCommand( (String) acl.getContentObject() ), conv );  	
  	
  }
	
    
//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------

  /**
   * Parses an ACL message content field for a wrapper command. This is
   * always in the same place. This method does not check that the command
   * is valid - it simply parses the string lexicographically. If the command
   * is not valid then the request will get passed to the <code>handleOther</code> method
   * by the dispatcher.
   *
   * @param content The ACL content to parse
   * @return The wrapper command, or null if no command was found
   */
  private static String getWrapperCommand( String content )
  {
  	
  	int action = content.indexOf( BASETOKENS.ACTION );
  	
  	if ( action >= 0 )
  	{
  		
  		int bracket = content.indexOf( BASETOKENS.OPENBRACKET, action + 1 );
  		
  		if ( bracket >= 0 )
  		{
  			
  			int space = content.indexOf( BASETOKENS.SPACE, bracket + 1 );
  			
  			if ( space >= 0 )
  			{
  				
  				return content.substring( bracket + 1, space );	
  				
  			}
  			
  		}
  		
  	}
  		
  	return null;
  	
  }
  
//----------------------------------------------------------------------  
  
  /**
	 * Capitalises the first character of a string
	 *
	 * @param string The string to capitalise
	 */
	private String capitaliseFirstChar( String string )
	{
		
		return string.substring( 0, 1 ).toUpperCase() + string.substring( 1 );		
		
	}
  
//----------------------------------------------------------------------	
	
  /**
   * This method uses reflection to deliver a wrapper request to the correct
   * handler in the wrapper agent implementation. If a handler method is not
   * present in the wrapper state concrete implementation then the message
   * is dispatched to the <code>handleOtherCommand</code> method.
   *
   * @param conv The incoming conversation
   */
	private void dispatch( String command, Conversation conv )
	{
		
		if ( command == null )
		{
			
			handleOtherCommand( conv );
			
			return;
			
		}
		
	  // Convert command to handleX method name
	  // This involves changing the name to be capitalised and removing '-' characters
	  // and capitalising the next letter after the '-'
	  // e.g. software-subscribe --> handleSoftwareSubscribe
	  // e.g. init --> handleInit
	  boolean more_dashes = true;
	  StringBuffer method = new StringBuffer( METHOD_PREFIX );
	  
	  while ( more_dashes )
	  {
		  
		  int dash = command.indexOf( BASETOKENS.DASH );
		  
		  if ( dash > 0 )
		  {
		  	
		  	method.append( capitaliseFirstChar( command.substring( 0, dash ) ) );
		  	command = command.substring( dash + 1 );
		  	
		  }
		  else
		  {
		  	
		  	method.append( capitaliseFirstChar( command ) );	
		  	more_dashes = false;
		  	
		  }
		  
	  }
	  
	  // Use reflection to see if this state has that handle method
	  // Call it if it does otherwise call the catch all handler
	  try
	  {
	  	
	  	Class method_args[] = { Conversation.class };
	    Object invoke_args[] = { conv };

			this.getClass().getMethod( method.toString(), method_args ).invoke( this, invoke_args );
			
	  }
	  catch ( NoSuchMethodException nsme )
	  {

	  	handleOtherCommand( conv );	
	  	
	  }
	  catch ( Exception e ) { e.printStackTrace(); }
		
	}
  

//----------------------------------------------------------------------
// PROTECTED METHODS
//----------------------------------------------------------------------

  /**
   * Returns a unique service instance ID for a wrapper session.
   * If your wrapper needs to provide a different ID then you should
   * override this method.
   *
   * @param prefix An optional string prefix for the service instance ID - 
   *        if you do not wish to use the prefix pass in a null parameter
   */
  protected String generateServiceID( String prefix )
  {
    
    StringBuffer output = new StringBuffer();
    
    if ( prefix != null && ( ! prefix.equals( BASETOKENS.BLANK ) ) )
    {
    	
      output.append( prefix );
      output.append( BASETOKENS.DASH );
            
    }
    
    output.append( System.currentTimeMillis() );
    output.append( BASETOKENS.DASH );
    output.append( _service_id_counter );
    
    _service_id_counter++;
    
    // Cycle the number when we reach the end of the int range
    // Don't worry - this gives us over 2 billion numbers
    if ( _service_id_counter >= Integer.MAX_VALUE )
    {
    	
      _service_id_counter = 0;
      
    }
    
    return output.toString(); 
    
  }
  
//----------------------------------------------------------------------  
  
  /**
   * Replies back to the sender with a default error.
   * If your wrapper needs to handle unknown/incorrect wrapper commands in a
   * different way then you will need to override this method.
   *
   * @param conv The conversation to reply into
   */
  protected void handleOtherCommand( Conversation conv )
  {
  	
    ACL acl = conv.getACL( conv.getLatestMessageIndex() );
    
    acl.setReceiverAID( acl.getSenderAID() );
    acl.setSenderAID( _owner.getAID() );
    acl.setPerformative( FIPACONSTANTS.REFUSE ); // Always valid as next message as a first reply in FIPA-Query and FIPA-Request
    acl.setInReplyTo( acl.getReplyWith() );
    
		String content = (String) acl.getContentObject();
		String command = getWrapperCommand( content );
		
		if ( command != null )
		{
			
			int index = content.indexOf( command );
			StringBuffer c = new StringBuffer( content.substring( 1, index + command.length() ) );
			c.append( BASETOKENS.SPACE );
			c.append( BASETOKENS.OPENBRACKET );
			c.append( WRAPPERCONSTANTS.UNAUTHORISED );
			c.append( BASETOKENS.CLOSEBRACKET );
			acl.setContentObject( c.toString() );
			
		}
    
    acl.setContentObject( WRAPPERCONSTANTS.UNAUTHORISED ); // Default error reason - always valid (but not always correct)
    
    try
    {
    
    	forward( acl );
    	
    }
    catch ( Throwable mse ) {}
  	
  }
  
  
} // End FIPAWrapperTask

	
