//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Emorphia are Copyright (C) 2001 Emorphia Limited
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.test.agent;

import fipaos.agent.profile.*;
import fipaos.agent.*;
import fipaos.mts.*;
import fipaos.util.*;
import fipaos.ont.fipa.fipaman.AgentID;

import java.util.*;

import junit.framework.*;

/**
 * Test-framework for testing an Agent at the ACL level
 */
public abstract class AgentTestFramework extends TestCase
{
	/**
	 * AgentProfile to be used by Agents
	 */
	private AgentProfile _ap;
	
	/**
	 * PlatformProfile to be used by Agents
	 */
	private PlatformProfile _pp;
	
	/**
	 * Other Agents that this Agent depends upon
	 */
	private List _agents = new LinkedList();
	
	/** 
	 * Creates new AgentTestFramework
	 */
	public AgentTestFramework(String name)
	{
		super( name );
	}
	
	/**
	 * Setup the tests
	 */
	protected void setUp()
	{
		// Set-up dummy Platform Profile
		_pp = getPlatformProfile();
		
		// Setup dummy Agent Profile
		_ap = getAgentProfile();
	}
	
	
	/**
	 * Clean-up the tests
	 */
	protected void tearDown()	
	{
		// Kill off all Agents
		Iterator i = _agents.iterator();
		while( i.hasNext() )
		{
			FIPAOSAgent agent = (FIPAOSAgent) i.next();
			try
			{
				agent.shutdown();
			}
			catch( Throwable t )
			{
			}
		}
		
		_agents.clear();
	}
	
	
	/**
	 * Start the given Agent
	 */
	protected boolean startAgent( String name, String class_name )
	{
		FIPAOSAgent agent = null;
		
		// Start Agent
		try
		{
			Class args[] = new Class[] { PlatformProfile.class, AgentProfile.class,
																		String.class, String.class };
																		
			Object params[] = new Object[] { _pp, _ap, name, "fipaos-test" };
				
			DynamicClass dc = new DynamicClass( class_name, args, params );
			
			agent = (FIPAOSAgent) dc.getObject();
		}
		catch( Throwable t )
		{
			t.printStackTrace();
			return false;
		}
		
		// Save reference to Agent
		_agents.add( agent );
		
		return true;
	}
	
	
	/**
	 * Send message to the test Agent
	 */
	protected boolean sendMessage( Message msg )
	{
		return TestingMTP.sendMessage( getTestAgentName(), msg );
	}
	
	/**
	 * Get next message from the test Agent
	 */
	protected Message getIncomingMessage()
	{
		return TestingMTP.getNextIncomingMessage();
	}
	
	/**
	 * AID to use for testing
	 */
	protected AgentID getAID()
	{
		return TestingMTP.getAID();
	}
	
	/**
	 * Get the name of the AP
	 */
	protected abstract String getAPName();
	
	/**
	 * Get the name of the test-agent
	 */
	protected abstract String getTestAgentName();
	
	/**
	 * Replaces all occurences of a String with another String within a String
	 *
	 * @param str String to modify
	 * @param find String to find
	 * @param replace String to replace found occurences with
	 * @return The modified String
	 */
	public static String findAndReplace( String str, String find, String replace )
	{
		StringBuffer buff = new StringBuffer();
		
		int index = str.indexOf( find );
		int last_index = 0;

		while( index > -1 )
		{
			buff.append( str.substring( last_index, index ) );
			buff.append( replace );
			last_index = index + find.length();
			index = str.indexOf( find, last_index );
		}
		buff.append( str.substring( last_index ) );

		return buff.toString();
	}
	
	/**
	 * Returns a bog-standard PlatformProfile
	 */
	public static PlatformProfile getPlatformProfile()
	{
		PlatformProfile pp = new PlatformProfileImpl();
		pp.setDynamic( false );
		pp.setHAPName( "localap" );
		pp.setAMSAddress( (new URL( "fipaos-rmi://localhost/ams" )).toString() );
		pp.setMobility( false );
		pp.setProfileDirectory( "./" );
		return pp;
	}
	
	/**
	 * Returns a bog-standard AgentProfile
	 */
	public static AgentProfile getAgentProfile()
	{
		AgentProfile ap = new AgentProfileImpl();
		InternalAddress ia = new InternalAddressImpl();
		ia.setAddress( (new URL( "test://localhost" )).toString() );
		ap.addInternalAddress( ia );
		DatabaseProfile dbp = new DatabaseProfileImpl();
		dbp.setDatabaseLocation( "./" );
		dbp.setDatabaseType( "MemoryDatabase" );
		ap.setDatabaseProfile( dbp );
		ProtocolProfile pp = new ProtocolProfileImpl();
		ProtocolMapping pm1 = new ProtocolMappingImpl();
		pm1.setProtocolClass( "fipaos.agent.conversation.standard.protocol.FIPARequest" );
		pm1.setProtocolName( "fipa-request" );
		pp.addProtocolMapping( pm1 );
		pm1 = new ProtocolMappingImpl();
		pm1.setProtocolClass( "fipaos.agent.conversation.standard.protocol.FIPAAuctionDutch" );
		pm1.setProtocolName( "fipa-auction-dutch" );
		pp.addProtocolMapping( pm1 );
		pm1 = new ProtocolMappingImpl();
		pm1.setProtocolClass( "fipaos.agent.conversation.standard.protocol.FIPAAuctionEnglish" );
		pm1.setProtocolName( "fipa-auction-english" );
		pp.addProtocolMapping( pm1 );
		pm1 = new ProtocolMappingImpl();
		pm1.setProtocolClass( "fipaos.agent.conversation.standard.protocol.FIPABrokering" );
		pm1.setProtocolName( "fipa-brokering" );
		pp.addProtocolMapping( pm1 );
		pm1 = new ProtocolMappingImpl();
		pm1.setProtocolClass( "fipaos.agent.conversation.standard.protocol.FIPAContractNet" );
		pm1.setProtocolName( "fipa-contract-net" );
		pp.addProtocolMapping( pm1 );
		pm1 = new ProtocolMappingImpl();
		pm1.setProtocolClass( "fipaos.agent.conversation.standard.protocol.FIPAIteratedContractNet" );
		pm1.setProtocolName( "fipa-iterated-contract-net" );
		pp.addProtocolMapping( pm1 );
		pm1 = new ProtocolMappingImpl();
		pm1.setProtocolClass( "fipaos.agent.conversation.standard.protocol.FIPAQuery" );
		pm1.setProtocolName( "fipa-query" );
		pp.addProtocolMapping( pm1 );
		pm1 = new ProtocolMappingImpl();
		pm1.setProtocolClass( "fipaos.agent.conversation.standard.protocol.FIPARecruiting" );
		pm1.setProtocolName( "fipa-recruiting" );
		pp.addProtocolMapping( pm1 );
		pm1 = new ProtocolMappingImpl();
		pm1.setProtocolClass( "fipaos.agent.conversation.standard.protocol.FIPARequestWhen" );
		pm1.setProtocolName( "fipa-request-when" );
		pp.addProtocolMapping( pm1 );
		pm1 = new ProtocolMappingImpl();
		pm1.setProtocolClass( "fipaos.agent.conversation.standard.protocol.NoProtocol" );
		pm1.setProtocolName( "no-protocol" );
		pp.addProtocolMapping( pm1 );
		pm1 = new ProtocolMappingImpl();
		pm1.setProtocolClass( "fipaos.agent.conversation.standard.protocol.RuntimeProtocol" );
		pm1.setProtocolName( "runtime-protocol" );
		pp.addProtocolMapping( pm1 );
		pm1 = new ProtocolMappingImpl();
		pm1.setProtocolClass( "fipaos.agent.conversation.standard.protocol.FIPASubscribe" );
		pm1.setProtocolName( "fipa-subscribe" );
		pp.addProtocolMapping( pm1 );
		ap.setProtocolProfile( pp );
		
		// Setup TM profile
		TaskManagerProfile tmp = new TaskManagerProfileImpl();
		tmp.setUseGlobalThreads( false );
		tmp.setLocalThreads( 5 );
		ap.setTaskManagerProfile( tmp );
		
		return ap;
	}
}
