//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Nortel Networks Corporation are Copyright (C) 1999-2000 
//  Nortel Networks Corporation licensed under the Nortel Networks FIPA-OS 
//  Public License, available at http://fipa-os.sourceforge.net/. 
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.test.agent;

import fipaos.agent.*;
import fipaos.agent.task.*;
import fipaos.agent.conversation.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.util.*;
import fipaos.mts.service.Service;

import java.util.*;
import javax.swing.*;

/**
 * This agent aims to test the round-trip time for sending messages to a counter-part
 * agent
 */
public class MessageTimingAgent extends fipaos.agent.FIPAOSAgent
{
	private static final int MAX_ITER = 100;
	private static final String ROUND_TRIP = "rt";
	private static final int _payload_sizes[] = new int[] { 1 , 100, 1000, 5000 };
	
	private AgentID _counterpart;
	private Map _results[] = new Map[ _payload_sizes.length ];
	private int _current_payload = 0;
	
	/** Creates new MessageTiming */
	public MessageTimingAgent( String platform, String agent, String owner )
	{
		super( platform, agent, owner );
		
		// Start new idle-task
		setListenerTask( new IdleTask() );
		
	}
	
	/**
	 * Retrieves the given result list for the current payload
	 */
	List getList( String name )
	{
		Map current_map = _results[ _current_payload ];
		
		if( current_map == null )
		{
			current_map = _results[ _current_payload ] = new HashMap();
		}
		
		List list = (List) current_map.get( name );
		
		if( list == null )
		{
			list = new LinkedList() ;
			current_map.put( name, list );
		}
		
		return list;
	}
	
	/**
	 * Calculate the average of a list of Long's
	 */
	long calcAverage( List l )
	{
		long v = 0;
		Iterator i = l.iterator();
		while( i.hasNext() )
		{
			v += ( (Long) i.next() ).longValue();
		}
		v /= l.size();
		return v;
	}
	
	/**
	 * Promote getCM() status
	 */
	public ConversationManager getCM()
	{
		return super.getCM();
	}
	
	/**
	 * Simple Task to respond to incoming request messages, and just reply with an
	 * inform - it performs the minimum of work in doing so
	 */
	public class IdleTask extends Task
	{
		public void startTask()
		{
			// Get name of counterpart agent
			String agent_name = JOptionPane.showInputDialog( "Counter-part agent name? (or hit cancel to just listen)" );
			if( agent_name != null && agent_name.length() > 0 )
			{
				_counterpart = new AgentID();
				_counterpart.setName( agent_name );
			}
			
			// If we know our counterparts details, start the testing
			if( _counterpart != null )
			{
				this._tm.newTask( new TestingTask( _counterpart ) );
			}
		}
		
		public void handleRequest( Conversation conv )
		{
			// Just reply with an inform
			ACL acl = (ACL) conv.getACL( conv.getLatestMessageIndex() ).clone();
			acl.setReceiverAID( acl.getSenderAID() );
			acl.setSenderAID( getAID() );
			acl.setPerformative( FIPACONSTANTS.INFORM );
			acl.setInReplyTo( acl.getReplyWith() );
			acl.setReplyWith( null );
			forward( acl );

			try
			{
				// Remove conversation from CM
				getCM().removeConversation( acl.getConversationID() );
			}
			catch( Throwable t )
			{
			}			
		}
	}
	
	/**
	 * Task to carry out the actual job of testing
	 */
	public class TestingTask extends Task
	{
		private AgentID _target;
		private boolean _reachable;
		private int _iteration;
		private String _payload;
		
		public TestingTask( AgentID target )
		{
			_target = target;
		}
		
		public void startTask()
		{
			DIAGNOSTICS.println( "Checking counter-part is alive", this, DIAGNOSTICS.LEVEL_MAX );
				
			// Start by sending initial message just to check agent is reachable
			newTask( new MessageTestTask( _target, "( inital-test )" ), 30000 );
		}
		
		public void startTesting()
		{
			if( _current_payload >= _payload_sizes.length )
			{
				// Were finished!
				finishedTesting();
			}
			else
			{
				DIAGNOSTICS.println( "Starting testing with payload of " +
							_payload_sizes[ _current_payload ], this, DIAGNOSTICS.LEVEL_MAX );
				
				// Create List to contain results
				_iteration = 0;
				
				// Create payload
				StringBuffer buff = new StringBuffer();
				buff.append( "( " );
				for( int i = 0; i < _payload_sizes[ _current_payload ]; i++ )
				{
					buff.append( "t" );
				}
				buff.append( " )" );
				_payload = buff.toString();
				
				// Continue testing
				continueTesting();
			}
		}
		
		public void continueTesting()
		{
			_iteration++;
			
			if( _iteration > MAX_ITER )
			{
				_iteration = 0;
				_current_payload++;
				startTesting();
			}
			else
			{
				newTask( new MessageTestTask( _target, _payload ), 15000 );				
			}
		}
		
		public void finishedTesting()
		{
			DIAGNOSTICS.println( "Testing finished, compiling results...", this, DIAGNOSTICS.LEVEL_MAX );
			
			// Display results
			StringBuffer out = new StringBuffer( "\nResults\n=======\n\n" );
			for( int i = 0; i < _payload_sizes.length; i++ )
			{
				out.append( "Payload = " + _payload_sizes[ i ] + "\n" );
				Map m = _results[ i ];
				Iterator j = m.keySet().iterator();
				while( j.hasNext() )
				{
					String name = (String) j.next();
					long av = calcAverage( (List) m.get( name ) );
					out.append( "\t" + name + "\t" + av + "\n"  );
				}
				out.append( "\n\n" );
			}
			DIAGNOSTICS.println( out.toString(), this, DIAGNOSTICS.LEVEL_MAX );
			
			done();
		}
		
		public void timeoutMessageTimingAgent_MessageTestTask( Task t )
		{
			DIAGNOSTICS.println( "Couldn't send message to couter-part agent, giving up....", this,
														DIAGNOSTICS.LEVEL_MAX);
			
			finishedTesting();
		}
		
		public void doneMessageTimingAgent_MessageTestTask( Object result )
		{
			if( !_reachable )
			{
				DIAGNOSTICS.println( "Counter-part responded OK", this, DIAGNOSTICS.LEVEL_MAX );
				
				_reachable = true;
				startTesting();
			}
			else
			{
				getList( ROUND_TRIP ).add( result );
				continueTesting();
			}
		}
	}
	
	/**
	 * Task to test message sending & timing
	 */
	public class MessageTestTask extends Task
	{				
		private long _start;
		
		/**
		 * Message to send & await response to
		 */
		private ACL _message;
		
		/**
		 * Round-trip time (time to receive agree)
		 */
		private long _round_trip;
		
		public MessageTestTask( AgentID aid, Object content )
		{
			_message = new ACL();
			_message.setPerformative( FIPACONSTANTS.REQUEST );
			_message.setSenderAID( getAID() );
			_message.setReceiverAID( aid );
			_message.setContentObject( content );
		}
		
		public void startTask()
		{
			_start = System.currentTimeMillis();
			forward( _message );
		}
		
		public void handleInform( Conversation conv )
		{
			long _round_trip = System.currentTimeMillis() - _start;
			
			// Extract propogation times
			Envelope env = conv.getEnvelope( conv.getLatestMessageIndex() );
			Set slots = env.getUserSlots();
			Iterator i = slots.iterator();
			long lowest = Long.MAX_VALUE;
			while( i.hasNext() )
			{
				try
				{
					// Get next object
					String slot = (String) i.next();
					Object next = env.getSlotValue( slot );
					
					if( slot.startsWith( Service.FIPAOS_DEBUG ) )
					{
						Long val = null; 
						if( next instanceof String )
						{
							val = new Long( (String) next );
						}
						else if( next instanceof Long )
						{
							val = (Long) next;
						}
						
						if( val != null && val.longValue() < lowest )
						{
							lowest = val.longValue();
						}
					}
				}
				catch( Throwable t )
				{
				}
			}
			
			slots = env.getUserSlots();
			i = slots.iterator();
			while( i.hasNext() )
			{
				try
				{
					// Get next object
					String slot = (String) i.next();
					Object next = env.getSlotValue( slot );
					
					if( slot.startsWith( Service.FIPAOS_DEBUG ) )
					{
						slot = slot.substring( Service.FIPAOS_DEBUG.length() );
						if( next instanceof String )
						{
							long val = Long.parseLong((String) next ) - lowest;
							getList( slot ).add( new Long( val ) );
						}
						else if( next instanceof Long )
						{
							long val = ((Long) next ).longValue() - lowest;
							getList( slot ).add( new Long( val ) );
						}
					}
				}
				catch( Throwable t )
				{
				}
			}
			
			try
			{
				// Remove conversation from CM
				getCM().removeConversation( conv.getConversationID() );
			}
			catch( Throwable t )
			{
			}
			
			done( new Long( _round_trip ) );
		}

		public void handleOther( Conversation conv )
		{
			throw new RuntimeException( "Unexpected response from target agent" );
		}
	}
	
	public static void main( String args[] )
	{
		new MessageTimingAgent( args[1], args[0],  "fipaos" );
	}
}