//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Emorphia are Copyright (C) 2001 Emorphia Limited
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.test.parser.envelope.xml;

import java.util.*;

import junit.framework.*;

import fipaos.parser.envelope.xml.Parser;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.util.*;

/**
 * Test-harness for fipaos.parser.acl.bitefficient package
 */
public class ParserTest extends TestCase
{
	private static String FIPA_EXAMPLE_ENVELOPE_1 = 
	"<?xml version=\"1.0\"?>\n" +
	"<envelope>\n" +
	"\t<params index=\"1\">\n" +
	"\t\t<to>\n" +
	"\t\t\t<agent-identifier>\n" +
	"\t\t\t\t<name>receiver@foo.com</name>\n" +
	"\t\t\t\t<addresses>\n" +
	"\t\t\t\t\t<url>http://foo.com/acc</url>\n" +
	"\t\t\t\t</addresses>\n" +
	"\t\t\t</agent-identifier>\n" +
	"\t\t</to>\n" +
	"\t\t<from>\n" +
	"\t\t\t<agent-identifier>\n" +
	"\t\t\t\t<name>sender@bar.com</name>\n" +
	"\t\t\t\t<addresses>\n" +
	"\t\t\t\t\t<url>http://bar.com/acc</url>\n" +
	"\t\t\t\t</addresses>\n" +
	"\t\t\t</agent-identifier>\n" +
	"\t\t</from>\n" +
	"\t\t<acl-representation>fipa.acl.rep.xml.std</acl-representation>\n" +
	"\t\t<date>20000508T042651481</date>\n" +
	"\t\t<encrypted>no encryption</encrypted>\n" +
	"\t\t<received >\n" +
	"\t\t\t<received-by value=\"http://foo.com/acc\" />\n" +
	"\t\t\t<received-date value=\"20000508T042651481\" />\n" +
	"\t\t\t<received-id value=\"123456789\" />\n" +
	"\t\t</received>\n" +
	"\t</params>\n" +
	"</envelope>\n";
	
	/**
	 * Default constructor
	 */
	public ParserTest( String name )
	{
		super( name );
	}
	
	/**
	 * Setup the tests
	 */
	protected void setUp()
	{
	}
	
	/**
	 * Overrides method in parent to enable running of our tests
	 */
	protected void runTest() throws Throwable
	{
//		testParseAndDeparse();
		testComparePredefinedMessage();
	}
	
	/**
	 * Executed when comparison with existing examples should be performed
	 */
	private void testComparePredefinedMessage() throws Throwable
	{
		// Test parsing
		Envelope env1 = Parser.parse( FIPA_EXAMPLE_ENVELOPE_1.getBytes() );
		
		// Check contents of Envelope
		testPredefinedEnvelope1( env1 );
		
		// Now deparse, reparse and re-check
		testPredefinedEnvelope1( Parser.parse( Parser.deparse( env1 ) ) );
	}
	
	/**
	 * Compares the given Envelope against the expected values contained in
	 * FIPA_EXAMPLE_ENVELOPE_1
	 */
	private void testPredefinedEnvelope1( Envelope env1 )
	{
		// Check receiver AID
		assert( "Receiver AID name not parsed correctly - not found in Envelope receiver field",
						env1.getReceivers() != null &&
						env1.getReceivers().size() > 0 );
		AgentID aid = (AgentID) env1.getReceivers().get( 0 );
		assert( "Receiver AID name not parsed correctly - name part not as expected", 
						aid.getName().equals( "receiver@foo.com" ) );
		assert( "Receiver AID addresses not parsed correctly - they're missing", 
						aid.getAddresses() != null ||
						aid.getAddresses().size() > 0 );
		URL url = (URL) aid.getAddresses().get(0);
		assert( "Receiver AID addresses not parsed correctly - address incorrect (\"" + url + "\")", 
						url.equals( new URL( "http://foo.com/acc" ) ) );
		
		// Check sender AID
		assert( "Sender AID name not parsed correctly - not found in Envelope sender field",
						env1.getSender() != null );
		aid = (AgentID) env1.getSender();
		assert( "Sender AID name not parsed correctly - name part not as expected", 
						aid.getName().equals( "sender@bar.com" ) );
		assert( "Sender AID addresses not parsed correctly - they're missing", 
						aid.getAddresses() != null || 
						aid.getAddresses().size() > 0 );
		url = (URL) aid.getAddresses().get(0);
		assert( "Sender AID addresses not parsed correctly - address incorrect (\"" + url + "\")", 
						url.equals( new URL( "http://bar.com/acc" ) ) );
		
		// Check received field
		assert( "Received parameter doesn't exist!", env1.getReceived() != null );
		assert( "Received.by value is incorrect", 
						env1.getReceived().getBy() != null &&
						env1.getReceived().getBy().equals( new URL( "http://foo.com/acc" ) ) );
		assert( "Received.date value is incorrect", 
						env1.getReceived().getDateTime() != null &&
						env1.getReceived().getDateTime().equals( new UTCTime( "20000508T042651481" ) ) );
		assert( "Received.id value is incorrect", 
						env1.getReceived().getId() != null &&
						env1.getReceived().getId().equals( "123456789" ) );
		
		// Check simple fields
		assert( "ACL representation not parsed correctly (\"" + env1.getACLRepresentation() + "\")",
						env1.getACLRepresentation() != null && 
						env1.getACLRepresentation().equals( "fipa.acl.rep.xml.std" ) );
		assert( "Date not parsed correctly (\"" + env1.getDate() + "\")",
						env1.getDate() != null && 
						env1.getDate().equals( new UTCTime( "20000508T042651481" ) ) );
		assert( "Encryption settings not parsed correctly (\"" + env1.getEncrypted() + "\")",
						env1.getEncrypted() != null && 
						env1.getEncrypted().size() > 0 &&
						env1.getEncrypted().get( 0 ).equals( "no encryption" ) );
	}
	
	/**
	 * Executed when Parse then Deparse test should be performed
	 */
	private void testParseAndDeparse() throws Throwable
	{
/*		ACL acl = new ACL( ACL_STRING );
		
		byte acl2[] = Parser.deparse( acl );
		
		ACL acl3 = Parser.parse( acl2 );
		
		assert( "AbstractEnv->XML->AbstractEnv isn't reflexive!", acl.toString().equals( acl3.toString() ) );
*/
	}
}