//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Emorphia are Copyright (C) 2001 Emorphia Limited
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.test.platform;

import fipaos.test.agent.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.mts.*;

/**
 * Test-harness for the AMS
 */
public class AMSAgentTest extends AgentTestFramework
{
	/**
	 * Standard test messages
	 */
	public final String REGISTRATION_MSG = "(action %ams% (register (ams-agent-description :name %test% :state active))))";
	public final String DEREGISTRATION_MSG = "(action %ams% (deregister (ams-agent-description :name %test%))))";
	public final String MODIFICATION_MSG = "(action %ams% (modify (ams-agent-description :name %test% :state wait))))";
	
	/**
	 * Conversation ID value
	 */
	private long _conv_id_count = 0;
	
	/**
	 * AID of AMS to be tested
	 */
	private AgentID _ams_aid;
	
	/** Creates new AMSAgentTest */
	public AMSAgentTest( String name )
	{
		super( name );
	}
	

	/**
	 * Override - start-up method to start our test Agent
	 */
	protected void setUp()
	{
		// Let super-class setup
		super.setUp();

		// Setup AMS AID
		_ams_aid = new AgentID();
		_ams_aid.setName( getTestAgentName() + "@" + getAPName() );
		
		// Start AMS
		startAgent( getTestAgentName(), "fipaos.platform.AgentManagementSystem" );
	}
	

	/**
	 * Overrides method in parent to enable running of our tests
	 */
	protected void runTest() throws Throwable
	{
		testFirstRegistrationDeregistration();
		testDeregisterWithoutRegister();
		testAlreadyRegisted();
		testModification();
		
	}
	
	
	/**
	 * Test registration & deregistration
	 */
	private void testFirstRegistrationDeregistration() throws Throwable
	{
		// Send registration message
		sendRegistrationMessage();
		
		// Wait for agree then inform
		Message resp1 = getIncomingMessage();
		assert( "AMS didn't agree as expected when registering",
						resp1.getACL().getPerformative().equals( FIPACONSTANTS.AGREE ) );
		Message resp2 = getIncomingMessage();
		assert( "AMS didn't succeed in registration as expected",
						resp2.getACL().getPerformative().equals( FIPACONSTANTS.INFORM ) );
		
		// Send deregistration message
		sendDeregistrationMessage();
		
		// Wait for agree then inform
		Message resp3 = getIncomingMessage();
		assert( "AMS didn't agree as expected when deregistering",
						resp3.getACL().getPerformative().equals( FIPACONSTANTS.AGREE ) );
		Message resp4 = getIncomingMessage();
		assert( "AMS didn't succeed in deregistration as expected",
						resp4.getACL().getPerformative().equals( FIPACONSTANTS.INFORM ) );		
	}
	
	/**
	 * Test just deregistration - should fail
	 */
	private void testDeregisterWithoutRegister() throws Throwable
	{
		// Send deregistration message
		sendDeregistrationMessage();
		
		// Wait for agree then failure
		Message resp1 = getIncomingMessage();
		assert( "AMS didn't agree as expected when deregistering",
						resp1.getACL().getPerformative().equals( FIPACONSTANTS.AGREE ) );
		Message resp2 = getIncomingMessage();
		assert( "AMS didn't indicate failure as expected",
						resp2.getACL().getPerformative().equals( FIPACONSTANTS.FAILURE ) );		
		AgentManagementException ame = (AgentManagementException) resp2.getACL().getContentObject();
		assert( "AMS failure reason not \"not-registered\" as expected",
						ame.getException().equals( "not-registered" ) );
	}
	
	/**
	 * Test registration of an already registered Agent - should fail
	 */
	private void testAlreadyRegisted() throws Throwable
	{
		// Send registration message
		sendRegistrationMessage();
		
		// Wait for agree then inform
		Message resp1 = getIncomingMessage();
		assert( "AMS didn't agree as expected when deregistering",
						resp1.getACL().getPerformative().equals( FIPACONSTANTS.AGREE ) );
		Message resp2 = getIncomingMessage();
		assert( "AMS didn't succeed in registration as expected",
						resp2.getACL().getPerformative().equals( FIPACONSTANTS.INFORM ) );
		
		// Send second registration
		sendRegistrationMessage();
		
		// Wait for agree then failure
		Message resp3 = getIncomingMessage();
		assert( "AMS didn't agree as expected when deregistering",
						resp3.getACL().getPerformative().equals( FIPACONSTANTS.AGREE ) );
		Message resp4 = getIncomingMessage();
		assert( "AMS didn't indicate failure as expected",
						resp4.getACL().getPerformative().equals( FIPACONSTANTS.FAILURE ) );		
		AgentManagementException ame = (AgentManagementException) resp4.getACL().getContentObject();
		assert( "AMS failure reason not \"not-registered\" as expected",
						ame.getException().equals( "already-registered" ) );
		
		
		// Send deregistration message
		sendDeregistrationMessage();
		
		// Wait for agree then inform
		Message resp5 = getIncomingMessage();
		assert( "AMS didn't agree as expected when deregistering",
						resp5.getACL().getPerformative().equals( FIPACONSTANTS.AGREE ) );
		Message resp6 = getIncomingMessage();
		assert( "AMS didn't succeed in deregistration as expected",
						resp6.getACL().getPerformative().equals( FIPACONSTANTS.INFORM ) );		
	}

	
	/**
	 * Test registration & deregistration
	 */
	private void testModification() throws Throwable
	{
		// Send modification message
		sendModificationMessage();
		Message resp0 = getIncomingMessage();
		assert( "AMS didn't agree as expected when modifying",
						resp0.getACL().getPerformative().equals( FIPACONSTANTS.AGREE ) );
		Message resp01 = getIncomingMessage();
		assert( "AMS didn't fail as expected when modifying",
						resp01.getACL().getPerformative().equals( FIPACONSTANTS.FAILURE ) );
		AgentManagementException ame = (AgentManagementException) resp01.getACL().getContentObject();
		assert( "AMS failure reason not \"not-registered\" as expected",
						ame.getException().equals( "not-registered" ) );		
		
		
		// Send registration message
		sendRegistrationMessage();
		
		// Wait for agree then inform
		Message resp1 = getIncomingMessage();
		assert( "AMS didn't agree as expected when registering",
						resp1.getACL().getPerformative().equals( FIPACONSTANTS.AGREE ) );
		Message resp2 = getIncomingMessage();
		assert( "AMS didn't succeed in registration as expected",
						resp2.getACL().getPerformative().equals( FIPACONSTANTS.INFORM ) );

		
		// Send modification message
		sendModificationMessage();
		
		// Wait for agree then inform
		Message resp3 = getIncomingMessage();
		assert( "AMS didn't agree as expected when modifying",
						resp3.getACL().getPerformative().equals( FIPACONSTANTS.AGREE ) );
		Message resp4 = getIncomingMessage();
		assert( "AMS didn't succeed in modifying as expected",
						resp4.getACL().getPerformative().equals( FIPACONSTANTS.INFORM ) );

		
		// Send deregistration message
		sendDeregistrationMessage();
		
		// Wait for agree then inform
		Message resp5 = getIncomingMessage();
		assert( "AMS didn't agree as expected when deregistering",
						resp5.getACL().getPerformative().equals( FIPACONSTANTS.AGREE ) );
		Message resp6 = getIncomingMessage();
		assert( "AMS didn't succeed in deregistration as expected",
						resp6.getACL().getPerformative().equals( FIPACONSTANTS.INFORM ) );		
	}
	
	
	/**
	 * Send registration message
	 */
	private void sendRegistrationMessage()
	{
		// Setup content
		String content = findAndReplace( REGISTRATION_MSG, "%ams%", _ams_aid.toString() );
		content = findAndReplace( content, "%test%", getAID().toString() );

		// Send message to our test Agent
		ACL msg = new ACL();
		msg.setContentObject( content );
		msg.setSenderAID( getAID() );
		msg.setReceiverAID( _ams_aid );
		msg.setProtocol( FIPACONSTANTS.FIPA_REQUEST );
		msg.setPerformative( FIPACONSTANTS.REQUEST );
		msg.setOntology( "fipa-agent-management" );
		msg.setLanguage( FIPACONSTANTS.SL0 );
		msg.setConversationID( Long.toString( _conv_id_count++ ) );
		sendMessage( new Message( msg ) );		
	}
	
	
	/**
	 * Send deregistration message
	 */
	private void sendDeregistrationMessage()
	{
		// Setup content
		String content = findAndReplace( DEREGISTRATION_MSG, "%ams%", _ams_aid.toString() );
		content = findAndReplace( content, "%test%", getAID().toString() );

		// Send message to our test Agent
		ACL msg = new ACL();
		msg.setContentObject( content );
		msg.setSenderAID( getAID() );
		msg.setReceiverAID( _ams_aid );
		msg.setProtocol( FIPACONSTANTS.FIPA_REQUEST );
		msg.setPerformative( FIPACONSTANTS.REQUEST );
		msg.setOntology( "fipa-agent-management" );
		msg.setLanguage( FIPACONSTANTS.SL0 );
		msg.setConversationID( Long.toString( _conv_id_count++ ) );
		sendMessage( new Message( msg ) );		
	}
	

	/**
	 * Send modification message
	 */
	private void sendModificationMessage()
	{
		// Setup content
		String content = findAndReplace( MODIFICATION_MSG, "%ams%", _ams_aid.toString() );
		content = findAndReplace( content, "%test%", getAID().toString() );

		// Send message to our test Agent
		ACL msg = new ACL();
		msg.setContentObject( content );
		msg.setSenderAID( getAID() );
		msg.setReceiverAID( _ams_aid );
		msg.setProtocol( FIPACONSTANTS.FIPA_REQUEST );
		msg.setPerformative( FIPACONSTANTS.REQUEST );
		msg.setOntology( "fipa-agent-management" );
		msg.setLanguage( FIPACONSTANTS.SL0 );
		msg.setConversationID( Long.toString( _conv_id_count++ ) );
		sendMessage( new Message( msg ) );		
	}
	

	protected java.lang.String getAPName()
	{
		return "testap";
	}
	
	
	protected java.lang.String getTestAgentName()
	{
		return "ams";
	}
}
