//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Emorphia are Copyright (C) 2001 Emorphia Limited
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.tool.configurator;

import java.util.*;
import javax.swing.*;
import java.io.*;

public class ScriptEditor implements ScriptVars
{
	
	private static final String _separator = System.getProperty("file.separator");
	
	// store the script keys and values in arrays to preserve ordering
	private String _script_variables[];
	private String _script_values[];
	private String _script_lines[];
	
	// use a Map to convieniently pass the script data to the gui
	private Map _script_keys_values;
	
	// keep track of the line numbers that variables are stored on
	private Map _vars_to_lines;
	
	public static final String SETUP_WINDOWS	= "SetupFIPAOS.bat";
	public static final String SETUP_UNIX		= "SetupFIPAOS";
	
	private boolean _is_windows;
	
	private String _setupscript_filename;
	private String _real_fipaos_location;
	private String _setupscript;
	
	// LinkedList to allow quick access to editible vars
	private List _editable_vars;
	
	// Indicates if we're working in LEGACLY mode
	private boolean _legacy;
	
	/*********************************************************
	 *
	 * Constructor
	 *
	 *********************************************************/
	public ScriptEditor()
	{
		
		_legacy = true;
		
		_is_windows = isWindowsOS();
		
		// copy editable script var names into the linked list for quick access
		
		_editable_vars = new LinkedList();
		
		_editable_vars.addAll( SETUPFIPAOS_VARS );
		
	}
	
	/**
	 * Generic constructor to load given bat/script file
	 */
	public ScriptEditor( File batch_file, List variables ) throws IOException
	{
		_editable_vars = new LinkedList( variables );		
		_setupscript_filename =  batch_file.getCanonicalPath();
		_is_windows = _setupscript_filename.endsWith( ".bat" );
		_setupscript = FIPAOSConfiguratorTools.loadFileIntoString( _setupscript_filename );
		extractSetupParameters();
	}
	
	
	/*********************************************************
	 *
	 * isWindowsOS()
	 *
	 * @return boolean. True if this is a windows OS. False if not.
	 *
	 *********************************************************/
	public boolean isWindowsOS()
	{
		
		String os = System.getProperty("os.name");
		
		if (os.toLowerCase().indexOf("windows") == -1)
		{
			
			return false;
			
		}
		else
		{
			
			return true;
			
		}
		
	} // end isWindowsOS()
	
	/*********************************************************
	 *
	 * populate()
	 *
	 * @param location. A string containing the location to attempt
	 * to load the setup script from.
	 *
	 *********************************************************/
	public boolean populate(String fipaos_location)
	{
		
		_real_fipaos_location = fipaos_location;
		
		if (fipaos_location == null) // use JFileChooser to locate script
		{
			JFileChooser chooser = new JFileChooser(fipaos_location);
			
			chooser.setDialogTitle("Select location of setup script");
			
			int returnVal = chooser.showDialog(null, "Select");
			
			if(returnVal == JFileChooser.APPROVE_OPTION)
			{
				
				_setupscript_filename = chooser.getSelectedFile().getAbsolutePath();
				
			}
		}
		else
		{
			_setupscript_filename = fipaos_location + _separator + "bat" + _separator + SETUP_UNIX;
			
			if (_is_windows == true)
			{
				
				_setupscript_filename = _setupscript_filename + ".bat";
				
			}
			
		}
		
		try
		{
			
			_setupscript = FIPAOSConfiguratorTools.loadFileIntoString(_setupscript_filename);
					
		}
		catch (Throwable t)
		{
			t.printStackTrace();
			
			return populate(null);
			
		} // end try
		
		extractSetupParameters();

		return true;		
		
	} // end populate()
	
	/*********************************************************
	 *
	 * getValueOfScriptVariable()
	 *
	 * @param variable. The variable name to get the value for
	 *
	 * @return String. The value associated with the given variable name from the script. null if not found
	 *
	 *********************************************************/
	private String getValueOfScriptVariable(String variable)
	{
		
		for (int i = 0; i < _script_variables.length; i++)
		{
			
			// if this variable position contains the requires variable name
			if (_script_variables[i] != null && _script_variables[i].equals(variable))
			{
				
				return _script_values[i];
				
			} // end if
			
		} // end for
		
		// not found
		return null;
		
	} // end getValueOfScriptVariable()
	
	/*********************************************************
	 *
	 * extractSetupParameters()
	 *
	 * Parses the setup script and extracts the parameters and values for later use.
	 *
	 *********************************************************/
	private void extractSetupParameters()
	{
		
		StringTokenizer setuplines = new StringTokenizer(_setupscript, "\n", false);
		
		String line;
		
		String prefix = "";
		
		if (_is_windows == true)
		{
			
			prefix = "set ";
			
		}
		
		String name;
		String value;
		
		int set_present;
		int equals_present;
		
		int num_tokens = setuplines.countTokens();
		
		_vars_to_lines = new HashMap();
		
		_script_keys_values = new HashMap();
		
		_script_variables 	= new String[num_tokens];
		_script_values		= new String[num_tokens];
		_script_lines		= new String[num_tokens];
		
		int pos = 0;
		
		int line_number = 0;
		
		while (setuplines.hasMoreTokens())
		{
			
			line = setuplines.nextToken();
			
			_script_lines[line_number] = line;
			
			// if the line contains the set prefix and an = sign
			// then this is a parameter line (possibly rem'd out)
			set_present = line.toLowerCase().indexOf(prefix);
			equals_present = line.indexOf('=', set_present);
			
			if (set_present != -1 && equals_present != -1)
			{ // this is a parameter line
				
				name  = getName(line);
				value = getValue(line);
				
				// if variable name is not null
				if (name != null)
				{
					
					// if it is an editable variable
					if (_editable_vars.contains(name))
					{
						
						// override value in script with real location
						if (name.equals(SETUPFIPAOS_FIPAOS) && _legacy)
						{
							
							value = _real_fipaos_location;
							
						}
						
						_script_variables[pos] 	= name;
						_script_values[pos]	= value;
						
						_script_keys_values.put(name, value);
						
						_vars_to_lines.put(name, new Integer(line_number));
						
						pos++;
						
					} // end if editable
					
				} // end if valid variable
				
			} // end if variable line
			
			line_number++;
			
		} // end while
		
	} // end extractSetupParameters()
	
	/*********************************************************
	 *
	 * getScriptVariables()
	 *
	 * @return Map. The variables parsed from the setup script as key-value pairs
	 *
	 *********************************************************/
	public Map getScriptVariables()
	{
		
		return _script_keys_values;
		
	} // end getScriptVariables()
	
	/*********************************************************
	 *
	 * getName()
	 *
	 * @param setup_line. The line in the script to examine.
	 *
	 * @return String. The variable name found on that line. null if none found.
	 *
	 *********************************************************/
	private String getName(String setup_line)
	{
		
		int start;
		int end;
		
		String prefix = "";
		
		if (_is_windows == true)
		{
			
			prefix = "set ";
			
			start = setup_line.toLowerCase().indexOf(prefix);
			
			end 	= setup_line.indexOf('=', start);
			
		}
		else
		{ // no 'set' used to set unix environment vars
			
			start = 0;
			
			end 	= setup_line.indexOf('=', start);
			
		}
		
		if (start != -1 && end != -1)
		{
			
			return setup_line.substring(start + prefix.length(), end);
			
		}
		else
		{
			
			return null;
			
		}
		
	} // end getName()
	
	/*********************************************************
	 *
	 * getValue()
	 *
	 * @param setup_line. The line in the script to examine.
	 *
	 * @return String. The value from the given line. null if none found.
	 *
	 *********************************************************/
	private String getValue(String setup_line)
	{
		
		int start = setup_line.indexOf('=');
		
		if (start != -1)
		{
			
			return setup_line.substring(start + 1, setup_line.length());
			
		}
		else
		{
			
			return null;
			
		}
		
	}	// end getValue()
	
	/*********************************************************
	 *
	 * setScriptVariables()
	 *
	 * @param variables. The values to write back to the script file
	 *
	 *********************************************************/
	public void setScriptVariables(Map variables)
	{
		
		Iterator keys = variables.keySet().iterator();
		
		String key;
		String value;
		Integer to_replace;
		int line_number;
		
		String new_line;
		
		// for each variable returned
		while (keys.hasNext())
		{
			
			// get the variable name
			key = (String)keys.next();
			
			// get the value
			value = (String)variables.get(key);
			
			// get the script line number to replace
			to_replace = (Integer)_vars_to_lines.get(key);
			
			// this can occur for variables that are used in the configurator but not in the GUI
			if (to_replace != null)
			{
				
				// convert to an int
				line_number = to_replace.intValue();
				
				if (_is_windows == true)
				{
					
					new_line = "SET " + key + "=" + value;
					
				}
				else
				{
					
					new_line = key + "=" + value;
					
				}
				
				// replace the line in the script
				_script_lines[line_number] = new_line;
				
			}
			
		} // end while
		
		StringBuffer updated_script = new StringBuffer();
		
		String linebreak = System.getProperty("line.separator");
		
		for (int i = 0; i < _script_lines.length; i++)
		{
			
			updated_script.append(_script_lines[i]);
			updated_script.append(linebreak);
			
		}
		
		_setupscript = updated_script.toString();
		
	} // end setScriptVariables()
	
	/*********************************************************
	 *
	 * saveNewConfiguration()
	 *
	 * Save the updated information to the configuration files
	 *
	 *********************************************************/
	public void saveNewConfiguration()
	{
		
		// Save the setup script
		try
		{
			FIPAOSConfiguratorTools.saveStringIntoFile(_setupscript_filename, _setupscript);
		}
		catch (Exception e)
		{
			System.out.println(e);
			System.out.println("Could not save SetupFIPAOS script");
		}
		
	} // end saveNewConfiguration()
	
} // end ScriptEditor.java
