//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.tool.loader;

//----------------------------------------------------------------------
import fipaos.agent.FIPAOSAgent;

import fipaos.agent.profile.*;
import fipaos.ont.fipa.FIPACONSTANTS;
import fipaos.ont.BASETOKENS;
import fipaos.util.DynamicClass;
import fipaos.util.DIAGNOSTICS;
import fipaos.util.XMLDataBinding;

import java.util.*;
import java.io.*;
import javax.swing.*;

//----------------------------------------------------------------------

/** The Agent Loader will instantiate agents specified in the loader profile
 * in the same java virtual machine.
 */
public class AgentLoader
{
	
	//----------------------------------------------------------------------
	// CONSTANTS
	//----------------------------------------------------------------------
	
	//----------------------------------------------------------------------
	// STATIC VARIABLES
	//----------------------------------------------------------------------
	
	//----------------------------------------------------------------------
	// INSTANCE VARIABLES
	//----------------------------------------------------------------------
	
	/** 
	 * Map of instances of started agents
	 */
	private Map _started_agents = new HashMap();
	
	/** 
	 * Map of all startable agents against AgentDescription
	 */
	private Map _agents = new HashMap();
	
	/**
	 * List of started Agents (in order they were started)
	 */
	private List _started_agents_list = new LinkedList();
	
	/**
	 * List of all startable Agents
	 */
	private List _agents_list = new LinkedList();
	
	/**
	 * LoaderProfile for this
	 */
	private LoaderProfile _loader_profile;
	
	/**
	 * Location of platform profile for use when creating agents
	 */
	private String _platform_profile;
	
	//----------------------------------------------------------------------
	// CONSTRUCTOR
	//----------------------------------------------------------------------
	
	/** Constructs the agent loader and instantiates the specified agents.
	 * @param profile_location Directory where the agent loader profile is stored
	 * @param file_name the filename of the profile
	 */
	public AgentLoader( String platform_profile_location, String loader_profile_name, boolean gui )
	{
		// Save platform profile location
		_platform_profile = platform_profile_location;
		
		// Attempt to set the default L&F for the platform we're running on
		try
		{
			UIManager.setLookAndFeel ( UIManager.getSystemLookAndFeelClassName() );
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Couldn't set the default Look & Feel for this OS", this, DIAGNOSTICS.LEVEL_3 );
		}
		
		try
		{
			// load the platform profile as we need to get the profiles directory from it
			PlatformProfile platform_profile = (PlatformProfile)XMLDataBinding.unmarshallFile( platform_profile_location );
			String profiles_dir = platform_profile.getProfileDirectory();
			String loader_file = profiles_dir + loader_profile_name;
			_loader_profile = (LoaderProfile)XMLDataBinding.unmarshallFile(loader_file);
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Error reading profile - aborting", t, DIAGNOSTICS.LEVEL_MAX );
			return;
		}
		
		// Get AgentDescriptions from LoaderProfile
		Iterator i = _loader_profile.getAgentDescriptions().iterator();
		while( i.hasNext() )
		{
			AgentDescription ad = (AgentDescription) i.next();
			
			// Check an Agent name has been given
			if( ad.getAgentName() == null )
			{
				DIAGNOSTICS.println( "WARNING: An agent specified in \"" + loader_profile_name + "\" has no name - not added to list of agents", 
															DIAGNOSTICS.LEVEL_MAX );
			}
			else
			{
				// Agent name given - add Agent to our list/map of Agents
				_agents.put( ad.getAgentName(), ad ); 
				_agents_list.add( ad.getAgentName() );
			}
		}
		
		startAgents();
		
		if( gui )
		{
			try
			{
				LoaderGUI loaderGUI = new LoaderGUI( this );
				loaderGUI.show();
			}
			catch( Throwable t )
			{
				DIAGNOSTICS.println( "Error starting GUI, continuing without...", this, DIAGNOSTICS.LEVEL_MAX );
			}
		}
		
		checkAgentState();
	}
	
	//----------------------------------------------------------------------
	// PUBLIC METHODS
	//----------------------------------------------------------------------
	
	/**
	 * Get a list of "alive" agents
	 */
	public List getAliveAgents()
	{
		return new LinkedList( _started_agents_list );
	}
	
	/**
	 * Get a list of Agents that can be started
	 */
	public List getAgents()
	{
		return new LinkedList( _agents_list );
	}
	
	/**
	 * Get AgentDesciption for given Agent
	 */
	public AgentDescription getAgentDescription( String agent )
	{
		return (AgentDescription) _agents.get( agent );
	}
	
	/**
	 * Get the state of the Agent with the given name - returns null if no such agent
	 * is alive
	 */
	public String getAgentState( String agent_name )
	{
		// Get FIPAOSAgent from _started_agents
		Object o = _started_agents.get( agent_name );
		if( o instanceof FIPAOSAgent )
		{
			FIPAOSAgent agent = (FIPAOSAgent) o;

			// Check if we got a reference
			if( agent != null )
			{
				return agent.getState();
			}
			else
			{
				return null;
			}
		}
		else
		{
			return null;
		}
	}
	
	/** Shut down the agent with the given name
	 * @param agent The agent to be shutdown
	 */
	public void shutdownAgent( final String agent_name )
	{
		shutdownAgent( agent_name, true );
	}
	
	/** Shut down the agent with the given name
	 * @param agent The agent to be shutdown
	 * @param async Do this asyncronously (with timeout!)
	 */
	public void shutdownAgent( final String agent_name, boolean async )
	{
		// See if such an Agent exists
		if( getAgentState( agent_name ) != null )
		{
			final FIPAOSAgent agent = (FIPAOSAgent) _started_agents.get( agent_name );
			
			// If Agent doesn't exist
			if( agent == null )
			{
				return;
			}
			
			// Start Async shutdown?
			if( async )
			{
				Thread t = new Thread( new Runnable()
				{
					public void run()
					{
						if( agent != null )
						{
							try
							{
								DIAGNOSTICS.println( "Stopping " + agent_name );
								agent.shutdown();
							}
							catch( Throwable x )
							{
								DIAGNOSTICS.println( "Error shuting-down " + agent_name, x, DIAGNOSTICS.LEVEL_3 );
							}

							_started_agents_list.remove( agent_name );
							_started_agents.remove( agent_name );
						}
					}
				} );
				t.start();
			}
			else
			{
				try
				{
					agent.shutdown();
				}
				catch( Throwable t )
				{
					DIAGNOSTICS.println( "Error shuting-down " + agent_name, t, DIAGNOSTICS.LEVEL_3 );
				}
			}
		}		
	}
	
	/** Starts an agent
	 * @param agent_name Name of the Agent (defined in the Loader Profile) to start
	 * @param async Start the Agent asyncronously?
	 */
	public void startAgent( String agent_name, boolean async )
	{
		// Extract AgentDescitpion for the Agent to be started
		AgentDescription agent_desc = (AgentDescription) _agents.get( agent_name ); 
		
		// Check that the given agent has been defined
		if( !_agents.containsKey( agent_name ) )
		{
			JOptionPane.showMessageDialog( null, "Agent \"" + agent_name + "\" is not defined - cannot load Agent!" );
			return;
		}
		
		startAgent( agent_name, agent_desc.getClassName(), agent_desc.getOwner(), async );
	}

	/** Starts an agent (async)
	 * @param agent_name Name of the Agent to start
	 * @param class_name Name of the class for this Agent
	 */
	public void startAgent( String agent_name, String class_name )
	{
		startAgent( agent_name, class_name, "fipaos", true );
	}
		
	/** Starts an agent
	 * @param agent_name Name of the Agent to start
	 * @param class_name Name of the class for this Agent
	 * @param owner Owner of the Agent
	 * @param async Start the Agent asyncronously?
	 */
	public void startAgent( final String agent_name, final String class_name,
																				final String owner, boolean async )
	{
		// Check that no agents with the same name are currently alive
		if( _started_agents.containsKey( agent_name ) )
		{
			JOptionPane.showMessageDialog( null, "Agent \"" + agent_name + "\" is already running!" );
			return;
		}

		// Check if name is valid (contains no spaces!)
		if( agent_name.indexOf( " " ) != -1 )
		{
			JOptionPane.showMessageDialog( null, "Agent names cannot contain spaces!" );
			return;
		}
		
		// A thread group so that the threads belonging to an agent can be kept track of
		ThreadGroup new_tg = new ThreadGroup( agent_name );
		
		// Object to lock on for sync. execution
		final boolean done[] = new boolean[1];
		
		// Start agent in new thread
		Thread t = new Thread( new_tg, new Runnable()
		{
			public void run()
			{
				try
				{
					// Add place-holder to list of active agents
					_started_agents_list.add( agent_name );
					_started_agents.put( agent_name, new Object() );
					
					// Setup expected parameters
					Class args[] = new Class[] { String.class, String.class, String.class };
					Object params[] = new Object[] { _platform_profile, agent_name,
																						owner };
					
					DynamicClass dc = new DynamicClass( class_name, args, params );
					
					// Add real FIPAOSAgent instance to map of started Agents
					_started_agents.put( agent_name, dc.getObject() );
				}
				catch(Throwable x)
				{
					_started_agents_list.remove( agent_name );
					_started_agents.remove( agent_name );
					DIAGNOSTICS.println( "Something went wrong while starting Agent", x,
					DIAGNOSTICS.LEVEL_MAX );
				}
				finally
				{
					synchronized( done )
					{
						done[0] = true;
						done.notifyAll();
					}
				}
			}
		} );
		
		// Start thread async?
		if( !async )
		{
			// Wait for other thread to stop
			synchronized( done )
			{
				t.start();
				
				while( !done[0] )
				{
					try
					{
						done.wait();
					}
					catch( InterruptedException e )
					{
					}
				}
			}
		}
		else
		{
			// Just start other thread and continue
			t.start();
		}
	}
	
	/** 
	 * Method that starts all Agent that are marked as startable at AgentLoader initialisation
	 */
	public void startAgents()
	{
		AgentDescription ams = null;
		AgentDescription df = null;
		
		// Check if AMS or DF have been specified - if so start these first
		Iterator i = _agents_list.iterator();
		while( i.hasNext() )
		{
			String next = (String) i.next();
			if( next.equalsIgnoreCase( "df" ) )
			{
				df = (AgentDescription) _agents.get( next );
			}
			else if( next.equalsIgnoreCase( "ams" ) )
			{
				ams = (AgentDescription) _agents.get( next );
			}
		}
		
		// Start AMS and DF if necessary
		if( ams != null && ams.getStart() )
		{
			startAgent( ams.getAgentName(), false );
		}
		if( df != null && df.getStart() )
		{
			startAgent( df.getAgentName(), false );
		}
		
		// Iterate though AgentDescriptions
		i = _agents_list.iterator();
		while( i.hasNext() )
		{
			AgentDescription ad = (AgentDescription) _agents.get( i.next() );
			
			// Check if we should start this Agent
			if( ad.getStart() && 
					!( ad.getAgentName().equalsIgnoreCase( "df" ) ||
							ad.getAgentName().equalsIgnoreCase( "ams" ) ) )
			{
				startAgent( ad.getAgentName(), false );
			}
		}
	}
	
	/**
	 * Stops all executing agents
	 */
	public void shutdownAgents()
	{
		// Get a list of started Agent - move AMS & DF to the end since they must
		// be stopped last
		List to_be_stopped = new LinkedList( _started_agents_list );
		boolean moved_ams = false;
		boolean moved_df = false;
		
		// Iterate through list, and move AMS and DF to end of list (DF should come before
		// AMS)
		Iterator i = new LinkedList( to_be_stopped ).iterator();
		while( i.hasNext() )
		{
			String agent = (String) i.next();
			if( agent.equalsIgnoreCase( "df" ) && !moved_df )
			{
				moved_df = true;
				to_be_stopped.remove( agent );				
				if( moved_ams )
				{
					to_be_stopped.add( to_be_stopped.size() - 1, agent );
				}
				else
				{
					to_be_stopped.add( agent );
				}
			}
			else if( agent.equalsIgnoreCase( "ams" ) && !moved_ams ) 
			{
				moved_ams = true;
				to_be_stopped.remove( agent );				
				to_be_stopped.add( agent );
			}
		}
		
		// Now remove Agents
		i = to_be_stopped.iterator();
		while( i.hasNext() )
		{
			String agent = (String) i.next();
			shutdownAgent( agent, false );
		}
		
		// Flush DIAGNOSTICS
		DIAGNOSTICS.flush();
	}
	
	/**
	 * Loop to remove dead Agents
	 */
	private void checkAgentState()
	{
		while( true )
		{
			try
			{
				Thread.sleep( 250 );
				
				List l = new LinkedList( _started_agents_list );
				Iterator i = l.iterator();
				while( i.hasNext() )
				{
					String agent = (String) i.next();
					String state = getAgentState( agent );
					Object agent_instance = _started_agents.get( agent );
					if( agent_instance instanceof FIPAOSAgent &&
							 ( (FIPAOSAgent) agent_instance ).isShutdown() )
					{
						_started_agents_list.remove( agent );
						_started_agents.remove( agent );
					}
				}
			}
			catch( Throwable t )
			{
			}
		}
	}
	
	/**
	 * "Activate" the given Agent
	 */
	public void activate( String agent_name )
	{
		// See if such an Agent exists
		if( getAgentState( agent_name ) != null )
		{
			FIPAOSAgent agent = (FIPAOSAgent) _started_agents.get( agent_name );
			
			if( agent != null )
			{
				agent.activate();
			}
		}
	}
	
	//----------------------------------------------------------------------
	// MAIN/TESTING METHOD
	//----------------------------------------------------------------------
	
	public static void main(String[] args)
	{
		boolean gui = true;
		String platform = null;
		String loader = null;
			
		try
		{
			if( args != null && args.length > 1 )
			{
				platform = args[0];
				loader = args[1];
				
				if( args[0].equalsIgnoreCase( "-nogui" ) )
				{
					gui = false;
					platform = args[1];
					loader = args[2];
				}
			}
			else
			{
				platform = "profiles/platform.profile";
				loader = "loader.profile";
			}
			
		}
		catch( Throwable t )
		{
			t.printStackTrace();
			System.out.println( "Usage: " + AgentLoader.class + " [-nogui] <platform-profile-location> <loader-profile-name>" );
			System.exit( -10 );
		}

		AgentLoader agent_loader = new AgentLoader( platform, loader, gui );
	}
	
} // End AgentLoader


