//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Emorphia are Copyright (C) 2001 Emorphia Limited
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.util;

// import Enhydra data binding classes
import org.enhydra.xml.binding.*;

// import JDOM classes
import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;

// import Zeus classes
import org.enhydra.zeus.source.StreamSource;
import org.enhydra.zeus.Source;

import java.io.*;

public class XMLDataBinding
{
	public static final String ENHDRA_PACKAGE_HEADER_START = "<?enhydra-unmarshall package=\"";
	public static final String ENHDRA_PACKAGE_HEADER_END = "\"?>";
	public static final String XML_HEADER_START = "<?xml version=\"";
	public static final String XML_HEADER_END = "?>";
	
/** Marshalls a Data-Binding compatible object into an XML instance document
 * @param obj The Object to be marshalled
 * @param strip_enhydra_tags Indicates if Enhydra specific tags should be removed
 * @return A String containing the XML instance document
 */
	public static String marshall(Object obj, boolean strip_enhydra_tags ) throws IOException
	{
		String xml_output = null;
		Document xml = null;
		
		StringWriter writer = new StringWriter();
		xml = Marshaller.marshall(obj);
		
		if (xml != null)
		{
			
			XMLOutputter outputter = new XMLOutputter("\t", true);
			outputter.output(xml, writer);			
			xml_output = writer.toString();
			
			// Check if we need to strip Enhydra specific tags
			if( strip_enhydra_tags )
			{
				StringBuffer out = new StringBuffer();
				
				// Check for header
				int index_start_header = xml_output.indexOf( ENHDRA_PACKAGE_HEADER_START );
				int index_end_header = xml_output.indexOf( ENHDRA_PACKAGE_HEADER_END, index_start_header + 1 );
				
				// Strip header (if it exists)
				if( index_start_header > -1 )
				{
					out.append( xml_output.substring( 0, index_start_header ) );
					out.append( xml_output.substring( index_end_header + ENHDRA_PACKAGE_HEADER_END.length() ) );
				}
				
				// Get modified output
				xml_output = out.toString();
			}
			
			return xml_output;
		}

		return xml_output;
	} // end marshall()
	
	
/** Marshalls a Data-Binding compatible object into an XML instance document
 * @param obj The Object to be marshalled
 * @return A String containing the XMLhjkhjkhjkhjk instance document
 */
	public static String marshall(Object obj) throws IOException
	{
		return marshall( obj, false );
	} // end marshall()
	
	
/** Marshalls a Data-Binding compatible object into an XML instance document and
 * saves the document into a file
 * @param obj The Object to be marshalled
 * @param filename The file to save into
 */
	public static void marshallFile(Object obj, String filename) throws IOException
	{
		
		String instance = marshall(obj);
		
		FileWriter writer = new FileWriter(filename);
		
		writer.write(instance, 0, instance.length());
		
		writer.flush();
						
	} // end marshallFile()	
	
	
/** Marshalls a Data-Binding compatible object into an XML instance document and
 * saves the document into a file
 * @param obj The Object to be marshalled
 * @param file The file to save into
 */
	public static void marshallFile(Object obj, File file ) throws IOException
	{
		marshallFile( obj, file.getCanonicalPath() );
	}
		
	
/** Unmarshalls a string back into a Data-Binding object
 * @param xml The String to be unmarshalled
 * @return The reconstructed object
 */
	public static Object unmarshall(String xml) throws IOException, org.jdom.JDOMException
	{
		
		Document doc = null;
		
		Object obj = null;
		
		SAXBuilder saxbuilder = new SAXBuilder();
		
		doc = saxbuilder.build(new StringReader(xml));
		
		obj = Unmarshaller.unmarshall(doc);
		
		return obj;
		
	} // end unmarshall()
	
	
/** Unmarshalls a string back into a Data-Binding object, and inserts Enhydra specific tags
 * back into XML if necessary
 * @param xml The String to be unmarshalled
 * @param default_package Default package to find objects to create to represent the XML content
 * @return The reconstructed object
 */
	public static Object unmarshall(String xml, String default_package ) throws IOException, org.jdom.JDOMException
	{
		// Check if we need to add Enhydra specific tags
		if( default_package != null )
		{
			StringBuffer out = new StringBuffer();

			// Check for header
			int index_start_header = xml.indexOf( ENHDRA_PACKAGE_HEADER_START );

			// Add header (if it doesn't exist)
			if( index_start_header == -1 )
			{
				int index_start_xml_header = xml.indexOf( XML_HEADER_START );
				int index_end_xml_header = xml.indexOf( XML_HEADER_END, index_start_xml_header + 1 );
				
				out.append( xml.substring( 0, index_end_xml_header + XML_HEADER_END.length() ) );
				out.append( '\n' );
				out.append( ENHDRA_PACKAGE_HEADER_START );
				out.append( default_package );
				out.append( ENHDRA_PACKAGE_HEADER_END );
				out.append( xml.substring( index_end_xml_header + XML_HEADER_END.length() ) );
			}

			// Get modified output
			xml = out.toString();
		}
		
		System.out.println( xml );
		
		return unmarshall( xml );
	} // end unmarshall()
	
	
/** Create a completely distinct clone of a data bound object by
 * marshalling it to a String and then unmarshalling it back into
 * a new databound object.
 *
 *@param obj. The object to clone.
 *@return Object. An exact copy of obj.
 */	
	public static Object cloneDataboundObject(Object obj) throws IOException, org.jdom.JDOMException
	{
	
		return unmarshall(marshall(obj));	
		
	}
	
	
/** Unmarshalls an XML document in a file back into a Data-Binding object
 * @param filename The File to be unmarshalled
 * @return The reconstructed object
 */
	public static Object unmarshallFile(String filename) throws IOException, org.jdom.JDOMException
	{
		
		String contents;
		
		contents = DocumentRetriever.getDoc(filename);
			
		return unmarshall(contents);

	} // end unmarshall()
	
	
/** Unmarshalls a file back into a Data-Binding object
 * @param file The String to be unmarshalled
 * @return The reconstructed object
 */
	public static Object unmarshallFile( File file ) throws IOException, org.jdom.JDOMException
	{
		return unmarshallFile( file.getCanonicalPath() );
	}
	
	
	/**
	 * Unmarshall the given XML using the Zeus marshaller
	 *
	 * @param xml XML document to unmarshall
	 * @param def_package Default package to look for classes to unmarshall the XML into
	 * @return An object representing the XML
	 */
	public static Object zeusUnmarshall( String xml, String def_package ) throws java.io.IOException
	{
		org.enhydra.zeus.Unmarshaller unmarshaller = new org.enhydra.zeus.Unmarshaller();
		unmarshaller.setJavaPackage( def_package );
		return unmarshaller.unmarshal( (Source) new StreamSource( new StringReader( xml ) ) );
	}
	
	/**
	 * Marshall the given Object using the Zeus marshaller
	 *
	 * @param obj Object to marshall
	 * @return The XML document representing the Object
	 */
	public static String zeusMarshall( Object obj ) throws java.io.IOException
	{
		org.enhydra.zeus.Marshaller marshaller = new org.enhydra.zeus.Marshaller();
		StringWriter writer = new StringWriter();
		org.enhydra.zeus.result.StreamResult result = new org.enhydra.zeus.result.StreamResult( writer );
		marshaller.marshal( obj, result );
		return writer.getBuffer().toString();
	}
} // end class XMLDataBinding