//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis,
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Emorphia are Copyright (C) 2001 Emorphia Limited
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.util.objecteditor;

import fipaos.util.*;
import fipaos.util.objecteditor.primitive.*;
import java.util.*;

/**
 * Factory to produce ObjectEditorComponent instances that can modify the
 * properties of an object
 */
public class ObjectEditorComponentFactory
{
	
	/**
	 * Generate a JPanel sub-class which implements the ObjectEditorComponent
	 * interface for the given Object
	 */
	public static ObjectEditorComponent getComponent( Object obj )
	{
		ObjectEditorComponent component = null;
		
		// Attempt to create ObjectEditorComponent to deal with given property
		Class editor_type = getPrimitiveEditor( obj.getClass() );
		
		// Attempt to dynamically instantiate class
		// Try Object as argument to constructor
		try
		{
			Class arg_types[] = new Class[]
			{ Object.class };
			Object args[] = new Object[]
			{ obj };
			DynamicClass dc = new DynamicClass( editor_type.getName(), arg_types, args );
			component = (ObjectEditorComponent) dc.getObject();
		}
		catch( DynamicClassException e )
		{
			DIAGNOSTICS.println( "Error creating GUI component", e.getThrowable(), DIAGNOSTICS.LEVEL_MAX );
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Couldn't instantiate class " + editor_type, t );
		}
		
		return component;
	}
	
	
	/**
	 * Generate a JPanel sub-class which implements the ObjectEditorComponent
	 * interface for the given Property
	 */
	public static ObjectEditorComponent getComponent( DynamicClass.Property property )
	{
		ObjectEditorComponent component = null;
		Class editor_type = null;
		
		// Check if we can auto-determine what GUI component to use
		if( property.getHint() instanceof DynamicClass.ConstrainedSelectionHint )
		{
			editor_type = ConstrainedChoicePanel.class;
		}
		else
		{
			// Attempt to create ObjectEditorComponent to deal with given property
			editor_type = getPrimitiveEditor( property.getType() );
		}
		
		// If type is complex, we just want to add a ComplexPropertyEditorPanel
		if( editor_type == ComplexEditorPanel.class )
		{
			editor_type = ComplexPropertyEditorPanel.class;
		}
		
		// Attempt to dynamically instantiate class
		// Try Object as argument first, and if that fails try the property
		// object
		try
		{
			Class arg_types[] = new Class[]
			{ Object.class };
			Object args[] = new Object[]
			{ property.getValue() };
			DynamicClass dc = new DynamicClass( editor_type.getName(), arg_types, args );
			component = (ObjectEditorComponent) dc.getObject();
		}
		catch( DynamicClassException e )
		{
			DIAGNOSTICS.println( "Error creating GUI component", e.getThrowable(), DIAGNOSTICS.LEVEL_MAX );
		}
		catch( Throwable t )
		{
			// Something bad happened - try again using the property object
			try
			{
				Class arg_types[] = new Class[]
				{ DynamicClass.Property.class };
				Object args[] = new Object[]
				{ property };
				DynamicClass dc = new DynamicClass( editor_type.getName(), arg_types, args );
				component = (ObjectEditorComponent) dc.getObject();
			}
			catch( DynamicClassException e )
			{
				DIAGNOSTICS.println( "Error creating GUI component", e.getThrowable(), DIAGNOSTICS.LEVEL_MAX );
			}
			catch( Throwable s )
			{
				DIAGNOSTICS.println( "Couldn't instantiate class " + editor_type, t );
			}
		}
		
		return component;
	}
	
	/**
	 * Generate a JDialog sub-class which implements the ObjectEditorComponent
	 * interface for the given Property
	 */
	public static ObjectEditorComponent getDialog( java.awt.Frame parent, DynamicClass.Property property )
	{
		return new ObjectEditorComponentDialog( parent, property );
	}
	
	/**
	 * Generate a JDialog sub-class which implements the ObjectEditorComponent
	 * interface for the given Object
	 */
	public static ObjectEditorComponent getDialog( java.awt.Frame parent, Object obj )
	{
		if( obj instanceof String )
		{
			return new JOptionPaneDialog( parent, (String) obj );
		}
		else
		{
			return new ObjectEditorComponentDialog( parent, obj );
		}
	}
	
	/**
	 * Returns the type of object to instantiate to edit the given type
	 */
	private static Class getPrimitiveEditor( Class c )
	{
		// Search through the possibilities, and return an appropriate component
		if( c.equals( String.class ) )
		{
			return StringEditorPanel.class;
		}
		else if( c.equals( Boolean.class ) )
		{
			return BooleanEditorPanel.class;
		}
		else if( c.equals( Integer.class ) )
		{
			return IntegerEditorPanel.class;
		}
		else if( c.equals( Long.class ) )
		{
			return LongEditorPanel.class;
		}
		else if( c.equals( Float.class ) )
		{
			return FloatEditorPanel.class;
		}
		else if( c.equals( Double.class ) )
		{
			return DoubleEditorPanel.class;
		}
/*		else if( c.equals( Array.class ) )
		{
			return ArrayEditorPanel( property );
		}
 */
		else if( c.isAssignableFrom( List.class ) )
		{
			return ListEditorPanel.class;
		}
		else if( c.isAssignableFrom( Set.class ) )
		{
			return SetEditorPanel.class;
		}
		else if( c.isAssignableFrom( Map.class ) )
		{
			return MapEditorPanel.class;
		}
		else
		{
			return ComplexEditorPanel.class;
		}
	}
	
	private static class JOptionPaneDialog implements ObjectEditorComponent
	{
		private String _orig_value;
		private String _new_value;
		
		public JOptionPaneDialog( java.awt.Frame parent, String value )
		{
			javax.swing.Icon i = (parent != null ? new javax.swing.ImageIcon( parent.getIconImage() ) : null );
			_orig_value = value;
			_new_value = (String) javax.swing.JOptionPane.showInputDialog( parent, "Enter new value:", "Input", javax.swing.JOptionPane.OK_CANCEL_OPTION, i, null, value );
		}
		
		/**
		 * @return The new value
		 */
		public Object getValue()
		{
			if( _new_value == null )
			{
				return _orig_value;
			}
			else
			{
				return _new_value;
			}
		}
	}
}
